/**
 * 
 */
package com.cosylab.application;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.net.URL;
import java.util.Properties;
import java.util.logging.ConsoleHandler;
import java.util.logging.Level;
import java.util.logging.Logger;



/**
 * This class searches for configuration file in several location in well defined order.
 * 
 * @author ikriznar
 *
 */
public final class ConfigurationLoader {
	
	private ConfigurationLoader() {
		super();
	}
	
	/**
	 * Tries to return class loader from provided class. If not possible, 
	 * tries to return default class loader. If no usable class loader is found, then <code>null</code>
	 * is returned.
	 * 
	 * @param c class for witch class loader is searched
	 * @return class loader instance or <code>null</code>
	 */
	public static ClassLoader findClassLoader(Class c) {
		Logger l= Logger.getLogger(ConfigurationLoader.class.getName());
		ClassLoader cl=null;
		
		if (c == null) {
			c= ConfigurationLoader.class;
		}
			
		try {
			cl= c.getClassLoader();
		} catch (Exception e) {
			l.log(Level.FINEST,"Failed to load class loader for '"+c.getName()+"'",e);
		}
		
		if (cl!=null) {
			return cl;
		}
		
		try {
			cl= ClassLoader.getSystemClassLoader();
		} catch (Exception e) {
			l.log(Level.FINEST,"Failed to load system class loader for '"+c.getName()+"'",e);
		}
		
		return cl;
	}
	
	public static Properties loadConfiguration(Class c, String name) {
		return loadConfigurationFile(c,name+".properties");
	}
	public static Properties loadConfiguration(String name) {
		return loadConfigurationFile(null,name+".properties");
	}
	public static Properties loadConfigurationFile(Class c, String propFileName) {
		Logger l= Logger.getLogger(ConfigurationLoader.class.getName());
		
		Properties p= new Properties();
		
		/*
		 * Tries to load from classpath (eg. configuration in jar)
		 */
		
		ClassLoader cl = findClassLoader(c);
		if (cl!=null) {
			
			URL url= cl.getResource(".");
			InputStream is=null;
			File f= null;
			try {
				f= new File(url.toURI());
				f= new File(f,propFileName);
				if (f.exists()) {
					is= new BufferedInputStream(new FileInputStream(f));
					p.load(is);
					is.close();
					l.log(Level.CONFIG,"Configuration '"+propFileName+"' loaded from file '"+f.getAbsolutePath()+"'.");
				} else {
					l.log(Level.FINEST,"Configuration '"+propFileName+"' does not exist as file '"+f.getAbsolutePath()+"'.");
				}
			} catch (Exception e) {
				l.log(Level.FINEST,"Failed to load '"+propFileName+"' from file '"+f+"'.",e);
			} finally {
				try {
					if (is!=null) is.close();
				} catch (Exception e) {
					// ignore
				}
			}
			
			is= cl.getResourceAsStream(propFileName);
			l.log(Level.FINE,"Configuration '"+propFileName+"' not found on classpath.");
			if (is!=null) {
				try {
					p.load(is);
					l.log(Level.CONFIG,"Configuration '"+propFileName+"' loaded from classpath.");
					is.close();
				} catch (Exception e) {
					l.log(Level.FINEST,"Failed to load '"+propFileName+"' from classpath.",e);
				} finally {
					try {
						if (is!=null) is.close();
					} catch (Exception e) {
						// ignore
					}
				}
			}
		}
		
		/*
		 * Tries to load from current directory.
		 */
		File f= new File(System.getProperty("user.dir"));
		InputStream is=null;
		try {
			if (f.exists()) {
				f= new File(f,propFileName);
				if (f.exists()) {
					is= new BufferedInputStream(new FileInputStream(f));
					p.load(is);
					is.close();
					l.log(Level.CONFIG,"Configuration '"+propFileName+"' loaded from file '"+f.getAbsolutePath()+"'.");
				}
			} else {
				l.log(Level.FINEST,"Configuration '"+propFileName+"' does not exist as file '"+f.getAbsolutePath()+"'.");
			}
		} catch (Exception e) {
			l.log(Level.FINEST,"Failed to load '"+propFileName+"' from file '"+f.getAbsolutePath()+"'.",e);
		} finally {
			try {
				if (is!=null) is.close();
			} catch (Exception e) {
				// ignore
			}
		}
		
		
		return p;
	}
	
	public static void main(String[] args) {
		Logger l= Logger.getLogger(ConfigurationLoader.class.getName());
		l.setLevel(Level.ALL);
		ConsoleHandler c= new ConsoleHandler();
		c.setLevel(Level.ALL);
		l.addHandler(c);
		
		Properties p= ConfigurationLoader.loadConfiguration("project");
		c.flush();
		p.list(System.out);
		
	}

}
