/*
 * Copyright (c) 2003 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */
package com.cosylab.application.state.impl;

import com.cosylab.application.state.*;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintWriter;

import java.util.Iterator;


/**
 * This class is an implementation of the <code>StateStorage</code> that uses
 * simple text file to store the array of the <code>State</code> objects.
 * Stored file is the plain <code>java.util.Properties</code> file format with
 * the few 'weird' properties added which are used for fast hierarchy
 * resolving.
 *
 * @author dvitas
 */
public class TextStateStorage extends DefaultStateStorage {
    protected final static String stateStart = "<State>";
    protected final static String stateEnd = "<EndState>";
    protected final static String childDelimiter = "<Child>";

    /**
                     *
                     */
    public TextStateStorage() {
        super();
    }

    /**
     * Creates a new TextStateStorage object.
     *
     * @param ss DOCUMENT ME!
     */
    public TextStateStorage(StateStorage ss) {
        super(ss);
    }

    private void readState(BufferedReader br, State state)
        throws IOException {
        String line;

        while (true) {
            line = br.readLine();

            if (line == null) {
                break;
            }

            line = line.trim();

            if (line.startsWith(stateEnd)) {
                break;
            }

            if (line.startsWith(childDelimiter)) {
                String childKey = line.substring(childDelimiter.length());
                State s = state.createState(childKey);
                br.readLine(); // eat state header
                readState(br, s);

                continue;
            }

            int index = line.indexOf('=');
            String key = line.substring(0, index);
            String value = line.substring(index + 1);
            state.putString(key, value);
        }
    }

    /* (non-Javadoc)
     * @see com.cosylab.application.state.StateStorage#load(java.io.InputStream)
     */
    public void load(String filePath, String applicationName)
        throws IOException {
        if (delegate != null) {
            delegate.load(filePath, applicationName);
            addAll(delegate.getStates());

            return;
        }

        // get the InputStream if file exists
        InputStream is = getInputStream(filePath, applicationName + ".txt");
        load(is);
    }

    /* (non-Javadoc)
	 * @see com.cosylab.application.state.StateStorage#load(java.io.InputStream)
	 */
	public void load(InputStream is) throws IOException {
		if (is == null) {
			return;
		}

		BufferedReader br = new BufferedReader(new InputStreamReader(is));
		String line;

		while (true) {
			line = br.readLine();

			if (line == null) {
				break;
			}

			line = line.trim();

			if (line.startsWith(stateStart)) {
				State state = StateFactory.createState();
				readState(br, state);
				states.add(state);
			}
		}
	}
    
    private void printIndent(PrintWriter pw, int indent) {
        for (int i = 0; i < indent; i++) {
            pw.print("  ");
        }
    }

    private void print(PrintWriter pw, State s, int indent) {
        int i;
        Iterator iter = s.keySet().iterator();
        printIndent(pw, indent);
        pw.println(stateStart);

        while (iter.hasNext()) {
            String key = (String) iter.next();
            Object value = s.getObject(key);

            if (value instanceof State) {
                printIndent(pw, indent + 1);
                pw.println(childDelimiter + key);
                print(pw, (State) value, indent + 1);

                continue;
            }

            printIndent(pw, indent);
            pw.println(key + "=" + value.toString());
        }

        printIndent(pw, indent);
        pw.println(stateEnd);
    }

    /* (non-Javadoc)
     * @see com.cosylab.application.state.StateStorage#store(java.io.OutputStream)
     */
    public void store(String filePath, String applicationName)
        throws IOException {
        OutputStream os = getOutputStream(filePath, applicationName + ".txt");
        store(os);
    }
    
    /* (non-Javadoc)
	 * @see com.cosylab.application.state.StateStorage#store(java.io.OutputStream, java.lang.String)
	 */
	public void store(OutputStream os) throws IOException {
		if (os == null) {
			return;
		}

		PrintWriter pw = new PrintWriter(os);

		for (int i = 0; i < states.size(); i++) {
			State s = (State) states.get(i);
			print(pw, s, 0);
		}

		pw.flush();
		pw.close();
	}
}


/* __oOo__ */
