/*
 * Copyright (c) 2003 by Cosylab d.o.o.
 * 
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab,
 * d.o.o.
 * 
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.swing.AbstractAction;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.SwingConstants;

import com.cosylab.gui.components.util.ColorHelper;
import com.cosylab.gui.components.util.IconHelper;
import com.cosylab.gui.components.util.PopupManager;
import com.cosylab.gui.components.util.RunnerHelper;
import com.cosylab.gui.displayers.AbstractDisplayerPanel;
import com.cosylab.gui.displayers.CommonDisplayer;
import com.cosylab.gui.displayers.DataConsumer;
import com.cosylab.gui.displayers.LongConsumerMulticaster;
import com.cosylab.gui.util.UserSettingsProtection;
import com.cosylab.util.CommonException;

/** 
 * <code>IconDisplayer</code> is a displayer that displays icons associated with
 * different values.
 * 
 * @author Jaka Bobnar, Cosylab
 * 
 */
public class IconDisplayer extends AbstractDisplayerPanel implements
        com.cosylab.gui.displayers.IconDisplayer {
	
	public static final String ICONS_PROPERTY = "icons";
	public static final String DEFAULT_ICON_PROPERTY = "defaultIcon";
	public static final String DEFAULT_ICON_NAME_PROPERTY = "defaultIconName";

	private static final long serialVersionUID = 1L;

	private JLabel label;

	private Icon defaultIcon;
	private String defaultIconName;
	private long value;
	private PopupManager popupManager;
	private InfoDialog infoDialog;
	private boolean scaleIcons=false;
	
	private Map<Long, ValueIconPair> valueIcons;

	/**
     * Creates new IconDisplayer.
     */
	public IconDisplayer() {
		super();
		initialize();
	}

	/**
     * Creates new IconDisplayer and defines default icon.
     * 
     * @param defaultIcon
     *        Default icon to set.
     */
	public IconDisplayer(Icon defaultIcon) {
		super();
		this.defaultIcon=defaultIcon;
		defaultIconName = defaultIcon.toString();
		initialize();
	}

	/**
     * Creates new IconDisplayer and new default icon.
     * 
     * @param defaultIcon the name of default icon in classpath to use
     */
	public IconDisplayer(String defaultIcon) {
		this(IconHelper.createIcon(defaultIcon));
		defaultIconName = defaultIcon;
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.displayers.IconDisplayer#getLoadedIcons()
	 */
	public ValueIconPair[] getIcons() {
		ValueIconPair[] icons = new ValueIconPair[valueIcons.size()];
		Iterator<ValueIconPair> it = valueIcons.values().iterator();
		int i = 0;
		while (it.hasNext()) {
			icons[i] = it.next();
			i++;
		}
		return icons;
	
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.displayers.IconDisplayer#setLoadedIcons(java.util.Map)
	 */
	public void setIcons(ValueIconPair[] icons) {
   		ValueIconPair[] old = getIcons();
		valueIcons.clear();
		for (ValueIconPair pair : icons) {
			valueIcons.put(pair.getValue(), pair);
		}
		firePropertyChange(ICONS_PROPERTY, old, getIcons());
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.displayers.IconDisplayer#addIcons(java.util.Map)
	 */
	public void addIcons(ValueIconPair[] icons) {
		ValueIconPair[] old = getIcons();
		for (ValueIconPair pair : icons) {
			valueIcons.put(pair.getValue(), pair);
		}       
		firePropertyChange(ICONS_PROPERTY, old, getIcons());
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.displayers.IconDisplayer#removeIcons(com.cosylab.gui.IconValuepair[])
	 */
	public void removeIcons(ValueIconPair[] icons) {
		ValueIconPair[] old = getIcons();
		for (ValueIconPair pair : icons) {
			valueIcons.remove(pair.getValue());
		}   
		firePropertyChange(ICONS_PROPERTY, old, getIcons());
		
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.displayers.IconDisplayer#removeIcon(com.cosylab.gui.IconValuepair)
	 */
	public void removeIcon(ValueIconPair icon) {
		ValueIconPair[] old = getIcons();
		valueIcons.remove(icon.getValue());
		firePropertyChange(ICONS_PROPERTY, old, getIcons());
	}
	 
	public void addIcon(long value, Icon icon) {
		addIcon(new ValueIconPair(value,icon));
	}

	public void addIcon(ValueIconPair pair) {
		ValueIconPair[] old = getIcons();
		valueIcons.put(pair.getValue(), pair);
		firePropertyChange(ICONS_PROPERTY, old, getIcons());
	}
	
	/**
     * Preforms GUI initialization.
     */
	private void initialize() {

		setLayout(new BorderLayout());
		add(getTitleLabel(), BorderLayout.NORTH);

		
		add(getLabel(), BorderLayout.CENTER);
		
		Dimension size = new Dimension(50, 50);
		setPreferredSize(size);
		setSize(size); 
		setBackground(ColorHelper.getCosyControl());
		getLabel().setIcon(getDefaultIcon());
		valueIcons = new HashMap<Long, ValueIconPair>();
		
		setTitleVisible(false);
	}
	
	@Override
	public void setBackground(Color bg) {
		super.setBackground(bg);
		getLabel().setBackground(bg);
	}
	
	@Override
	public void setForeground(Color fg) {
		super.setForeground(fg);
		getLabel().setForeground(fg);
	}
	
	private JLabel getLabel() {
		if (label == null) {
			label = new JLabel();
			label.setHorizontalAlignment(SwingConstants.CENTER);
			label.setVerticalAlignment(SwingConstants.CENTER);
			label.addComponentListener(new ComponentAdapter() {
			
				@Override
				public void componentResized(ComponentEvent e) {
					updateIcon();
				}
			
			});
		}
		return label;
	}

	/**
     * Sets the value of this displayer.
     * 
     * @see com.cosylab.gui.displayers.LongDisplayer#setValue(long)
     */
	public void setValue(long value) {
		this.value = value;
		updateIcon();
	}
	
	protected void updateIcon() {
		Icon icon = null;
		if (valueIcons.get(value) != null) {
			icon = valueIcons.get(value).getIcon();
		} else {
			icon = getDefaultIcon();
		}
		
		if (icon instanceof ImageIcon && scaleIcons) {
			label.setIcon(IconCustomizer.getScaledIcon((ImageIcon)icon, getHeight(), getWidth()));
		} else {
			label.setIcon(icon);
		}

	}

	/**
     * Returns current value of this displayer.
     * 
     * @see com.cosylab.gui.displayers.LongDisplayer#getValue()
     */
	public long getValue() {
		return value;
	}

	/**
     * Creates and sets default icon from file name. Default icon will be used
     * when no condition is met.
     * 
     * @param filename
     */
	public void setDefaultIconName(String filename) {
		if (filename != null && filename.equals(defaultIconName) || (filename == null && defaultIconName == null)) 
			return;
		String old = defaultIconName;
		setDefaultIcon(IconHelper.createIcon(filename));
		defaultIconName = filename;
		firePropertyChange(DEFAULT_ICON_NAME_PROPERTY, old, defaultIconName);
	}

	/**
     * Sets default icon. Default icon will be used when no condition is met.
     * 
     * @param newIcon
     *        New icon to use. May be null.
     */
	public void setDefaultIcon(Icon newIcon) {
		Icon old = defaultIcon;
		defaultIcon = newIcon;
		updateIcon();
		if (defaultIcon != null)
			defaultIconName = defaultIcon.toString();
		else 
			defaultIconName = null;
		firePropertyChange(DEFAULT_ICON_PROPERTY, old, this.defaultIcon);
	}
	
	public String getDefaultIconName() {
		return defaultIconName;
	}

	/**
     * Returns default icon. Default icon is used when no condition is met.
     * 
     * @return Instance of default icon or null.
     */
	public Icon getDefaultIcon() {
//		if (defaultIcon==null) {
//			defaultIcon= new Icon() {
//			
//				public int getIconHeight() {
//					return 32;
//				}
//			
//				public int getIconWidth() {
//					return 32;
//				}
//			
//				public void paintIcon(Component c, Graphics g, int x, int y) {
//					PaintHelper.paintDisabled(g,4,4,g.getClipBounds().width-8,g.getClipBounds().height-8);
//				}
//			
//			};
//		}
		return defaultIcon;
	}

	/*
     * (non-Javadoc)
     * 
     * @see com.cosylab.gui.displayers.Displayer#isEditable()
     */
	public boolean isEditable() {
		return false;
	}

	/*
     * (non-Javadoc)
     * 
     * @see com.cosylab.gui.displayers.LongConsumer#updateValue(long, long)
     */
	public void updateValue(long timestamp, long value) throws CommonException {
		setValue(value);
	}

	/*
     * (non-Javadoc)
     * 
     * @see com.cosylab.gui.displayers.DataConsumer#setCharacteristics(java.util.Map)
     */
	public void setCharacteristics(Map characteristics) {
		suspend();
		Object o = characteristics.get(CommonDisplayer.C_DISPLAY_NAME);

		if (o != null) {
			try {
				UserSettingsProtection.setUnprotected(this,"title",o.toString());
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		resume();
	}

	/*
     * (non-Javadoc)
     * 
     * @see com.cosylab.gui.displayers.DataConsumer#getSupportedConsumerTypes()
     */
	public Class[] getSupportedConsumerTypes() {
		return LongConsumerMulticaster.SUPPORTED_CONSUMER_TYPES;
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDataConsumer(java.lang.Class)
	 */
	public <D extends DataConsumer> D getDataConsumer(Class<D> type) {
		return (D) LongConsumerMulticaster.createDataConsumer(type, this);
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.displayers.AbstractDisplayerPanel#cleanup()
	 */
	public void cleanup()
	{
		super.cleanup();
		valueIcons.clear();
	}
	
	public InfoDialog getInfoDialog() {
		if (infoDialog == null) {
			infoDialog = new InfoDialog(this);
		}
		return infoDialog;
	}
	
	public PopupManager getPopupManager()
	{
		if (popupManager == null) {
			popupManager = super.getPopupManager();
			popupManager.addAction(new AbstractAction("Info...") {
	
					public void actionPerformed(ActionEvent e)
					{
						getInfoDialog().setVisible(true);
					}
				});
		}
		
		return popupManager;
	}
	
	public static void main(String[] args) {
		final IconDisplayer id = new IconDisplayer();
		ValueIconPair[] icons = new ValueIconPair[]{
		new ValueIconPair(0L, IconHelper.createIcon("icons/navigation/ArrowLeftLeft16.gif")),
		        new ValueIconPair(1L, IconHelper.createIcon("icons/navigation/ArrowLeft16.gif")),
		        new ValueIconPair(2L, IconHelper.createIcon("icons/navigation/ArrowRight16.gif")),
		        new ValueIconPair(3L, IconHelper.createIcon("icons/navigation/ArrowRightRight16.gif"))
		};
		id.setIcons(icons);
		
		Thread th = new Thread(new Runnable(){
			public void run() {
				try {
					int i = 0;

					while (true) {
						id.setValue(i);
						i = (i + 1) % 5;
						Thread.sleep(5000);
					}
				} catch (InterruptedException e) {
				}				
			}
		});
		id.setTitle("Demo");
		RunnerHelper.runComponent(id, 300,300);
		th.start();
	}

	
	/**
	 * Returns the scaleIcons.
	 * @return Returns the scaleIcons.
	 */
	public boolean isScaleIcons() {
		return scaleIcons;
	}

	
	/**
	 * Sets new scaleIcons.
	 * @param scaleIcons The scaleIcons to set.
	 */
	public void setScaleIcons(boolean scaleIcons) {
		boolean old= this.scaleIcons;
		this.scaleIcons = scaleIcons;
		setValue(getValue());
		firePropertyChange("scaleIcons",old,scaleIcons);
	}
}

/* __oOo__ */
