/*
 * Copyright (c) 2004 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui;

import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.beans.Beans;
import java.beans.PropertyVetoException;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.swing.AbstractAction;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.SwingConstants;
import javax.swing.TransferHandler;

import com.cosylab.gui.components.AbstractDisplayerPanel;
import com.cosylab.gui.components.ResizableTextLabel;
import com.cosylab.gui.components.customizer.AbstractCustomizerPanel;
import com.cosylab.gui.components.util.CosyTransferHandler;
import com.cosylab.gui.components.util.CosyUIElements;
import com.cosylab.gui.components.util.PopupManageable;
import com.cosylab.gui.components.util.PopupManager;
import com.cosylab.gui.displayers.DataConsumer;
import com.cosylab.gui.displayers.DataSource;
import com.cosylab.gui.displayers.DataState;
import com.cosylab.gui.displayers.StringConsumerMulticaster;
import com.cosylab.gui.displayers.StringDisplayer;
import com.cosylab.gui.util.UserSettingsProtection;
import com.cosylab.logging.DebugLogger;
import com.cosylab.util.CommonException;


/**
 * A displayer GUI widget that can uniformly display changable string value.
 * Displayer is build primarely as <code>StringConsumer</code> but can act as
 * any other consumer by help of <code>StringConsumerMulticaster</code>
 * utility class. By look is similar to number displayer. User can display
 * arbytrary value by implementing own way to turn it in to String.
 *
 * @author <a href="mailto:igor.verstovsek@cosylab.com">Igor Verstovsek</a>
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 * @version $id$
 */
public class LabelDisplayer extends AbstractDisplayerPanel
	implements StringDisplayer, PopupManageable
{
	private static final long serialVersionUID = 1L;
	private DataSource dataSource;
	private DataState dataState = new DataState(DataState.UNDEFINED);
	private ResizableTextLabel label = null;
	private int suspendCount;
	private String value;
	private final Logger logger = DebugLogger.getLogger("LabelDisplayer",
		    Level.OFF);
	private PopupManager popupManager;
	private AbstractCustomizerPanel customizer;
	private InfoDialog infoDialog;
	
	/**
	 * Constructor for LabelDisplayer.
	 */
	public LabelDisplayer()
	{
		super();
		internalCleanup();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.CommonDisplayer#cleanup()
	 */
	public void cleanup()
	{
		internalCleanup();
		updateDataState(new DataState(DataState.NOT_INITIALIZED));
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.core.CosyComponent#destroy()
	 */
	public void destroy()
	{
		internalCleanup();
	}

	/**
	 * Returns number of character columns. Fixes the length of displayed string to fixed value.  
	 *
	 * @return number of characters in displayed string
	 */
	public int getColumns()
	{
		return getValueLabel().getColumns();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDataConsumer(java.lang.Class)
	 */
	public DataConsumer getDataConsumer(Class type)
	{
		return StringConsumerMulticaster.createDataConsumer(type, this);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.Displayer#getDataSource()
	 */
	public DataSource getDataSource()
	{
		return dataSource;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.CommonDisplayer#getDataState()
	 */
	public DataState getDataState()
	{
		return dataState;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDefaultDataConsumer()
	 */
	public DataConsumer getDefaultDataConsumer()
	{
		return this;
	}

	/**
	 * A label which is used to display value. Extended class can access
	 * directly to customize how label will display value. Note that some
	 * label properties are overriden by default displayer policies when
	 * policy is changed.
	 *
	 * @return the label which displays string value
	 */
	protected ResizableTextLabel getValueLabel()
	{
		if (label == null) {
			label = new ResizableTextLabel();
			label.setHorizontalAlignment(SwingConstants.CENTER);
			label.setEnabled(isEnabled());
			label.setEnhanced(isEnhanced());
			label.setResizable(isResizable());
			internalSetValue();
			label.setOpaque(true);
			label.setBorder(CosyUIElements.getFlushBorder());
			label.setBackground(Color.LIGHT_GRAY);
		}

		return label;
	}
	
	@Override
	public void setBackground(Color bg) {
		Color labelBg = getValueLabel().getBackground();
		super.setBackground(bg);
		getValueLabel().setBackground(labelBg);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedCharacteristics()
	 */
	public String[] getSupportedCharacteristics()
	{
		return new String[]{ C_DISPLAY_NAME };
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedConsumerTypes()
	 */
	public Class[] getSupportedConsumerTypes()
	{
		return StringConsumerMulticaster.SUPPORTED_CONSUMER_TYPES;
	}

	/**
	 * Returns the string value.
	 *
	 * @return the string value
	 */
	public String getValue()
	{
		return value;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.components.DisplayerPanel#getValueComponent()
	 */
	protected JComponent getValueComponent()
	{
		return getValueLabel();
	}

	private void internalCleanup()
	{
		logger.info("Cleanup");

		//setLayoutOrientation(AbstractDisplayerPanel.DYNAMIC_LAYOUT);
		///setResizable(true);
		//setEnhanced(true);
		//setTitleVisible(true);
		//setTitle("Label Displayer");
		//setValue("value");
		setEnhanced(true);
		setResizable(true);
		setValue("value");
		setTitle("Label Displayer");
		setLayoutOrientation(AbstractDisplayerPanel.DYNAMIC_LAYOUT);
		setTitleVisible(true);
		setColumns(5);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.components.DisplayerPanel#internalSetEnabled()
	 */
	protected void internalSetEnabled()
	{
		super.internalSetEnabled();
		getValueLabel().setEnabled(isEnabled());
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.components.DisplayerPanel#internalSetEnhanced()
	 */
	protected void internalSetEnhanced()
	{
		super.internalSetEnhanced();
		getValueLabel().setEnhanced(isEnhanced());
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.components.DisplayerPanel#internalSetResizable()
	 */
	protected void internalSetResizable()
	{
		super.internalSetResizable();
		getValueLabel().setResizable(isResizable());
	}

	private void internalSetValue()
	{
		String tValue = value;

		if (tValue == null && Beans.isDesignTime()) {
			tValue = "<value>";
		}

		if (getValueLabel().isEnabled()) {
			if (logger!=null) logger.fine("'" + tValue + "'");
			getValueLabel().setText(tValue);
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.Displayer#isEditable()
	 */
	public boolean isEditable()
	{
		return false;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.CommonDisplayer#isSuspended()
	 */
	public boolean isSuspended()
	{
		return suspendCount > 0;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.CommonDisplayer#resume()
	 */
	public void resume()
	{
		if (suspendCount > 0) {
			suspendCount--;
		}

		if (suspendCount == 0) {
			setEnabled(true);
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#setCharacteristics(java.util.Map)
	 */
	public void setCharacteristics(Map characteristics)
	{
		Object o = characteristics.get(C_DISPLAY_NAME);

		if (o != null) {
			try {
				UserSettingsProtection.setUnprotected(this,"title",o.toString());
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @param c DOCUMENT ME!
	 */
	public void setColumns(int c)
	{
		getValueLabel().setColumns(c);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.Displayer#setDataSource(com.cosylab.gui.displayers.DataSource)
	 */
	public void setDataSource(DataSource dataSource)
		throws PropertyVetoException
	{
		if (dataSource != null) {
			dataSource.removeConsumer(this);
		}

		if (dataSource != null) {
			dataSource.addConsumer(this);
		}

		DataSource old = this.dataSource;
		this.dataSource = dataSource;

		firePropertyChange(DATA_SOURCE, old, dataSource);
	}

	/**
	 * @see com.cosylab.gui.displayers.StringDisplayer#getValue()
	 */
	public void setValue(String value)
	{
		logger.info("'" + value + "'");

		if (value == null && this.value == null
		    || value != null && value.equals(this.value)) {
			return;
		}

		String oldValue = this.value;
		this.value = value;
		internalSetValue();
		firePropertyChange("value", oldValue, value);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.CommonDisplayer#suspend()
	 */
	public void suspend()
	{
		setEnabled(false);
		suspendCount++;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#updateDataState(com.cosylab.gui.displayers.DataState)
	 */
	public void updateDataState(DataState state)
	{
		DataState old = dataState;
		dataState = state;
		firePropertyChange(DATA_STATE, old, dataState);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.StringConsumer#updateValue(long, java.lang.String)
	 */
	public void updateValue(long timestamp, String value)
		throws CommonException
	{
		logger.info(timestamp + ",'" + value + "'");
		setValue(value);
	}
	
	public PopupManager getPopupManager()
	{
		if (popupManager == null) {
			popupManager = new PopupManager(this, true);
			popupManager.addAction(new AbstractAction("Preferences...") {
				private static final long serialVersionUID = 1L;
				public void actionPerformed(ActionEvent e)
				{
					getCustomizer().showDialog();
				}
			});
			popupManager.addAction(new AbstractAction("Info...") {
				
				public void actionPerformed(ActionEvent e)
				{
					getInfoDialog().setVisible(true);
				}
			});
		}

		return popupManager;
	}
	
	public InfoDialog getInfoDialog() {
		if (infoDialog == null) {
			infoDialog = new InfoDialog(this);
		}
		return infoDialog;
	}
	
	/**
	 * Loads the default
	 * <code>com.cosylab.gui.components.customizer.Customizer</code> for this
	 * displayer and returns it.
	 *
	 * @return the Customizer intance for this displayer
	 */
	public AbstractCustomizerPanel getCustomizer()
	{
		if (customizer == null) {
			customizer = AbstractCustomizerPanel.findCustomizer(this);
		}

		return customizer;
	}
	
	public static void main(String[] args) {
		try {
			JFrame frame= new JFrame();
			frame.addWindowListener(new WindowAdapter() {
				public void windowClosing(WindowEvent e) {
					System.exit(0);
				}
			});
			frame.setSize(200,200);
			
			LabelDisplayer d= new LabelDisplayer();
			
			frame.setContentPane(d);
			
			//frame.show();
			frame.setVisible(true);
			
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
    /* (non-Javadoc)
     * @see javax.swing.JComponent#setTransferHandler(javax.swing.TransferHandler)
     */
    @Override
    public void setTransferHandler(TransferHandler newHandler) {
    	super.setTransferHandler(newHandler);
    	CosyTransferHandler.registerTransferHandler(this,newHandler,new JComponent[]{getValueComponent()});
    }

}

/* __oOo__ */
