/*
 * Copyright (c) 2004 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.adapters;

import java.beans.PropertyVetoException;
import java.util.HashMap;
import java.util.Map;

import com.cosylab.gui.displayers.DataConsumer;
import com.cosylab.gui.displayers.DataSource;
import com.cosylab.gui.displayers.DataState;
import com.cosylab.gui.displayers.NonblockingNumberConsumer;


/**
 * <code>AbstractConverter</code> is implementation of
 * <code>Converter</code> interface intended for convenience when implementing new converter.
 * 
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 * @version $Id: AbstractConverter.java,v 1.4 2007-12-11 13:33:02 jbobnar Exp $
 *
 * @see com.cosylab.gui.adapters.Converter
 */
public abstract class AbstractConverter extends DataConsumerDispatcher
	implements Converter, DataConsumer
{
	// flag indicating if user has called cache characteristics
	private boolean userCache = false;

	// dispatcher which intercepts feedback value from peer data consumer and dispatch it to peer data sources

	/** DOCUMENT ME! */
	transient protected NonblockingNumberDispatcher nonblDispatcher;

	// last dispatched characteristics, used to be redispatched
	// when multiplication factor is changed or new consumer added
	transient private Map lastCharacteristics;

	// last dispatched data state, used to be redispatched when
	// new consumer added
	private DataState lastDataState;

	/**
	 * Creates new converter.
	 */
	public AbstractConverter(Class[] types)
	{
		// we dalagate to super suppport class only the consumer type/types we want to suppport
		super(types);

		// the name of this converter when it acts as a consumer
		name = "AbstractConverter";
	}

	/**
	 * This is called by peer data source. Call is delegated to contained
	 * consumers. In the process it is modified with transformation function.
	 *
	 * @see com.cosylab.gui.displayers.DataConsumer#setCharacteristics(java.util.Map)
	 */
	public void setCharacteristics(Map characteristics)
	{
		if (!userCache) {
			// We store in chache characteristic if implementation does not.
			// We are assuming that user implementation did no changed them.
			lastCharacteristics = new HashMap(characteristics);
		}

		super.setCharacteristics(characteristics);
	}

	/**
	 * Override this method to provide own dispatcher, which forwards revers data flow (from consumer to source).
	 * @return nonblocking reverse dispatcher
	 */
	protected NonblockingNumberDispatcher getNonblDisplatecher()
	{
		if (nonblDispatcher == null) {
			nonblDispatcher = new NonblockingNumberDispatcher();
		}

		return nonblDispatcher;
	}

	/**
	 * In this method we investigate the following cases:
	 * 
	 * <ul>
	 * <li>
	 * if consumer is <code>NonblockingNumberConsumer</code>, the it comes from
	 * peer data source and wants to receive notifications from peer cosnumer
	 * about user change requests.
	 * </li>
	 * <li>
	 * if consumer is peer consumer and perhatps implements
	 * <code>DataSource</code> interface. It it does, we must notify listener
	 * for user data change requests.
	 * </li>
	 * </ul>
	 * 
	 *
	 * @see com.cosylab.gui.displayers.DataSource#addConsumer(com.cosylab.gui.displayers.DataConsumer)
	 */
	public void addConsumer(DataConsumer consumer) throws PropertyVetoException
	{
		if (consumer instanceof NonblockingNumberConsumer) {
			// this one comes from peer data source, to this one we delegate calls from peer data source
			getNonblDisplatecher().addConsumer(consumer);

			return;
		}

		super.addConsumer(consumer);
		consumerAdded(consumer);

		// new consumer did not receive characteristics or state, if converter
		// already have already received them, they must be delegated
		if (lastCharacteristics != null) {
			consumer.setCharacteristics(lastCharacteristics);
		}

		if (lastDataState != null) {
			consumer.updateDataState(lastDataState);
		}

		if (consumer instanceof DataSource) {
			Class[] cl = ((DataSource)consumer).getAcceptableConsumerTypes();

			for (int i = 0; i < cl.length; i++) {
				if (cl[i].equals(NonblockingNumberConsumer.class)) {
					/* here we reqister our nonblocking dispatcher, which dispatches
					 * nonblocking requests to peer data source.
					 */
					((DataSource)consumer).addConsumer(getNonblDisplatecher());

					return;
				}
			}
		}
	}

	/**
	 * This method is called when consumer has been added to this converter. You may
	 * override this method to implement own handling of added consumer.
	 * @param consumer the consumer which was added
	 */
	protected void consumerAdded(DataConsumer consumer) {
		//
	}

	/* (non-Javadoc)
	     * @see com.cosylab.gui.displayers.DataConsumer#updateDataState(com.cosylab.gui.displayers.DataState)
	     */
	public void updateDataState(DataState state)
	{
		lastDataState = state;
		super.updateDataState(state);
	}

	/**
	 * If converter implementation decides to change characteristics it is very
	 * advisable to cache unmodified characteristics. This chached
	 * characteristics are used to initialized new consumers. They can be also
	 * used by implementation to reinitialize consumers if transformation
	 * function has changed.
	 * 
	 * <p>
	 * If implementation does not change characteristic, than this method
	 * doesen't have to be used.
	 * </p>
	 *
	 * @param characteristics the unmodified characcteristics before
	 *        transformation
	 */
	protected void cacheLastCharacteristics(Map characteristics)
	{
		lastCharacteristics = new HashMap(characteristics);
		userCache = true;
	}

	/**
	 * Returnes characcteristics whic was last time used on this data
	 * converter. Use <code>cacheLastCharacteristics(Map)</code> to store
	 * them.
	 *
	 * @return last used characcteristics from internal chache
	 *
	 * @see #cacheLastCharacteristics(Map)
	 */
	protected Map getLastCharacteristics()
	{
		return lastCharacteristics;
	}

	/**
	 * Clears converter to initial state.
	 */
	public void clear()
	{
		getNonblDisplatecher().clear();

		DataConsumer[] c = getConsumers();

		for (int i = 0; i < c.length; i++) {
			if (c[i] instanceof DataSource) {
				((DataSource)c[i]).removeConsumer(getNonblDisplatecher());
			}
		}

		super.clear();
	}
	
	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataSourceSupport#clone()
	 */
	@Override
	public Object clone() throws CloneNotSupportedException {
		AbstractConverter c= (AbstractConverter)super.clone();
		c.lastCharacteristics=null;
		c.lastDataState=null;
		c.nonblDispatcher=null;
		return c;
	}
}

/* __oOo__ */
