/*
 * Copyright (c) 2003 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.adapters;

import java.io.Serializable;

import com.cosylab.gui.displayers.DataConsumer;
import com.cosylab.gui.displayers.DataSource;


/**
 * <code>Converter</code> interface distinguishes objects that wish to act as
 * data converters for data flow betwean some arbitrary
 * <code>DataSource</code> and <code>DataConsumer</code>. To accomplish this
 * they must represent itself as data consumer to peer data source, thus
 * implementing <code>DataConsumer</code> interface, and on the other side
 * accept peer data consumer as data source, thus implementing at the same
 * time <code>DataSource</code> interface. With such approach it is possible
 * to chain multiple <code>Converter</code>s in a single data flow line from
 * source to consumer.
 * 
 * <p>
 * Besides implementing <code>DataSource</code> and <code>DataConsumer</code>
 * interfaces converter must comply to the following contract:
 * </p>
 * 
 * <ul>
 * <li>
 * When implementing method
 * <code>DataCosnumer::addConsumer(DataConsumer)</code> the consumer parameter
 * must be examined for the following:
 * 
 * <ul>
 * <li>
 * If the consumer is <code>NonblockingNumberConsumer</code>, then it comes
 * from peer data source and wants to receive notifications from this
 * converter as data source of feedback requests. See <code>Displayer</code>
 * documentation for details. At the moment only
 * <code>NonblockingNumberConsumer</code> is supported, later will be also
 * other <code>Nonblocking&lt;TYPPE&gt;Consumer</code>,
 * <code>Synchronous&lt;TYPPE&gt;Consumer</code> and
 * <code>Asynchronous&lt;TYPPE&gt;Consumer</code> interfaces from
 * <code>con.cosylab.gui.displayers</code> package.
 * </li>
 * <li>
 * Any other consumer is treated as it would normaly be in data source:
 * </li>
 * <li>
 * consumer is added to the list of consumers which will be notified about data
 * flow events.
 * </li>
 * <li>
 * if consumer implements <code>DataSource</code> interface we must add
 * <code>NonblockingNumberConsumer</code> implementation, which will dispatch
 * value change notifications to set of consumers of same type, the ones that
 * was mentioned above in first bullet.
 * </li>
 * </ul>
 * 
 * </li>
 * </ul>
 * 
 * <p>
 * To clear the things there is simple exemplatory implementation
 * <code>MultiplierConverter</code>. This example is fully functional and
 * performs convertions in which any incoming number is multiplied by some
 * user defined factor. To see how this example converter is used run
 * <code>con.cosylab.showcase.CalculationPanel</code> showcase demo.
 * </p>
 * 
 * <p>
 * The <code>SimpleConverterSupport</code> class could be used as
 * implementation base for most of transformations.
 * </p>
 *
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 * @version $Id: Converter.java,v 1.8 2007-12-10 10:10:30 jbobnar Exp $
 *
 * @see com.cosylab.gui.displayers.Displayer
 * @see com.cosylab.gui.displayers.DataSource
 * @see com.cosylab.gui.displayers.DataConsumer
 * @see com.cosylab.gui.adapters.SimpleConverterSupport
 * @see com.cosylab.gui.adapters.MultiplierConverter
 * @since Feb 14, 2004.
 */
public interface Converter extends DataSource, DataConsumer, Cloneable, Serializable
{
	public Object clone() throws CloneNotSupportedException;
}

/* __oOo__ */
