/*
 * Copyright (c) 2004 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.adapters;

import com.cosylab.gui.displayers.DataConsumer;
import com.cosylab.gui.displayers.DisplayerUtilities;
import com.cosylab.gui.displayers.DoubleConsumer;
import com.cosylab.gui.displayers.DoubleSeqConsumer;
import com.cosylab.util.CommonException;


/**
 * <code>SimpleConverterSupport</code> is implementation of
 * <code>Converter</code> interface intended for simple implementation, which
 * use transfromation function on double data flow.
 * 
 * <p>
 * <b>Note!</b> This implementation works only with data sources and consumers,
 * whcih support <code>double</code> as value type.
 * </p>
 *
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 * @version $Id: SimpleConverterSupport.java,v 1.10 2007-12-10 10:10:30 jbobnar Exp $
 *
 * @see com.cosylab.gui.adapters.Converter
 * @since Feb 14, 2004.
 */
public abstract class SimpleConverterSupport extends AbstractConverter
	implements Converter, DoubleConsumer, DoubleSeqConsumer
{
	public static final String SHORT_NAME = "conv";
	protected int arrayIndex=0;

	
	/**
	 * Returns the arrayIndex.
	 * This index is used to extract value if this converter receives double array update.
	 * @return Returns the arrayIndex.
	 */
	public int getArrayIndex() {
		return arrayIndex;
	}

	
	/**
	 * Sets new arrayIndex. 
	 * This index is used to extract value if this converter receives double array update.
	 * @param arrayIndex The arrayIndex to set.
	 */
	public void setArrayIndex(int arrayIndex) {
		this.arrayIndex = arrayIndex;
	}

	/**
	 * Creates new converter.
	 */
	public SimpleConverterSupport()
	{
		// we dalagate to super suppport class only the consumer type/types we want to suppport
		super(new Class[]{ DoubleConsumer.class });

		// the name of this converter when it acts as a consumer
		name = "SimpleConverter";
	}

	/**
	 * This is called by peer data source. Call is delegated to contained
	 * consumers.
	 *
	 * @see com.cosylab.gui.displayers.DoubleConsumer#updateValue(long, double)
	 */
	public void updateValue(long timestamp, double value)
		throws CommonException
	{

		DataConsumer[] delegates = getConsumers();

		// transform value
		final double newValue = transform(value);

		for (int i = 0; i < delegates.length; i++) {
			try {
				if (delegates[i] instanceof DoubleSeqConsumer) ((DoubleSeqConsumer)delegates[i]).updateValue(timestamp, new double[]{newValue});
				else if (delegates[i] instanceof DoubleConsumer)((DoubleConsumer)delegates[i]).updateValue(timestamp, newValue);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}

	protected NonblockingNumberDispatcher getNonblDisplatecher()
	{
		if (nonblDispatcher == null) {
			nonblDispatcher = new NonblockingNumberDispatcher() {		
				private static final long serialVersionUID = -2136890381501631542L;

						public void updateNonblocking(Number value)
						{
							// we delegate with inverse to multiplication
							super.updateNonblocking(new Double(inverseTransform(
							            value.doubleValue())));
						}
					};
		}

		return nonblDispatcher;
	}

	/**
	 * User must implement this method and from prvided parameter calculate
	 * with own transfromatin function new value and return it.
	 *
	 * @param value the value to be transfirmed
	 *
	 * @return the transformed value
	 */
	abstract protected double transform(double value);

	/**
	 * User must implement this method to make inverse transformation as in
	 * <code>transfirm(double)</code> function. Following must be true
	 * <code>some_value == inverseTransform(transfirm(some_value))</code>.
	 *
	 * @param value the value to be transfirmed with inverse function
	 *
	 * @return the inversed value
	 *
	 * @see #transform(double)
	 */
	abstract protected double inverseTransform(double value);

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DoubleSeqConsumer#updateValue(long, double[])
	 */
	public void updateValue(long timestamp, double[] value) throws CommonException {
		DataConsumer[] delegates = getConsumers();

		for (int i = 0; i < delegates.length; i++) {
			try {
				if (delegates[i] instanceof DoubleSeqConsumer) ((DoubleSeqConsumer)delegates[i]).updateValue(timestamp, transform(value));
				else if (delegates[i] instanceof DoubleConsumer)((DoubleConsumer)delegates[i]).updateValue(timestamp, transform(DisplayerUtilities.extract(arrayIndex,value)));
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}


	private double[] transform(double[] value) {
		double[] d = new double[value.length];
		for (int i = 0; i < value.length; i++) d[i] = transform(value[i]);
		return d;
	}
}

/* __oOo__ */
