/*
 * Copyright (c) 2003 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.components;

import java.util.ArrayList;


/**
 * Convenience implementation of <code>ProgressMonitor</code>. Contains
 * methods for firing events.
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public abstract class AbstractProgressTask extends Thread
	implements ProgressMonitor
{
	private final ArrayList listeners = new ArrayList();

	/**
	 * AbstractProgressTask constructor comment.
	 */
	public AbstractProgressTask()
	{
		super();
	}

	/**
	 * Adds listener. Listeners will be notified when task reports progress.
	 *
	 * @param listener to add.
	 */
	public void addProgressListener(ProgressListener listener)
	{
		listeners.add(listener);
	}

	/**
	 * Notifies listeners that current progress has changed. There is no
	 * prescribed frequency with which these events should be fired.
	 *
	 * @param status Description of current status.
	 * @param current progress, should be less than total
	 * @param total ammount.
	 */
	protected void fireProgressEvent(String status, int current, int total)
	{
		yield();

		ProgressEvent e = new ProgressEvent(this, status, current, total);

		for (int i = 0; i < listeners.size(); i++) {
			((ProgressListener)listeners.get(i)).progress(e);
		}

		yield();
	}

	/**
	 * Notifies listeners that a task has been completed.
	 */
	protected void fireTaskComplete()
	{
		yield();

		ProgressEvent e = new ProgressEvent(this, getStatus(), getCurrent(),
			    getTotal());

		for (int i = 0; i < listeners.size(); i++) {
			((ProgressListener)listeners.get(i)).taskComplete(e);
		}

		yield();
	}

	/**
	 * Notifies listeners that a task has been interrupted during execution.
	 *
	 * @param reason Description of reason for interruption.
	 */
	protected void fireTaskInterrupted(String reason)
	{
		yield();

		ProgressEvent e = new ProgressEvent(this, reason, 0, -1);

		for (int i = 0; i < listeners.size(); i++) {
			((ProgressListener)listeners.get(i)).taskInterruped(e);
		}

		yield();
	}

	/**
	 * Notify listeners that task has started execution.
	 */
	protected void fireTaskStarted()
	{
		yield();

		ProgressEvent e = new ProgressEvent(this, getStatus(), getCurrent(),
			    getTotal());

		for (int i = 0; i < listeners.size(); i++) {
			((ProgressListener)listeners.get(i)).taskStarted(e);
		}

		yield();
	}

	/**
	 * Force listeners to update immediately.
	 */
	protected void immediateUpdate()
	{
		fireProgressEvent(getStatus(), getCurrent(), getTotal());
	}

	/**
	 * Removes listener.
	 *
	 * @param listener to remove.
	 */
	public void removeProgressListener(ProgressListener listener)
	{
		listeners.remove(listener);
	}
}

/* __oOo__ */
