/*
 * Copyright (c) 2004 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.components;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.BufferedImage;

import javax.swing.AbstractAction;
import javax.swing.JPanel;

import com.cosylab.application.state.State;
import com.cosylab.application.state.StateFactory;
import com.cosylab.application.state.StateOriginator;
import com.cosylab.gui.components.customizer.AbstractCustomizerPanel;
import com.cosylab.gui.components.gauger.Needle;
import com.cosylab.gui.components.gauger.ScaleTransform;
import com.cosylab.gui.components.gauger.ScaleTransformFactory;
import com.cosylab.gui.components.gauger.ScaleTransformRenderer;
import com.cosylab.gui.components.gauger.ValueLabel;
import com.cosylab.gui.components.range2.LinearRange;
import com.cosylab.gui.components.range2.LogarithmicRange;
import com.cosylab.gui.components.range2.RangedValueController;
import com.cosylab.gui.components.range2.RangedValue;
import com.cosylab.gui.components.range2.RangedValueEvent;
import com.cosylab.gui.components.range2.RangedValueListener;
import com.cosylab.gui.components.range2.RangedValuePolicy;
import com.cosylab.gui.components.range2.RescalingValuePolicy;
import com.cosylab.gui.components.range2.ShiftValuePolicy;
import com.cosylab.gui.components.range2.Tick;
import com.cosylab.gui.components.range2.TickParameters;
import com.cosylab.gui.components.range2.TrimValuePolicy;
import com.cosylab.gui.components.util.ColorHelper;
import com.cosylab.gui.components.util.PaintHelper;
import com.cosylab.gui.components.util.PopupManageable;
import com.cosylab.gui.components.util.PopupManager;
import com.cosylab.gui.components.util.ScreenCapturer;
import com.cosylab.util.PrintfFormat;


/**
 * Gauger is a visual component that displays a value on a scale between
 * minimum and maximum. Shape and visual layout of the component will change
 * to best fit into the available space. Behaviour of the scale can be
 * adjusted to fixed (lower and upper bound cannot change), constant span
 * (scale will always span fixed ammount) or rescaling (lower and upper bound
 * will adjust to accomodate current value. Supported scales are linear and
 * logarithmic.
 * 
 * <p>
 * For accurate readouts, the currently set value is also displayed in custom
 * formatted label.
 * </p>
 * 
 * <p>
 * Gauger is aimed to provide fastest possible response to value changes.
 * </p>
 *
 * @author Ales Pucelj
 * @version $id$
 */
public class Gauger extends JPanel implements TickParameters, PopupManageable, /* implements TickCallback - edit apikus */
	StateOriginator
{
	private static final long serialVersionUID = 1L;
	
	private AbstractCustomizerPanel customizer;
	
	private PopupManager popupManager;
	
	/**
	 * Gauger preferences
	 */
	private Color scaleColor = new Color(239, 239, 239);
	private Color needleColor = new Color(0, 192, 0);
	private Color alarmColor = new Color(239,0,0);
	private Color warningColor = new Color(0xEEEC300);
	private Color outOfBoundsColor = new Color(239,0,0);

	private Font tickTextFont = new Font("Serif", Font.PLAIN, 12);
	private Font valueLabelFont = new Font("dialog", Font.PLAIN, 12);
	
	/* the size of range values for scale limit */

	private double lowAlarmLimit = 0.05;
	private double lowWarningLimit= 0.2;
	private double highWarningLimit= 0.2;
	private double highAlarmLimit= 0.05;
	private PrintfFormat formatter = new PrintfFormat("%3.2f");

	/**
	 * Fixed limits scale.
	 *
	 * @see #setScaleMode(int)
	 */
	public static final int FIXED_SCALE = 0;

	/**
	 * Constant span scale.
	 *
	 * @see #setScaleMode(int)
	 */
	public static final int CONSTANT_SPAN = 1;

	/**
	 * Resizable scale.
	 *
	 * @see #setScaleMode(int)
	 */
	public static final int STRETCH_SCALE = 2;
	protected RangedValueController userRangedValue; /* edit apikus */
	private BufferedImage buffer = null;
	private ScaleTransformRenderer scaleRenderer = null;
	private ScaleTransform transform = null;
	protected Needle needle = null;
	private ValueLabel valueLabel = null;
	private double labelValue;
	private Tick[] ticks = null;
	private String title = "";
	private boolean titleVisible = false;
	private int titleMinimumFontSize = 2;
	private int titleMaximumFontSize = 20;
	private boolean resizable = true;
	
	private class ComponentListener extends ComponentAdapter
	{
		/**
		 * Clears background and tick buffer whenever the component is resized.
		 * Buffer will be regenerated on next call to repaint.
		 *
		 * @param e ComponentEvent
		 *
		 * @see java.awt.event.ComponentListener#componentResized(ComponentEvent)
		 */
		public void componentResized(ComponentEvent e)
		{
			if (e.getSource() == Gauger.this) {
				synchronized (Gauger.this) {
					transform = null;
					clearBuffer();
					repaint();
				}
			}
		}

		/**
		 * Disposes the buffer whenever the component is hidded to minimize
		 * memory use whenever the component is not visible.
		 *
		 * @param e ComponentEvent
		 *
		 * @see ComponentListener#componentHidden(ComponentEvent)
		 */
		public void componentHidden(ComponentEvent e)
		{
			super.componentHidden(e);
			clearBuffer();
		}
	}

	private class ValueListener implements RangedValueListener
	{
		/**
		 * Responds to value changes by either repainint the background or only
		 * updating the needle position.
		 *
		 * @see com.cosylab.gui.components.range.RangedValueListener#valueChanged(RangedValueEvent)
		 */
		public void valueChanged(RangedValueEvent event)
		{
			
			synchronized (Gauger.this) {
				if (event.isMinimumChanged() || event.isMaximumChanged()) {
					clearBuffer();
					repaint();
				}

			
				Needle needle = getNeedle();
				Rectangle oldArea = needle.getBounds();


				
				if(getValue()>getMaximum()) {
					needle.setPosition(userRangedValue.getRelativeValue()+0.05); /* edit apikus */
				} else if (getValue()<getMinimum()){
					needle.setPosition(userRangedValue.getRelativeValue()-0.05);
				} else {
					needle.setPosition(userRangedValue.getRelativeValue());
				}
			
				if (needle.isChangeSignificant()) {
					repaint(oldArea);
					repaint(needle.getBounds());
				}

				ValueLabel vl = getValueLabel();
				oldArea = vl.getBounds();

				if (vl.isChangeSignificant()) {
					repaint(oldArea);
					repaint(vl.getBounds());
				}
			}
		}
	}

	/**
	 * Creates a new gauger and sets the default parameters.
	 */
	public Gauger()
	{
		super();

		userRangedValue = new RangedValueController(); /* edit apikus */
		userRangedValue.setRange(new LinearRange());
		userRangedValue.addPolicy(new RescalingValuePolicy());
		userRangedValue.addRangedValueListener(new ValueListener());
		
		//default is fixed_scale
		setScaleMode(Gauger.FIXED_SCALE);
		setBackground(ColorHelper.getCosyControl());
		scaleRenderer = new ScaleTransformRenderer(this);

		setDoubleBuffered(false);
		setOpaque(true);
		setDropTarget(null);

		addComponentListener(new ComponentListener());
	}

	/**
	 * Creates a new buffer on which the scale background is rendered.  Buffer
	 * should always be accessed via this method, since the buffer can  be
	 * disposed if the component is not visible.
	 *
	 * @return BufferedImage
	 */
	private synchronized BufferedImage getBuffer()
	{
		if (buffer == null) {
			int w = Math.max(getWidth(), 1);
			int h = Math.max(getHeight(), 1);

			buffer = new BufferedImage(w, h, BufferedImage.TYPE_4BYTE_ABGR);

			Graphics2D g2 = buffer.createGraphics();
			g2.setRenderingHints(PaintHelper.getAntialiasingHints());

			g2.setColor(getBackground());
			g2.fillRect(0, 0, w, h);
			scaleRenderer.render(getTransform(), g2);
		}

		return buffer;
	}

	/**
	 * Sets behaviour of the scale. This can take one of 3 possible modes.<br>
	 * 
	 * <ul>
	 * <li>
	 * Gauger.FIXED_SCALE
	 * </li>
	 * <li>
	 * Gauger.CONSTANT_SPAN
	 * </li>
	 * <li>
	 * Gauger.STRETCH_SCALE
	 * </li>
	 * </ul>
	 * 
	 * <p>
	 * FIXED_SCALE will fix the limits to any value regardless of the actual
	 * value displayed by needle. If value displayed is outside limits, it
	 * will be limited to either bound.
	 * </p>
	 * 
	 * <p>
	 * CONSTANT_SPAN will automatically move the scale based on needle
	 * movement. If needle moves outside current limits, they will move in the
	 * same  direction, but span (maximum-minimum) of the scale will remain
	 * unchanged.
	 * </p>
	 * 
	 * <p>
	 * STRETCH_SCALE will change only one limit at the time whenever needle
	 * moves outside current limits. Scale will not shrink back when needle is
	 * within old bounds.
	 * </p>
	 *
	 * @param mode New mode to set.
	 */
	public void setScaleMode(int mode)
	{
		switch (mode) {
		case FIXED_SCALE:
			getRangedValue().setPolicy(new TrimValuePolicy());
			getRangedValue().setPolicy(new TrimValuePolicy());

			break;

		case CONSTANT_SPAN:
			getRangedValue().setPolicy(new ShiftValuePolicy());

			break;

		case STRETCH_SCALE:default:
			getRangedValue().setPolicy(new RescalingValuePolicy());

			break;
		}
	}

	/**
	 * Returns current scale mode. Return value will be one of <code>
	 * Gauger.FIXED_SCALE</code>, <code>Gauger.CONSTANT_SPAN</code> or <code>
	 * Gauger.STRETCH_SCALE</code>.
	 *
	 * @return One of scale mode constants.
	 *
	 * @see #setScaleMode(int)
	 */
	public int getScaleMode()
	{
		RangedValuePolicy p = getRangedValue().getPolicy();

		if (p instanceof TrimValuePolicy) {
			return FIXED_SCALE;
		}

		if (p instanceof ShiftValuePolicy) {
			return CONSTANT_SPAN;
		}

		return STRETCH_SCALE;
	}

	/**
	 * Clears the background buffer and precalculated ticks.
	 */
	private synchronized void clearBuffer()
	{
		ticks = null;
		buffer = null;
		transform = null;
	}

	/**
	 * Returns current maximum value of the scale.
	 *
	 * @return double
	 */
	public double getMaximum()
	{
		return userRangedValue.getMaximum(); /* edit apikus */
	}

	/**
	 * Returns current minimum value of the scale.
	 *
	 * @return double
	 */
	public double getMinimum()
	{
		return userRangedValue.getMinimum(); /* edit apikus */
	}

	/**
	 * Returns needle renderer. If needle does not exist, it will be created.
	 * This method uses lazy initialization, so setting the needle field to
	 * null will cause it to be recreated on next call.
	 *
	 * @return Needle needle renderer implementation.
	 */
	protected synchronized Needle getNeedle()
	{
		if (needle == null) {
			needle = new Needle(this);
			needle.setPosition(0.5);
		}

		return needle;
	}

	/**
	 * Return <code>RangedValue</code> implementation. This method can be used
	 * to gain access to more advanced aspects of <code>RangedValue</code>
	 * interface. Instance returned is not replacable.
	 *
	 * @return Instance of <code>RangedValue</code>
	 */
	public RangedValueController getRangedValue()
	{
		return userRangedValue;
	}

	/**
	 * Returns the current value of the scale. This is the value displayed by
	 * the needle. Regardles of the scale, this value is never log based.
	 *
	 * @return double
	 */
	public double getValue()
	{
		return labelValue;
	}

	/**
	 * Gets the instance of the value label used to display current value.
	 *
	 * @return ValueLabel
	 */
	protected ValueLabel getValueLabel()
	{
		if (valueLabel == null) {
			valueLabel = new ValueLabel(this);
		}

		return valueLabel;
	}

	/**
	 * Sets format string of value label.
	 *
	 * @param format C-Style format of value label.
	 */
	public void setFormat(String format)
	{
		String old = getValueLabel().getFormat();

		try {
			getValueLabel().setFormat(format);
		} catch (Exception e) {
			e.printStackTrace();
		}
		formatter = new PrintfFormat(format);
		repaint();

		firePropertyChange("format", old, format);
	}

	/**
	 * Returns format of value label.
	 *
	 * @return C-Style format of value label.
	 */
	public String getFormat()
	{
		return getValueLabel().getFormat();
	}

	/**
	 * Sets units of value label.
	 *
	 * @param units String to be appended to value label. Can be null.
	 */
	public void setUnits(String units)
	{
		getValueLabel().setUnits(units);
		repaint();
	}

	/**
	 * Returns units string of value label.
	 *
	 * @return String representing the units. Can be null.
	 */
	public String getUnits()
	{
		return getValueLabel().getUnits();
	}

	/**
	 * Sets units visible of value label.
	 *
	 * @param b whether units should be displayed.
	 */
	public void setUnitsVisible(boolean b)
	{
		getValueLabel().setUnitsVisible(b);
		repaint();
	}

	/**
	 * Returns unitsVisible of value label.
	 *
	 * @return true if units are visible
	 */
	public boolean isUnitsVisible()
	{
		return getValueLabel().isUnitsVisible();
	}


	/**
	 * Sets the new value to be displayed on this scale. In this implementation
	 * the value is shown by the needle and value label. Method also ensures
	 * that the value is in the valid range as specified by minimum and
	 * maximum. If this is not true, minimum and maximum will be adjusted
	 * according to the rescaleMode parameter. After performing all these
	 * operations the scale  is redrawn accordingly.
	 *
	 * @param newValue double
	 */
	public void setValue(double newValue)
	{
		double old = labelValue;
		labelValue = newValue;
		getValueLabel().setValue(labelValue);

		try {
			userRangedValue.setValue(newValue); /* edit apikus */
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		repaint();

		firePropertyChange("value", new Double(old), new Double(newValue));
	}

	/**
	 * Returns the information about the scale ticks. This method is used
	 * internaly by renderers
	 *
	 * @return Tick[]
	 */
	public Tick[] getTicks()
	{
		if (ticks == null) {
			//ticks = rangedValue.getRange().getTicks((int)getTransform().scaleWidth(0.7)).toTicks(this);
			ticks = userRangedValue.calculateTicks((int)getTransform().scaleWidth(0.7), this); /* edit apikus */
		}
		return ticks;
	}

	/**
	 * Returns current transform used to render the scale. This transform is
	 * selected based on width and height of the component and is used when
	 * rendering Gauger.
	 *
	 * @return ScaleTransform
	 */
	public ScaleTransform getTransform()
	{
		if (transform == null) {
			transform = ScaleTransformFactory.createTransform(getWidth(),
				    getHeight());

			transform.setParameters(getWidth(), getHeight(), 0,
			    getValueLabel().getBounds().height, Tick.TICKS_OFFSET_PIXELS);

			getNeedle().setTransform(transform);
			setValueLabelPosition(getTransform().getLabelPosition());
		}

		return transform;
	}

	/**
	 * Sets the lower limit for the display. This value cannot be set above
	 * current maximum, it can be however set above current value.
	 *
	 * @param newMin
	 */
	public void setMinimum(double newMin)
	{
		userRangedValue.setValue(newMin, getMaximum(), getValue());
		repaint();
	}

	/**
	 * Sets the upper limit for the display. This value cannot be set below
	 * current minimum, it can be however set below current value.
	 *
	 * @param newMax
	 */
	public void setMaximum(double newMax)
	{
		userRangedValue.setValue(getMinimum(), newMax, getValue());
		repaint();
	}

	/**
	 * Sets the limits for the displayer.
	 *
	 * @param newMin New minimum.
	 * @param newMax New maximum.
	 * @param newVal New value.
	 */
	public void setValue(double newMin, double newMax, double newVal)
	{
		double old = labelValue;
		labelValue = newVal;
		getValueLabel().setValue(newVal);
		userRangedValue.setValue(newMin, newMax, newVal); /* edit apikus */
		firePropertyChange("value", new Double(old), new Double(newVal));
	}

	/**
	 * Sets position of value label.
	 *
	 * @param p Center point of value label.
	 */
	public void setValueLabelPosition(Point p)
	{
		getValueLabel().setPosition(p.x, p.y);
	}

	/**
	 * Paints the component. This method only paints the gauger from background
	 * buffer, which is initialized elsewhere. If neccessary, active elements
	 * such as needle and value label are updated as well.
	 *
	 * @param g Context to paint to.
	 */
	protected void paintComponent(Graphics g)
	{
		clearBuffer();
		//System.out.println(" SIZE "+getSize());
		//System.out.println("GSIZE "+g.getClipBounds());
		try {
			//if (getTransform() != null) {
				Graphics2D g2 = (Graphics2D)g;
				
				// Rectangle clip = g2.getClipBounds();
				// Image image = getBuffer().getSubimage(clip.x, clip.y, clip.width, clip.height);
				g.drawImage(getBuffer(), 0, 0, null);
				g2.addRenderingHints(PaintHelper.getAntialiasingHints());
				getNeedle().setTransform(getTransform());
				getNeedle().render(g2);
				g2.setColor(getForeground());
				getValueLabel().draw(g2);
				
			//}
		} catch (Exception e) {
			// Weird synhronization error has been noticed to occur in this
			// method.
		}
	}

	/**
	 * Sets scale to logarithmic.
	 */
	public void setLogarithmicScale()
	{
		userRangedValue.setRange(new LogarithmicRange()); /* edit apikus */
	}

	/**
	 * Sets scale to linear.
	 */
	public void setLinearScale()
	{
		userRangedValue.setRange(new LinearRange()); /* edit apikus */
	}

	/**
	 * Returns whether the scale is logarithmic.
	 *
	 * @return True if scale is logarithmic.
	 */
	public boolean isLogarithmicScale()
	{
		return userRangedValue.getRange() instanceof LogarithmicRange; /* edit apikus */
	}

	/**
	 * Returns whether the scale is linear.
	 *
	 * @return True if scale is linear.
	 */
	public boolean isLinearScale()
	{
		return userRangedValue.getRange() instanceof LinearRange; /* edit apikus */
	}

	/*
	 *
	 *
	 * @param position PropertyManager
	 * @param text
	 *
	 * @return PropertyGroup[]
	 *
	 * @see PropertyProvider#getProperties(PropertyManager)     public
	 *      PropertyGroup[] getProperties(PropertyManager manager) {
	 *      PropertyGroup limits = new DefaultPropertyGroup("Display limits",
	 *      0);      JComponent pe;      //        manager.setTitle("Settings
	 *      for PBEND_I.01.Readback");      pe =
	 *      manager.getEditorFactory().getEditorForClass( "Lower limit",
	 *      double.class); manager.addAction(    (PropertyEditor) pe,     new
	 *      MethodInvocationAction(this, "getMinimum", "setMinimum"));
	 *      limits.addProperty(pe);      pe =
	 *      manager.getEditorFactory().getEditorForClass( "Upper limit",
	 *      double.class); manager.addAction(    (PropertyEditor) pe,     new
	 *      MethodInvocationAction(this, "getMaximum", "setMaximum"));
	 *      limits.addProperty(pe);      pe =
	 *      manager.getEditorFactory().getEditorForClass( "Allow dynamic limit
	 *      change", boolean.class); manager.addAction(    (PropertyEditor)
	 *      pe,     new MethodInvocationAction( rangedValue,
	 *      "getRescalingAllowed", "setRescalingAllowed")); EnabledAction ea =
	 *      new EnabledAction();      JComponent pe2 =
	 *      manager.getEditorFactory().getEditorForClass( "Dynamic change
	 *      ammount", double.class); manager.addAction(    (PropertyEditor)
	 *      pe2,     new MethodInvocationAction( rangedValue,
	 *      "getRescaleAmmount", "setRescaleAmmount")); ea.bind(pe2);
	 *      ea.attachAction((PropertyEditor) pe);      limits.addProperty(pe);
	 *      limits.addProperty(pe2);      //
	 *      ((LimitedRangeModel)getModel()).getRes;      PropertyGroup vlGroup
	 *      = new DefaultPropertyGroup("Value label", 1);      pe =
	 *      manager.getEditorFactory().getEditorForClass( "Value format",
	 *      String.class); manager.addAction(    (PropertyEditor) pe,     new
	 *      MethodInvocationAction( getValueLabel(), "getFormat",
	 *      "setFormat")); vlGroup.addProperty(pe);      pe =
	 *      manager.getEditorFactory().getEditorForClass( "Value font size",
	 *      String.class); vlGroup.addProperty(pe);      PropertyGroup[]
	 *      returnValue = { limits, vlGroup };      return returnValue;     }
	 */
	/*
	 * public String getProviderDescription() {
	 *         return "Gauger";
	 * }
	 */

	/**
	 * Returns size of tick label when renderer. Return value is ammount in
	 * pixels. This is the size of label in the direction along the scale.
	 *
	 * @see com.cosylab.gui.components.range.TickCallback#measureTick(double,
	 *      String)
	 */
	public int measureTick(double position, String text)
	{
		return getTransform().measureTick(getBuffer().getGraphics(), position,
		    text);

		//		int w = getBuffer().getGraphics().getFontMetrics().stringWidth(text);
		//		int x = (int)(getTransform().scaleWidth(0.7) * position);
		//		return new TickSize(x - w / 2, w);
	}

	/**
	 * Returns popup manager for adding popup actions.
	 *
	 * @see com.cosylab.gui.components.util.PopupManageable#getPopupManager()
	 */
	public PopupManager getPopupManager()
	{
		if (popupManager == null) {
			popupManager = new PopupManager(this, true);
			popupManager.addAction(new AbstractAction("Preferences...") {
				private static final long serialVersionUID = 1L;
					public void actionPerformed(ActionEvent e)
					{
						getCustomizer().showDialog();
					}
				});
			popupManager.addAction(new AbstractAction("Capture screen...") {
				private static final long serialVersionUID = 1L;
				public void actionPerformed(ActionEvent e)
				{
					ScreenCapturer sc = new ScreenCapturer(Gauger.this);
					sc.showScreenDialog();
				
				}
			});
		}

		return popupManager;
	}

	/**
	 * Returns the Customizer intance for this displayer.
	 *
	 * @return the Customizer intance for this displayer
	 */
	public AbstractCustomizerPanel getCustomizer()
	{
		if (customizer == null) {
			customizer = AbstractCustomizerPanel.findCustomizer(this);
		}
		return customizer;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.StateOriginator#getState()
	 */
	public State getState()
	{
		State s = StateFactory.createState();
		s.putDouble("Min", getMinimum());
		s.putDouble("Max", getMaximum());
		s.putString("Format", getFormat());
		s.putString("Units", getUnits());
		s.putBoolean("LogarithmicScale", isLogarithmicScale());
		s.putBoolean("LinearScale", isLinearScale());
		s.putInt("ScaleMode", getScaleMode());
		s.putFont("Font", getValueLabelFont());

		return s;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.StateOriginator#setState(com.cosylab.application.state.State)
	 */
	public void setState(State state)
	{
		setMinimum(state.getDouble("Min", getMinimum()));
		setMaximum(state.getDouble("Max", getMaximum()));
		setFormat(state.getString("Format", getFormat()));
		setUnits(state.getString("Units", getUnits()));

		if (state.getBoolean("LogarithmicScale", false)) {
			setLogarithmicScale();
		}

		if (state.getBoolean("LinearScale", false)) {
			setLinearScale();
		}

		setScaleMode(state.getInt("ScaleMode", getScaleMode()));
		setValueLabelFont(state.getFont("Font"));
	}
	
	
//	/**
//	 * Returns background color of Gauger. 
//	 *
//	 * @return Background color.
//	 */
//	public Color getBackgroundColor()
//	{
//		return backgroundColor;
//	}
//
//	/**
//	 * Sets the background color.
//	 *
//	 * @param color New background color.
//	 */
//	public void setBackgroundColor(Color color)
//	{
//		backgroundColor = color;
//	}

	/**
	 * Sets the font used to display tick labels.
	 *
	 * @param tickText Font for tick labels.
	 */
	public void setTickTextFont(Font tickText)
	{
		this.tickTextFont = tickText;
	}

	/**
	 * Returns font used to display tick labels.
	 *
	 * @return Current font for ticks.
	 */
	public Font getTickTextFont()
	{
		return tickTextFont;
	}

	/**
	 * Returns color of scale background.
	 *
	 * @return Color of scale background.
	 */
	public Color getScaleColor()
	{
		return scaleColor;
	}

	/**
	 * Sets color of scale background.
	 *
	 * @param newScaleColor Color of scale background.
	 */
	public void setScaleColor(Color newScaleColor)
	{
		scaleColor = newScaleColor;
	}

	/**
	 * Returns color of needle.
	 *
	 * @return Color of needle.
	 */
	public Color getNeedleColor()
	{
		return needleColor;
	}
	

	/**
	 * Sets color of needle.
	 *
	 * @param newNeedleColor New color of needle.
	 */
	public void setNeedleColor(Color newNeedleColor)
	{
		needleColor = newNeedleColor;
	}

	/**
	 * Returns font for value label.
	 *
	 * @return Font for value label.
	 */
	public Font getValueLabelFont()
	{
		
		return valueLabelFont;
	}

	/**
	 * Sets font for value label.
	 *
	 * @param value New font for value label.
	 */
	public void setValueLabelFont(Font value)
	{
		valueLabelFont = value;
	}
	
	public void setLowWarningLimit(double v){
		lowWarningLimit = v;
	}
	
	public double getLowWarningLimit(){
		return lowWarningLimit;
	}
	
	public void setLowAlarmLimit(double v){
		lowAlarmLimit = v;
	}
	
	public double getLowAlarmLimit(){
		return lowAlarmLimit;
	}
	
	public void setHighWarningLimit(double v){
		highWarningLimit = v;
	}
	
	public double getHighWarningLimit(){
		return highWarningLimit;
	}
	
	public void setHighAlarmLimit(double v){
		highAlarmLimit = v;
	}
	
	public double getHighAlarmLimit(){
		return highAlarmLimit;
	}
	
	/**
	 * Simple test method for gauger.
	 *
	 * @param args String[]
	 */
	public static void main(String[] args)
	{
		Thread t = new Thread() {
				public void run()
				{
					javax.swing.JFrame frame = new javax.swing.JFrame();
					frame.addWindowListener(new WindowAdapter() {
							public void windowClosing(WindowEvent e)
							{
								System.exit(0);
							}
						});

					frame.setSize(new java.awt.Dimension(150, 120));

					Gauger sp = new Gauger();

					sp.setScaleMode(Gauger.FIXED_SCALE);
					sp.setValue(-30.0, 30.0, 0);
					sp.setUnits("A");
					sp.setFormat("%4.4f");
					frame.setContentPane(sp);
					//frame.show();
					frame.setVisible(true);

					//double x = 5;
					int i = 1;

					//long time1 = System.currentTimeMillis();
					//long time2 = 0;
					//boolean b= false;
					while (true) {
						//time2 = System.currentTimeMillis() - time1;
						//x = 2.0
						//	+ NoiseMaker.perlinNoise((double)time2 / 100.0 * Math.PI / 180);
						sp.setValue(Math.random() * 80 -40 );
					
						try {
							i++;

							if (i > 50) {
								//	time2 = System.currentTimeMillis();
								//	delta = time2 - time1;
								//	time1 = time2;
								//	System.out.println(i * 1000 / delta + " FPS");
								i = 0;
							}

							sleep(500);
						} catch (Exception e) {
							e.printStackTrace();
						}
					}
				}
			};

		t.run();
	}

	/**
	 * @return Returns the alarmColor.
	 */
	public Color getAlarmColor() {
		return alarmColor;
	}

	/**
	 * @param alarmColor The alarmColor to set.
	 */
	public void setAlarmColor(Color alarmColor) {
		this.alarmColor = alarmColor;
	}

	/**
	 * @return Returns the outOfBoundsColor.
	 */
	public Color getOutOfBoundsColor() {
		return outOfBoundsColor;
	}

	/**
	 * @param outOfBoundsColor The outOfBoundsColor to set.
	 */
	public void setOutOfBoundsColor(Color outOfBoundsColor) {
		this.outOfBoundsColor = outOfBoundsColor;
	}

	/**
	 * @return Returns the warningColor.
	 */
	public Color getWarningColor() {
		return warningColor;
	}

	/**
	 * @param warningColor The warningColor to set.
	 */
	public void setWarningColor(Color warningColor) {
		this.warningColor = warningColor;
	}

	public String getTitle() {
		return title;
	}

	public void setTitle(String title) {
		this.title = title;
	}

	public boolean isTitleVisible() {
		return titleVisible;
	}

	public void setTitleVisible(boolean titleVisible) {
		this.titleVisible = titleVisible;
	}

	public boolean isResizable() {
		return resizable;
	}

	public void setResizable(boolean resizable) {
		this.resizable = resizable;
	}

	public int getTitleMaximumFontSize() {
		return titleMaximumFontSize;
	}

	public void setTitleMaximumFontSize(int titleMaximumFontSize) {
		this.titleMaximumFontSize = titleMaximumFontSize;
	}

	public int getTitleMinimumFontSize() {
		return titleMinimumFontSize;
	}

	public void setTitleMinimumFontSize(int titleMinimumFontSize) {
		this.titleMinimumFontSize = titleMinimumFontSize;
	}
	
	public String formatNumber(double x) {
		return formatter.sprintf(x);
	}
}

/* __oOo__ */
