/*
 * Copyright (c) 2003 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.components;

import java.awt.Container;
import java.awt.Window;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;


/**
 * InterceptorPane is specialized GUI component that installs itself into any
 * descendant of the <code>java.awt.Window</code>. It acts as a mouse click
 * interceptor whenever the window is not active. This is a safety mechanism
 * to prevent accidental clicks to be performed when the user only wants to
 * bring  the window into focus.
 * 
 * <p>
 * To install InterceptorPane onto a JFrame or JDialog, simply create a new
 * instance using the constructor:<br>
 * <code> JFrame frame; new InterceptorFrame(frame); </code>
 * </p>
 * 
 * <p>
 * Creation date: (4/8/2002 11:51:40)
 * </p>
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public final class InterceptorPane extends JComponent
{
	// Window instance interceptor is installed on
	protected Window window = null;

	// Flag indicating whether to ignore windows owned by this window
	protected boolean ignoreOwned = false;

	// Mouse click interceptor
	private ClickListener clickListener = null;

	// Monitoring of changes in focus
	private FocusMonitor focusMonitor = null;

	/**
	 * DOCUMENT ME!
	 *
	 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
	 * @version $id$
	 */
	protected final class ClickListener extends MouseAdapter
	{
		/**
		 * DOCUMENT ME!
		 *
		 * @param e MouseEvent
		 */
		private final void ignoreEvent(MouseEvent e)
		{
			e.consume();
			setVisible(false);
		}

		/**
		 * DOCUMENT ME!
		 *
		 * @param e MouseEvent
		 *
		 * @see java.awt.event.MouseListener#mouseClicked(MouseEvent)
		 */
		public final void mouseClicked(MouseEvent e)
		{
			ignoreEvent(e);
		}

		/**
		 * DOCUMENT ME!
		 *
		 * @param e MouseEvent
		 *
		 * @see java.awt.event.MouseListener#mousePressed(MouseEvent)
		 */
		public final void mousePressed(MouseEvent e)
		{
			ignoreEvent(e);
		}
	}

	/**
	 * This listener monitors changes in focus. When a window with an
	 * interceptor pane is deactivated, its interceptor pane is activated, and
	 * intercepts the mouse click in its area.
	 *
	 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
	 * @version $id$
	 */
	protected final class FocusMonitor extends WindowAdapter
	{
		/**
		 * Installs the glass pane only if the
		 *
		 * @param e WindowEvent
		 *
		 * @see java.awt.event.WindowListener#windowDeactivated(WindowEvent)
		 */
		public final void windowDeactivated(WindowEvent e)
		{
			if (ignoreOwned
			    && (e.getOppositeWindow().getOwner() == getWindow())) {
				return;
			}

			setSize(getWindow().getSize());
			setVisible(true);
		}

		/**
		 * Uneccesary, but still implemented.
		 *
		 * @param e WindowEvent
		 *
		 * @see java.awt.event.WindowListener#windowClosing(WindowEvent)
		 */
		public final void windowClosing(WindowEvent e)
		{
			uninstall();
		}
	}

	/**
	 * Default constructor for InterceptorPane. Creation date: (10.4.2002
	 * 15:15:26)
	 *
	 * @param container Container
	 */
	public InterceptorPane(Container container)
	{
		this(container, false);
	}

	/**
	 * Creates new InterceptorPane with option to not ignore clicks when
	 * switching to and from windows owned by this window.
	 *
	 * @param container Container
	 * @param ignoreOwnedWindows boolean
	 */
	public InterceptorPane(Container container, boolean ignoreOwnedWindows)
	{
		super();
		initialize(container);
		ignoreOwned = ignoreOwnedWindows;
	}

	/**
	 * Installs the interceptor pane into the component specified. If the
	 * specified container is not a valid target, and exception will be
	 * thrown.
	 *
	 * @param container Container
	 *
	 * @throws UnsupportedOperationException
	 */
	private void initialize(Container container)
	{
		boolean initialized = false;

		if (container instanceof JFrame) {
			window = (JFrame)container;
			((JFrame)window).setGlassPane(this);

			initialized = true;
		}

		if (container instanceof JDialog) {
			window = (JDialog)container;
			((JDialog)window).setGlassPane(this);

			initialized = true;
		}

		if (!initialized) {
			throw new UnsupportedOperationException("Window class "
			    + container.getClass().getName()
			    + " does not support glass panes.");
		}

		clickListener = new ClickListener();
		addMouseListener(clickListener);

		// addWindowListener is used instead of addWindowFocusListener
		// for backward compatibility
		focusMonitor = new FocusMonitor();
		window.addWindowListener(focusMonitor);
	}

	/**
	 * Removes InterceptorPane from the window. This method uninstalls all
	 * listeners from the window and removes itself from the glass pane. To
	 * reenable the pane, a new instance must be created.
	 */
	public void uninstall()
	{
		if (window == null) {
			return;
		}

		window.removeMouseListener(clickListener);

		window.removeWindowListener(focusMonitor);

		if (window instanceof JFrame) {
			((JFrame)window).setGlassPane(new JLabel());
		}

		if (window instanceof JDialog) {
			((JDialog)window).setGlassPane(new JLabel());
		}

		window = null;
		clickListener = null;
		focusMonitor = null;
	}

	/**
	 * Returns instance of the window this interceptor is installed on.
	 *
	 * @return Window
	 */
	public Window getWindow()
	{
		return window;
	}
}

/* __oOo__ */
