/*
 * Copyright (c) 2004 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.components;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JTextField;
import javax.swing.SwingConstants;
import javax.swing.border.Border;

import com.cosylab.application.state.State;
import com.cosylab.events.SetListener;
import com.cosylab.gui.components.numberfield.NumberDescriptor;


/**
 * <code>LabelledNumberField</code> is extended <code>NumberField</code> with
 * title label and unit label.
 *
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 * @author <a href="mailto:jernej.kamenik@cosylab.com">Jernej Kamenik</a>
 * @version $id$
 */
public class LabelledNumberField extends AbstractNumericDisplayerPanel
{
	private NumberField number;

	/**
	 * Creates a new LabelledNumberField object.
	 */
	public LabelledNumberField()
	{
		this(null, "3.2f", null, null, false);
	}

	/**
	 * Creates a new LabelledNumberField object.
	 *
	 * @param format
	 */
	public LabelledNumberField(String format)
	{
		this(null, format, null, null, false);
	}

	/**
	 * Creates a new LabelledNumberField object.
	 *
	 * @param value
	 */
	public LabelledNumberField(Number value)
	{
		this(value, "3.2f", null, null, false);
	}

	/**
	 * Creates a new LabelledNumberField object.
	 *
	 * @param value
	 * @param format
	 */
	public LabelledNumberField(Number value, String format)
	{
		this(value, format, null, null);
	}

	/**
	 * Creates a new LabelledNumberField object.
	 *
	 * @param value
	 * @param format
	 * @param title
	 * @param units
	 */
	public LabelledNumberField(Number value, String format, String title,
	    String units)
	{
		this(value, format, title, units, false);
	}

	/**
	 * Creates a new LabelledNumberField object.
	 *
	 * @param value initial value
	 * @param format initial format
	 * @param title initial title
	 * @param units initial units
	 * @param unitsWithTitle if <code>true</code> units shown in title label
	 *        and not in separate label
	 */
	public LabelledNumberField(Number value, String format, String title,
	    String units, boolean unitsWithTitle)
	{
		this(value, format, title, units, true, true, true, unitsWithTitle);
	}

	/**
	 * Creates a new LabelledNumberField object.
	 *
	 * @param value initial value
	 * @param format initial format
	 * @param title initial title
	 * @param units initial units
	 * @param resizable inital resizable
	 * @param enchanced inital enchanced
	 * @param editable initial editable
	 * @param unitsWithTitle if <code>true</code> units shown in title label
	 *        and not in separate label
	 */
	public LabelledNumberField(Number value, String format, String title,
	    String units, boolean resizable, boolean enchanced, boolean editable,
	    boolean unitsWithTitle)
	{
		super(title, true, HORIZONTAL_LAYOUT, resizable, true, enchanced, 1,
		    Integer.MAX_VALUE, units, !unitsWithTitle, unitsWithTitle);

		setEditable(editable);
		setFormat(format);
		setMaximumValue(new Double(Double.MAX_VALUE));
		setMinimumValue(new Double(-Double.MAX_VALUE));

		if (value != null) {
			setNumberType(value.getClass());
			setNumberValue(value);
		}

		
	}

	/**
	 * Sets number of columns on the <code>NumberField</code>.
	 *
	 * @see NumberField#setColumns(int)
	 */
	public void setColumns(int columns)
	{
		int oldValue = getColumns();
		getNumber().setColumns(columns);
		firePropertyChange("columns", oldValue, columns);
	}

	/**
	 * Returns number of columns on the <code>NumberField</code>.
	 *
	 * @see JTextField#getColumns()
	 */
	public int getColumns()
	{
		return getNumber().getColumns();
	}

	/**
	 * Sets the value of type <code>double</code> to be displayed.
	 *
	 * @param newValue
	 */
	public void setDoubleValue(double newValue)
	{
		double oldValue = getDoubleValue();
		getNumber().setDoubleValue(newValue);
		firePropertyChange("doubleValue", oldValue, newValue);
	}

	/**
	 * Returns the value of type <code>double</code>
	 *
	 * @return double
	 */
	public double getDoubleValue()
	{
		return getNumber().getDoubleValue();
	}

	/**
	 * Enables/disables user editing of number.
	 *
	 * @param editable if <code>true</code> user is aloved to input number
	 */
	public void setEditable(boolean editable)
	{
		boolean oldValue = isEditable();
		getNumber().setEditable(editable);
		firePropertyChange("editable", oldValue, editable);
	}

	/**
	 * Returns <code>true</code> if use is allowed to input number.
	 *
	 * @return boolean <code>true</code> if use is allowed to input number
	 */
	public boolean isEditable()
	{
		return getNumber().isEditable();
	}

	/**
	 * Sets the value of type <code>long</code> to be displayed.
	 *
	 * @param newValue
	 */
	public void setLongValue(long newValue)
	{
		long oldValue = getLongValue();
		getNumber().setLongValue(newValue);
		firePropertyChange("longValue", oldValue, newValue);
	}

	/**
	 * Returns the value of type <code>long</code>
	 *
	 * @return long
	 */
	public long getLongValue()
	{
		return getNumber().getLongValue();
	}

	/**
	 * Sets the maximum allowed value that can be entered in this NumberField.
	 * Setting maximum to null means no maximum is set. This method does not
	 * check for validity of the limit. If maximum is set less than minimum
	 * then no value can be set.
	 *
	 * @param newMaximum Number
	 */
	public void setMaximumValue(Number newMaximum)
	{
		Number oldValue = getMaximumValue();
		getNumber().setMaximum(newMaximum);
		super.setMaximumValue(number.getMaximum());
		firePropertyChange("maximumValue", oldValue, newMaximum);
	}

	/**
	 * Returns the maximum value allowed by this NumberField. Can be null if no
	 * maximum is set.
	 *
	 * @return Number maximum allowed value.
	 */
	public Number getMaximumValue()
	{
		return getNumber().getMaximum();
	}

	/**
	 * Sets the minimum allowed value that can be entered in this NumberField.
	 * Setting minimum to null means no minimum is set. This method does not
	 * check for validity of the limit. If maximum is set less than minimum
	 * then no value can be set.
	 *
	 * @param newMinimum Number
	 */
	public void setMinimumValue(Number newMinimum)
	{
		Number oldValue = getMinimumValue();
		getNumber().setMinimum(newMinimum);
		super.setMinimumValue(number.getMinimum());
		firePropertyChange("minimumValue", oldValue, newMinimum);
	}

	/**
	 * Returns the minimum value allowed by this NumberField. Can be null if no
	 * miminim is set.
	 *
	 * @return Number minimum allowed value
	 */
	public Number getMinimumValue()
	{
		return getNumber().getMinimum();
	}

	/**
	 * Sets the number format.
	 *
	 * @param newNumberFormat
	 */
	public void setNumberType(Class newNumberFormat)
	{
		Class oldValue = getNumberType();
		getNumber().setNumberType(newNumberFormat);
		super.setNumberType(newNumberFormat);
		firePropertyChange("numberType", oldValue, newNumberFormat);
	}

	/**
	 * Returns type of displayed number.
	 *
	 * @return type of displayed number
	 */
	public Class getNumberType()
	{
		return getNumber().getNumberType();
	}

	/**
	 * Sets the numeric value for this component. New value is subject to
	 * min/max test if any of them is set. If new value does not fall within
	 * these bounds, setValue will do nothing.
	 *
	 * @param newValue Number
	 */
	public void setNumberValue(Number newValue)
	{
		Number oldValue = getNumberValue();
		getNumber().setValue(newValue);
		firePropertyChange("numberValue", oldValue, newValue);
	}

	/**
	 * Returns generic representation of the value using the
	 * <code>Number</code> class.
	 *
	 * @return java.lang.Number
	 */
	public Number getNumberValue()
	{
		return getNumber().getValue();
	}

	/**
	 * Run test applet.
	 *
	 * @param args command line parameters
	 */
	public static void main(String[] args)
	{
		JFrame frame = new JFrame("NumberField Testing Applet");
		frame.getContentPane().setLayout(new BorderLayout(10, 10));

		//frame.getContentPane().add(new LabelledNumberField(new Long(10), "%3d",
		//        "Teddy", "Bears"), BorderLayout.CENTER);
		LabelledNumberField field= new LabelledNumberField(new Double(10),
        "%1.0f", "Current", "A",true);
		field.setTitleMaximumFontSize(12);
		field.setUnitsVisible(true);
		
		frame.getContentPane().add(field, BorderLayout.CENTER);
		frame.setSize(300, 100);
		frame.addWindowListener(new WindowAdapter() {
				public void windowClosing(WindowEvent e)
				{
					System.exit(0);
				}
			});
		frame.setVisible(true);
	}

	/**
	 * Overriden to implement additional features loading.
	 *
	 * @param state to be loaded.
	 */
	public void setState(State state)
	{
		super.setState(state);

		setEditable(state.getBoolean("editable", isEditable()));
		setColumns(state.getInt("columns", getColumns()));
		setTiltingEnabled(state.getBoolean("tiltingEnabled", isTiltingEnabled()));
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @param b
	 */
	public void setTiltingEnabled(boolean b)
	{
		if (isTiltingEnabled() == b) return;
		getNumber().setTiltingEnabled(b);
		firePropertyChange("tiltingEnabled", !b, b);
	}

	/**
	 * Overriden to implement additional features saving.
	 *
	 * @return state of the component.
	 */
	public State getState()
	{
		State state = super.getState();

		state.putBoolean("editable", isEditable());
		state.putBoolean("tiltingEnabled", isTiltingEnabled());
		state.putInt("columns", getColumns());

		return state;
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @return
	 */
	public boolean isTiltingEnabled()
	{
		return getNumber().isTiltingEnabled();
	}

	/**
	 * Adds action listener on <code>NumberField</code>.
	 *
	 * @param l
	 */
	public synchronized void addActionListener(ActionListener l)
	{
		number.addActionListener(l);
	}

	/**
	 * Adds set listener to the <code>NumberField</code>.
	 *
	 * @see NumberField#addSetListener(SetListener)
	 */
	public void addSetListener(SetListener l)
	{
		getNumber().addSetListener(l);
	}

	/**
	 * Removes action listener from <code>NumberField</code>.
	 *
	 * @param l
	 */
	public synchronized void removeActionListener(ActionListener l)
	{
		number.removeActionListener(l);
	}

	/**
	 * Removes set listener from the <code>NumberField</code>.
	 *
	 * @see NumberField#removeSetListener(SetListener)
	 */
	public void removeSetListener(SetListener l)
	{
		getNumber().removeSetListener(l);
	}

	/**
	 * The <code>NumberField</code> component, that does number editing and
	 * rendering.
	 *
	 * @return the <code>NumberField</code> component
	 */
	protected NumberField getNumber()
	{
		if (number == null) {
			number = new NumberField();
			number.setResizable(true);
			number.setSettingPolicy(NumberField.SET_ON_APPLY);
			number.setHorizontalAlignment(SwingConstants.RIGHT);

			//			number.addPropertyChangeListener(new PropertyChangeListener() {
			//					public void propertyChange(PropertyChangeEvent evt)
			//					{
			//						if (("preferredSize").equals(evt.getPropertyName())) {
			//							setPreferredSize(new Dimension(
			//							        (int)(getNumber().getPreferredSize().width * 1.05),
			//							        (int)(getNumber().getPreferredSize().height * 1.05)));
			//						}
			//
			//						firePropertyChange(evt.getPropertyName(),
			//						    evt.getOldValue(), evt.getNewValue());
			//
			//						if (NumberField.VALUE.equals(evt.getPropertyName())) {
			//							LabelledNumberField.this.doLayout();
			//						}
			//					}
			//				});
			number.addMouseListener(getPopupManager().getMouseHook());
		}

		return number;
	}
	
	/* (non-Javadoc)
	 * @see com.cosylab.gui.components.DisplayerPanel#getValueComponent()
	 */
	protected JComponent getValueComponent()
	{
		return getNumber();
	}

	protected void internalSetEnabled()
	{
		super.internalSetEnabled();
		getNumber().setEnabled(isEnabled());
	}

	/**
	 * Overriden to implement additional functionality.
	 *
	 * @see com.cosylab.gui.components.AbstractDisplayerPanel#setEnhanced(boolean)
	 */
	protected void internalSetEnhanced()
	{
		super.internalSetEnhanced();
		getNumber().setEnhanced(isEnhanced());
	}

	protected void internalSetFormat()
	{
		super.internalSetFormat();

		String tFormat = getFormat();

		if (tFormat == null) {
			tFormat = NumberField.createDefaultFormat(getNumberType());
		}

		getNumber().setFormat(tFormat);
	}

	/**
	 * Overriden to implement additional functionality.
	 *
	 * @see com.cosylab.gui.components.AbstractDisplayerPanel#setResizable(boolean)
	 */
	protected void internalSetResizable()
	{
		super.internalSetResizable();
		getNumber().setResizable(isResizable());
		getNumber().revalidate();
	}

	/**
	 * Returns descriptor object, to which visualization of number as strring
	 * is dlegated.
	 *
	 * @return number visualization
	 */
	public NumberDescriptor getNumberDescriptor()
	{
		return getNumber().getNumberDescriptor();
	}

	/**
	 * When this is set with non-null value, this NumberField uses this
	 * descriptor to convert Number to string and back. By this user can
	 * define ovn visualization of Number.
	 *
	 * @param numberDescriptor new Number visualization, can be
	 *        <code>null</code>
	 */
	public void setNumberDescriptor(NumberDescriptor numberDescriptor)
	{
		NumberDescriptor oldValue = getNumberDescriptor();
		getNumber().setNumberDescriptor(numberDescriptor);
		firePropertyChange("numberDescriptor", oldValue, numberDescriptor);
	}
	/**
	 * Sets the setting policy of thi Numberdisplayer. SET_AS_TYPED updates the
	 * Numberfield value when typed. SET_ON_APPLY updates the Numberfield only
	 * on actionPerformed events (when the user presses ENTER) SET_ON_EXIT
	 * updates the numberfield when it loses focus
	 *
	 * @param newPolicy
	 *
	 * @throws IllegalArgumentException DOCUMENT ME!
	 */
	public void setSettingPolicy(int newPolicy)
	{
		int oldValue = getSettingPolicy();
		getNumber().setSettingPolicy(newPolicy);
		firePropertyChange("settingPolicy", oldValue, newPolicy);
	}
	/**
	 * Returns the currently used seting policy by this Numberfield
	 *
	 * @return int currently used setting policy.
	 */
	public int getSettingPolicy()
	{
		return getNumber().getSettingPolicy();
	}
	
	/**
	 * DOCUMENT ME!
	 *
	 * @param border DOCUMENT ME!
	 */
	public void setStaticBorder(Border border)
	{
		Border oldValue = getStaticBorder();
		getNumber().setStaticBorder(border);
		firePropertyChange("staticBorder", oldValue, border);
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @return DOCUMENT ME!
	 */
	public Border getStaticBorder()
	{
		return getNumber().getStaticBorder();
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @param border DOCUMENT ME!
	 */
	public void setEditBorder(Border border)
	{
		Border oldValue = getEditBorder();
		getNumber().setEditBorder(border);
		firePropertyChange("editBorder", oldValue, border);
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @return DOCUMENT ME!
	 */
	public Border getEditBorder()
	{
		return getNumber().getEditBorder();
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @param policy DOCUMENT ME!
	 */
	public void setEnterPolicy(int policy)
	{
		int oldValue = getEnterPolicy();
		getNumber().setEnterPolicy(policy);
		firePropertyChange("enterPolicy", oldValue, policy);
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @return DOCUMENT ME!
	 */
	public int getEnterPolicy()
	{
		return getNumber().getEnterPolicy();
	}
	
	public void setWarnOutOfBounds(Boolean b) {
		Boolean oldValue = isWarnOutOfBounds();
		getNumber().setWarnOutOfBounds(b);
		firePropertyChange("warnOutOfBounds", oldValue, b);
	}
	
	public boolean isWarnOutOfBounds() {
		return getNumber().isWarnOutOfBounds();
	}
	
	public void setOutOfBoundsColor(Color c) {
		Color oldValue = getOutOfBoundsColor();
		getNumber().setOutOfBoundsColor(c);
		firePropertyChange("outOfBoundColor", oldValue, c);
	}
	
	public Color getOutOfBoundsColor() {
		return getNumber().getWarningColor();
	}
	
	@Override
	public void setBackground(Color bg) {
		Color fieldBg = getNumber().getBackground();
		super.setBackground(bg);
		getNumber().setBackground(fieldBg);
	}
}

/* __oOo__ */
