/*
 * Copyright (c) 2004 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.components;

import com.cosylab.gui.components.table.DefaultObjectTableEngine;
import com.cosylab.gui.components.table.MutableTableModelRows;
import com.cosylab.gui.components.table.ObjectTableEngine;
import com.cosylab.gui.components.table.TableRow;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;


/**
 * <code>TableModel</code>, which accepts new items as objects and converts
 * them to new rows. This model acts as <code>Set</code> for Objects, it does
 * not hold them in constant order and does not permit multiple entries with
 * same object.
 *
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 */
public class ObjectTableModel extends MutableTableModelRows implements Set
{
	private ObjectTableEngine objectTableEngine;
	private Map objects2rows = new HashMap();

	/**
	 * Creates new empty model.
	 */
	public ObjectTableModel()
	{
		super();
	}

	/**
	 * Creates new model with engine.
	 *
	 * @param engine the table engine
	 */
	public ObjectTableModel(ObjectTableEngine engine)
	{
		this();
		setObjectTableEngine(engine);
	}

	/* (non-Javadoc)
	 * @see java.util.Set#size()
	 */
	public int size()
	{
		return objects2rows.size();
	}

	/* (non-Javadoc)
	 * @see java.util.Set#clear()
	 */
	public synchronized void clear()
	{
		Object[] o = objects2rows.keySet().toArray();

		for (int i = 0; i < o.length; i++) {
			remove(o[i]);
		}
	}

	/* (non-Javadoc)
	 * @see java.util.Set#isEmpty()
	 */
	public boolean isEmpty()
	{
		return objects2rows.isEmpty();
	}

	/* (non-Javadoc)
	 * @see java.util.Set#toArray()
	 */
	public synchronized Object[] toArray()
	{
		return objects2rows.keySet().toArray();
	}

	/* (non-Javadoc)
	 * @see java.util.Set#add(java.lang.Object)
	 */
	public synchronized boolean add(Object o)
	{
		if (objects2rows.containsKey(o)) {
			return false;
		}

		TableRow row = getObjectTableEngine().getTableRow(o);

		if (row == null) {
			return false;
		}

		addRow(row);
		objects2rows.put(o, row);

		return true;
	}

	/* (non-Javadoc)
	 * @see java.util.Set#contains(java.lang.Object)
	 */
	public synchronized boolean contains(Object o)
	{
		return objects2rows.containsKey(o);
	}

	/* (non-Javadoc)
	 * @see java.util.Set#remove(java.lang.Object)
	 */
	public synchronized boolean remove(Object o)
	{
		TableRow r = (TableRow)objects2rows.get(o);

		if (r == null) {
			return false;
		}

		removeRow(r);
		objects2rows.remove(o);
		getObjectTableEngine().releaseTableRow(o, r);

		return true;
	}

	/* (non-Javadoc)
	 * @see java.util.Set#addAll(java.util.Collection)
	 */
	public synchronized boolean addAll(Collection c)
	{
		boolean b = false;
		Iterator it = c.iterator();

		while (it.hasNext()) {
			Object o = it.next();

			if (!objects2rows.containsKey(o)) {
				b = true;
				add(o);
			}
		}

		return b;
	}

	/* (non-Javadoc)
	 * @see java.util.Set#containsAll(java.util.Collection)
	 */
	public synchronized boolean containsAll(Collection c)
	{
		return objects2rows.keySet().containsAll(c);
	}

	/* (non-Javadoc)
	 * @see java.util.Set#removeAll(java.util.Collection)
	 */
	public synchronized boolean removeAll(Collection c)
	{
		if (c == null || c.isEmpty()) {
			return false;
		}

		Iterator it = c.iterator();

		while (it.hasNext()) {
			remove(it.next());
		}

		return true;
	}

	/* (non-Javadoc)
	 * @see java.util.Set#retainAll(java.util.Collection)
	 */
	public synchronized boolean retainAll(Collection c)
	{
		ArrayList rem = new ArrayList(objects2rows.size());
		Iterator it = objects2rows.keySet().iterator();

		while (it.hasNext()) {
			Object key = it.next();

			if (!c.contains(key)) {
				rem.add(key);
			}
		}

		return removeAll(rem);
	}

	/* (non-Javadoc)
	 * @see java.util.Set#iterator()
	 */
	public Iterator iterator()
	{
		return objects2rows.keySet().iterator();
	}

	/* (non-Javadoc)
	 * @see java.util.Set#toArray(java.lang.Object[])
	 */
	public synchronized Object[] toArray(Object[] a)
	{
		return objects2rows.keySet().toArray(a);
	}

	/**
	 * Return engine, which converts objects to table rows. If engine was not
	 * set before, call to this method initializes this model to
	 * <code>DefaultObjectTableEngine</code>.
	 *
	 * @return object to row converter
	 */
	public ObjectTableEngine getObjectTableEngine()
	{
		if (objectTableEngine == null) {
			objectTableEngine = new DefaultObjectTableEngine();
		}

		return objectTableEngine;
	}

	/**
	 * Sets new object to for converter. Setting new factory is only alllowed
	 * with empty model.
	 *
	 * @param objectTableEngine new object to row converter
	 *
	 * @throws IllegalStateException if model is not empty
	 */
	public void setObjectTableEngine(ObjectTableEngine objectTableEngine)
	{
		if (size() > 0) {
			throw new IllegalStateException(
			    "objectTableEngine can be changed only on empty model.");
		}

		this.objectTableEngine = objectTableEngine;
	}

	/* (non-Javadoc)
	 * @see javax.swing.table.AbstractTableModel#setValueAt(java.lang.Object, int, int)
	 */
	public synchronized void setValueAt(Object aValue, int rowIndex,
	    int columnIndex)
	{
		if (rowIndex < 0 || rowIndex >= getRowCount() || columnIndex < 0
		    || columnIndex >= getColumnCount()) {
			return;
		}

		TableRow tr = getRow(rowIndex);

		Iterator it = objects2rows.keySet().iterator();

		while (it.hasNext()) {
			Object o = it.next();

			if (tr == objects2rows.get(o)) {
				getObjectTableEngine().setValue(aValue, o, tr, columnIndex);

				return;
			}
		}
	}
	/* (non-Javadoc)
	 * @see com.cosylab.gui.components.table.TableModelRows#getColumnName(int)
	 */
	public String getColumnName(int column) {
		return getObjectTableEngine().getColumnName(column);
	}
}

/* __oOo__ */
