/*
 * Copyright (c) 2004 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.components;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GradientPaint;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GridLayout;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.BufferedImage;
import java.beans.Beans;
import java.util.Timer;
import java.util.TimerTask;
import javax.swing.AbstractAction;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;
import com.cosylab.application.state.State;
import com.cosylab.application.state.StateFactory;
import com.cosylab.application.state.StateOriginator;
import com.cosylab.gui.components.customizer.AbstractCustomizerPanel;
import com.cosylab.gui.components.range2.RangedValueController;
import com.cosylab.gui.components.range2.LinearRange;
import com.cosylab.gui.components.range2.LogarithmicRange;
import com.cosylab.gui.components.range2.RangedValueEvent;
import com.cosylab.gui.components.range2.RangedValueListener;
import com.cosylab.gui.components.range2.RangedValuePolicy;
import com.cosylab.gui.components.range2.Tick;
import com.cosylab.gui.components.range2.TickParameters;
import com.cosylab.gui.components.range2.TrimValuePolicy;
import com.cosylab.gui.components.util.ActionList;
import com.cosylab.gui.components.util.ColorHelper;
import com.cosylab.gui.components.util.FontHelper;
import com.cosylab.gui.components.util.PaintHelper;
import com.cosylab.gui.components.util.PopupManageable;
import com.cosylab.gui.components.util.PopupManager;
import com.cosylab.gui.components.util.ScreenCapturer;
import com.cosylab.util.PrintfFormat;


/**
 * A visual component for displaying numerical (<code>double</code>) values
 * using a liquid partially filled vertical tube visualisation. A value label
 * above the pipe displays the textural value together vith possible units.
 *
 * @author <a href="mailto:jernej.kamenik@cosylab.com">Jernej Kamenik</a>
 * @version $id$
 */
public class Piper extends JComponent implements PopupManageable,
	StateOriginator
{
	private static final long serialVersionUID = 1L;
	/*
	 * An extension of Timer used for periodic tilting of the Wheelswitch.
	 *
	 * @author <a href="mailto:jernej.kamenik@cosylab.com">Jernej Kamenik</a>
	 * @version $id$
	 *
	 */
	protected class TiltHandler extends Timer
	{
		private class TiltTask extends TimerTask
		{
			/**
			 * DOCUMENT ME!
			 */
			public void run()
			{
				if (numberOfTilts >= MAX_NUMBER_OF_TILTS) {
					cancel();
					tilting = false;
				} else {
					numberOfTilts++;
					tilting = !tilting;
				}

				renderer.clearBuffer();
				repaint();
			}
		}

		private final int MAX_NUMBER_OF_TILTS = 3;
		private final long TILT_RATE = 200;
		private int numberOfTilts = MAX_NUMBER_OF_TILTS;

		/**
		 * Scedules a new tilting task if the user value equals any of the
		 * bounds.
		 */
		public void tilt()
		{
			if (numberOfTilts >= MAX_NUMBER_OF_TILTS) {
				numberOfTilts = 0;
				schedule(new TiltTask(), 0, TILT_RATE);
			} else {
				numberOfTilts = 0;
			}
		}
	}

	private class PiperRenderer implements TickParameters /* implements TickCallback - edit apikus */
	{
		private Graphics2D g;
		private Rectangle pipeRect;
		private int arcRadius;
		private BufferedImage pipeImage;
		private BufferedImage scaleImage;
		private Color DARK_COLOR = new Color(0, 0, 10);
		private Color LIGHT_COLOR = new Color(50, 150, 250);
		private Color OUT_OF_BOUNDS_DARK_COLOR = new Color(250, 120, 10);
		private Color OUT_OF_BOUNDS_LIGHT_COLOR = new Color(250, 250, 50);
		private Tick[] ticks;

		/**
		 * Clears image buffers.
		 */
		
		/* edit apikus */
		/*public void clearBuffer()
		{
			SwingUtilities.invokeLater(new Runnable() {
				public void run() {
					pipeImage = null;
					scaleImage = null;
					pipeRect = null;
					ticks = null;
					arcRadius = 0;
				}
			});
		}*/

		public void clearBuffer() { /* edit apikus */
			pipeImage = null;
			scaleImage = null;
			pipeRect = null;
			ticks = null;
			arcRadius = 0;
		}
		
		/**
		 * Gets the size of the scale text.
		 *
		 * @param position double
		 * @param text String
		 *
		 * @return int
		 *
		 * @see com.cosylab.gui.components.range.TickCallback#measureTick(double,
		 *      java.lang.String)
		 */
		public int measureTick(double position, String text)
		{
			return g.getFontMetrics().getHeight();
		}

		/**
		 * Paints the liquid fill buffered image on the Graphics supplied. If
		 * no buffer exists, new one is created.
		 *
		 * @param g Graphics2D
		 */
		public void paintFill(Graphics2D g)
		{
			//setPipeRect(g);

			//System.out.println("FILL " + Piper.this.getSize() + " " + pipeRect);

			int fillHeight = Math.max(1,
				    (int)Math.round(
				        rangedValue.getRelativeValue() * (pipeRect.height
				        - 2 * arcRadius)));
			int fillWidth = pipeRect.width;

			BufferedImage fillImage = new BufferedImage(fillWidth,
				    fillHeight + arcRadius, BufferedImage.TYPE_4BYTE_ABGR);
			Graphics2D g2D = (Graphics2D)fillImage.getGraphics();
			prepareGraphics(g2D);

			boolean checkValue = (getValue() >= getMinimum()
				&& getValue() <= getMaximum());

			Color lightColor = (checkValue ? LIGHT_COLOR
				: OUT_OF_BOUNDS_LIGHT_COLOR);
			Color darkColor = (checkValue ? DARK_COLOR : OUT_OF_BOUNDS_DARK_COLOR);

			if (enhanced) {
				g2D.setPaint(new GradientPaint(0.f, 0.f, lightColor,
				        fillWidth, fillHeight, darkColor));
			} else {
				g2D.setColor(lightColor);
			}

			g2D.fillRect(0, 0, fillWidth, fillHeight);
			g2D.fillArc(0, fillHeight - arcRadius, 2 * arcRadius,
			    2 * arcRadius, 180, 270);
			g2D.fillArc(fillWidth - 2 * arcRadius, fillHeight - arcRadius,
			    2 * arcRadius, 2 * arcRadius, 270, 360);

			if (arcRadius < fillWidth) {
				g2D.fillRect(arcRadius, fillHeight, fillWidth - 2 * arcRadius,
				    arcRadius);
			}

			if (enhanced) {
				g.setComposite(AlphaComposite.getInstance(
				        AlphaComposite.SRC_OVER, 0.5f));
			}

			g.drawImage(fillImage, null, pipeRect.x,
			    pipeRect.y + pipeRect.height - arcRadius - fillHeight);

			//System.out.println("VALUE "+rangedValue.getRelativeValue()+" "+fillHeight+" "+pipeRect.y+" "+ pipeRect.height +" "+ arcRadius +" "+ fillHeight);
			if (enhanced) {
				g.setComposite(AlphaComposite.getInstance(
				        AlphaComposite.SRC_OVER, 1.f));
			}
		}

		/**
		 * Paints the value label showing the actual value dipalyed on the
		 * Graphics supplied.
		 *
		 * @param g Graphics2D
		 */
		public void paintLabel(Graphics2D g)
		{
			//setPipeRect(g);

			String unitText = units;

			if (units == null && Beans.isDesignTime()) {
				unitText = "<u>";
			}

			if (unitText == null || !unitsVisible) {
				unitText = "";
			}

			String text;

			try {
				//text = formatter.sprintf(labelValue); /* edit apikus */
				text = this.formatNumber(labelValue);
			} catch (Exception e) {
				text = Double.toString(labelValue);
			}

			text += (units == null ? "" : " " + unitText);

			int textWidth = g.getFontMetrics().stringWidth(text);
			int textHeight = g.getFontMetrics().getAscent();

			// workaround - Java bug
			textHeight = (int)(textHeight * 0.7) + 1;

			int botLeftX = pipeRect.x + (pipeRect.width - textWidth) / 2;
			int botLeftY = pipeRect.y + pipeRect.height
				+ (textHeight + pipeRect.y) / 2;

			g.setColor((isEnabled() ? Color.BLACK : Color.GRAY));
			g.drawString(text, botLeftX, botLeftY);
		}

		/**
		 * Paints the pipe buffered image on the Graphics supplied. If no
		 * buffer exists, new one is created.
		 *
		 * @param g Graphics2D
		 */
		public void paintPipe(Graphics2D g)
		{
			//setPipeRect(g);

			if (pipeImage == null) {
				//Debug.out("New pipe image created.");
				pipeImage = new BufferedImage(pipeRect.width, pipeRect.height,
					    BufferedImage.TYPE_4BYTE_ABGR);

				Graphics2D g2D = (Graphics2D)pipeImage.getGraphics();
				prepareGraphics(g2D);
				g2D.setColor((tilting ? ColorHelper.getAlarm() : Color.WHITE));
				g2D.fillRoundRect(0, 0, pipeRect.width, pipeRect.height,
				    2 * arcRadius, 2 * arcRadius);
			}

			g.drawImage(pipeImage, null, pipeRect.x, pipeRect.y);
		}

		private Dimension getTickDim(Graphics g, int pipeHeigth)
		{
			Tick[] tcks = getTicks(pipeHeigth);
			int tckW = 0;
			int maxLen = 0;

			for (int i = 0; i < tcks.length; i++) {
				if (tcks[i] != null && tcks[i].text != null
				    && tcks[i].text.length() > maxLen) {
					int newW = g.getFontMetrics().stringWidth(tcks[i].text);
					tckW = Math.max(tckW, newW);
				}
			}

			return new Dimension(tckW, g.getFontMetrics().getHeight());
		}

		private void setPipeRect(Graphics2D g)
		{
			if (pipeRect == null) {

				this.g=g;

				int compW = Piper.this.getWidth();
				int compH = Piper.this.getHeight();
				int x;
				int y;
				int w;
				int h;

				y = g.getFontMetrics().getHeight()
					+ Math.max(1, (int)(6 * Math.atan(compH / 100)));

				h = Math.max(1, compH - 2 * y);

				Dimension tickDim = getTickDim(g, h);

				if (stretch) {
					x = 10;
					w = Math.max(1, compW - tickDim.width * 3 / 2 - 2 * x);
				} else {
					x = compW * 1 / 5;

					if (10 + tickDim.width > x) {
						x = (2 * x - 10 - tickDim.width) / 2 + 5;
						w = Math.max(1, compW - x - 10 - tickDim.width);
					} else {
						w = Math.max(1, compW - 2 * x);
					}
				}

				pipeRect = new Rectangle(x, y, w, h);

				g=null;

				setArcRadius(pipeRect);
			}
		}

		private void setArcRadius(Rectangle pipeRect)
		{
			double ratio = pipeRect.height / (double)pipeRect.width;

			if (ratio > 5) {
				arcRadius = pipeRect.width / 2;
			} else {
				arcRadius = (int)Math.max(ratio / 5 * pipeRect.width / 2, 2);
			}
		}

		/**
		 * Paints the scale buffered image on the Graphics supplied. If no
		 * buffer exists, new one is created.
		 *
		 * @param g Graphics2D
		 */
		public void paintScale(Graphics2D g)
		{
			g.drawImage(getScaleImage(), null, 0, 0);
		}

		/**
		 * Paints the title.
		 *
		 * @param g
		 */
		public void paintTitle(Graphics2D g)
		{
			//setPipeRect(g);

			String text = title;

			if (title == null && Beans.isDesignTime()) {
				text = "<title>";
			}

			if (text == null || text.length() == 0) {
				return;
			}

			int textWidth = g.getFontMetrics().stringWidth(text);
			int textHeight = g.getFontMetrics().getAscent();

			// workaround - Java bug
			textHeight = (int)(textHeight * 0.7) + 1;

			int botLeftX = pipeRect.x + (pipeRect.width - textWidth) / 2;
			int botLeftY = (textHeight + pipeRect.y) / 2;

			g.setColor((isEnabled() ? Color.BLACK : Color.GRAY));

			//g.setFont(FontHelper.getFontWithSize(g.getFont().getSize()*12/10,g.getFont()));
			g.drawString(text, botLeftX, botLeftY);
		}

		protected void prepareGraphics(Graphics2D g)
		{
			Font f = FontHelper.getDefaultFont();

			if (resizable) {
				int fontHeight = 5 + Math.min(getWidth(), getHeight() / 2) / 15;
				f = FontHelper.getFontWithSize(fontHeight, f);
			}

			g.setFont(f);

			if (enhanced) {
				g.addRenderingHints(PaintHelper.getAntialiasingHints());
			}
		}

		private BufferedImage getScaleImage()
		{
			if (scaleImage == null) {
				//Debug.out("New scale image created.");
				int width = Piper.this.getWidth();
				int height = Piper.this.getHeight();

				scaleImage = new BufferedImage(width, height,
					    BufferedImage.TYPE_4BYTE_ABGR);

				Graphics2D g2D = (Graphics2D)scaleImage.getGraphics();
				prepareGraphics(g2D);
				g2D.setColor((isEnabled() ? Color.BLACK : Color.GRAY));

				Font oldFont = g2D.getFont();
				g2D.setFont(oldFont.deriveFont(Font.PLAIN,
				        oldFont.getSize2D() * 0.8f));

				//setPipeRect(g2D);

				/*System.out.println("TICKS " + Piper.this.getSize() + " "
				    + pipeRect);*/

				int spaceWidth = width - pipeRect.x - pipeRect.width;

				int scaleHeight = pipeRect.height - 2 * arcRadius;

				int tickLength = Math.max(pipeRect.width * 3 / 4, 5);
				int minorTickLength = tickLength / 2;

				if (pipeRect.width > 200) {
					tickLength = 150 + (pipeRect.width - 200) * 1 / 3;
					minorTickLength = 75;
				}

				// Offset from the right edge of the pipe
				Dimension tickDim = getTickDim(g2D, pipeRect.height);
				int offsetX = tickDim.height / 2;

				if (2 * offsetX + tickDim.width > spaceWidth) {
					offsetX = Math.max(1, (spaceWidth - tickDim.width) / 2);
				}

				Tick[] ticks = getTicks(scaleHeight);

				if (ticks == null) {
					return scaleImage;
				}

				for (int i = 0; i < ticks.length; i++) {
					if (ticks[i].major) {
						g2D.drawLine(pipeRect.x
						    + (pipeRect.width - tickLength) / 2,
						    pipeRect.y + arcRadius
						    + (int)(scaleHeight * (1. - ticks[i].proportional)),
						    pipeRect.x + (pipeRect.width + tickLength) / 2 - 1,
						    pipeRect.y + arcRadius
						    + (int)(scaleHeight * (1. - ticks[i].proportional)));
						g2D.drawString(ticks[i].text,
						    pipeRect.x + pipeRect.width + offsetX,
						    pipeRect.y + arcRadius
						    + (int)(scaleHeight * (1. - ticks[i].proportional))
						    + 5);

						//						System.out.println("TICK "+ticks[i].text+" "+ticks[i].proportional+" "+pipeRect.y +" "+ arcHeight +" "+ pipeRect.height+" " + arcHeight +" " + ticks[i].proportional);
						//  				System.out.println("VALUE "+rangedValue.getRelativeValue()+" "+fillHeight+" "+pipeRect.y+" "+ pipeRect.height +" "+ arcRadius +" "+ fillHeight);
					} else {
						g2D.drawLine(pipeRect.x
						    + (pipeRect.width - minorTickLength) / 2,
						    pipeRect.y + arcRadius
						    + (int)(scaleHeight * (1. - ticks[i].proportional)),
						    pipeRect.x + (pipeRect.width + minorTickLength) / 2
						    - 1,
						    pipeRect.y + arcRadius
						    + (int)(scaleHeight * (1. - ticks[i].proportional)));
					}
				}
			}

			return scaleImage;
		}

		private Tick[] getTicks(int height)
		{
			if (ticks == null) {
				//ticks = rangedValue.getRange().getTicks(height).toTicks(this); /* edit apikus */
				ticks = rangedValue.calculateTicks(height,this);
			}

			return ticks;
		}
		
		public String formatNumber(double x) { /* Use this method - edit apikus */
			return formatter.sprintf(x);
		}
	}

	private class ResizeListener extends ComponentAdapter
	{
		/**
		 * Listens for component events and resizes the component accordingly.
		 *
		 * @param e ComponentEvent
		 *
		 * @see java.awt.event.ComponentListener#componentResized(ComponentEvent)
		 */
		public void componentResized(ComponentEvent e)
		{
			Piper.this.setSize(getBounds().width, getBounds().height);
			renderer.clearBuffer(); /* edit apikus */
			repaint();
		}
	}

	private class ValueListener implements RangedValueListener
	{
		/**
		 * Listens for value changes and redraws the component accordingly.
		 *
		 * @param event RangedValueEvent
		 *
		 * @see com.cosylab.gui.components.range.RangedValueListener#valueChanged(RangedValueEvent)
		 */
		public void valueChanged(RangedValueEvent event)
		{
			if (event.isMinimumChanged() || event.isMaximumChanged()) {
				renderer.clearBuffer();

				if (!isEnabled()) {
					return;
				}
				repaint();
			}

			if (event.isValueChanged()) {
				if (!isEnabled()) {
					return;
				}
				repaint();
			}
		}
	}

	private RangedValuePolicy policy;
	private AbstractCustomizerPanel customizer;
	private PiperRenderer renderer;
	private PopupManager popupManager;
	private PrintfFormat formatter;
	//private RangedValue rangedValue; /* edit apikus */
	private RangedValueController rangedValue;
	private String format;
	private String title;
	private boolean titleVisible;
	private int titleMinimumFontSize;
	private int titleMaximumFontSize;
	private String units;
	private TiltHandler tiltHandler;
	private boolean enhanced = true;
	private boolean logarithmic;
	private boolean popupEnabled;
	private boolean resizable = true;
	private boolean stretch = false;
	private boolean tilting;
	private boolean tiltingEnabled;
	private double labelValue;
	private boolean unitsVisible = true;

	/**
	 * Creates a new Piper displaying the specified value.
	 *
	 * @param newValue double
	 */
	public Piper(double newValue)
	{
		super();

		renderer = new PiperRenderer();

		format = "%3.2f";
		formatter = new PrintfFormat(format);

		policy = new TrimValuePolicy();

		labelValue = newValue;

		rangedValue = new RangedValueController(); /* edit apikus */
		rangedValue.setValue(0, 1, newValue);
		rangedValue.setRange(new LinearRange());
		rangedValue.addPolicy(policy);
		rangedValue.addRangedValueListener(new ValueListener());

		addComponentListener(new ResizeListener());

		setBackground(ColorHelper.getCosyControl());
		setPreferredSize(new Dimension(150, 300));
		setMinimumSize(new Dimension(60, 100));
		setVisible(true);
		setOpaque(true);
	}

	/**
	 * Creates a new Piper with zero value.
	 */
	public Piper()
	{
		this(0);
	}

	/**
	 * Returns the Customizer intance for this displayer.
	 *
	 * @return the Customizer intance for this displayer
	 */
	public AbstractCustomizerPanel getCustomizer()
	{
		if (customizer == null) {
			customizer = AbstractCustomizerPanel.findCustomizer(this);
		}

		return customizer;
	}

	/**
	 * Triggers on and off the visually enhanced mode of the Piper. When
	 * enhanced, the liquid filling is painted with a gradient color and
	 * anti-aliasing is used.
	 *
	 * @param enhanced
	 */
	public void setEnhanced(boolean enhanced)
	{
		if (this.enhanced == enhanced) {
			return;
		}

		boolean oldEnhanced = this.enhanced;
		this.enhanced = enhanced;
		renderer.clearBuffer(); /* edit apikus */
		repaint();
		firePropertyChange("enhanced", oldEnhanced, enhanced);
	}

	/**
	 * Returns wether the component visualisation is enhanced.
	 *
	 * @return boolean
	 */
	public boolean isEnhanced()
	{
		return enhanced;
	}

	/**
	 * Sets the format of the value label text.
	 *
	 * @param format The format to set
	 */
	public void setFormat(String format)
	{
		String oldFormat = this.format;

		if (format == null && oldFormat == null
		    || format != null && format.equals(oldFormat)) {
			return;
		}

		this.format = format;
		formatter = new PrintfFormat(format);
		firePropertyChange("format", oldFormat, format);
		renderer.clearBuffer();
		if (!isEnabled()) {
			return;
		}
		
		repaint();
	}

	/**
	 * Returns the format of the value label text.
	 *
	 * @return String
	 */
	public String getFormat()
	{
		return format;
	}

	/**
	 * Sets the new maximum value displayed at the top of the scale of the
	 * Piper. This method delegates all logic to the RangedValue
	 * implementation.
	 *
	 * @param newMaximum double
	 */
	public void setMaximum(double newMaximum)
	{
		double oldMaximum = rangedValue.getMaximum();

		//if (oldMaximum == newMaximum) {
		//	return;
		//}
		rangedValue.setValue(rangedValue.getMinimum(), newMaximum,
		    rangedValue.getValue());
		firePropertyChange("maximum", oldMaximum, newMaximum);
	}

	/**
	 * Returns the maximum value displayed at the top of the scale of the
	 * Piper.
	 *
	 * @return double
	 */
	public double getMaximum()
	{
		return rangedValue.getMaximum();
	}

	/**
	 * Sets the new minimum value displayed at the bottom of the scale of the
	 * Piper. This method delegates all logic to the RangedValue
	 * implementation.
	 *
	 * @param newMinimum double
	 */
	public void setMinimum(double newMinimum)
	{
		double oldMinimum = rangedValue.getMinimum();

		//if (oldMinimum == newMinimum) {
		//	return;
		//}
		rangedValue.setValue(newMinimum, rangedValue.getMaximum(),
		    rangedValue.getValue());
		firePropertyChange("minimum", oldMinimum, newMinimum);
	}

	/**
	 * Gets the minimum value displayed at the bottom of the scale of the
	 * Piper.
	 *
	 * @return double
	 */
	public double getMinimum()
	{
		return rangedValue.getMinimum();
	}

	/**
	 * Switches between linear and logarithmic scales displayed. This method
	 * delegates all logic to the RangedValue implementation.
	 *
	 * @param logScale boolean
	 */
	public void setLogarithmic(boolean logScale)
	{
		if (logScale == logarithmic) {
			return;
		}

		logarithmic = logScale;

		if (logScale) {
			rangedValue.setRange(new LogarithmicRange());
		} else {
			rangedValue.setRange(new LinearRange());
		}

		firePropertyChange("logarithmic", !logScale, logScale);
	}

	/**
	 * Returns <code>true</code> if the display is in logarithmic scale.
	 *
	 * @return boolean
	 */
	public boolean isLogarithmic()
	{
		return logarithmic;
	}

	/**
	 * Sets both minimum and maximum displayed values displayed  at the bottom
	 * and at the top of the Pipers's scale. This method delegates all logic
	 * to the RangedValue implementation.
	 *
	 * @param newMax double
	 * @param newMin double
	 */
	public void setMaxMin(double newMax, double newMin)
	{
		double oldMaximum = rangedValue.getMaximum();
		double oldMinimum = rangedValue.getMinimum();

		if (oldMaximum == newMax && oldMinimum == newMin) {
			return;
		}

		rangedValue.setValue(newMin, newMax, rangedValue.getValue());
		firePropertyChange("maximum", oldMaximum, newMax);
		firePropertyChange("minimum", oldMinimum, newMin);
	}

	/**
	 * Enables or disables the popup capabilities of the component.
	 *
	 * @param b the component should bring up popup dialog on mouse click.
	 */
	public void setPopupEnabled(boolean b)
	{
		if (popupEnabled != b) {
			if (b) {
				addMouseListener(getPopupManager().getMouseHook());
			} else {
				removeMouseListener(getPopupManager().getMouseHook());
			}
		}

		popupEnabled = b;
		firePropertyChange("popupEnabled", !b, b);
	}

	/**
	 * Returns whether the user can trigger the component's popup menu through
	 * mouse click.
	 *
	 * @return true if popup is enabled
	 */
	public boolean isPopupEnabled()
	{
		return popupEnabled;
	}

	/**
	 * Sets the new value to be displayed. This method delegates all logic to
	 * the RangedValue implementation.
	 *
	 * @param newValue double
	 */
	public void setValue(double newValue)
	{
		double oldValue = rangedValue.getValue();

		labelValue = newValue;

		//if (oldValue == newValue) {
		//	return;
		//}
		rangedValue.setValue(newValue);

		if ((newValue > rangedValue.getMaximum()
		    || newValue < rangedValue.getMinimum()) && tiltingEnabled) {
			getTiltHandler().tilt();
		}

		firePropertyChange("value", oldValue, newValue);
	}

	/**
	 * Returns the value currently displayed.
	 *
	 * @return double
	 */
	public double getValue()
	{
		return labelValue;
	}

	/**
	 * For demonstration purposes.
	 *
	 * @param args String[]
	 */
	public static void main(String[] args)
	{
		Thread t = new Thread() {
				public void run()
				{
					javax.swing.JFrame frame = new javax.swing.JFrame();
					frame.addWindowListener(new WindowAdapter() {
							public void windowClosing(WindowEvent e)
							{
								System.exit(0);
							}
						});

					JPanel panel = new JPanel();
					frame.setContentPane(panel);
					frame.setSize(new java.awt.Dimension(100, 220));

					Piper sp = new Piper();
					sp.setEnhanced(true);
					sp.setPopupEnabled(true);
//					sp.setStretch(false);
					sp.setMinimum(-10.0);
					sp.setMaximum(20.0);
					sp.setValue(-10.0);

					panel.setLayout(new GridLayout());
					panel.add(sp);
					
					frame.setVisible(true);

					//double x = 5;
					//int i = 1;
					//long time1 = System.currentTimeMillis();
					//long time2 = 0;

					/*while (true) {
					    time2 = System.currentTimeMillis() - time1;
					    x = 0.5
					        + NoiseMaker.perlinNoise((double)time2 / 100.0 * Math.PI / 180);

					    //sp.setValue(x);
					    //sp.setValue(0.5);
					    sp.repaint();

					    try {
					        i++;

					        if (i > 50) {
					            i = 0;
					        }

					        sleep(10);
					    } catch (Exception e) {
					    }
					}*/
				}
			};

		t.run();
	}

	/**
	 * Overriden to implement visual notification of enabled state.
	 *
	 * @see java.awt.Component#setEnabled(boolean)
	 */
	public void setEnabled(boolean enabled)
	{
		super.setEnabled(enabled);

		renderer.clearBuffer();
		repaint();
	}

	/**
	 * Returns popup manager for adding popup actions.
	 *
	 * @see com.cosylab.gui.components.util.PopupManageable#getPopupManager()
	 */
	public PopupManager getPopupManager()
	{
		if (popupManager == null) {
			popupManager = new PopupManager(this, true);
			popupManager.addAction(new AbstractAction("Preferences...") {
				private static final long serialVersionUID = 1L;

					public void actionPerformed(ActionEvent e)
					{
						getCustomizer().showDialog();
					}
				});

			//JMenu submenu = new JMenu("Capture screen");

			ActionList al=new ActionList("Capture screen");

			al.addAction(new AbstractAction("To file...") {
				private static final long serialVersionUID = 1L;

					public void actionPerformed(ActionEvent e)
					{
						ScreenCapturer sc = new ScreenCapturer(Piper.this);
						sc.showScreenDialog();
					}
				});
			/*al.addAction(new AbstractAction("To clipboard...") {
				public void actionPerformed(ActionEvent e)
				{
					ScreenCapturer sc = new ScreenCapturer();
					sc.saveImageToSysClipBoard(Piper.this);
				}
			});*/
		   popupManager.addAction(al);

		}

		return popupManager;
	}

	/**
	 * Sets the resizable property of the component. If set, all components
	 * text is rendered in font size, adjusted to the size of the whole
	 * component.
	 *
	 * @param b
	 */
	public void setResizable(boolean b)
	{
		if (b == resizable) {
			return;
		}
		boolean oldResizable = resizable;

		resizable = b;

		firePropertyChange("resizable", oldResizable, resizable);
		renderer.clearBuffer();

		if (!isEnabled()) {
			return;
		}

		repaint();
	}

	/**
	 * Returns the resizable property of the component.
	 *
	 * @return resizable
	 */
	public boolean isResizable()
	{
		return resizable;
	}

	/**
	 * Sets the state to the component.
	 *
	 * @param state to set.
	 *
	 * @see com.cosylab.application.state.StateOriginator#setState(com.cosylab.application.state.State)
	 */
	public void setState(State state)
	{
		setFormat(state.getString("format", getFormat()));
		setMaximum(state.getDouble("maximum", getMaximum()));
		setMinimum(state.getDouble("minimum", getMinimum()));
		setTitle(state.getString("title", getTitle()));
		setUnits(state.getString("units", getUnits()));
		setLogarithmic(state.getBoolean("logarithmic", isLogarithmic()));
		setStretch(state.getBoolean("stretch", isStretch()));
		setEnhanced(state.getBoolean("enhanced", isEnhanced()));
		setResizable(state.getBoolean("resizable", isResizable()));
		setPopupEnabled(state.getBoolean("popupEnabled", isPopupEnabled()));
		setTiltingEnabled(state.getBoolean("tiltingEnabled", isTiltingEnabled()));
	}

	/**
	 * Returns the current state of the component.
	 *
	 * @return current state.
	 *
	 * @see com.cosylab.application.state.StateOriginator#getState()
	 */
	public State getState()
	{
		State state = StateFactory.createState();

		state.putString("format", getFormat());
		state.putDouble("graphMax", getMaximum());
		state.putDouble("graphMin", getMinimum());
		state.putString("title", getTitle());
		state.putString("units", getUnits());
		state.putBoolean("logarithmic", isLogarithmic());
		state.putBoolean("stretch", isStretch());
		state.putBoolean("enhanced", isEnhanced());
		state.putBoolean("resizable", isResizable());
		state.putBoolean("popupEnabled", isPopupEnabled());
		state.putBoolean("tiltingEnabled", isTiltingEnabled());

		return state;
	}

	/**
	 * Sets if the pipe should strech all across the width of the component or
	 * not.
	 *
	 * @param b
	 */
	public void setStretch(boolean b)
	{
		if (b == stretch) {
			return;
		}

		boolean oldStretch = stretch;
		stretch = b;
		renderer.clearBuffer();
		firePropertyChange("stretch", oldStretch, stretch);

		if (!isEnabled()) {
			return;
		}
		repaint();
	}

	/**
	 * Returns wether the pipe should strech all across the width of the
	 * component or not.
	 *
	 * @return strech
	 */
	public boolean isStretch()
	{
		return stretch;
	}

	/**
	 * Sets the tilitng enabled property.
	 *
	 * @param b whether the component should tilt when value is out of bounds.
	 */
	public void setTiltingEnabled(boolean b)
	{
		if (tiltingEnabled == b) {
			return;
		}

		tiltingEnabled = b;
		firePropertyChange("tiltingEnabled", !b, b);
	}

	/**
	 * Returns whether the component should indicate value out of bounds
	 * condition by visually tilting its border.
	 *
	 * @return boolean
	 */
	public boolean isTiltingEnabled()
	{
		return tiltingEnabled;
	}

	/**
	 * Sets the title to be rendered on top of the component when not
	 * stretched. Can be null in which case nothing is rendered.
	 *
	 * @param string title
	 */
	public void setTitle(String string)
	{
		if (string == null && title == null
		    || string != null && string.equals(title)) {
			return;
		}

		String oldTitle = title;

		title = string;
		firePropertyChange("title", oldTitle, title);

		if (!isEnabled()) {
			return;
		}

		repaint();
	}

	/**
	 * Returns the current title of the component or null is none is set.
	 *
	 * @return title
	 */
	public String getTitle()
	{
		return title;
	}

	/**
	 * Sets the units to be displayed in the value label. If null, nothing is
	 * rendered.
	 *
	 * @param units The units to set, can be null
	 */
	public void setUnits(String units)
	{
		if (this.units == null && units == null
		    || units != null && units.equals(this.units)) {
			return;
		}

		String oldUnits = this.units;
		this.units = units;
		firePropertyChange("units", oldUnits, units);
		
		if (!isEnabled()) {
			return;
		}
		repaint();
	}

	/**
	 * Returns the units to be displayed in the value label or null is none is
	 * set.
	 *
	 * @return String
	 */
	public String getUnits()
	{
		return units;
	}

	/**
	 * Returns the rangedValue object containing all the value logic.
	 *
	 * @return RangedValue
	 */
	
//	protected RangedValue getRangedValue()
//	{
//		return rangedValue;
//	}
	
	protected RangedValueController getRangedValue()
	{
		return rangedValue;
	}

	/**
	 * This method has been overriden to implement custom painting of the
	 * Piper.
	 *
	 * @param g Graphics
	 *
	 * @see JComponent#paintComponent(java.awt.Graphics)
	 */
	protected void paintComponent(Graphics g)
	{
		if (isOpaque()) {
			g.setColor(getBackground());
			g.fillRect(0, 0, getWidth(), getHeight());
		}

		Graphics2D g2D = (Graphics2D)g;
		g2D.setBackground(getBackground());
		renderer.prepareGraphics(g2D);

		renderer.setPipeRect(g2D);

		renderer.paintPipe(g2D);
		renderer.paintFill(g2D);
		renderer.paintScale(g2D);

		renderer.paintLabel(g2D);
		if (titleVisible)
			renderer.paintTitle(g2D);
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @return DOCUMENT ME!
	 */
	private TiltHandler getTiltHandler()
	{
		if (tiltHandler == null) {
			tiltHandler = new TiltHandler();
		}

		return tiltHandler;
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @param p DOCUMENT ME!
	 */
	public void setValuePolicy(RangedValuePolicy p)
	{
		rangedValue.removePolicy(policy);
		this.policy = p;
		rangedValue.addPolicy(p);
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @return
	 */
	public RangedValuePolicy getValuePolicy()
	{
		return policy;
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @return
	 */
	public boolean isUnitsVisible()
	{
		return unitsVisible;
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @param b
	 */
	public void setUnitsVisible(boolean b)
	{
		if (unitsVisible == b) {
			return;
		}

		boolean oldB = unitsVisible;
		unitsVisible = b;
		firePropertyChange("unitsVisible", oldB, b);
		renderer.clearBuffer(); /* edit apikus */
		repaint();
	}

	public boolean isTitleVisible() {
		return titleVisible;
	}

	public void setTitleVisible(boolean titleVisible) {
		if (this.titleVisible == titleVisible) {
			return;
		}

		this.titleVisible = titleVisible;
		firePropertyChange("titleVisible", !titleVisible, titleVisible);
		repaint();
	}

	public int getTitleMaximumFontSize() {
		return titleMaximumFontSize;
	}

	public void setTitleMaximumFontSize(int titleMaximumFontSize) {

		if (this.titleMaximumFontSize == titleMaximumFontSize) return;
		int i = this.titleMaximumFontSize;
		this.titleMaximumFontSize = titleMaximumFontSize;
		firePropertyChange("titleMaximumFontSize", i, titleMaximumFontSize);
	}

	public int getTitleMinimumFontSize() {
		return titleMinimumFontSize;
	}

	public void setTitleMinimumFontSize(int titleMinimumFontSize) {
		if (this.titleMinimumFontSize == titleMinimumFontSize) return;
		int i = this.titleMinimumFontSize;
		this.titleMinimumFontSize = titleMinimumFontSize;
		firePropertyChange("titleMinimumFontSize", i, titleMinimumFontSize);
	}
}

/* __oOo__ */
