/*
 * Copyright (c) 2003 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.components.gauger;

import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.geom.Point2D;


/**
 * This transform defines shape of horizontal linear gauger. It imposes
 * constraints on minimum and maximum sizes and defines borders, as well as
 * position of value label.
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public class LinearHorizontalTransform extends ScaleTransform
{
	/** Maximum height to which the scale will grow */
	private static final int MAXIMUM_HEIGHT = 80;

	/** Minimum height to which the scale will shrink */
	private static final int MINIMUM_HEIGHT = 5;

	/** Horizontal offset of scale from border */
	private static final double RELATIVE_X_OFFSET = 0.05;

	/** Vertical offset of scale from border */
	private static final double RELATIVE_Y_OFFSET = 0.15;

	/**	 */
	private int width = 0;

	/**	 */
	private int height = 0;

	/**	 */
	private double xOffset = 0;

	/**	 */
	private double yOffset = 0;

	/**	 */
	private double scaleWidth;

	/**	 */
	private int tickOffset = 0;

	/**
	 * Sets the dimensions of this scale and calculates internal parameters.
	 *
	 * @param w int Gauger width.
	 * @param h int Gauger height.
	 * @param marginX int Horizontal margin from border in pixels.
	 * @param marginY int Vertical margin from border in pixels.
	 * @param tickOffset int Minimum offset of first and last tick from scale
	 *        in pixels.
	 *
	 * @see ScaleTransform#setParameters(int, int, int, int)
	 */
	public void setParameters(int w, int h, int marginX, int marginY,
	    int tickOffset)
	{
		this.tickOffset = tickOffset;

		width = w - 2 * (marginX + tickOffset);
		height = h - 2 * marginY;

		//		width = w - (int) (2 * w * RELATIVE_X_OFFSET) - 5;
		//		height = h - (int) (2 * h * RELATIVE_Y_OFFSET) - 5;
		height = Math.min(height, MAXIMUM_HEIGHT);
		height = Math.max(height, MINIMUM_HEIGHT);

		xOffset = (w - width) / 2.0;
		yOffset = (h - height - marginY) / 2.0;

		scaleWidth = Math.abs(width - 2 * tickOffset);

		Rectangle outline = new Rectangle((int)xOffset, (int)yOffset, width,
			    height);
		addSegment(new RectangleSegment(outline));

		setLabelPosition(w / 2, (int)(h - (h - height - yOffset) / 2));
	}

	/**
	 * Converts point in scale space to point in gauger space. Conversion is
	 * linear.
	 *
	 * @param scaleSpace PointDouble Point in scale space.
	 * @param cartesianSpace Point Point in gauger space.
	 *
	 * @see ScaleTransform#mapUVtoXY(PointDouble, Point)
	 */
	public void mapUVtoXY(Point2D scaleSpace, Point2D cartesianSpace)
	{
		double x = (xOffset + scaleWidth(0.0) * scaleSpace.getX() + tickOffset);
		double y = (yOffset + (height - 1) * (1 - scaleSpace.getY()) - 1);
		cartesianSpace.setLocation(x, y);
	}

	/**
	 * Not implemented.
	 *
	 * @param cartesianSpace Point
	 * @param scaleScape PointDouble
	 *
	 * @see ScaleTransform#maxXYtoUV(Point, PointDouble)
	 */
	public void mapXYtoUV(Point2D cartesianSpace, Point2D scaleScape)
	{
	}

	/**
	 * Returns width of scale at relative vertical position. Returns constant
	 * width.
	 *
	 * @param v double Relative vertical position in range 0.0 to 1.0
	 *
	 * @return double Width in pixels.
	 *
	 * @see ScaleTransform#scaleWidth(double)
	 */
	public double scaleWidth(double v)
	{
		return Math.abs(width - 2 * tickOffset);
	}

	/**
	 * Returns scale height at relative horizontal position. Returns constant
	 * height.
	 *
	 * @param u double Relative horizontal position in range 0.0 to 1.0.
	 *
	 * @return double Heightin pixels.
	 *
	 * @see ScaleTransform#scaleHeight(double)
	 */
	public double scaleHeight(double u)
	{
		return height;
	}

	/**
	 * Calculates tick label width when rendered.
	 *
	 * @param g Graphics Context to render to.
	 * @param x double Relative value of tick position.
	 * @param text String Text to render.
	 *
	 * @return int Width in pixels.
	 *
	 * @see com.cosylab.gui.components.gauger.ScaleTransform#measureTick(Graphics,
	 *      double, String)
	 */
	public int measureTick(Graphics g, double x, String text)
	{
		return g.getFontMetrics().stringWidth(text);
	}
}

/* __oOo__ */
