/*
 * Copyright (c) 2003 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.components.gauger;

import java.awt.Point;
import java.awt.geom.Point2D;


/**
 * Class representing polar point defined by radius and angle.
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public class PolarPoint
{
	private static final double DEG_TO_RAD = Math.PI / 180.0;
	private static final double RAD_TO_DEG = 180.0 / Math.PI;

	private double r;

	private double phi;

	/**
	 * Creates a new PolarPoint object.
	 */
	public PolarPoint()
	{
	}

	/**
	 * Creates a new PolarPoint object.
	 *
	 * @param r Radius parameter.
	 * @param phi Angle parameter in radians.
	 */
	public PolarPoint(double r, double phi)
	{
		this.r = r;
		this.phi = phi;
	}

	/**
	 * Converts degrees to radians.
	 *
	 * @param angle angle in degrees.
	 *
	 * @return Angle in radians.
	 */
	public static final double degToRad(double angle)
	{
		return angle * DEG_TO_RAD;
	}

	/**
	 * Converts radians to degrees.
	 *
	 * @param angle in radians.
	 *
	 * @return angle in degrees.
	 */
	public static final double radToDeg(double angle)
	{
		return angle * RAD_TO_DEG;
	}

	/**
	 * Converts polar point to cartesian coordinates.
	 *
	 * @param r Radius parameter.
	 * @param phi Angle parameter.
	 *
	 * @return Point in cartesian space.
	 */
	public static final Point2D polarToCartesian(double r, double phi)
	{
		double phiInRadians = degToRad(phi);
		double x = (r * Math.cos(phiInRadians));
		double y = -(r * Math.sin(phiInRadians));

		return new Point2D.Double(x, y);
	}

	/**
	 * Converts polar point to cartesian coordinates. Result will be placed in
	 * point specified. This method will not create new instance of point to
	 * hold the result, but the provided instance will be returned.
	 *
	 * @param p Point which will hold the result.
	 * @param r Radius parameter.
	 * @param phi Angle parameter.
	 *
	 * @return p parameter.
	 */
	public static final Point2D polarToCartesian(Point2D p, double r, double phi)
	{
		double rPhi = degToRad(phi);
		p.setLocation((r * Math.cos(rPhi)), -(r * Math.sin(rPhi)));

		return p;
	}

	/**
	 * Converts polar point to cartesian coordinates.
	 *
	 * @param pp Point in polar coordinates.
	 * @param p Point in cartesian coordinates.
	 */
	public static final void polarToCartesian(PolarPoint pp, Point2D p)
	{
		double rPhi = degToRad(pp.phi);

		p.setLocation(pp.r * Math.cos(rPhi), pp.r * Math.sin(rPhi));
	}

	/**
	 * Converts this point to cartesian coordinate representation.
	 *
	 * @return Point in cartesian coordinates.
	 */
	public Point2D toPoint()
	{
		return polarToCartesian(r, phi);
	}

	/**
	 * Converts this point to cartesian coordinates and puts result in provided
	 * instance of point.
	 *
	 * @param p Object that will hold the result.
	 *
	 * @return Resulting point, which is the same object as one passed as 
	 * 		   parameter.
	 */
	public Point2D toPoint(Point p)
	{
		return polarToCartesian(p, r, phi);
	}
}

/* __oOo__ */
