/*
 * Copyright (c) 2003 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.components.gauger;

/**
 * Factory class that creates transform implementations based on dimensions of
 * the gauger.
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public class ScaleTransformFactory
{
	/** Scale shape */
	public static final short AUTOMATIC_SHAPE = 0;

	/** Horizontal linear scale */
	public static final short HORIZONTAL_LINEAR_SHAPE = 1;

	/** Horizontal polar scale */
	public static final short HORIZONTAL_CIRCULAR_SHAPE = 2;

	/** Circular polar scale */
	public static final short FULL_CIRCULAR_SHAPE = 3;

	/** Vertical polar scale */
	public static final short VERTICAL_CIRCULAR_SHAPE = 4;

	/** Vertical linear scale */
	public static final short VERTICAL_LINEAR_SHAPE = 5;

	/** Constants used when changing scale shapes */
	private static final double FROM_LINEAR_TO_HALF_CIRCULAR = 3.5;

	/** */
	private static final double FROM_HALF_CIRCULAR_TO_FULL_CIRCULAR = 1.7;

	/** */
	private static final double FROM_FULL_CIRCULAR_TO_HALF_CIRCULAR = 0.65;

	/** */
	private static final int HEIGHT_SMALLER_THAN_THIS_ALWAYS_LINEAR = 95;

	/** */
	private static final int WIDTH_SMALLER_THAN_THIS_ALWAYS_LINEAR = 120;

	/**
	 * Determines the optimal renderer for this scale, given the width and
	 * height of the area.
	 *
	 * @param w int
	 * @param h int
	 *
	 * @return int
	 */
	protected static int getOptimalRenderer(int w, int h)
	{
		double ratio = (1.0 * w) / h;
		short shape;

		if(ratio > FROM_LINEAR_TO_HALF_CIRCULAR) {
			shape = HORIZONTAL_LINEAR_SHAPE;
		} else if(ratio > FROM_HALF_CIRCULAR_TO_FULL_CIRCULAR) {
			shape = HORIZONTAL_CIRCULAR_SHAPE;
		} else if(ratio > FROM_FULL_CIRCULAR_TO_HALF_CIRCULAR) {
			shape = FULL_CIRCULAR_SHAPE;
		} else if(ratio > 1 / FROM_LINEAR_TO_HALF_CIRCULAR) {
			shape = VERTICAL_CIRCULAR_SHAPE;
		} else {
			shape = VERTICAL_LINEAR_SHAPE;
		}

		if((w < 1.5 * WIDTH_SMALLER_THAN_THIS_ALWAYS_LINEAR)
		    && (h < 1.5 * HEIGHT_SMALLER_THAN_THIS_ALWAYS_LINEAR)) {
			shape = HORIZONTAL_LINEAR_SHAPE;
		}

		if(w < WIDTH_SMALLER_THAN_THIS_ALWAYS_LINEAR) {
			shape = VERTICAL_LINEAR_SHAPE;
		}

		if(h < HEIGHT_SMALLER_THAN_THIS_ALWAYS_LINEAR) {
			shape = HORIZONTAL_LINEAR_SHAPE;
		}

		return shape;
	}

	/**
	 * Creates a new instance of transform that is best suited for given size.
	 *
	 * @param w int
	 * @param h int
	 *
	 * @return ScaleTransform
	 */
	public static ScaleTransform createTransform(int w, int h)
	{
		int newValue = getOptimalRenderer(w, h);

		switch(newValue) {
		case HORIZONTAL_LINEAR_SHAPE:
			return new LinearHorizontalTransform();

		case HORIZONTAL_CIRCULAR_SHAPE:
			return new PolarHorizontalTransform();

		case FULL_CIRCULAR_SHAPE:
			return new PolarFullTransform();

		case VERTICAL_CIRCULAR_SHAPE:
			return new PolarVerticalTransform();

		case VERTICAL_LINEAR_SHAPE:
			return new LinearVerticalTransform();

		default:
			return new LinearVerticalTransform();
		}
	}
}

/* __oOo__ */
