/*
 * Copyright (c) 2003 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.components.gauger;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.geom.Line2D;

import com.cosylab.gui.components.Gauger;


/**
 * Component defining and rendering the needle.
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public class ThinNeedle implements ActiveElement
{
	/* Scale appearance */

	/** Relative length of needle as fraction of scale space height */
	private final static double RELATIVE_NEEDLE_SIZE = 0.6;
	private Gauger gauger;
	private boolean changeSignificant;
	private Point p;
	private double angle;
	private Rectangle bounds = new Rectangle(0, 0, 0, 0);
	private double currentPosition = 0.5;
	protected ScaleTransform transform = null;
	//private Line poly = new Polygon(new int[3], new int[3], 3);
	
	private Point p1 = new Point(0, 0);
	private Point p2 = new Point(0, 0);
	
	private Line2D line = new Line2D.Double();

	/**
	 * Needle constructor comment.
	 *
	 * @param gauger on which the needle will be rendered.
	 */
	public ThinNeedle(Gauger gauger)
	{
		p = new Point();
		angle = 0;
		this.gauger = gauger;
		changeSignificant = true;
	}

	/**
	 * Updates the polygon representation of the needle.
	 */
	protected synchronized void calculatePolygon()
	{
		double sin = Math.sin(angle-Math.PI/2.0);
		double cos = Math.cos(angle-Math.PI/2.0);

		double x1 = (int)(p1.x * sin - p1.y * cos + p.x);
		double y1 = (int)(p1.x * cos + p1.y * sin + p.y);

		double x2 = (int)(p2.x * sin - p2.y * cos + p.x);
		double y2 = (int)(p2.x * cos + p2.y * sin + p.y);
		
		line.setLine(x1, y1, x2, y2);
		
		bounds = line.getBounds();
		bounds.grow(3, 3);
	}

	private synchronized final void invalidate()
	{
		double size = transform.scaleHeight(0.5) * RELATIVE_NEEDLE_SIZE;
		p2.x = 0;
		p2.y = (int)size;
	}

	/**
	 * Returns bounding rectangle of needle.
	 *
	 * @return Bounding rectangle of needle.
	 */
	public Rectangle getBounds()
	{
		return bounds;
	}

	/**
	 * Returns true if needle has moved enough to require immediate redraw.
	 *
	 * @return True if needle should be redrawn immediatelly.
	 */
	public boolean isChangeSignificant()
	{
		return changeSignificant;
	}

	protected void updatePosition(double x)
	{
		if((transform != null)) {
			transform.mapUVtoXY(x, 0, p);
			angle = transform.getAngle(x);
			calculatePolygon();
		}
	}

	/**
	 * Sets position of needle in scale space coordinates. Value is relative
	 * position between 0.0 and 1.0 inclusively.
	 *
	 * @param proportionalPosition New position of needle.
	 */
	public void setPosition(double proportionalPosition)
	{
		updatePosition(proportionalPosition);

		if(changeSignificant) {
			currentPosition = proportionalPosition;
		}
	}

	/**
	 * Renders needle.
	 *
	 * @see ActiveElement#render(Graphics2D)
	 */
	public synchronized void render(Graphics2D g)
	{
		g.setColor(gauger.getNeedleColor());

		g.setColor(Color.BLACK);
		BasicStroke bs = new BasicStroke(2.0f);
		g.setStroke(bs);
		g.draw(line);

	}

	/**
	 * Sets transform defining the gauger. This transform is used to render the
	 * needle at appropriate position and angle.
	 *
	 * @see ActiveElement#setTransform(ScaleTransform)
	 */
	public void setTransform(ScaleTransform transform)
	{
		this.transform = transform;
		invalidate();
		updatePosition(currentPosition);
	}
}

/* __oOo__ */
