/*
 * Copyright (c) 2003 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.components.introspection;

import java.util.HashMap;

import com.cosylab.gui.components.customizer.PropertiesTableCustomizer;
import com.cosylab.introspection.DataFormatter;

import javax.swing.table.AbstractTableModel;


/**
 * A table model, which help edit arbitrary set of properties. Each property
 * has arbitrary name, value and type. In connection with
 * <code>PropertiesTable</code> this model is capabel of JavaBeans editing
 * based on introspection.
 *
 * @author <a href="mailto:miha.kadunc@cosylab.com">Miha Kadunc</a>
 * @version $id$
 */
public class PropertiesTableModel extends AbstractTableModel
{
	private static final long serialVersionUID = 1L;
	private HashMap<String,Integer> namesToIndex;
	private String[] names = null;
	private Object[] values = null;
	private Class[] types = null;
	private boolean typeShowing = false;
	int valCol = 1;
	private int rows = 0;
	private boolean editable= true;

	/**
	 * Constructor for PropertiesTableModel.
	 */
	public PropertiesTableModel()
	{
		super();
		names = new String[0];
		types = new Class[0];
		values = new Object[0];
		rows = 0;
		namesToIndex= new HashMap<String,Integer>(16);
	}

	/**
	 * Creates new model with specified property names, types and values.
	 *
	 * @param names String[]
	 * @param types Class[]
	 * @param values Object[]
	 */
	public PropertiesTableModel(String[] names, Class[] types, Object[] values)
	{
		this();
		setPropertyNames(names);
		setPropertyTypes(types);
		setPropertyValues(values);
	}

	/*
	 * @see TableModel#getRowCount()
	 */
	public int getRowCount()
	{
		return rows;
	}

	/**
	 * Sets the parameter names
	 *
	 * @param names String[]
	 */
	public synchronized void setPropertyNames(String[] names)
	{
		if (names == null) {
			this.names = new String[0];
		} else {
			this.names = names;
			namesToIndex.clear();
			for (int i = 0; i < names.length; i++) {
				namesToIndex.put(names[i],i);
			}
		}

		rows = this.names.length;

		if (values.length != rows) {
			values = new Object[rows];
		}

		if (types.length != rows) {
			types = new Class[rows];
		}
	}

	/**
	 * Sets the parameters
	 *
	 * @param values Object[]
	 */
	public synchronized void setPropertyValues(Object[] values)
	{
		if (values == null || values.length != rows) {
			this.values = new Object[rows];
		} else {
			this.values = values;
		}
	}

	/**
	 * Sets the parameter types
	 *
	 * @param types Class[]
	 */
	public synchronized void setPropertyTypes(Class[] types)
	{
		if (types == null || types.length != rows) {
			this.types = new Class[rows];
		} else {
			this.types = types;
		}
	}

	/*
	 * @see TableModel#getColumnCount()
	 */
	public int getColumnCount()
	{
		return valCol + 1;
	}

	/*
	 * @see TableModel#getValueAt(int, int)
	 */
	public synchronized Object getValueAt(int rowIndex, int columnIndex)
	{
		if (rowIndex >= getRowCount() || columnIndex >= getColumnCount()) {
			return null;
		}

		if (columnIndex == 0) {
			return names[rowIndex];
		} else if (columnIndex == 1 && typeShowing) {
			return DataFormatter.toString(types[rowIndex]);
		} else if (columnIndex == valCol) {
			return values[rowIndex];
		} else {
			return null;
		}
	}

	/*
	 * @see AbstractTableModel#findColumn(String)
	 */
	public int findColumn(String columnName)
	{
		if (columnName == "property") {
			return 0;
		}

		if (columnName == "type" && typeShowing) {
			return 1;
		}

		if (columnName == "value") {
			return valCol;
		} else {
			return -1;
		}
	}

	/*
	 * @see TableModel#getColumnName(int)
	 */
	public String getColumnName(int columnIndex)
	{
		if (columnIndex == 0) {
			return "property";
		}

		if (columnIndex == valCol) {
			return "value";
		}

		if (columnIndex == 1 && typeShowing) {
			return "type";
		} else {
			return super.getColumnName(columnIndex);
		}
	}

	/*
	 * @see TableModel#isCellEditable(int, int)
	 */
	public boolean isCellEditable(int rowIndex, int columnIndex)
	{
		return (columnIndex == valCol) && isEditable();
	}

	/*
	 * @see TableModel#setValueAt(Object, int, int)
	 */
	public synchronized void setValueAt(Object aValue, int rowIndex,
	    int columnIndex)
	{
		if (rowIndex >= getRowCount() || columnIndex >= getColumnCount()) {
			return;
		}

		if (columnIndex == 0 && aValue instanceof String) {
			names[rowIndex] = (String)aValue;
		} else if (columnIndex == valCol) {
			if (values[rowIndex] != aValue) {
				Object oldValue = values[rowIndex];

				try {
					values[rowIndex] = EditorsHelper.convert(aValue,
						    types[rowIndex]);
					fireTableCellUpdated(rowIndex, typeShowing ? 2 : 1);
				} catch (Exception e) {
					e.printStackTrace();
					values[rowIndex] = oldValue;
				}
			}
		} else if (columnIndex == 1 && typeShowing && aValue instanceof Class) {
			types[rowIndex] = (Class)aValue;
		}

		return;
	}

	synchronized void setValueForProperty(String name, Object value)
	{
		Integer i = namesToIndex.get(name);

		if (i==null) {
			return;
		}

		values[i] = value;

		//System.out.println(name+" "+names[i]+" "+i+" "+value);
		fireTableCellUpdated(i, valCol);
	}

	/**
	 * Returns array with names of properties.
	 *
	 * @return String[]
	 */
	public String[] getPropertyNames()
	{
		return names;
	}

	/**
	 * Returns array with types of properties.
	 *
	 * @return Class[]
	 */
	public Class[] getPropertyTypes()
	{
		return types;
	}

	/**
	 * Returns array with values of properties.
	 *
	 * @return Object[]
	 */
	public Object[] getPropertyValues()
	{
		return values;
	}

	/**
	 * Returns <code>true</code> if table shows additional column with property
	 * types.
	 *
	 * @return boolean
	 */
	public boolean isTypeShowing()
	{
		return typeShowing;
	}

	/**
	 * Enables table to show additional column with property types.
	 *
	 * @param typeShowing The typeShowing to set
	 */
	public void setTypeShowing(boolean typeShowing)
	{
		this.typeShowing = typeShowing;
		valCol = typeShowing ? 2 : 1;
	}

	/**
	 * Returns type of property in specified row.
	 *
	 * @param row
	 *
	 * @return
	 */
	public Class getTypeForRow(int row)
	{
		return types[row];
	}
	
	public boolean containsProperty(String name) {
		return namesToIndex.containsKey(name);
	}
	
	public boolean isEditable() {
		return editable;
	}

	public void setEditable(boolean editable) {
		this.editable = editable;
	}
}

/* __oOo__ */
