/*
 * Copyright (c) 2004 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.components.ledder;

import java.awt.Color;

import java.util.ArrayList;


/**
 * Default implementation of <code>BitDescriptor</code>.
 *
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 */
public class DefaultBitDescriptor implements BitDescriptor
{
	class BitDescription
	{
		/**
		 * Creates a new BitDescription object.
		 *
		 * @param on DOCUMENT ME!
		 * @param off DOCUMENT ME!
		 * @param desc DOCUMENT ME!
		 */
		public BitDescription(Color on, Color off, String desc)
		{
			this.on = on;
			this.off = off;
			this.desc = desc;
		}

		/** ON bit color. */
		public Color on;

		/** OFF bit color. */
		public Color off;

		/** Bit description. */
		public String desc;
	}

	private BitDescription defaultBitDesc = new BitDescription(Color.GREEN,
		    Color.GRAY, null);
	private ArrayList bits = new ArrayList(32);

	/**
	 * Creates new BitDescriptor with default colors set to <code>GREEN</code>
	 * for on and <code>GRAY</code> for off.
	 */
	public DefaultBitDescriptor()
	{
		super();
	}

	/**
	 * Returns the number of led descriptions.
	 *
	 * @return int
	 */
	public int size()
	{
		return bits.size();
	}

	protected synchronized BitDescription getBitDescription(int i)
	{
		if (i > -1 && i < bits.size()) {
			return (BitDescription)bits.get(i);
		}

		return null;
	}

	/**
	 * Appends new bit descriptions to the end of  the descriptions lists.
	 *
	 * @param label String
	 * @param colorOn Color
	 * @param colorOff Color
	 */
	public synchronized void add(String label, Color colorOn, Color colorOff)
	{
		bits.add(new BitDescription(colorOn, colorOff, label));
	}

	/**
	 * Removes bit description at the specified index from the descriptor. If
	 * index is invalid or the model is empty, this method does nothing. Order
	 * of the remaining leds will not change.
	 *
	 * @param index int
	 */
	public synchronized void remove(int index)
	{
		if ((index < 0) || (index >= size())) {
			return;
		}

		bits.remove(index);
	}

	/**
	 * Removes all bit descriptions from the descriptor.
	 */
	public synchronized void clear()
	{
		bits.clear();
	}

	/**
	 * Returns the default description of the bit at the specified index or
	 * default string if no description is defined.
	 *
	 * @param index int
	 *
	 * @return String Description of led.
	 */
	public String getDescription(int index)
	{
		BitDescription bd = getBitDescription(index);

		if (bd != null && bd.desc != null) {
			return bd.desc;
		}

		StringBuffer sb = new StringBuffer();
		sb.append("<bit");
		sb.append(index);
		sb.append('>');

		return sb.toString();
	}

	/**
	 * Sets description for the bit at the specified index. If the index
	 * greater than size, method does nothing.
	 *
	 * @param index int
	 * @param description String
	 */
	public void setDescription(int index, String description)
	{
		BitDescription bd = getBitDescription(index);

		if (bd != null) {
			bd.desc = description;
		}
	}

	/**
	 * Returns the default color representing ON state for bit at specified
	 * index or default value if color is not defined.
	 *
	 * @param index int
	 *
	 * @return Color Color of bit or default value
	 */
	public Color getColorWhenOn(int index)
	{
		BitDescription bd = getBitDescription(index);

		if (bd != null && bd.on != null) {
			return bd.on;
		}

		return defaultBitDesc.on;
	}

	/**
	 * Returns the default color representing ON state for bit used when no
	 * specific description is defined.
	 *
	 * @return Color default value
	 */
	public Color getDefaultColorWhenOn()
	{
		return defaultBitDesc.on;
	}

	/**
	 * Sets new color indicating the ON state for the specified bit index. If
	 * the  index is invalid, method does nothing.
	 *
	 * @param index int
	 * @param color Color
	 */
	public void setColorWhenOn(int index, Color color)
	{
		BitDescription bd = getBitDescription(index);

		if (bd != null) {
			bd.on = color;
		}
	}

	/**
	 * Sets default color indicating the ON state.
	 *
	 * @param color Color
	 */
	public void setDefaultColorWhenOn(Color color)
	{
		defaultBitDesc.on = color;
	}

	/**
	 * Returns the color representing OFF state for bit at specified index or
	 * default OFF value if there is no description.
	 *
	 * @param index int
	 *
	 * @return Color
	 */
	public Color getColorWhenOff(int index)
	{
		BitDescription bd = getBitDescription(index);

		if (bd != null && bd.off!=null) {
			return bd.off;
		}

		return defaultBitDesc.off;
	}

	/**
	 * Returns the default color representing OFF state for bit used when no
	 * specific description is defined.
	 *
	 * @return Color
	 */
	public Color getDefaultColorWhenOff()
	{
		return defaultBitDesc.off;
	}

	/**
	 * Sets new color indicating OFF state for the specified bit. If the index
	 * is larger or equal to size, method does nothing.
	 *
	 * @param index int
	 * @param color Color
	 */
	public void setColorWhenOff(int index, Color color)
	{
		BitDescription bd = getBitDescription(index);

		if (bd != null) {
			bd.off = color;
		}
	}

	/**
	 * Sets default color indicating OFF state.
	 *
	 * @param color Color
	 */
	public void setDefaultColorWhenOff(Color color)
	{
		defaultBitDesc.off = color;
	}
}

/* __oOo__ */
