/*
 * @@COPYRIGHT@@
 */
package com.cosylab.gui.components.numberfield;

import javax.swing.text.AttributeSet;
import javax.swing.text.BadLocationException;
import javax.swing.text.PlainDocument;

/**
 * Abstract base for implementation of number parsers, hiding the details of 
 * the <code>javax.swing.text.Document</code> interface. This class handles 
 * parsing of the string and caching last value. It ensures that only 
 * acceptable numbers are entered  (within the numeric range of the desired 
 * class).
 * 
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public abstract class AbstractNumberDocument extends PlainDocument {

	private Number value;
    protected Number min=null;
    protected Number max=null;
    private boolean updated = false;

	/**
	 * Constructor for AbstractNumberDocument.
	 */
	public AbstractNumberDocument() {
		super();
	}

	/**
	 * Inserts the string at str parameter into existing string, but only if
	 * the result is valid number.
	 * 
	 * @param offs int
	 * @param str String
	 * @param attributes AttributeSet
	 * @throws BadLocationException
	 * @see Document#insertString(int, String, AttributeSet)
	 */
	public void insertString(int offs, String str, AttributeSet attributes)
		throws BadLocationException {
		
		updated = true;
		
		String currentText = getText(0, getLength());
		String beforeOffset = currentText.substring(0, offs);
		String afterOffset = currentText.substring(offs, currentText.length());

		String proposedResult = beforeOffset + str + afterOffset;

		Number newValue = parseNumber(proposedResult);

		if (doChecks(proposedResult,newValue)) {
            value = newValue;
			super.insertString(offs, str, attributes);
		}

	}
    
    private boolean doChecks(String proposedResult, Number newValue) {
        return (newValue != null && checkBounds(newValue)) || proposedResult.equals("") || proposedResult.equals("-");
    }
    
    /*
     * Returnd true if OK.
     */
    protected boolean checkBounds(Number newVal) {
        if (newVal==null) return true;
        if (min==null || min.doubleValue()<=newVal.doubleValue()) {
            if (max==null || max.doubleValue()>=newVal.doubleValue()) {
                return true;
            }
        }
        return false;
    }

	/**
	 * Removes the string specified by offs and len, but only if the resulting
	 * string is a valid number.
	 * 
	 * @param offs int
	 * @param len int
	 * @throws BadLocationException
	 * @see Document#remove(int, int)
	 */
	public void remove(int offs, int len) throws BadLocationException {
		
		updated = true;
		
		String currentText = getText(0, getLength());
		if (currentText == null) currentText = "";
		String beforeOffset = currentText.substring(0, offs);
		int s = (offs + len) < 0 ? 0 : offs + len;
		String afterOffset = "";
		if (s < currentText.length()) {
			afterOffset =
				currentText.substring(s, currentText.length());
		}

		String proposedResult = beforeOffset + afterOffset;

		Number newValue = parseNumber(proposedResult);

		if (doChecks(proposedResult,newValue)) {
            value = newValue;
			super.remove(offs, len);
		}
	}

	/**
	 * Parses the given string and returns the number represented. If the string
	 * is not a valid representation of the number, return value will be null.
	 * This method throws no exceptions, since all parse errors result in 
	 * invalid number and must be handled internally.
	 * 
	 * @param s String
	 * @return Number value of the number represented by string or null if 
	 *                 no or invalid number
	 */
	protected abstract Number parseNumber(String s);

	/**
	 * Returns last successfully entered number. This value will not change 
	 * if the string passed to the document is invalid. 
	 * 
	 * @return Number number parsed by this Document
	 */
	public Number getValue() {
		return value;
	}

    public Number getMaxValue() {
        return max;
    }

    public void setMaxValue(Number max) {
        this.max = max;
    }

    public Number getMinValue() {
        return min;
    }

    public void setMinValue(Number min) {
        this.min = min;
    }

	/**
	 * @return Returns the updated.
	 */
	public boolean isUpdated() {
		return updated;
	}

	/**
	 * @param updated The updated to set.
	 */
	public void setUpdated(boolean updated) {
		this.updated = updated;
	}

}
