package com.cosylab.gui.components.range2;

/**
 * This is the implementation of <code>RangedValuePolicy</code> that allows 
 * values of <code>RangedValue</code> to be only within the specified range.
 * This range is defined by allowedMinimum and allowedMaximum. These values
 * are independent from minimum and maximum defined in <code>RangedValue</code>
 * in that no value can be outside these bounds, including minimum and maximum,
 * but will be cropped to these values.
 * 
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public class LimitedRangePolicy extends RangedValuePolicy {

	// Absolute allowed minimum for all three parameter in RangedValue.
	private double allowedMin = -Double.MAX_VALUE;

	// Absolute allowed maximum for all three parameter in RangedValue.
	private double allowedMax = Double.MAX_VALUE;

	/**
	 * Constructor for LimitedRangedValuePolicy.
	 * @param rangedValue
	 */
	public LimitedRangePolicy(double allowedMinimum, double allowedMaximum) {
		if (allowedMinimum > allowedMaximum) {
			double temp = allowedMinimum;
			allowedMinimum = allowedMaximum;
			allowedMaximum = temp;
		}
		allowedMin = allowedMinimum;
		allowedMax = allowedMaximum;
	}

	/**
	 * Convenience method to check for bounds acceptability.
	 * 
	 * @param value to check against bounds.
	 * @return adjusted value.
	 */
	private final double checkBounds(double value) {
		if (value < allowedMin) {
			value = allowedMin;
		}
		if (value > allowedMax) {
			value = allowedMax;
		}
		return value;
	}

	/**
	 * @see com.cosylab.gui.components.range.RangedValuePolicy#validate(RangedValueHolder)
	 */
	public void validate(RangedValue params) {
		params.setMinimum(checkBounds(params.getMinimum()));
		params.setMaximum(checkBounds(params.getMaximum()));
		params.setValue(checkBounds(params.getValue()));
		
		validatePeerPolicy(params);
	}
}
