/*
 * Copyright (c) 2004 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.components.range2;


/**
 * Uses Range to find optimal tick distribution for LinearRange.
 *
 * @author <a href="mailto:nejc.kosnik@kgb.ijs.si">Nejc Kosnik</a>
 * @version $id$
 */
public class LinearTickCollector
{
	/**
	 * Creates a new TickCollector object.
	 *
	 * @param range Definition of range.
	 * @param tickSpacing Minimum space between minor ticks.
	 * @param length Available space in pixels.
	 */
	public LinearTickCollector(Range range, RangedValue rangedVal, int tickSpacing, int length, boolean boundTicksVisible)
	{
		this.range = range;
		this.tickSpacing = tickSpacing;
		this.length = length;
		this.rangedVal= rangedVal;
		min = rangedVal.getMinimum();
		max = rangedVal.getMaximum();
		this.boundTicksVisible = boundTicksVisible;
	}

	/**
	 * Returns new ticks using default callback.
	 *
	 * @return Array of ticks.
	 */
	public Tick[] toTicks()
	{
		zeroOnScale = isInRange(0);
		
		return calculateTicks();
	}

	/**
	 * Creates new ticks using given callback.
	 *
	 * @param measurer Callback used to measure available space.
	 *
	 * @return Array of ticks.
	 */
	public Tick[] toTicks(TickParameters measurer)
	{
		this.measurer = measurer;
		zeroOnScale = isInRange(0);
		return calculateTicks();
	}

	/** Factors for stretching and shrinking tickstep */
	public static final double[] FACTORS = { 2, 2.5, 2 };

	private class TickStep
	{
		/** Current tickstep */
		public double tep;
		private int i;

		/** If the last performed stepchange has been stretching. */
		public boolean multipliedBefore;

		/** DOCUMENT ME! */
		public double x;

		/**
		 * Shrinks tickstep.
		 */
		public void divStep()
		{
			if (multipliedBefore) {
				i = change(-1);
				multipliedBefore = false;
			}

			tep /= FACTORS[i];
			i = change(-1);
		}

		/**
		 * Stretches tickstep.
		 */
		public void mulStep()
		{
			if (!multipliedBefore) {
				i = change(1);
				multipliedBefore = true;
			}

			tep *= FACTORS[i];
			i = change(+1);
		}

		/**
		 * Creates a new TickStep object.
		 *
		 * @param step Initial step
		 * @param start Which index in <code>FACTORS</code> should be used for
		 *        stretching first.
		 * @param multipliedBefore Should be current
		 *        <code>FACTORS[start]</code> used for first stretching.
		 */
		public TickStep(double step, int start, boolean multipliedBefore)
		{
			i = start;
			this.tep = step;
			this.multipliedBefore = multipliedBefore;
		}

		private int change(int di)
		{
			if (i == FACTORS.length - 1 && di == 1) {
				return 0;
			}

			if (i == 0 && di == -1) {
				return 2;
			}

			return i + di;
		}
	}

	private RangedValue rangedVal;
	private Range range;
	private TickParameters measurer;
	private int tickSpacing;
	private int length;
	private double min;
	private double max;
	private boolean boundTicksVisible;
	private boolean zeroOnScale;

	private Tick[] calculateTicks()
	{	
		if (measurer == null) measurer = new DefaultTickParameters(50, "%0.6g");
		TickStep s = new TickStep(getRStep(), 0, true);

		double rmin = getRmin(min, s);
		double rmax = getRmax(rmin, s);

		int nticks = (int)((rmax - rmin) / s.tep) + 1;

		double tickSpace = (range.toRelative(rmax,rangedVal)
			- range.toRelative(rmax - s.tep,rangedVal)) * length;

		// try to increase major tick density
		while (tickSpace != 0 && nticks < 6 /* || tickSpace > 20*tickSpacing*/) {
			s.divStep();
			rmin = getRmin(min, s);
			rmax = getRmax(rmin, s);
			nticks = (int)((rmax - rmin) / s.tep) + 1;
			tickSpace = (range.toRelative(rmax,rangedVal)
				- range.toRelative(rmax - s.tep,rangedVal)) * length;

			//	tickSpace = ((rmax - rmin) / (max - min) * length - nticks) / (nticks
			//		- 1);
		}

		//we might have to decrease major ticks count
		while (tickSpace < tickSpacing || nticks > 25) {
			s.mulStep();
			rmin = getRmin(rmin, s);
			rmax = getRmax(rmin, s);

			if (rmin == rmax || Double.isNaN(rmin) || Double.isNaN(rmax)
			    || Double.isInfinite(rmin) || Double.isInfinite(rmax)) {
				if (boundTicksVisible) {
					return new Tick[]{
						new Tick(min, 0, true, ""), new Tick(max, 1, true, "")
					};
				} else if (zeroOnScale) {
					return new Tick[]{
						new Tick(0,range.toRelative(0,rangedVal), true,
						    this.formatNumber(0.0))
					};
				} else {
					return new Tick[]{
						new Tick(rmin,range.toRelative(rmin,rangedVal), true,
						    this.formatNumber(rmin))
					};
				}
			}

			nticks = (int)((rmax - rmin) / s.tep) + 1;
			tickSpace = (range.toRelative(rmax,rangedVal)
				- range.toRelative(rmax - s.tep,rangedVal)) * length;
		}

		//from now on we have to deal with labels
		double rstep = s.tep;
		TickStep ls = new TickStep(s.tep, s.i, s.multipliedBefore);
		double rminL = rmin;
		double rmaxL = rmax;

		//measuring tick labels
		double x;

		while (true) {
			// rstepL/10 is instead of zero to prevent round-off errors 
			for (x = rminL; x < rmaxL - ls.tep / 10; x += ls.tep) {
				if (measurer.measureTick(range.toRelative(x,rangedVal),
						this.formatNumber(x))
				    + measurer.measureTick(range.toRelative(x + ls.tep,rangedVal),
				        this.formatNumber(x + ls.tep)) > 1.3 * (range
				    .toRelative(x + ls.tep,rangedVal) - range.toRelative(x,rangedVal)) * length) {
					break;
				}
			}

			if (x >= rmaxL - ls.tep / 10) {
				break;
			} else {
				ls.mulStep();
				rminL = getRmin(min, ls);
				rmaxL = getRmax(rminL, ls);
			}
		}

		if ((int)((rmaxL - rminL) / ls.tep) + 1 > 12) { //to many labeled ticks
			ls.mulStep();
			rminL = getRmin(min, ls);
			rmaxL = getRmax(rminL, ls);
		}

		rstep = ls.tep / 2;
		rmin = getRmin(min, new TickStep(rstep, 0, true));
		rmax = getRmax(rmin, new TickStep(rstep, 0, true));

		//minor ticks
		double minorStep = rstep / 5;
		double minorMin = getRmin(min, new TickStep(minorStep, 0, true));
		double minorMax = getRmax(minorMin, new TickStep(minorStep, 0, true));

		if ((range.toRelative(minorMax,rangedVal)
		    - range.toRelative(minorMax - minorStep,rangedVal)) * length > tickSpacing) {
			//everything done already
		} else {
			minorStep = rstep / 2;
			minorMin = getRmin(min, new TickStep(minorStep, 0, true));
			minorMax = getRmax(minorMin, new TickStep(minorStep, 0, true));

			if ((range.toRelative(minorMax,rangedVal)
			    - range.toRelative(minorMax - minorStep,rangedVal)) * length > tickSpacing) {
				//everything done already
			} else {
				minorStep = rstep;
			}
		}

		minorMin = getRmin(min, new TickStep(minorStep, 0, true));
		minorMax = getRmax(minorMin, new TickStep(minorStep, 0, true));

		nticks = (int)Math.round(Math.round((rmax - rmin) / rstep) * Math.round(
			        rstep / minorStep)
			    + Math.round((minorMax - rmax + rmin - minorMin) / minorStep))
			+ 1;

		if (nticks == 0) {
			Tick[] ticks = new Tick[1];
			ticks[0] = new Tick(0,range.toRelative(0,rangedVal), true, this.formatNumber(0));

			return ticks;
		}

		int i = 0;
		int di = 0;

		if (boundTicksVisible) {
			if ((rmin - min) / (max - min) > 0.01
			    || (max - rmin) / (max - min) > 0.01) {
				nticks += 2;
				i = 1;
				di = 1;
			}
		}

		Tick[] ticks = new Tick[nticks];

		if (i == 1) {
			//			this tick was unlabelled originally
			ticks[0] = new Tick(rmin, 0.0, true, "");
		}

		int ratio2 = (int)Math.round(ls.tep / minorStep);
		int ratio1 = (int)Math.round(rstep / minorStep);
		x = minorMin;

		for (; x < rmin - minorStep / 10; i++, x += minorStep) {
			ticks[i] = new Tick(x, range.toRelative(x,rangedVal), false);
		}

		int cont = 0;
		
		for (int k = 0; x < rminL - minorStep / 10; i++, k++, x += minorStep) {
			if (k % ratio1 == 0) {
				x = rmin + (cont++) * rstep; //to prevent round off errors
				ticks[i] = new Tick(x, range.toRelative(x,rangedVal), true, "");
			} else {
				ticks[i] = new Tick(x, range.toRelative(x,rangedVal), false);
			}
		}

		for (int k = 0; i < nticks - di; i++, k++, x += minorStep) {
			if (k % ratio2 == 0) {
				x = rmin + (cont++) * rstep; //to prevent round off errors
				x = (Math.abs(x) != 0.0 ? x : Math.abs(x)); //to prevent ugly negative zero representation -0.0 
				ticks[i] = new Tick(x, range.toRelative(x,rangedVal), true,
					    /*formatter.sprintf(x)*/ this.formatNumber(x));
			} else if (k % ratio1 == 0) {
				x = rmin + (cont++) * rstep; //to prevent round off errors
				ticks[i] = new Tick(x, range.toRelative(x,rangedVal), true, "");
			} else {
				ticks[i] = new Tick(x, range.toRelative(x,rangedVal), false);
			}
		}

		if (di == 1) {
			//this tick was unlabelled originally
			ticks[i] = new Tick(rmax, 1.0, true, "");
		}
		
		return ticks;
	}
	
	private String formatNumber(double x) {
		return measurer.formatNumber(x);
	}

	private double getRmin(double min, TickStep s)
	{
		double rmin;

		if (zeroOnScale) {
			rmin = -Math.floor(-min / s.tep) * s.tep;

			if (rmin == -0.0) {
				rmin = 0.0;
			}
		} else {
			rmin = (min < 0 ? -Math.floor(-min / s.tep) * s.tep
				: Math.ceil(min / s.tep) * s.tep);
		}

		return rmin;
	}

	private double getRmax(double rmin, TickStep s)
	{
		// conversion to float makes rounder values
		return rmin + Math.floor((float)((max - rmin) / s.tep)) * s.tep;
	}

	private double getRStep()
	{
		double rstep;

		if (Math.abs(max) > Math.abs(min)) {
			rstep = Math.pow(10,
				    Math.floor(Math.log(Math.abs(max)) / Math.log(10)));
		} else {
			rstep = Math.pow(10,
				    Math.floor(Math.log(Math.abs(min)) / Math.log(10)));
		}

		return rstep;
	}

	private boolean isInRange(double value)
	{
		return (value >= min && value <= max);
	}
}

/* __oOo__ */
