/*
 * Copyright (c) 2003 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.components.range2;

/**
 * Implementation of logarithmic range.
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public class LogarithmicRange implements Range
{
	/**
	 * Returns absolute value. Function is only guaranteed to return for values
	 * larger than 0.
	 *
	 * @param relative value in 0.0 to 1.0 range.
	 *
	 * @return absolute value.
	 *
	 * @see com.cosylab.gui.components.range.Range#toAbsolute(double)
	 */
	public double toAbsolute(double relative, RangedValue val)
	{
		double logMin = Math.log(val.getMinimum());
		double logMax = Math.log(val.getMaximum());

		return Math.exp(logMin + relative * (logMax - logMin));
	}

	/**
	 * Returns relative value in 0.0 to 1.0 range inclusively.
	 *
	 * @param absolute value, must be larger than 0.0.
	 *
	 * @return value in 0.0 to 1.0 range.
	 *
	 * @see com.cosylab.gui.components.range.Range#toRelative(double)
	 */
	public double toRelative(double absolute, RangedValue val)
	{
		double logMin = Math.log(val.getMinimum());
		double logMax = Math.log(val.getMaximum());

		double range = logMax - logMin;

		if (range == 0.0) {
			return 0.5;
		}

		return (Math.log(absolute) - logMin) / range;
	}

	/** Constant defining smallest number that can be defined within range. */
	public static final double SMALLEST = 1e-8;

	/**
	 * Checks whether value is positive. If value is smaller than value defined
	 * by <code>SMALLEST</code> constant, this constant is returned. This
	 * prevents setting of values that cannot be representated on logarithmic
	 * scale.
	 *
	 * @param value to check.
	 *
	 * @return same as value parameter if valid, <code>SMALLEST</code>
	 *         otherwise.
	 */
	public final double validate(double value)
	{
		return (value > SMALLEST) ? value : SMALLEST;
	}

	public Tick[] calculateTicks(RangedValue val, int length) {
		return new LogarithmicTickCollector(this,val,3,length,false).toTicks();
	}
	
	public Tick[] calculateTicks(RangedValue val, int length, TickParameters measurer) {
		return new LogarithmicTickCollector(this,val,3,length,false).toTicks(measurer);
	}
}

/* __oOo__ */
