/*
 * Copyright (c) 2003 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.components.range2;

import java.util.ArrayList;
import java.util.List;


/**
 * <p>
 * RangedValue is used as pasive data representation, where each change to the
 * RangedValue is reported via RangedValueListener. This ensures centralized
 * handling of all changes and modifications.<br>
 * This is also the central class to be used when working with ranged value.
 * Typical use can be shown as:<br>
 * <code> class Example implements RangedValueListener {<br>
 * ...<br>
 * RangedValue value = new RangedValue();<br>
 * value.addRangeValueListener(this);<br>
 * value.addChangePolicy(new RescalingValuePolicy()); value.setRange(new
 * LogarithmicRange()); ... void valueChanged(RangedValueEvent event) { if
 * (event.minOrMaxChanged()) { // repaint scale } if (event.valueChanged()) {
 * // repaint value }}</code>
 * </p>
 * 
 * <p>
 * RangedValue will ensure that value vill always be consistant within min, max in range and
 * policies.
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public class RangedValueController 
{
	private Range range = null;
	private double rawValue;
	private final ArrayList listeners = new ArrayList();
	private double scalingFactor=1.0;

	private RangedValue rangedValue= new RangedValue();
	private RangedValuePolicy policy;

	/**
	 * Creates default range implementation. Override this method to provide
	 * different default implementation. By default, <code>LinearRange</code>
	 * class is created.
	 *
	 * @return Range range Range interface implementation.
	 */
	protected Range createDefaultRange()
	{
		return new LinearRange();
	}
	/**
	 * Returns current range definition.
	 *
	 * @see com.cosylab.gui.components.range2.RangedValueController#getRange()
	 */
	public Range getRange()
	{
		if (range == null) {
			range = createDefaultRange();
		}
		return range;
	}

	/**
	 * Sets the new definition of <code>Range</code>.
	 *
	 * @see com.cosylab.gui.components.range2.RangedValueController#setRange(Range)
	 */
	public void setRange(Range newRange)
	{
		// Save current range in case parameters cannot be adjusted to new
		// range
		Range oldRange = range;
		range = newRange;

		if (range != null) {
			// Try tranferring the state
			setRangedValue(new RangedValue(getMinimum(),getMaximum(),getRawValue()));
		}
	}

	public void setRangedValue(RangedValue v) {
		final double raw=v.getValue();

		// Make sure new values are valid in range
		final Range r = getRange();
		
		v.setMinimum(r.validate(v.getMinimum()));
		v.setMaximum(r.validate(v.getMaximum()));
		v.setValue(r.validate(v.getValue()));

		// Apply policies
		validatePolicies(v);

		// Validate again if policieas has not moved ouside range
		v.setMinimum(r.validate(v.getMinimum()));
		v.setMaximum(r.validate(v.getMaximum()));
		v.setValue(r.validate(v.getValue()));
		
		// find out what has changed
		boolean chMin = (getMinimum() != v.getMinimum());
		boolean chMax = (getMaximum() != v.getMaximum());
		boolean chVal = (getValue() != v.getValue());
		
		
		if (chMin || chMax || chVal) {
			rangedValue=new RangedValue(v.getMinimum(),v.getMaximum(),v.getValue());
		}
		
		boolean chRaw= raw!=rawValue;
		if (chRaw) {
			rawValue=raw;
		}
		
		
		fireValueChange(chMin, chMax, chVal, chRaw);

	}
	public RangedValue getRangedValue() {
		if (rangedValue==null) {
			rangedValue= new RangedValue();
		}
		return rangedValue;
	}

	/**
	 * Returns current minimum the value can take. Depending on effective
	 * policies, value could still be set below minimum.
	 *
	 * @return double current minimum.
	 *
	 * @see com.cosylab.gui.components.range2.RangedValueController#getMinimum()
	 */
	public double getMinimum()
	{
		return getRangedValue().getMinimum();
	}

	/**
	 * Returns current maximum the value can take. Depending on effective
	 * policies, value could still be set above maximum.
	 *
	 * @return double current maximum.
	 *
	 * @see com.cosylab.gui.components.range2.RangedValueController#getMaximum()
	 */
	public double getMaximum()
	{
		return getRangedValue().getMaximum();
	}

	/**
	 * Returns current value.
	 *
	 * @return double current value.
	 *
	 * @see com.cosylab.gui.components.range2.RangedValueController#getValue()
	 */
	public double getValue()
	{
		return getRangedValue().getValue();
	}

	/**
	 * Sets new value. This will attempt to set new value. If new value is
	 * outside of range or policies are set, new value might not be the same
	 * as the one set. Notification about this change will be reported via
	 * RangedValueListener, where event can be queried about the actual  changes.<br>
	 * setValue(double, double, double) should be called before using this
	 * method for the first time to ensure the values. This is not neccessary
	 * if no policies are in effect or values are known to conform to with
	 * them.
	 *
	 * @see com.cosylab.gui.components.range2.RangedValueController#setValue()
	 */
	public void setValue(double value)
	{
		setValue(getMinimum(), getMaximum(), value);
	}

	/**
	 * Sets new minimum. Setting the minimum may have side effects based on
	 * range definition, policies and current value or maximum. Actual change
	 * will be reported via RangedValueListener.<br>
	 * setValue(double, double, double) should be called before using this
	 * method for the first time to ensure the values. This is not neccessary
	 * if no policies are in effect or values are known to conform to with
	 * them.
	 *
	 * @param value double
	 */
	public void setMinimum(double value)
	{
		setValue(value, getMaximum(), getValue());
	}

	/**
	 * Sets new maximum. Setting the maximum may have side effects based on
	 * range definition, policies and current value or minimum. Actual change
	 * will be reported via RangedValueListener.<br>
	 * setValue(double, double, double) should be called before using this
	 * method for the first time to ensure the values. This is not neccessary
	 * if no policies are in effect or values are known to conform to with
	 * them.
	 *
	 * @param value double
	 */
	public void setMaximum(double value)
	{
		setValue(getMinimum(), value, getValue());
	}

	/**
	 * Checks all policies to give them oportunity to modify the values.
	 *
	 * @param values holder of values.
	 */
	public void validatePolicies(RangedValue values)
	{

		if (policy!=null) {
			policy.validate(values);
		}

		// this is mandatory policy, keeps value visthin min/max.
		TrimValuePolicy.trim(values);

	}

	/**
	 * Initializes all values to new values. Minimum, value and maximum should
	 * be in this order when compared to each other.
	 *
	 * @param minimum double
	 * @param maximum double
	 * @param value double
	 *
	 * @throws InvalidBoundException In case the parameters do not conform with
	 *         policies defined by this range.
	 */
	public synchronized void setValue(double minimum, double maximum,
	    double value)
	{
		setRangedValue(new RangedValue(minimum,maximum,value));
	}

	/**
	 * Notify listeners of the change.
	 *
	 * @param event
	 */
	protected void fireValueChange(RangedValueEvent event)
	{
		if (event == null) {
			return;
		}

		synchronized (this) {
			final int n = listeners.size();

			for (int i = 0; i < n; i++) {
				((RangedValueListener)listeners.get(i)).valueChanged(event);
			}
		}
	}

	protected void fireValueChange(boolean min, boolean max, boolean val, boolean raw)
	{
		if (min || max || val || raw) {
			RangedValueEvent event = new RangedValueEvent(this, min, max, val, raw);
			fireValueChange(event);
		}
	}

	protected void firePolicyAdded()
	{
		fireValueChange(new RangedValueEvent(this, true));
	}

	protected void firePolicyRemoved()
	{
		fireValueChange(new RangedValueEvent(this, false));
	}

	/**
	 * Adds new <code>RangedValuePolicy</code>. Polices will be checked in the
	 * same order they are added.
	 *
	 * @see com.cosylab.gui.components.range2.RangedValueController#addChangePolicy(RangedValuePolicy)
	 */
	public synchronized void setPolicy(RangedValuePolicy policy)
	{
		this.policy= policy;

		firePolicyAdded();
	}

	/**
	 * @see com.cosylab.gui.components.range2.RangedValueController#addRangedValueListener(RangedValueListener)
	 */
	public void addRangedValueListener(RangedValueListener listener)
	{
		if (listener == null) {
			return;
		}

		if (listeners.contains(listener)) {
			return;
		}

		synchronized (listeners) {
			listeners.add(listener);
		}
	}

	/**
	 * @see com.cosylab.gui.components.range2.RangedValueController#removeRangedValueListener(RangedValueListener)
	 */
	public void removeRangedValueListener(RangedValueListener listener)
	{
		if (listener == null) {
			return;
		}

		synchronized (listeners) {
			listeners.remove(listener);
		}
	}

	/**
	 * @see com.cosylab.gui.components.range2.RangedValueController#getRelativeValue()
	 */
	public synchronized double getRelativeValue()
	{
		if (scalingFactor==1.0) {
			return getRange().toRelative(getValue(),getRangedValue());
		}
		return getRange().toRelative(getValue(),getRangedValue())*scalingFactor;
	}

	/**
	 * @see com.cosylab.gui.components.range2.RangedValueController#setRelativeValue(double)
	 */
	public void setRelativeValue(double value)
	{
		if (scalingFactor==1.0) {
			setValue(getRange().toAbsolute(value,getRangedValue()));
		} else {
			setValue(getRange().toAbsolute(value/scalingFactor,getRangedValue()));
		}
	}

	public double getRawValue() {
		return rawValue;
	}
	
	public double validate(double value) {
		double v= getRange().validate(value);
		RangedValue val= new RangedValue(getMinimum(),getMaximum(),v);
		validatePolicies(val);
		return val.getValue();
	}

	
	/**
	 * Returns the scalingFactor.
	 * @return Returns the scalingFactor.
	 */
	public double getScalingFactor() {
		return scalingFactor;
	}

	
	/**
	 * Sets new scalingFactor. Scaling factor scales relative value from 0 to 1.0 or whatever scaling factor is.
	 * @param scalingFactor The scalingFactor to set.
	 */
	public void setScalingFactor(double scalingFactor) {
		this.scalingFactor = scalingFactor;
	}
	
	public RangedValuePolicy getPolicy() {
		return policy;
	}
	
	public synchronized void addPolicy(RangedValuePolicy p) {
		if (p==null) return;
		
		if (policy==null) {
			policy=p;
		} else {
			policy.addPeerPolicy(p);
		}

		setRangedValue(new RangedValue(getMinimum(),getMaximum(),getRawValue()));
		
		firePolicyAdded();
	}
	
	public void removePolicy(RangedValuePolicy p) {
		if (p==null) {
			return;
		}
		if (policy!=null) {
			RangedValuePolicy pp= policy;
			policy= policy.removePeerPolicy(p);
			if (pp!=policy) {
				setRangedValue(new RangedValue(getMinimum(),getMaximum(),getRawValue()));
				firePolicyAdded();
			}
		}
	}
	public void removePolicyByType(Class<? extends RangedValuePolicy> c) {
		if (c==null) {
			return;
		}
		if (policy!=null) {
			RangedValuePolicy pp= policy;
			policy= policy.removePeerPolicyByType(c);
			if (pp!=policy) {
				setRangedValue(new RangedValue(getMinimum(),getMaximum(),getRawValue()));
				firePolicyAdded();
			}
		}
	}
	public double toRelative(double d) {
		if (scalingFactor==1.0) {
			return getRange().toRelative(d,getRangedValue());
		}
		return getRange().toRelative(d, getRangedValue())*scalingFactor;
	}

	public Tick[] calculateTicks(int w, TickParameters measurer) {
		return getRange().calculateTicks(getRangedValue(), w, measurer);
	}
	public double toAbsolute(double proportional) {
		if (scalingFactor==1.0) {
			return getRange().toAbsolute(proportional,getRangedValue());
		}
		return getRange().toAbsolute(proportional/scalingFactor, getRangedValue());
	}
	public Tick[] calculateTicks(int width) {
		return getRange().calculateTicks(getRangedValue(), width);
	}
}

/* __oOo__ */
