/*
 * Copyright (c) 2003 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.components.range2;

/**
 * This class represents adjustable policies for <code>RangedValue</code>.
 * Whenever a change any of the values (minimum, maximum or value) is made in
 * <code>RangedValue</code> validate method will be called giving the
 * implementor of this interface the chance to adjust the values before they
 * are set and property change events invoked.<br>
 * Note: Regardles of the changes proposed by validate method,
 * <code>RangedValue</code> will always ensure the proper relation between the
 * values: values must always form increasing order (minimum, value,
 * maximum).
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public abstract class RangedValuePolicy
{
	RangedValuePolicy peer;
	
	/**
	 * Passes the value triplet <code>params</code> that may be modified by
	 * validate method. Regardles of the changes made, all three values will
	 * still be made to comply with the basic definition of <code>
	 * RangedValue</code>.
	 * 
	 * <p>Before or after this policy has been executed it must call protected method validatePeerPolicy(RangedValue).
	 *
	 * @param params Triplet of values to validate.
	 */
	public abstract void validate(RangedValue params);
	
	/**
	 * Inserts provided polici as first policy in chain, here next peer policy is this policy.
	 * @param p new first policy in chain
	 * @return returns new first policy in chain, the p parameter.
	 */
	public RangedValuePolicy insertPeerPolicy(RangedValuePolicy p) {
		p.peer=this;
		return p;
	}

	/**
	 * Adds provided policy to the end of the chin
	 * @param p
	 */
	public void addPeerPolicy(RangedValuePolicy p) {
		if (peer!=null) {
			peer.addPeerPolicy(p);
		} else {
			peer=p;
		}
	}
	
	public RangedValuePolicy getPeerPolicy() {
		return peer;
	}
	
	/**
	 * Removes provided policy from the chain, and returnes first policy in chain.
	 * @param p policy to be removed from chain
	 * @return returns new first policy in chain, this or peer policy
	 */
	public RangedValuePolicy removePeerPolicy(RangedValuePolicy p) {
		if (p==this) {
			RangedValuePolicy pp= peer;
			p.peer=null;
			return peer;
		}
		if (peer != null) {
			peer=peer.removePeerPolicy(p);
		}
		return this;
	}
	
	/**
	 * Removes provided policy from the chain, and returnes first policy in chain.
	 * @param p policy to be removed from chain
	 * @return returns new first policy in chain, this or peer policy
	 */
	public RangedValuePolicy removePeerPolicyByType(Class<? extends RangedValuePolicy> c) {
		if (c.isAssignableFrom(this.getClass())) {
			RangedValuePolicy pp= peer;
			peer=null;
			return peer;
		}
		if (peer != null) {
			peer=peer.removePeerPolicyByType(c);
		}
		return this;
	}

	protected void validatePeerPolicy(RangedValue p) {
		if (peer !=null) {
			peer.validate(p);
		}
	}
	
}

/* __oOo__ */
