/*
 * Copyright (c) 2003 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.components.range2;

/**
 * Implementation of <code>RangedValuePolicy</code> that resizes the range
 * based on the value parameter.
 * 
 * <p>
 * Whenever the value is set below minimum, minimum will be changed to the the
 * value, decreased by factor  (maximum - minimum). Same rule applies for
 * maximum.
 * </p>
 * This policy is value dominant i.e., bounds will be adjusted to ensure that
 * value is always valid. If either bound is set incorrectly, it will be 
 * adjusted to conform with this condition.
 * <p>
 * Rule for adjusting bounds is as follows. Whenever value is either below
 * minimum ore above maximum, the corresponding bound will be adjusted using
 * value - range * factor or value + range * factor for minimum and maximum
 * respectively, where range is maximum - minimum.
 * 
 * 
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public class RescalingValuePolicy extends RangedValuePolicy
{
	private double rescaleFactor;

	/**
	 * Default constructor for RescalingValuePolicy. Initializes rescaling
	 * factor to 0.25.
	 */
	public RescalingValuePolicy()
	{
		this(0.25);
	}

	/**
	 * Constructor for RescalingValuePolicy. Allows custom rescaling factor.
	 *
	 * @param factor by which the range will be resized.
	 */
	public RescalingValuePolicy(double factor)
	{
		super();
		setFactor(factor);
	}

	/**
	 * Adjusts bounds as specified in class JavaDoc.
	 *
	 * @param params RangedValueHolder
	 *
	 * @see com.cosylab.gui.components.range.RangedValuePolicy#validate(RangedValueHolder)
	 */
	public void validate(RangedValue params)
	{
		double value = params.getValue();
		double range = params.getMaximum() - params.getMinimum();

		if(value < params.getMinimum()) {
			params.setMinimum(value - range * getFactor());

			validatePeerPolicy(params);
			return;
		}

		if(value > params.getMaximum()) {
			params.setMaximum(value + range * getFactor());

			validatePeerPolicy(params);
			return;
		}
		
		validatePeerPolicy(params);
	}

	/**
	 * Returns rescaling factor.
	 *
	 * @return factor.
	 */
	public double getFactor()
	{
		return rescaleFactor;
	}

	/**
	 * Sets new rescaling factor. Factor set will not be less that 0.
	 *
	 * @param factor new rescaling factor.
	 */
	public void setFactor(double factor)
	{
		rescaleFactor = Math.max(0.0, factor);
	}
}

/* __oOo__ */
