/*
 * Copyright (c) 2003 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.components.slider;

import com.cosylab.events.SetEvent;
import com.cosylab.events.SetListener;

import com.cosylab.gui.components.NumberField;
import com.cosylab.gui.components.ResizableTextLabel;
import com.cosylab.gui.components.Slider;
import com.cosylab.gui.components.Slider.Button;
import com.cosylab.gui.components.util.ColorHelper;
import com.cosylab.gui.components.util.IconHelper;

import com.cosylab.util.Debug;
import com.cosylab.util.PrintfFormat;
import com.sun.corba.se.spi.legacy.connection.GetEndPointInfoAgainException;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import java.beans.Beans;

import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.JTextField;


/**
 * Button bar for exact control over slider behaviour. Consists of six buttons
 * (3 for decrementing, 3 for incrementing the value) and editable text field
 * for entering exact values. Text field is updated to reflect current value.
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public class NavigationBar extends JPanel
{
	private boolean unitsVisible = true;
	private String units;
	private String format = "%1.4f";
	protected JButton[] buttons = new JButton[6];
	private NumberField valueEditor;
	private ResizableTextLabel unitsLabel;
	protected Slider slider;
	//private RangedValue rangedValue;
	private PrintfFormat formater;
	
	private JButton fastIncrement; 
	private JButton fastDecrement;
	private JButton bitIncrement;
	private JButton bitDecrement;
	private JButton slowIncrement;
	private JButton slowDecrement;
		
	private class ButtonListener implements ActionListener
	{
		/**
		 * Handles button presses.
		 *
		 * @param e ActionEvent
		 *
		 * @see java.awt.event.ActionListener#actionPerformed(ActionEvent)
		 */
		public void actionPerformed(ActionEvent e)
		{
			for (int i = 0; i < buttons.length; i++) {
				if (e.getSource() == buttons[i]) {
					switch (i) {
					case 0:
						slider.decrementLarge();

						return;

					case 1:
						slider.decrementSmall();

						return;

					case 2:
						slider.decrementBit();

						return;

					case 3:
						slider.incrementBit();

						return;

					case 4:
						slider.incrementSmall();

						return;

					case 5:
						slider.incrementLarge();

						return;

					default:}

					return;
				}
			}
		}
	}

	/**
	 * Constructor for NavigationBar.
	 *
	 * @param owner Slider
	 *
	 * @throws IllegalArgumentException if owner is null.
	 */
	public NavigationBar(Slider owner)
	{
		super();

		if (owner == null) {
			throw new IllegalArgumentException(
			    "Cannot create new NavigationBar with null parameter.");
		}

		slider = owner;
		//rangedValue = owner.getRangedValue();
		initialize();
	}

	private GridBagConstraints createConstraints(int x)
	{
		GridBagConstraints gc = new GridBagConstraints();
		gc.gridx = x;
		gc.fill = GridBagConstraints.HORIZONTAL;

		//		gc.fill = GridBagConstraints.NONE;
		gc.weightx = 1.0;
		gc.weighty = 0.0;
		gc.insets = new Insets(1, 1, 1, 1);

		return gc;
	}

	private JButton createButton(String description, String iconName, int index,
	    int x, int min)
	{
		JButton button = new JButton();
		button.setBackground(ColorHelper.getCosyControl());
		button.setToolTipText(description);
		button.setPreferredSize(new Dimension(25, 20));
		button.setMinimumSize(new Dimension(min, 18));
		//button.setName(name)

		try {
			button.setIcon(IconHelper.createIcon("icons/navigation/"
			        + iconName));
		} catch (Exception e) {
			System.out.println("Icon resource not found: " + iconName);
		}

		buttons[index] = button;
		button.addActionListener(new ButtonListener());
		add(button, createConstraints(x));
		return button;
	}

	/**
	 * Sets value in text field.
	 *
	 * @param value to set.
	 */
	
	public void setEnabled(boolean enabled){
		super.setEnabled(enabled);
		/*
		 * If Slider is disabled, valueEditor NumberField should still follow
		 * current value setting. On the other hand user should not be able to
		 * edit it.
		 */
		valueEditor.setEditable(enabled);
		/*
		 * Setting valueEditor to not focusable prevents updating slider value
		 * of a disabled slider when focus is lost.
		 */
		valueEditor.setFocusable(enabled);
	    unitsLabel.setEnabled(enabled);
	    for(int i=0;i<6;i++){
	        buttons[i].setEnabled(enabled);
	    }
	}
	public void setValue(double value)
	{
		valueEditor.setDoubleValue(value);
	}

	/**
	 * Returns text field value.
	 *
	 * @return current value of text field.
	 */
	public double getValue()
	{
		return valueEditor.getDoubleValue();
	}

	/**
	 * Returns current format of text field.
	 *
	 * @return current format of text field.
	 */
	public String getFormat()
	{
		return format;
	}

	/**
	 * Sets new format for text field. This format is used only when displaying
	 * the value.
	 *
	 * @param format to set for textfield.
	 */
	public void setFormat(String format)
	{
		this.format = format;

		if (format == null) {
			format = "%1.4f";
		}

		valueEditor.setFormat(format);
		formater = new PrintfFormat(format);
	}

	/**
	 * Sets units to be displayed next to text field.
	 *
	 * @param units string to set.
	 */
	public void setUnits(String units)
	{
		this.units = units;

		if (units == null && Beans.isDesignTime()) {
			units = "<u>";
		}

		unitsLabel.setText((unitsVisible ? units : ""));
	}

	/**
	 * Returns units currently displayed next to edit field.
	 *
	 * @return current units string.
	 */
	public String getUnits()
	{
		return units;
	}

	private void initialize()
	{
		setLayout(new GridBagLayout());
		setOpaque(false);

		formater = new PrintfFormat("%1.4f");

		fastDecrement = createButton("Fast decrement", "ArrowLeftLeft16.gif", 0, 0, 15);
		slowDecrement = createButton("Slow decrement", "ArrowLeft16.gif", 1, 1, 12);
		bitDecrement = createButton("Bit decrement", "Minus16.gif", 2, 2, 12);
		bitIncrement = createButton("Bit increment", "Plus16.gif", 3, 5, 12);
		slowIncrement = createButton("Slow increment", "ArrowRight16.gif", 4, 6, 12);
		fastIncrement = createButton("Fast increment", "ArrowRightRight16.gif", 5, 7, 15);

		valueEditor = new NumberField();
		valueEditor.setNumberType(Double.class);
		valueEditor.setFormat("%1.4f");
		valueEditor.setResizable(false);
		valueEditor.setHorizontalAlignment(JTextField.CENTER);

		GridBagConstraints gc = new GridBagConstraints();
		gc.gridx = 3;
		gc.weightx = 8.0;
		gc.insets.left = 3;
		gc.fill = GridBagConstraints.HORIZONTAL;

		//	valueEditor.setMinimumSize(new java.awt.Dimension(50, 20));
		valueEditor.setPreferredSize(new java.awt.Dimension(50, 20));

		//	valueEditor.setMaximumSize(new java.awt.Dimension(120, 100000));
		add(valueEditor, gc);

		unitsLabel = new ResizableTextLabel();
		unitsLabel.setOpaque(false);
		unitsLabel.setResizable(false);

		if (Beans.isDesignTime()) {
			unitsLabel.setText("<u>");
		}

		gc = new GridBagConstraints();
		gc.gridx = 4;
		gc.weightx = 0.0;
		gc.fill = GridBagConstraints.HORIZONTAL;
		gc.insets.left = 1;
		gc.insets.right = 3;
		add(unitsLabel, gc);
	}

	/**
	 * Convinience format with Slider format
	 *
	 * @param value the value
	 *
	 * @return formated value as String
	 */
	public String format(double value)
	{
		return formater.sprintf(value);
	}

	public void setMinimum(double min) {
		valueEditor.setMinimum(new Double(min));
	}

	public void setMaximum(double min) {
		valueEditor.setMaximum(new Double(min));
	}

	/**
	 * @param b
	 */
	public void setUnitsVisible(boolean b) {
		this.unitsVisible = b;
		unitsLabel.setText((unitsVisible ? units : ""));
	}

	/**
	 * @return
	 */
	public boolean isUnitsVisible() {
		return unitsVisible;
	}
	
	@Override
	public void setBackground(Color bg) {
		super.setBackground(bg);
		Component[] comp = getComponents();
		for (Component c: comp) {
			if (c instanceof JButton) {
				c.setBackground(bg);
			}
		}
	}
	
	@Override
	public void setForeground(Color fg) {
		super.setForeground(fg);
		Component[] comp = getComponents();
		for (Component c: comp) {
			c.setForeground(fg);
		}
	}
	
	public void addSetListener(SetListener sl) {
		valueEditor.addSetListener(sl);
	}
	public void removeSetListener(SetListener sl) {
		valueEditor.removeSetListener(sl);
	}
	
	public void setButtonTooltip(Button button, String tooltip) {
		if (button.equals(Button.FAST_INCREMENT)) {
			fastIncrement.setToolTipText(tooltip);
		} else if(button.equals(Button.FAST_DECREMENT)) {
			fastDecrement.setToolTipText(tooltip);
		} else if(button.equals(Button.SLOW_INCREMENT)) {
			slowIncrement.setToolTipText(tooltip);
		} else if(button.equals(Button.SLOW_DECREMENT)) {
			slowDecrement.setToolTipText(tooltip);
		} else if (button.equals(Button.BIT_INCREMENT)) {
			bitIncrement.setToolTipText(tooltip);
		} else if(button.equals(Button.BIT_DECREMENT)) {	
			bitDecrement.setToolTipText(tooltip);
		}
	}
	
	public String getButtonTooltip(Button button) {
		if (button.equals(Button.FAST_INCREMENT)) {
			return fastIncrement.getToolTipText();
		} else if(button.equals(Button.FAST_DECREMENT)) {
			return fastDecrement.getToolTipText();
		} else if(button.equals(Button.SLOW_INCREMENT)) {
			return slowIncrement.getToolTipText();
		} else if(button.equals(Button.SLOW_DECREMENT)) {
			return slowDecrement.getToolTipText();
		} else if (button.equals(Button.BIT_INCREMENT)) {
			return bitIncrement.getToolTipText();
		} else if(button.equals(Button.BIT_DECREMENT)) {	
			return bitDecrement.getToolTipText();
		}
		return null;
	}
	
	public void setUnitsFont(Font font) {
		unitsLabel.setFont(font);
	}
	
	public Font getUnitsFont() {
		return unitsLabel.getFont();
	}
}

/* __oOo__ */
