/*
 * Copyright (c) 2003 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.components.table;

import java.util.Arrays;
import java.util.EventObject;

import javax.swing.event.TableModelEvent;


/**
 * <code>TableRowModelEvent</code> contains information about changes in
 * <code>TableRowModel</code>.
 *
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 * @version $Id: TableRowModelEvent.java,v 1.3 2004/11/15 18:06:05 ikriznar Exp $
 *
 * @since May 3, 2004.
 */
public class TableRowModelEvent extends EventObject
{
	/** Identifies structural change. */
	public final static int HEADER_ROW = TableModelEvent.HEADER_ROW;
	/** Identifies element index which applies for all row elements. */
	public final static int ALL_VALUES = TableModelEvent.ALL_COLUMNS;

	/** Identifies the addtion of new rows. */
	public static final int INSERT = TableModelEvent.INSERT;

	/** Identifies a change to existing data. */
	public static final int UPDATE = TableModelEvent.UPDATE;

	/** Identifies the removal of rows. */
	public static final int DELETE = TableModelEvent.DELETE;
	private int value = ALL_VALUES;
	private int type = UPDATE;
	private static final TableRow[] DEFAULT_ROWS = new TableRow[0]; 
	private TableRow[] rows = DEFAULT_ROWS;
	private int rowIndex = -1;

	/**
	 * Creates new event with row as source, column value is by default
	 * <code>ALL_VALUES</code> and type is <code>UPDATE</code>.
	 *
	 * @param source
	 * @param index the index of row in parent model
	 */
	public TableRowModelEvent(TableRow source, int index)
	{
		super(source);
		rowIndex = index;
		rows = new TableRow[]{ source };
	}

	/**
	 * Creates new event with row as source and value index, type is
	 * <code>UPDATE</code>..
	 *
	 * @param source table row as source
	 * @param index the index of row in parent model
	 * @param value the value index
	 *
	 * @throws IllegalArgumentException if value isless then -1
	 */
	public TableRowModelEvent(TableRow source, int index, int value)
	{
		this(source, index);
		this.value = value;

		if (value < -1) {
			throw new IllegalArgumentException(
			    "Argument value is less then -1 :'" + value + "'");
		}
	}

	/**
	 * Creates new event.
	 *
	 * @param source table row as source
	 * @param index the index of row in parent model
	 * @param value the value index
	 * @param type the type of event
	 *
	 * @throws IllegalArgumentException if type is not one of
	 *         <code>INSERT</code>, <code>DELETE</code> or <code>UPDATE</code>
	 *         values
	 */
	public TableRowModelEvent(TableRow source, int index, int value, int type)
	{
		this(source, index, value);
		this.type = type;

		if (type < -1 || type > 1) {
			throw new IllegalArgumentException(
			    "Argument type has illegal value :'" + type + "'");
		}
	}

	/**
	 * Creates new event.
	 *
	 * @param sources table row as source
	 * @param value the value index
	 * @param type the type of event
	 */
	public TableRowModelEvent(TableRow[] sources, int value, int type)
	{
		this(sources[0], -1, value, type);

		rows = sources;
	}

	/**
	 * Returns index of value for which event applies
	 *
	 * @return the value index
	 */
	public int getValue()
	{
		return value;
	}

	/**
	 * Returns type of event, valid value is <code>INSERT</code>,
	 * <code>DELETE</code> and <code>UPDATE</code>.
	 *
	 * @return the type of event
	 */
	public int getType()
	{
		return type;
	}

	/**
	 * Returns row, which was source of event. If there is more rows as source,
	 * first row is returned
	 *
	 * @return the event source
	 */
	public TableRow getRow()
	{
		return rows[0];
	}

	/**
	 * Returns array of rows, which are source of this event.
	 *
	 * @return the sources of event
	 */
	public TableRow[] getRows()
	{
		return rows;
	}

	/**
	 * Returns the index of rowinside parent model.
	 *
	 * @return the index of rowinside parent model
	 */
	public int getRowIndex()
	{
		return rowIndex;
	}

	/* (non-Javadoc)
	 * @see java.util.EventObject#toString()
	 */
	public String toString()
	{
		StringBuffer sb = new StringBuffer();
		sb.append("TableRowModelEvent = { sources = ");
		sb.append(Arrays.asList(getRows()));
		sb.append(" rowIndex = ");
		sb.append(rowIndex);
		sb.append(" value = ");
		sb.append(value);
		sb.append("' type = ");
		sb.append(type);
		sb.append(" }");

		return sb.toString();
	}
}

/* __oOo__ */
