/*
 * Copyright (c) 2004 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.components.table.cells;

import com.cosylab.util.NameValueList;
import com.cosylab.util.ObjectList;

import java.util.Iterator;
import java.util.List;
import java.util.Map;


/**
 * TableCell is most basic kind of table cells, which has own renderers and
 * editors in <code>com.cosylab.gui.components.table.renderers</code>.
 * 
 * <p>
 * This cell may contain arbitrary set of characteristics. They hint renderers,
 * how to display particular cell. Consult JavaDoc of particular cell to see
 * which characteristics are  supported and compatible with available
 * renderers.
 * </p>
 * 
 * <p>
 * User of  this cell migth decide to define values for following
 * characteristics: TIMEOUT, ERROR, SUSPENDED, OUT_OF_BOUNDS.
 * </p>
 * 
 * <p>
 * Note that characteristic map is optimized for number of characteristics up
 * to 10 because is implemented around linear array.
 * </p>
 * 
 * <p>
 * Note that all characteristics are optional. If they are null, renderers will
 * try best to use value without them.
 * </p>
 *
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 * @version $id$
 */
public class TableCell implements Comparable
{
	/** Characteristic name for tooltip. */
	public static final String TOOLTIP = "tooltip";

	/**
	 * Characteristic name for out of bouds value state, when cell value is out
	 * of designated value range, characteristic value must be Boolean.TRUE.
	 */
	public static final String OUT_OF_BOUNDS = "outOfBounds";

	/**
	 * Characteristic name for timeout state, value must be Boolean.TRUE if
	 * data source in timeout.
	 */
	public static final String TIMEOUT = "timeout";

	/**
	 * Characteristic name for error state, value must be Boolean.TRUE if data
	 * source is in error condition.
	 */
	public static final String ERROR = "error";

	/**
	 * Characteristic name for alarm state, value must be Boolean.TRUE if data
	 * source is in alarm condition.
	 */
	public static final String ALARM = "alarm";

	/**
	 * Characteristic name for warning state, value must be Boolean.TRUE if
	 * value is out of bounds or source is in warning condition.
	 */
	public static final String WARNING = "warning";

	/**
	 * Characteristic name for severity of alarm or error state.
	 */
	public static final String SEVERITY = "severity";

	/**
	 * Characteristic name for suspended state, value must be Boolean.TRUE if
	 * data source is in suspended condition (eg. not initialized).
	 */
	public static final String SUSPENDED = "suspended";
	protected Object value = null;
	protected Object dataSource;
	protected Command[] commands = null;
	protected String identifier = null;
	private Map characteristics;
	private List supported = new ObjectList(String.class);
	protected long timestamp;

	/**
	 * Creates new noninitialized instance of cell proxy.
	 */
	protected TableCell()
	{
		super();
	}

	/**
	 * Creates a new TableCell object.
	 *
	 * @param identifier unique string within context of current table
	 * @param dataSource object which provides values for this cell
	 * @param value a cells value
	 * @param commands commands, which can be executed on this cell
	 */
	public TableCell(String identifier, Object dataSource, Object value,
	    Command[] commands)
	{
		this.identifier = identifier;
		this.dataSource = dataSource;
		this.value = value;
		this.commands = commands;
	}

	/**
	 * Returns commands, which can be executed on this cell.
	 *
	 * @return commands, which can be executed on this cell
	 */
	public Command[] getCommands()
	{
		return commands;
	}

	/**
	 * Object which deliveres values for this cell.
	 *
	 * @return objects which deliveres values to this cell
	 */
	public Object getDataSource()
	{
		return dataSource;
	}

	/**
	 * Generally expected to be unique string within context of current table.
	 *
	 * @return unique string within context of current table
	 */
	public String getIdentifier()
	{
		return identifier;
	}

	/**
	 * Return tooltip value or <code>null</code> if none is set.
	 *
	 * @return tooltip
	 */
	public String getTooltip()
	{
		return (String)characteristics.get(TOOLTIP);
	}

	/**
	 * Return tooltip value or <code>null</code> if none is set.
	 *
	 * @return tooltip
	 */
	public Object getValue()
	{
		return value;
	}

	/**
	 * Sets new value to this cell.
	 *
	 * @param newValue
	 */
	public void setValue(Object newValue)
	{
		value = newValue;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString()
	{
		return String.valueOf(value);
	}

	/**
	 * Return scharacteristic value with specific name.
	 *
	 * @param name a name of requested characteristic
	 *
	 * @return characteristic value, or <code>null</code> if not defined
	 */
	public Object getCharacteristic(String name)
	{
		return getCharacteristics().get(name);
	}

	/**
	 * Return scharacteristic value with specific name.
	 *
	 * @param name a name of requested characteristic
	 * @param defaultValue DOCUMENT ME!
	 *
	 * @return characteristic value, or <code>null</code> if not defined
	 */
	public Object getCharacteristic(String name, Object defaultValue)
	{
		Object o = getCharacteristics().get(name);

		if (o == null) {
			o = defaultValue;
		}

		return o;
	}

	/**
	 * Puts new characteristic value with provided name.
	 *
	 * @param name characteristic name
	 * @param value new characteristic value
	 *
	 * @return old characteristic value, or <code>null</code> if none
	 */
	public Object putCharacteristic(String name, Object value)
	{
		return getCharacteristics().put(name, value);
	}

	/**
	 * Puts all characteristics to this cell.
	 *
	 * @param characteristics Map with characteristics
	 */
	public void putAllCharacteristics(Map characteristics)
	{
		if (characteristics == null) return;
		Iterator it= characteristics.keySet().iterator();
		while (it.hasNext()) {
			Object key = it.next();
			Object val= characteristics.get(key);
			if (val!=null) {
				getCharacteristics().put(key,val);
			}
		}
	}

	/**
	 * Return map with all characteristics.
	 *
	 * @return map with all characteristics
	 */
	public Map getCharacteristics()
	{
		if (characteristics == null) {
			characteristics = new NameValueList(12);
		}

		return characteristics;
	}

	/**
	 * Returns time of latest change of value.
	 *
	 * @return time of latest change of value
	 */
	public long getTimestamp()
	{
		return timestamp;
	}

	/**
	 * Sets time of latest change of value.
	 *
	 * @param timestamp time of latest change
	 */
	public void setTimestamp(long timestamp)
	{
		this.timestamp = timestamp;
	}

	/**
	 * Returns <code>true</code> if ERROR characteristic is true.
	 *
	 * @return <code>true</code> if ERROR characteristic is true
	 */
	public boolean isError()
	{
		return Boolean.TRUE.equals(getCharacteristic(ERROR));
	}

	/**
	 * Returns <code>true</code> if ALARM characteristic is true.
	 *
	 * @return <code>true</code> if ALARM characteristic is true
	 */
	public boolean isAlarm()
	{
		return Boolean.TRUE.equals(getCharacteristic(ALARM));
	}

	/**
	 * Returns <code>true</code> if WARNING characteristic is true.
	 *
	 * @return <code>true</code> if WARNING characteristic is true
	 */
	public boolean isWarning()
	{
		return Boolean.TRUE.equals(getCharacteristic(WARNING));
	}

	/**
	 * Returns <code>true</code> if TIMEOUT characteristic is true.
	 *
	 * @return <code>true</code> if TIMEOUT characteristic is true
	 */
	public boolean isTimeout()
	{
		return Boolean.TRUE.equals(getCharacteristic(TIMEOUT));
	}

	/**
	 * Returns <code>true</code> if SUSPENDED characteristic is true.
	 *
	 * @return <code>true</code> if SUSPENDED characteristic is true
	 */
	public boolean isSuspended()
	{
		return Boolean.TRUE.equals(getCharacteristic(SUSPENDED));
	}

	/**
	 * Returns <code>true</code> if OUT_OF_BOUNDS characteristic is true.
	 *
	 * @return <code>true</code> if OUT_OF_BOUNDS characteristic is true
	 */
	public boolean isOutOfBounds()
	{
		return Boolean.TRUE.equals(getCharacteristic(OUT_OF_BOUNDS));
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @return DOCUMENT ME!
	 */
	public String[] getSupportedCharacteristics()
	{
		return (String[])supported.toArray(new String[supported.size()]);
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @param name DOCUMENT ME!
	 */
	public void addSupportedCharacteristic(String name)
	{
		if (!supported.contains(name)) {
			supported.add(name);
		}
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @param name DOCUMENT ME!
	 */
	public void removeSupportedCharacteristic(String name)
	{
		supported.remove(name);
	}

	/* (non-Javadoc)
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	public int compareTo(Object o)
	{
		if (o instanceof TableCell) {
			TableCell tc = (TableCell)o;

			if (getValue() != null && tc.getValue() != null) {
				if (getValue().getClass().equals(tc.getValue().getClass())
				    && (getValue() instanceof Comparable)) {
					return ((Comparable)value).compareTo(tc.getValue());
				} else if ((getValue() instanceof Number)
				    && (tc.getValue() instanceof Number)) {
					double d = ((Number)getValue()).doubleValue()
						- ((Number)tc.getValue()).doubleValue();

					if (d < -0.00001) {
						return -1;
					}

					if (d > 0.00001) {
						return 1;
					}
				}
			}
		}

		return 0;
	}
}

/* __oOo__ */
