/*
 * Copyright (c) 2004 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.components.util;

import java.awt.event.ActionEvent;

import java.util.ArrayList;
import java.util.Iterator;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.Icon;


/**
 * <p>
 * An action which holds a list of other <code>Action</code> objects.
 * Convenient for containment of functionally similar actions.
 * </p>
 * 
 * <p>
 * With asotiation with <code>Action</code> can be used for creation of menus.
 * Set <code>putValue(MENU_CONTEXT,MENU_CONTEXT_RADIO)</code> to produce group
 * of radio menu items.
 * </p>
 * 
 * <p>
 * <code>Action</code> will use separator menu item, where null
 * actionitemisplaced or Action has <code>SEPARATOR</code> boolean attribute.
 * </p>
 *
 * @author <a href="mailto:miha.kadunc@cosylab.com">Miha Kadunc</a>
 * @version $id$
 */
public class ActionList extends AbstractAction
{
	/**
	 * Key name for a Boolean value indicating that the added action should be
	 * represented in the GUI as a separator.
	 */
	public static final String SEPARATOR = "_group_separator_";

	/**
	 * Key name for a String value indicating if this list is to be presented
	 * in menu with buttons or radio buttons.
	 *
	 * @see javax.swing.Action#getValue(java.lang.String)
	 */
	public static final String MENU_CONTEXT = "menuContext";

	/** DOCUMENT ME! */
	public static final String MENU_CONTEXT_BUTTON = "button";

	/** DOCUMENT ME! */
	public static final String MENU_CONTEXT_RADIO = "radio";

	/** DOCUMENT ME! */
	public static String ACTIONS = "actions";
	private ArrayList actions = null;

	/**
	 * Creates a new ActionList object.
	 */
	public ActionList()
	{
		this(null, null);
	}

	/**
	 * Creates a new ActionList object.
	 *
	 * @param name DOCUMENT ME!
	 */
	public ActionList(String name)
	{
		this(name, null);
	}

	/**
	 * Creates a new ActionList object.
	 *
	 * @param name DOCUMENT ME!
	 * @param icon DOCUMENT ME!
	 */
	public ActionList(String name, Icon icon)
	{
		this(name, icon, null);
	}

	/**
	 * Creates a new ActionList object.
	 *
	 * @param name DOCUMENT ME!
	 * @param icon DOCUMENT ME!
	 * @param actions DOCUMENT ME!
	 */
	public ActionList(String name, Icon icon, Action[] actions)
	{
		super(name, icon);
		this.actions = new ArrayList();

		if (actions != null) {
			for (int i = 0; i < actions.length; i++) {
				this.actions.add(actions[i]);
			}

			putValue(ACTIONS, this.actions.toArray(new Action[size()]));
		}

		putValue(MENU_CONTEXT, MENU_CONTEXT_BUTTON);
	}

	/**
	 * Adds Action to end of the list.
	 *
	 * @param action Action to be added
	 */
	public void addAction(Action action)
	{
		addAction(size(), action);
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @param action DOCUMENT ME!
	 *
	 * @return DOCUMENT ME!
	 */
	public boolean contains(Action action)
	{
		return actions.contains(action);
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @param name DOCUMENT ME!
	 *
	 * @return DOCUMENT ME!
	 */
	public boolean containsActionWithName(String name)
	{
		return (firstIndexOfActionWithName(name) < 0);
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @param name DOCUMENT ME!
	 *
	 * @return DOCUMENT ME!
	 */
	public int firstIndexOfActionWithName(String name)
	{
		for (Iterator iter = actions.iterator(); iter.hasNext();) {
			Action element = (Action)iter.next();

			if ((element != null) && element.getValue(NAME).equals(name)) {
				return actions.indexOf(element);
			}
		}

		return -1;
	}

	/**
	 * Adds all actions from ActionList to end of the this list in specified
	 * order.
	 *
	 * @param action actions to be added
	 */
	public void addAll(ActionList action)
	{
		Action[] a = action.toArray();

		for (int i = 0; i < a.length; i++) {
			addAction(a[i]);
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.components.util.ActionList#addAction(javax.swing.Action, int)
	 */
	public synchronized void addAction(int index, Action action)
	{
		actions.add(index, action);
		putValue(ACTIONS, actions.toArray(new Action[size()]));
	}

	/**
	 * Returns Action under provided index.
	 *
	 * @param index DOCUMENT ME!
	 *
	 * @return DOCUMENT ME!
	 */
	public synchronized Action getAction(int index)
	{
		return (Action)actions.get(index);
	}

	/**
	 * Returns Action which command ID corresponds to the provided parameter. 
	 *
	 * @param ID the command key string of action
	 *
	 * @return Action which command ID corresponds to the provided parameter
	 * 
	 * @see Action#ACTION_COMMAND_KEY
	 */
	public synchronized Action getAction(String ID)
	{
		Iterator it = actions.iterator();

		while (it.hasNext()) {
			Action a = (Action)it.next();

			if (ID.equals(a.getValue(ACTION_COMMAND_KEY))) {
				return a;
			}
		}

		return null;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.components.util.ActionList#indexOf(javax.swing.Action)
	 */
	public int indexOf(Action action)
	{
		return actions.indexOf(action);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.components.util.ActionList#removeAction(javax.swing.Action)
	 */
	public void removeAction(Action action)
	{
		actions.remove(action);
		putValue(ACTIONS, actions.toArray(new Action[size()]));
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.components.util.ActionList#removeAction(int)
	 */
	public Action removeAction(int index)
	{
		Action a = (Action)actions.remove(index);
		putValue(ACTIONS, actions.toArray(new Action[size()]));

		return a;
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @return DOCUMENT ME!
	 */
	public int size()
	{
		return actions.size();
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @param e DOCUMENT ME!
	 */
	public void actionPerformed(ActionEvent e)
	{
	}

	/**
	 * Returns all contained actions in an array.
	 *
	 * @return returns all contained actions in an array
	 */
	public Action[] toArray()
	{
		return (Action[])actions.toArray(new Action[actions.size()]);
	}

	/**
	 * Returns iterator over contained actions.
	 *
	 * @return iterator over contained actions
	 */
	public Iterator iterator()
	{
		return actions.iterator();
	}

	/**
	 * DOCUMENT ME!
	 */
	public void clear()
	{
		actions.clear();
	}
}

/* __oOo__ */
