package com.cosylab.gui.components.util;

import java.awt.Color;
import java.util.ArrayList;

/**
 * ColorManager provides a sequence of colors which can be used in charts.
 * 
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 * @version $Id: ColorManager.java,v 1.3 2007-11-27 20:16:25 jbobnar Exp $
 *
 */
public class ColorManager {
	
	public static final Color[] DEFAULT_COLORS= new Color[]{Color.BLACK,Color.RED,Color.BLUE,Color.GREEN,Color.ORANGE,Color.DARK_GRAY,Color.RED.darker(),Color.BLUE.darker(),Color.GREEN.darker(),Color.ORANGE.darker()};
	private Color[] availableColors = DEFAULT_COLORS;
	private ArrayList<Color> unusedColors = new ArrayList<Color>(0);
	private int count = 0;
	
	/**
	 * Loads Default colors as the colors used by this manager.
	 * 
	 */
	public void loadDefaultColors() {
		loadColors(DEFAULT_COLORS);
	}
	
	/**
	 * Loads the colors among which manager will pick a color when requested.
	 * 
	 * @param colors
	 */
	public void loadColors(Color[] colors) {
		if (colors == null) throw new IllegalArgumentException("Colors cannot be null.");
		else if (colors.length == 0) throw new IllegalArgumentException("There should be at least one Color available.");
		synchronized (unusedColors) {
			availableColors = new Color[colors.length];
			System.arraycopy(colors, 0, availableColors, 0, colors.length);
			if (count >= availableColors.length) {
				count %= availableColors.length;
			}
		}
	}
	
	/**
	 * Picks the next color in the sequence.
	 * 
	 * @return
	 */
	public Color pickColor() {
		synchronized (unusedColors) {
			if (unusedColors.size() > 0) {
				Color c = unusedColors.remove(0);
				if (c != null) {
					count++;
					return c;
				}
			}
			return availableColors[count++%availableColors.length];
		}
	}
		
	/**
	 * Drops the color into a container with unused colors. When a new color will
	 * be picked, the manager will first choose among the dropped color and when
	 * there will be no more, it will continue the normal sequence as stated in the
	 * loaded colors.
	 * @param color
	 */
	public void dropColor(Color color) {
		if (color == null) throw new IllegalArgumentException("Dropped color cannot be null.");
		synchronized (unusedColors) {
			unusedColors.add(color);
			count--;
			if (count <=0 ) count = 0;
		}
	}
	
	/**
	 * Returns the current index at which the color is picked from the array.
	 * @return
	 */
	public int getColorIndex() {
		return count;
	}
	
	/**
	 * Sets the index at which the next color will be picked from array.
	 * @param index
	 */
	public void setColorIndex(int index) {
		this.count = index % availableColors.length;
	}
	
	/**
	 * Empties the container with all dropped colors.
	 *
	 */
	public void reset() {
		synchronized (unusedColors) {
			unusedColors.clear();
			count = 0;
		}
	}
	
	/**
	 * Previews the next color that will be picked.
	 * 
	 * @return
	 */
	public Color previewNextColor() {
		synchronized (unusedColors) {
			if (unusedColors.size() > 0) {
				return unusedColors.get(0);
			}
			return availableColors[count%availableColors.length];
		}
	}
	
}
