/*
 * Copyright (c) 2004 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.components.util;

import java.awt.Component;
import java.awt.Font;

import javax.swing.plaf.metal.MetalLookAndFeel;


/**
 * Helper class that provides simple access to Font UI resources
 *
 * @author <a href="mailto:jernej.kamenik@cosylab.com">Jernej Kamenik</a>
 * @version $id$
 *
 * @see javax.swing.plaf.metal.MetalLookAndFeel
 * @see java.awt.Font
 */
public final class FontHelper
{
	private static java.util.List fonts = new java.util.ArrayList();

	/**
	 * Returns the default Font name
	 *
	 * @return String name
	 *
	 * @see javax.swing.plaf.metal.MetalLookAndFeel#getControlTextFont()
	 */
	public static String getDefaultFontName()
	{
		return MetalLookAndFeel.getControlTextFont().getName();
	}

	/**
	 * Returns the default Font size
	 *
	 * @return int size
	 *
	 * @see javax.swing.plaf.metal.MetalLookAndFeel#getControlTextFont()
	 */
	public static int getDefaultFontSize()
	{
		return MetalLookAndFeel.getControlTextFont().getSize();
	}

	/**
	 * Returns the default Font
	 *
	 * @return Font
	 *
	 * @see javax.swing.plaf.metal.MetalLookAndFeel#getUserTextFont()
	 */
	public static Font getDefaultFont()
	{
		return MetalLookAndFeel.getUserTextFont();
	}

	/**
	 * Returns the specified Font with its style changed to the specified
	 * style.
	 *
	 * @param style style
	 * @param inFont inFont
	 *
	 * @return Font
	 */
	public static Font getFontWithStyle(int style, Font inFont)
	{
		for (int i = 0; i < fonts.size(); i++) {
			if (((Font)fonts.get(i)).getName().equals(inFont.getName())
			    && (((Font)fonts.get(i)).getSize() == inFont.getSize())
			    && (((Font)fonts.get(i)).getStyle() == style)) {
				return (Font)fonts.get(i);
			}
		}

		fonts.add(inFont.deriveFont(style));

		return (Font)fonts.get(fonts.size() - 1);
	}

	/**
	 * Returns the specified Font with its name changed to the specified name.
	 *
	 * @param name name
	 * @param inFont inFont
	 *
	 * @return Font
	 */
	public static Font getFontWithName(String name, Font inFont)
	{
		for (int i = 0; i < fonts.size(); i++) {
			if (((Font)fonts.get(i)).getName().equals(name)
			    && (((Font)fonts.get(i)).getSize() == inFont.getSize())
			    && (((Font)fonts.get(i)).getStyle() == inFont.getStyle())) {
				return (Font)fonts.get(i);
			}
		}

		fonts.add(new Font(name, inFont.getStyle(), inFont.getSize()));

		return (Font)fonts.get(fonts.size() - 1);
	}

	/**
	 * Returns the specified Font with its size changed to the specified size.
	 *
	 * @param size size
	 * @param inFont inFont
	 *
	 * @return Font
	 */
	public static Font getFontWithSize(int size, Font inFont)
	{
		for (int i = 0; i < fonts.size(); i++) {
			if (((Font)fonts.get(i)).getName().equals(inFont.getName())
			    && (((Font)fonts.get(i)).getSize() == size)
			    && (((Font)fonts.get(i)).getStyle() == inFont.getStyle())) {
				return (Font)fonts.get(i);
			}
		}

		fonts.add(inFont.deriveFont((float)size));

		return (Font)fonts.get(fonts.size() - 1);
	}

	/**
	 * Returns a new Font with the specified attributes
	 *
	 * @param name name
	 * @param style style
	 * @param size size
	 *
	 * @return Font
	 */
	public static Font getFont(String name, int style, int size)
	{
		for (int i = 0; i < fonts.size(); i++) {
			if (((Font)fonts.get(i)).getName().equals(name)
			    && (((Font)fonts.get(i)).getSize() == size)
			    && (((Font)fonts.get(i)).getStyle() == style)) {
				return (Font)fonts.get(i);
			}
		}

		fonts.add(new Font(name, style, size));

		return (Font)fonts.get(fonts.size() - 1);
	}

	/**
	 * Calculates new font, which would fit provided text inside component.
	 *
	 * @param comp the <code>Component</code> in which text should fit
	 * @param oldFont old font
	 * @param text a text to be fitted
	 * @param columns number of character columns to fit in component
	 * @param minFontSize minimal allowed font size
	 * @param maxFontSize maximal allowed font size
	 *
	 * @return new font, if calculation failes, old font is returned
	 */
	public static Font calculateFittingFont(Component comp, Font oldFont,
	    String text, int columns, int minFontSize, int maxFontSize)
	{
		int newSize = comp.getHeight() * 14 / 20;

		if (newSize <= 0) {
			return oldFont;
		}

		if (newSize > maxFontSize) {
			newSize = maxFontSize;
		} else if (newSize < minFontSize) {
			newSize = minFontSize;
		}

		/**
		 * stores data about text length. If columns equals zero, label font is
		 * immediately resized to fit text. If columns is non-zero, label min
		 * size is columns"m-width" (only use of columns!), but if text does
		 * not fit in it,  font size is changed.  Because newSize resets font
		 * size each time resize() is called, the columned-min width is
		 * defined as: length=getHeight()(14/20)columns"m-width-at-size-1";
		 * |---| why not (7/10)??    before the change, the code at CHECKPOINT
		 * 1 tested minWidth, that always retuned true for first #columns#
		 * characters, but didn't care about text length.
		 */
		Font font = getFontWithSize(newSize, oldFont);

		int textLength = comp.getFontMetrics(font).stringWidth("m") * columns;

		if (text != null) {
			int l = comp.getFontMetrics(font).stringWidth(text);

			if (l > textLength) {
				textLength = l;
			}
		}

		/*System.out.println("newSize=" + newSize + " textLength=" + textLength
		    + " columns=" + columns + " height=" + comp.getHeight() + " width="
		    + comp.getWidth());*/

		//CHECKPOINT 1_
		if (textLength > comp.getWidth() * 0.9 && textLength != 0) {
			//Debug.out("shrinking resizable label");
			newSize = (int)(newSize * comp.getWidth() * 0.9 / textLength);

			if (newSize < 1) {
				newSize = 1;
			}

			font = getFontWithSize(newSize, font);

			//System.out.println("shrinking newSize=" + newSize);
		}

		// we ignore if font is changed for less than 10%
		if ((double)Math.abs(oldFont.getSize() - font.getSize()) / (double)oldFont
		    .getSize() < 0.1) {
			/*System.out.println("change too small "
			    + (double)Math.abs(oldFont.getSize() - font.getSize()) / (double)oldFont
			    .getSize());*/
			return oldFont;
		}

		return font;
	}
}

/* __oOo__ */
