/*
 * @@COPYRIGHT@@
 */
package com.cosylab.gui.components.util;

import java.awt.Image;

import java.net.URL;
import java.util.HashMap;

import javax.swing.Icon;
import javax.swing.ImageIcon;

import sun.awt.image.InputStreamImageSource;


/**
 * IconHelper acts as an interface to the all icons currently in classpath.
 * IconHelper is a singleton class that maintains internal cache to all
 * accessed icons. This class may be used to minimize the memory consumption
 * where application uses static icons. If icons are changed independently,
 * this class should not be used.
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public final class IconHelper {
    /** cache for icons. */
    private static final HashMap cache = new HashMap();

    /** singleton instance of this class. */
	// notused private static IconHelper instance = null;

    /**
     * Protected constructor for creation of singleton class.
     */
    private IconHelper() {
        super();
    }

    /**
     * Returns instance of a singleton IconHelper.
     *
     * @return IconHelper
     */
    /* notused private static synchronized IconHelper getInstance() {
        if (instance == null) {
            instance = new IconHelper();
        }

        return instance;
    }*/

    /**
     * Returns icon from the cache. If icon cannot be found in cache, return
     * value will be <code>null</code>.
     *
     * @param name String
     * @return Icon instance of an icon or null if not found
     */
    private static Icon getCachedIcon(String name) {
        if (cache.containsKey(name)) {
            return (Icon) cache.get(name);
        }

        return null;
    }

    /**
     * Puts icon in the cache. If an icon with the specified name has already been stored,
     * current icon will be overwritten.
     *
     * @param name
     * @param icon
     */
    private static synchronized void setCachedIcon(String name, Icon icon) {
        cache.put(name, icon);
    }

    /**
     * Attempts to load the icon from the resources in classpath.
     * If the icon cannot be found or other error occurs, the return value is null.
     *
     * @param resource String
     * @param cl ClassLoader
     * @return Icon icon instance or null
     */
    private static final Icon loadResource(String resource, ClassLoader cl) {
        try {
        	URL url = cl.getResource(resource);
        	if (!(url.getContent() instanceof InputStreamImageSource)) return null;
            return new ImageIcon(url);
        } catch (Exception e) {
            return null;
        }
    }

    /**
     * Returns instance of the icon by loading it using the <code>
     * ClassLoader</code> instance provided. Icons are cached. If an icon with
     * the same name is requested several times, <code>createIcon</code> will
     * return the same instance each time. If icons will be manipulated
     * individually, do not use this method to load icons. If icon is not found
     * in the classpath, return value will be null.
     *
     * @param name String
     * @param cl ClassLoader Instance of class loader to use
     * @return Icon Instance of Icon or null if icon could not be found.
     */
    public static Icon createIcon(String name, ClassLoader cl) {
        Icon icon = getCachedIcon(name);

        if (icon == null) {
            icon = loadResource(name, cl);

            if (icon != null) {
                setCachedIcon(name, icon);
            }
        }

        return icon;
    }

    /**
     * Returns instance of the icon. Icons are cached. If an icon with the same
     * name is requested several times, <code>createIcon</code> will return the
     * same instance each time. If icons will be manipulated individually, do
     * not use this method to load icons. If icon is not found in the classpath,
     * return value will be null.
     *
     * @param name String
     * @return Icon Instance of Icon or null if icon could not be found.
     */
    public static Icon createIcon(String name) {
        return createIcon(name, IconHelper.class.getClassLoader());
    }

    /**
     *
     * @param name String
     * @return Image
     */
    public static Image createImage(String name) {
        return createImage(name, IconHelper.class.getClassLoader());
    }

    /**
     *
     * @param name String
     * @param cl ClassLoader
     * @return Image
     */
    public static Image createImage(String name, ClassLoader cl) {
        ImageIcon icon = ((ImageIcon) createIcon(name, cl));

        if (icon == null) {
            return null;
        }

        return icon.getImage();
    }
}
