/*
 * Created on Jun 23, 2003
 *
 * @@COPYRIGHT@@
 */
package com.cosylab.gui.components.util;

import java.awt.Component;
import java.awt.Point;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import javax.swing.Action;
import javax.swing.JComponent;
import javax.swing.JPopupMenu;
import javax.swing.SwingUtilities;


/**
 * <code>PopupManager</code> helps simple GUI widget to handle popup menu
 * in transparent and extensible way.
 *
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 * @version $id$
 */
public class PopupManager {
    private JComponent owner;
    private ActionList list = new ActionList();
    private JPopupMenu popup;
    private boolean enabled=true;

    /**
     *
     */
    public PopupManager(JComponent owner) {
        super();
        this.owner = owner;
    }

    /**
     * Creates manager, which can automatically show a popup menu.
     * @param owner owner, for which manager manages popup menu
     * @param auto <code>true</code> if menu popups automatically
     */
    public PopupManager(JComponent owner, boolean auto) {
        super();
        this.owner = owner;

        if (auto) {
            owner.addMouseListener(getMouseHook());
        }
    }



    /**
     * Returns owner of this manager. Usually it it GUI widgets which uses
     * this manager to manage popup menu.
     * @return owner of this manager
     */
    public JComponent getOwner() {
        return owner;
    }

    /**
     * Adds action to this manager. If action added is null, action is translateed to manu separator.
     * @param a action to be added, if null, action is translateed to manu separator.
     */
    public void addAction(Action a) {
        list.addAction(a);
        popup = null;
    }
    /**
     * Adds action to this manager. If action added is null, action is translateed to manu separator.
     * @param a action to be added, if null, action is translateed to manu separator.
     */
    public void addActions(Action[] a) {
    	for (int i = 0; i < a.length; i++) {
            list.addAction(a[i]);
		}
        popup = null;
    }

    public boolean containsAction(Action a) {
        return list.contains(a);
    }

    public boolean containsActionForName(String name) {
        return list.containsActionWithName(name);
    }

    public int firstIndexOf(Action a) {
        return list.indexOf(a);
    }

    public int firstIndexOfName(String name) {
        return list.firstIndexOfActionWithName(name);
    }

    public Action getAction(int index) {
        return list.getAction(index);
    }

    public void removeAction(Action a) {
        list.removeAction(a);
        popup = null;
    }

    public Action removeAction(int index) {
        popup = null;

        return list.removeAction(index);
    }
    public void removeActions(Action[] a) {
        popup = null;

        for (int i = 0; i < a.length; i++) {
			list.removeAction(a[i]);
		}
    }
    
    public Action[] removeAllActions() {
        popup = null;
        Action[] a= list.toArray();
        list.clear();
        return a;
    }
    
    public int actionCount() {
    	return list.size();
    }
    

    public Action[] getActions() {
        return list.toArray();
    }

    /**
     * Returns popup menu constructed from actions registered at the manager.
     * @return popup menu
     */
    public JPopupMenu getJPopupMenu() {
        if (popup == null) {
            popup = Actions.createPopupMenu(list);
        }

        return popup;
    }

    /**
     * Shows popup menu at given location.
     * @param x
     * @param y
     * @param c
     */
    public void showPopup(Component c, int x, int y) {
        if (list.size() > 0) {
            getJPopupMenu().show(c, x, y);
        }
    }

    /**
     * Convenience method for creating mouse listener, which fires
     * this manager to produce popup
     * @return new mouse listener, which shows popup
     */
    private MouseHook mouseHook;
    public MouseHook getMouseHook() {
    	if (mouseHook == null) {
	        mouseHook = new MouseHook();
        }

        return mouseHook;
    }

    private class MouseHook extends MouseAdapter {
        private Point lastPoint;
        private Object source;

        /**
         * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
         */
        public void mousePressed(MouseEvent e) {
            if (SwingUtilities.isRightMouseButton(e)) {
                lastPoint = e.getPoint();
                source = e.getSource();
            } else {
                source = null;
                lastPoint = null;
            }
        }

        /**
         * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
         */
        public void mouseReleased(MouseEvent e) {
            if (SwingUtilities.isRightMouseButton(e)) {
                if ((e.getSource() == source) && (lastPoint != null)) {
                    //if (lastPoint.distance(e.getPoint()) < 20) {
                    if (enabled) showPopup(e.getComponent(), e.getPoint().x, e.getPoint().y);

                    //}
                }

                lastPoint = null;
            }
        }
    }
	/**
	 * Returns <code>true</code> if popup manager is enabled and shows popup menu.
	 * @return
	 */
	public boolean isEnabled() {
		return enabled;
	}

	/**
	 * Enambes or disables popup manager.
	 * @param b
	 */
	public void setEnabled(boolean b) {
		enabled= b;
	}
}
