/*
 * Copyright (c) 2003 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */
package com.cosylab.gui.components.util;

import java.awt.event.ActionEvent;

import javax.swing.AbstractAction;
import javax.swing.Icon;

import com.cosylab.util.Debug;


/**
 * An extension of the <code>javax.swing.Action</code> interface which provides
 * <code>selected</code> bound property. It can be used to specify an action
 * which switches between two different states. GUI components will render
 * these actions with check boxes or toggle buttons.
 *
 * <p>
 * To use this action, either instantiate this class and register as a
 * propertyChangeListener, responding to events that denote a change in
 * "selected" property, or extend this class, overriding actionPerformed()
 * method.
 * </p>
 *
 * @author <a href="mailto:miha.kadunc@cosylab.com">Miha Kadunc</a>
 * @version $id$
 */
public class ToggleAction extends AbstractAction {
    /** Icon used when the action is selected */
    public static final String SMALL_SELECTED_ICON = "SmallSelectedIcon";
    /** Large icon used when the action is selected */
    public static final String LARGE_SELECTED_ICON = "LargeSelectedIcon";
    private boolean selected = false;
    private boolean internal = false;

    /**
     * Constructor with default values.
     */
    public ToggleAction() {
        super();
    }

    /**
     * Constructor with name.
     *
     * @param name the name of the action
     */
    public ToggleAction(String name) {
        super(name);
    }

    /**
     * Constructor with name and selection flag.
     *
     * @param name the name of the action
     * @param selected the selection flag
     */
    public ToggleAction(String name, boolean selected) {
        super(name);
        this.selected = selected;
    }

    /**
     * Contructor with name and icon.
     *
     * @param name
     * @param icon
     */
    public ToggleAction(String name, Icon icon) {
        super(name, icon);
    }

    /**
     * Contructor with name, icon and selection flag.
     *
     * @param name the name of action
     * @param icon the icon to be used GUI element
     * @param selected the selection flag
     */
    public ToggleAction(String name, Icon icon, boolean selected) {
        super(name, icon);
        this.selected = selected;
    }

    /**
     * Sets the <code>selected</code> property. Fires propertyChange events and
     * calls actionPerformed() method with <code>this</code> as event source.
     *
     * @param selected
     */
    public void setSelected(boolean selected) {
        if (this.selected != selected) {
            if (!internal) {
                actionPerformed(new ActionEvent(this, selected ? 1 : 0,
                        "selection_changed"));
            } else {
                internal = false;
                this.selected = selected;
                selectionChanged(selected);
                firePropertyChange("selected", new Boolean(!selected),
                    new Boolean(selected));
            }
        }
    }

    /**
     * @return whether this toggle action is selected.
     */
    public boolean isSelected() {
        return selected;
    }

    /**
     * Overriden to modify selection status. Called by the GUI when invoked.
     */
    public void actionPerformed(ActionEvent e) {
        internal = true;
        setSelected(!selected);
    }

    /**
     * Override this method to provide custom event handling when selection
     * changes.
     *
     * @param newSelected
     */
    protected void selectionChanged(boolean selected) {
    	Debug.out(""+selected);
    }
}


/* __oOo__ */
