/*
 * Copyright (c) 2003 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.displayers;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.beans.PropertyVetoException;

import javax.swing.AbstractAction;
import javax.swing.JLabel;
import javax.swing.JPanel;

import com.cosylab.gui.adapters.Converter;
import com.cosylab.gui.components.AbstractDisplayerPanelCustomizer;
import com.cosylab.gui.components.ResizableTextLabel;
import com.cosylab.gui.components.customizer.AbstractCustomizerPanel;
import com.cosylab.gui.components.util.ColorHelper;
import com.cosylab.gui.components.util.CosyUIElements;
import com.cosylab.gui.components.util.PopupManageable;
import com.cosylab.gui.components.util.PopupManager;
import com.cosylab.gui.components.util.ScreenCapturer;


/**
 * <code>com.cosylab.gui.displayers.DoubleDisplayer</code> based on the
 * <code>com.cosylab.gui.components.NumberField</code>  component.
 *
 * @author <a href="mailto:jernej.kamenik@cosylab.com">Jernej Kamenik</a>
 * @version $id$
 */
public abstract class AbstractDisplayerPanel extends JPanel implements Displayer,
	PopupManageable, ConvertibleDisplayer
{
	private DataState dataState = new DataState(DataState.UNDEFINED);
	private int suspendCount = 0;
	private PopupManager popupManager;
	private DataSource dataSource;
	private Converter converter;
	
	private ResizableTextLabel titleLabel;
	private AbstractCustomizerPanel customizer;

	/**
	 * Constructor for AbstractDisplayerPanel.
	 */
	public AbstractDisplayerPanel()
	{
		super();
		setBorder(CosyUIElements.getPlainBorder(true));

		Dimension size = new Dimension(200, 50);
		setPreferredSize(size);
	}
	
	public AbstractCustomizerPanel getCustomizer()
	{
		if (customizer == null) {
			customizer = AbstractCustomizerPanel.findCustomizer(this);
		}

		return customizer;
	}
	
	protected ResizableTextLabel getTitleLabel() {
		if (titleLabel == null) {
			titleLabel = new ResizableTextLabel();
			titleLabel.setHorizontalAlignment(JLabel.CENTER);
			titleLabel.addMouseListener(getPopupManager().getMouseHook());
			titleLabel.setEnabled(isEnabled());
		}

		return titleLabel;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDefaultDataConsumer()
	 */
	public DataConsumer getDefaultDataConsumer()
	{
		return this;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedCharacteristics()
	 */
	public String[] getSupportedCharacteristics()
	{
		return DisplayerUtilities.COMMON_NUMERIC_DISPLAYER_CHARACTERISTICS;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#updateDataState(com.cosylab.gui.displayers.DataState)
	 */
	public void updateDataState(DataState state)
	{
		DataState old = dataState;
		dataState = state;
		firePropertyChange(DATA_STATE, old, dataState);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.CommonDisplayer#cleanup()
	 */
	public void cleanup()
	{
		updateDataState(new DataState(DataState.NOT_INITIALIZED));
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.CommonDisplayer#getDataState()
	 */
	public DataState getDataState()
	{
		return dataState;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.CommonDisplayer#isSuspended()
	 */
	public boolean isSuspended()
	{
		return suspendCount > 0;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.CommonDisplayer#resume()
	 */
	public void resume()
	{
		suspendCount--;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.CommonDisplayer#suspend()
	 */
	public void suspend()
	{
		suspendCount++;
	}

	/**
	 * Returns popup manager for adding popup actions.
	 *
	 * @see com.cosylab.gui.components.util.PopupManageable#getPopupManager()
	 */
	public PopupManager getPopupManager()
	{
		if (popupManager == null) {
			popupManager = new PopupManager(this, true);
			popupManager.addAction(new AbstractAction("Preferences...") {
				private static final long serialVersionUID = 1L;
					public void actionPerformed(ActionEvent e)
					{
						getCustomizer().showDialog();
					}
				});
			popupManager.addAction(new AbstractAction("Capture screen..."){
				private static final long serialVersionUID = 1L;

				public void actionPerformed(ActionEvent e){
				   	ScreenCapturer sc = new ScreenCapturer(AbstractDisplayerPanel.this);
					sc.showScreenDialog();
				   }
			});
		}

		return popupManager;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.core.CosyComponent#destroy()
	 */
	public void destroy()
	{
		popupManager = null;
		dataState = null;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.Displayer#getDataSource()
	 */
	public DataSource getDataSource() {
		return dataSource;
	}
	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.Displayer#setDataSource(com.cosylab.gui.displayers.DataSource)
	 */
	public void setDataSource(DataSource dataSource)
		throws PropertyVetoException
	{
		DisplayerUtilities.prepareNewDataSource(dataSource,this);
		
		DataSource old= this.dataSource;
		this.dataSource = dataSource;
		
		firePropertyChange(DATA_SOURCE,old,dataSource);
	}

	
	/**
	 * Returns the converter.
	 * @return Returns the converter.
	 */
	public Converter getConverter() {
		return converter;
	}

	
	/**
	 * Sets new converter.
	 * @param converter The converter to set.
	 * @throws PropertyVetoException if set fails
	 */
	public void setConverter(Converter converter) throws PropertyVetoException {
		DisplayerUtilities.prepareNewConverter(converter,this);
		
		Converter old= this.converter;
		this.converter = converter;
		
		firePropertyChange(CONVERTER_PROPERTY,old,dataSource);
	}
	
	/**
	 * DOCUMENT ME!
	 *
	 * @return
	 */
	public int getTitleMaximumFontSize()
	{
		return getTitleLabel().getMaximumFontSize();
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @return
	 */
	public int getTitleMinimumFontSize()
	{
		return getTitleLabel().getMinimumFontSize();
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @param max
	 */
	public void setTitleMaximumFontSize(int max)
	{
		int old= getTitleMaximumFontSize();
		getTitleLabel().setMaximumFontSize(max);
		firePropertyChange(AbstractDisplayerPanelCustomizer.MAX_TITLE_FONT_SIZE,
		    old, max);
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @param min
	 */
	public void setTitleMinimumFontSize(int min)
	{
		int old= getTitleMinimumFontSize();
		getTitleLabel().setMaximumFontSize(min);
		firePropertyChange(AbstractDisplayerPanelCustomizer.MIN_TITLE_FONT_SIZE,
		    old, min);
	}
	/**
     * Sets visibility of title label.
     * 
     * @param value
     *        True if title should be visible.
     */
	public void setTitleVisible(boolean value) {
		getTitleLabel().setVisible(value);
	}

	/**
     * Returns visibility of title.
     * 
     * @return True if title is visible.
     */
	public boolean isTitleVisible() {
		return getTitleLabel().isVisible();
	}

	/**
     * Returns title of this component.
     * 
     * @see com.cosylab.gui.displayers.Displayer#getTitle()
     */
	public String getTitle() {
		return getTitleLabel().getText();
	}

	/**
     * Sets the title of component.
     * 
     * @see com.cosylab.gui.displayers.Displayer#setTitle(java.lang.String)
     */
	public void setTitle(String value) {
		getTitleLabel().setText(value);
	}
	
	/*
	 * (non-Javadoc)
	 * @see javax.swing.JComponent#setBackground(java.awt.Color)
	 */
	@Override
	public void setBackground(Color bg) {
		super.setBackground(bg);
		getTitleLabel().setBackground(bg);
		
	}
	
	/*
	 * (non-Javadoc)
	 * @see javax.swing.JComponent#setForeground(java.awt.Color)
	 */
	@Override
	public void setForeground(Color fg) {
		super.setForeground(fg);
		getTitleLabel().setForeground(fg);
	}


}

/* __oOo__ */
