/*
 * Copyright (c) 2004 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.displayers;

import java.util.BitSet;
import java.util.HashMap;
import java.util.Map;

import com.cosylab.util.BitSetUtilities;
import com.cosylab.util.CommonException;


/**
 * <code>LongConsumerMulticaster</code> casts different data consumers to
 * LongConsumer. ObjectConsumer cast must receive <code>Number</code>
 * instances. Delegate must be set, before any other method can be used.
 *
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 * @version $Id: LongConsumerMulticaster.java,v 1.5 2006/09/07 12:08:30 ikriznar Exp $
 *
 * @since Dec 4, 2003.
 */
public class LongConsumerMulticaster implements DoubleConsumer, NumberConsumer,
	LongConsumer, StringConsumer, ObjectConsumer, PatternConsumer
{
	/** Types of data consumers, that are supported by this multicaster. */
	public static final Class[] SUPPORTED_CONSUMER_TYPES = {
			DoubleConsumer.class, LongConsumer.class, NumberConsumer.class,
			ObjectConsumer.class, StringConsumer.class, PatternConsumer.class
		};
	LongConsumer delegate;

	/**
	 * If multicaster supports type, new multicaster is created, otherwise
	 * <code>null</code>is returned.
	 *
	 * @param type the requested consumer type
	 * @param delegate the consumer to which all call sare delegated
	 *
	 * @return new multicaster if type is supported, otherwise
	 *         <code>null</code>
	 */
	public static final LongConsumer createDataConsumer(Class type,
	    LongConsumer delegate)
	{
		if (type == LongConsumer.class) {
			return delegate;
		}

		if (type == NumberConsumer.class || type == StringConsumer.class
		    || type == ObjectConsumer.class || type == DoubleConsumer.class || type==PatternConsumer.class) {
			return new LongConsumerMulticaster(delegate);
		}

		return null;
	}

	/**
	 * Creates a new LongConsumerMulticaster object.
	 *
	 * @param delegate the consumer to which all call sare delegated
	 */
	public LongConsumerMulticaster(LongConsumer delegate)
	{
		this.delegate = delegate;
	}

	/**
	 * Creates a new LongConsumerMulticaster object.
	 */
	public LongConsumerMulticaster()
	{
	}

	/**
	 * Returns the consumer to which all call sare delegated.
	 *
	 * @return the consumer to which all call sare delegated
	 */
	public LongConsumer getDelegate()
	{
		return delegate;
	}

	/**
	 * Sets the consumer to which all call sare delegated.
	 *
	 * @param delegate the consumer to which all calls are delegated
	 */
	public void setDelegate(LongConsumer delegate)
	{
		this.delegate = delegate;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.LongConsumer#updateValue(long, double)
	 */
	public void updateValue(long timestamp, double value)
		throws CommonException
	{
		delegate.updateValue(timestamp, (long)value);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.NumberConsumer#updateValue(long, java.lang.Number)
	 */
	public void updateValue(long timestamp, Number value)
		throws CommonException
	{
		delegate.updateValue(timestamp, value.longValue());
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.LongConsumer#updateValue(long, long)
	 */
	public void updateValue(long timestamp, long value)
		throws CommonException
	{
		delegate.updateValue(timestamp, value);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDataConsumer(java.lang.Class)
	 */
	public DataConsumer getDataConsumer(Class type)
	{
		return delegate.getDataConsumer(type);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDefaultDataConsumer()
	 */
	public DataConsumer getDefaultDataConsumer()
	{
		return delegate.getDefaultDataConsumer();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getName()
	 */
	public String getName()
	{
		return delegate.getName();
	}

	/* (non-Javadoc)
	 * @see DataConsumer#setCharacteristics(Map)
	 */
	public void setCharacteristics(Map characteristics)
	{
		String f = (String)characteristics.get(CommonDisplayer.C_FORMAT);

		if (f != null) {
			HashMap map = new HashMap(characteristics);
			map.put(CommonDisplayer.C_FORMAT, DisplayerUtilities.toLongFormat(f));
			delegate.setCharacteristics(map);
		} else {
			delegate.setCharacteristics(characteristics);
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#updateDataState(com.cosylab.gui.displayers.DataState)
	 */
	public void updateDataState(DataState state)
	{
		delegate.updateDataState(state);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.StringConsumer#updateValue(long, java.lang.String)
	 */
	public void updateValue(long timestamp, String value)
		throws CommonException
	{
		delegate.updateValue(timestamp, Long.parseLong(value));
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.ObjectConsumer#updateValue(long, java.lang.Object)
	 */
	public void updateValue(long timestamp, Object value)
		throws CommonException
	{
		delegate.updateValue(timestamp, ((Number)value).longValue());
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.PatternConsumer#updateValue(long, java.util.BitSet)
	 */
	public void updateValue(long timestamp, BitSet value)
		throws CommonException
	{
		delegate.updateValue(timestamp, BitSetUtilities.toLong(value));
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedCharacteristics()
	 */
	public String[] getSupportedCharacteristics()
	{
		return delegate.getSupportedCharacteristics();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedConsumerTypes()
	 */
	public Class[] getSupportedConsumerTypes()
	{
		return SUPPORTED_CONSUMER_TYPES;
	}
}

/* __oOo__ */
