/*
 * Copyright (c) 2003 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.displayers;

import java.util.Map;

import com.cosylab.util.CommonException;


/**
 * <code>StringConsumerMulticaster</code> casts different data consumers to
 * StringConsumer. ObjectConsumer cast must receive <code>Number</code>
 * instances. Delegate must be set, before any other method can be used.
 *
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 * @version $Id: StringConsumerMulticaster.java,v 1.3 2007-02-16 07:26:12 ikriznar Exp $
 *
 * @since Dec 4, 2003.
 */
public class StringConsumerMulticaster implements DoubleConsumer,
	NumberConsumer, LongConsumer, StringConsumer, ObjectConsumer, 
	DoubleSeqConsumer, LongSeqConsumer, StringSeqConsumer, ObjectSeqConsumer
{
	StringConsumer delegate;

	/** Types of data consumers, that are supported by this multicaster. */
	public static final Class[] SUPPORTED_CONSUMER_TYPES = {
		LongConsumer.class, NumberConsumer.class, DoubleConsumer.class,
		ObjectConsumer.class, StringConsumer.class
	};

	/**
	 * If multicaster supports type, new multicaster is created, otherwise
	 * <code>null</code>is returned.
	 *
	 * @param type the requested consumer type
	 * @param delegate the consumer to which all call sare delegated
	 *
	 * @return new multicaster if type is supported, otherwise
	 *         <code>null</code>
	 */
	public static final StringConsumer createDataConsumer(Class type,
	    StringConsumer delegate)
	{
		if (type == StringConsumer.class) {
			return delegate;
		}

		if (type == LongConsumer.class || type == NumberConsumer.class
		    || type == DoubleConsumer.class || type == ObjectConsumer.class
		    || type == DoubleSeqConsumer.class || type == LongSeqConsumer.class
		    || type == StringSeqConsumer.class || type == ObjectSeqConsumer.class) {
			return new StringConsumerMulticaster(delegate);
		}

		return null;
	}

	/**
	 * Creates a new StringConsumerMulticaster object.
	 *
	 * @param delegate the consumer to which all call sare delegated
	 */
	public StringConsumerMulticaster(StringConsumer delegate)
	{
		this.delegate = delegate;
	}

	/**
	 * Creates a new StringConsumerMulticaster object.
	 */
	public StringConsumerMulticaster()
	{
	}

	/**
	 * Returns the consumer to which all call sare delegated.
	 *
	 * @return the consumer to which all call sare delegated
	 */
	public StringConsumer getDelegate()
	{
		return delegate;
	}

	/**
	 * Sets the consumer to which all call sare delegated.
	 *
	 * @param delegate the consumer to which all calls are delegated
	 */
	public void setDelegate(StringConsumer delegate)
	{
		this.delegate = delegate;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.StringConsumer#updateValue(long, double)
	 */
	public void updateValue(long timestamp, double value)
		throws CommonException
	{
		delegate.updateValue(timestamp, Double.toString(value));
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.NumberConsumer#updateValue(long, java.lang.Number)
	 */
	public void updateValue(long timestamp, Number value)
		throws CommonException
	{
		if (value==null) {
			delegate.updateValue(timestamp, "N/A");
			return;
		}
		delegate.updateValue(timestamp, value.toString());
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.LongConsumer#updateValue(long, long)
	 */
	public void updateValue(long timestamp, long value)
		throws CommonException
	{
		delegate.updateValue(timestamp, Long.toString(value));
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDataConsumer(java.lang.Class)
	 */
	public DataConsumer getDataConsumer(Class type)
	{
		return delegate.getDataConsumer(type);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDefaultDataConsumer()
	 */
	public DataConsumer getDefaultDataConsumer()
	{
		return delegate.getDefaultDataConsumer();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getName()
	 */
	public String getName()
	{
		return delegate.getName();
	}

	/* (non-Javadoc)
	 * @see DataConsumer#setCharacteristics(Map)
	 */
	public void setCharacteristics(Map characteristics)
	{
		delegate.setCharacteristics(characteristics);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#updateDataState(com.cosylab.gui.displayers.DataState)
	 */
	public void updateDataState(DataState state)
	{
		delegate.updateDataState(state);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.StringConsumer#updateValue(long, java.lang.String)
	 */
	public void updateValue(long timestamp, String value)
		throws CommonException
	{
		if (value==null) {
			delegate.updateValue(timestamp, "N/A");
			return;
		}
		delegate.updateValue(timestamp, value);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.ObjectConsumer#updateValue(long, java.lang.Object)
	 */
	public void updateValue(long timestamp, Object value)
		throws CommonException
	{
		if (value==null) {
			delegate.updateValue(timestamp, "N/A");
			return;
		}
		delegate.updateValue(timestamp, value.toString());
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedCharacteristics()
	 */
	public String[] getSupportedCharacteristics()
	{
		return delegate.getSupportedCharacteristics();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedConsumerTypes()
	 */
	public Class[] getSupportedConsumerTypes()
	{
		return SUPPORTED_CONSUMER_TYPES;
	}
	
	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.ObjectSeqConsumer#updateValue(long, java.lang.Object[])
	 */
	public void updateValue(long timestamp, Object[] value) throws CommonException {
		if (value==null) {
			delegate.updateValue(timestamp, "N/A");
			return;
		}
		delegate.updateValue(timestamp, DisplayerUtilities.toString(value));
	}
	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DoubleSeqConsumer#updateValue(long, double[])
	 */
	public void updateValue(long timestamp, double[] value) throws CommonException {
		if (value==null) {
			delegate.updateValue(timestamp, "N/A");
			return;
		}
		StringBuilder s= new StringBuilder(16*value.length);
		s.append('[');
		if (value.length>0) {
			s.append(value[0]);
		}
		for (int i = 1; i < value.length; i++) {
			s.append(',');
			s.append(value[i]);
		}
		s.append(']');
		delegate.updateValue(timestamp, s.toString());
	}
	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.LongSeqConsumer#updateValue(long, long[])
	 */
	public void updateValue(long timestamp, long[] value) throws CommonException {
		if (value==null) {
			delegate.updateValue(timestamp, "N/A");
			return;
		}
		StringBuilder s= new StringBuilder(16*value.length);
		s.append('[');
		if (value.length>0) {
			s.append(value[0]);
		}
		for (int i = 1; i < value.length; i++) {
			s.append(',');
			s.append(value[i]);
		}
		s.append(']');
		delegate.updateValue(timestamp, s.toString());
	}
	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.StringSeqConsumer#updateValue(long, java.lang.String[])
	 */
	public void updateValue(long timestamp, String[] value) throws CommonException {
		if (value==null) {
			delegate.updateValue(timestamp, "N/A");
			return;
		}
		delegate.updateValue(timestamp, DisplayerUtilities.toString(value));
	}
}

/* __oOo__ */
