/*
 * Copyright (c) 2003 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

package com.cosylab.gui.property.editors;

import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;

import javax.swing.JCheckBox;



/**
 * JBooleanCheckBox is a component which serves as an editor for boolean
 * values. It displays a checkBox and a string representation of the value,
 * i.e. either "true" or "false". It can also display an arbitrary user text
 * which labels the checkBox.
 *
 * @author <a href="mailto:miha.kadunc@cosylab.com">Miha Kadunc</a>
 * @version $id$
 */
public class BooleanEditor extends JCheckBox implements PropertyEditor
{
	private String description=null;
	/**
	 * Constructor for JBooleanCheckBox. Constructs a check box with default
	 * value set to false.
	 */
	public BooleanEditor()
	{
		this(false);
	}

	/**
	 * Constructor for JBooleanCheckBox. Constructs a check box with default
	 * value set to <code>value</code>.
	 *
	 * @param value the value to set.
	 */
	public BooleanEditor(boolean value)
	{
		super();
		setSelected(value);
		super.setText(Boolean.toString(value));
		init();
	}

	/**
	 * Constructor for JBooleanCheckBox. Constructs a check box with default
	 * value set to <code>value</code>.
	 *
	 * @param description desription of edited value
	 * @param value the value to set.
	 */
	public BooleanEditor(String description, boolean value)
	{
		this(value);
		setDescription(description);
	}

	/**
	 * returns a <code>Boolean</code> object with the value set. To retrieve a
	 * primitive <code>boolean</code>, use isSelected() method
	 *
	 * @return Object a Boolean containing the selected value.
	 */
	public Object getPropertyValue()
	{
		return new Boolean(isSelected());
	}

	/**
	 * If and only if the text is "true" this checkbox is selected.
	 *
	 * @param text the text to be set as boolean
	 *
	 * @see Boolean#getBoolean(java.lang.String)
	 * @see #setValue(boolean)
	 */
	public void setText(String text)
	{
		if ("true".equalsIgnoreCase(text)) {
			setSelected(true);
		} else if ("false".equalsIgnoreCase(text)) {
			setSelected(false);
		}
		super.setText(text);
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @param sel the value to be set
	 *
	 * @see setValue
	 */
	public void setSelected(boolean sel)
	{
		boolean oldValue = isSelected();

		if (oldValue != sel) {
			super.setSelected(sel);
			//super.setText(Boolean.toString(sel));
			firePropertyChange(PROPERTY_VALUE_NAME, oldValue, sel);
			firePropertyChange("value", oldValue, sel);
		}
	}


	private void init()
	{
		addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent e) {
				if (description==null) {
					BooleanEditor.super.setText(Boolean.toString(isSelected()));
				}
				firePropertyChange(PROPERTY_VALUE_NAME,!isSelected(),isSelected());
			}
		});
	}

	/**
	 * Sets the displayed value. Should be java.lang.Boolean
	 *
	 * @param value the value to be set.
	 *
	 * @return DOCUMENT ME!
	 *
	 * @see ValueEditor#setValue(Object)
	 */
	public boolean setPropertyValue(Object value)
	{
		if (value == null) {
			setSelected(false);
		} else if (value instanceof Boolean) {
			setSelected(((Boolean)value).booleanValue());

			return true;
		}

		return false;
	}

	/**
	 * @see com.cosylab.gui.property.editors.PropertyEditor#getDescription()
	 */
	public String getDescription()
	{
		return description;
	}

	/**
	 * @see com.cosylab.gui.property.editors.PropertyEditor#getEditor()
	 */
	public PropertyEditor getEditor()
	{
		return this;
	}

	/**
	 * @see com.cosylab.gui.property.editors.PropertyEditor#setDescription(String)
	 */
	public void setDescription(String description)
	{
		this.description= description; 
		setText(description);
	}
}

/* __oOo__ */
