package de.desy.acop.displayers;

import java.beans.PropertyVetoException;
import java.util.ArrayList;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import com.cosylab.gui.displayers.DataConsumer;
import com.cosylab.gui.displayers.DataSource;
import com.cosylab.util.CommonException;

import de.desy.acop.chart.AcopGraphStyleEnum;
import de.desy.acop.displayers.chart.AcopChartConsumer;
import de.desy.acop.displayers.chart.AcopTrendChartConsumer;
import de.desy.acop.displayers.chart.HistoryParameters;
import de.desy.acop.displayers.chart.THistoryConnector;
import de.desy.acop.displayers.tools.AcopTrendChartEvent;
import de.desy.acop.displayers.tools.AcopTrendChartListener;
import de.desy.acop.displayers.tools.AcopGraphParameters;
import de.desy.acop.displayers.tools.AcopGraphHistoryParameters;

public class AcopTrendChart extends AcopChartReorg {
	
	private static final long serialVersionUID = 1L;

	public static final String TIME_SPAN = "timeSpan";
	public static final String AUTOSCALE_TO_LIVE_DATA = "autoScaleToLiveData";
	public static final String CHECK_FOR_BAD_DATA = "checkForBadData";
	public static final String TIMEOUT = "timeout";
	
	private boolean scaleInitialized = false;
	private boolean autoScaleToLiveData = true;
	private double loadedHistoryStart;
	private double loadedHistoryStop;
	private boolean loading = false;
	private THistoryConnector tHistoryConnector;
	private ExecutorService executorService;
	private ArrayList<AcopTrendChartListener> acopTrendChartListeners = new ArrayList<AcopTrendChartListener>();
	
	public AcopTrendChart() {
		super();
		getAcop().setGraphStyle(AcopGraphStyleEnum.TimeLin.ordinal());
	}
	
	public THistoryConnector getTHistoryConnector() {
		if (tHistoryConnector == null) {
			tHistoryConnector = new THistoryConnector();
		}
		return tHistoryConnector;
	}
	
	/**
	 * Returns the single thread executor service.
	 * @return
	 */
	public ExecutorService getExecutorService() {
		if (executorService == null) {
			executorService = new ThreadPoolExecutor(1, 1, 0, TimeUnit.NANOSECONDS,
		              new LinkedBlockingQueue<Runnable>()){
				@Override
				protected void afterExecute(Runnable r, Throwable t) {
					super.afterExecute(r, t);
					if (t != null) {
						reportProgress(new AcopTrendChartEvent(AcopTrendChart.this, AcopTrendChartEvent.EventType.ERROR, t));
					}
				}
			};
		}
		return executorService;
	}
	
	private void loadHistory() {
		if (getLastTimestamp()-getTimeSpan() < loadedHistoryStart) {
			if (loading) {
				// TODO get list of runnables?
				getExecutorService().shutdownNow();
				executorService = null;
				loading = false;
			}
			obtainHistory();
		}
		else {
			loadedHistoryStart = getLastTimestamp()-getTimeSpan();
		}
	}
	
	private void obtainHistory() {
		loadedHistoryStart = getLastTimestamp()-getTimeSpan();
		loadedHistoryStop = getLastTimestamp();
		loading = true;
		getExecutorService().execute(new Runnable() {

			public void run() {
				double start = loadedHistoryStart;
				double stop = loadedHistoryStop;
				boolean reload;
				double[][] hd;
				reportProgress(new AcopTrendChartEvent(
						AcopTrendChart.this,
						AcopTrendChartEvent.EventType.MULTIPLE_OPERATION_STARTED,
						"Started loading histories."));
				for (AcopChartConsumer c : consumersList) {
					reload = true;
					if (c instanceof AcopTrendChartConsumer) {
						HistoryParameters hp = ((AcopTrendChartConsumer) c).getHistoryParameters();
						if (hp == null) continue;
						reportProgress(new AcopTrendChartEvent(
								AcopTrendChart.this,
								AcopTrendChartEvent.EventType.OPERATION_STARTED,
								"Started loading history for "+hp,
								(AcopGraphHistoryParameters) c.getDisplayerParameters()));
						while (reload) {
							hd = getTHistoryConnector().loadHistory(hp, start, stop);
							if (hd == null) {
								reload = reportError(new AcopTrendChartEvent(
										AcopTrendChart.this,
										AcopTrendChartEvent.EventType.ERROR,
										AcopTrendChartEvent.TIMEOUT,
										(AcopGraphHistoryParameters) c.getDisplayerParameters()));
							}
							else if (hd.length == 0) {
								reload = reportError(new AcopTrendChartEvent(
										AcopTrendChart.this,
										AcopTrendChartEvent.EventType.ERROR,
										AcopTrendChartEvent.NO_DATA,
										(AcopGraphHistoryParameters) c.getDisplayerParameters()));
							}
							else {
								reload = false;
								((AcopTrendChartConsumer) c).loadHistoryData(hd);
							}
						}
						reportProgress(new AcopTrendChartEvent(
								AcopTrendChart.this,
								AcopTrendChartEvent.EventType.OPERATION_ENDED,
								"Finished loading history for "+hp,
								(AcopGraphHistoryParameters) c.getDisplayerParameters()));
					}
				}
				loading = false;
				reportProgress(new AcopTrendChartEvent(
						AcopTrendChart.this,
						AcopTrendChartEvent.EventType.MULTIPLE_OPERATION_ENDED,
						"Finished loading histories."));
			}});
	}
	
	private void obtainHistory(final AcopTrendChartConsumer c) {
		final HistoryParameters hp = c.getHistoryParameters();
		if (hp == null) return;
		getExecutorService().execute(new Runnable() {

			public void run() {
				double start = getLastTimestamp()-getTimeSpan();
				double stop = getLastTimestamp();
				boolean reload = true;
				double[][] hd;
				reportProgress(new AcopTrendChartEvent(
						AcopTrendChart.this,
						AcopTrendChartEvent.EventType.OPERATION_STARTED,
						"Started loading history for "+hp,
						(AcopGraphHistoryParameters) c.getDisplayerParameters()));
				while (reload) {
					hd = getTHistoryConnector().loadHistory(hp, start, stop);
					if (hd == null) {
						reload = reportError(new AcopTrendChartEvent(
								AcopTrendChart.this,
								AcopTrendChartEvent.EventType.ERROR,
								AcopTrendChartEvent.TIMEOUT,
								(AcopGraphHistoryParameters) c.getDisplayerParameters()));
					}
					else if (hd.length == 0) {
						reload = reportError(new AcopTrendChartEvent(
								AcopTrendChart.this,
								AcopTrendChartEvent.EventType.ERROR,
								AcopTrendChartEvent.NO_DATA,
								(AcopGraphHistoryParameters) c.getDisplayerParameters()));
					}
					else {
						reload = false;
						c.loadHistoryData(hd);
					}
				}
				reportProgress(new AcopTrendChartEvent(
						AcopTrendChart.this,
						AcopTrendChartEvent.EventType.OPERATION_ENDED,
						"Finished loading history for "+hp,
						(AcopGraphHistoryParameters) c.getDisplayerParameters()));
			}});
	}
	
	protected void reportProgress(AcopTrendChartEvent event) {
		for (AcopTrendChartListener listener : getAcopTrendChartListeners()) {
			listener.reportProgress(event);
		}
	}
	
	protected boolean reportError(AcopTrendChartEvent event) {
		for (AcopTrendChartListener listener : getAcopTrendChartListeners()) {
			if (listener.reportError(event)) {
				String operation = (event.getMessage() == null) ? "" : ": "+event.getMessage();
				AcopTrendChartEvent e = new AcopTrendChartEvent(listener, AcopTrendChartEvent.EventType.PROGRESS, "Operation "+operation+" aborted.");
				reportProgress(e);
				return true;
			}
		}
		return false;
	}
	
	public void addAcopTrendChartListener(AcopTrendChartListener listener) {
		if (acopTrendChartListeners.contains(listener)) return;
		acopTrendChartListeners.add(listener);
	}
	
	public void removeAcopTrendChartListener(AcopTrendChartListener listener) {
		acopTrendChartListeners.remove(listener);
	}
	
	public AcopTrendChartListener[] getAcopTrendChartListeners() {
		return acopTrendChartListeners.toArray(new AcopTrendChartListener[acopTrendChartListeners.size()]);
	}

	/* (non-Javadoc)
	 * @see de.desy.acop.displayers.AcopChart2#getConsumer(java.lang.String, java.lang.Class)
	 */
	@Override
	public <D extends DataConsumer> D getConsumer(String name, Class<D> type) {
		if (type.isAssignableFrom(AcopChartConsumer.class)) {
			for (AcopChartConsumer c : consumersList) {
				if (c.getName().equals(name)) return type.cast(c);
			}
			
			DataConsumer[] oldConumers = getConsumers();
			AcopTrendChartConsumer c= new AcopTrendChartConsumer(this, name);
			c.setColor(colorManager.pickColor());
			maxLinkIndex++;
			consumersList.add(c);
			firePropertyChange(CONSUMERS, oldConumers, getConsumers());
			return type.cast(c);
		}
		return null;
	}

	/* (non-Javadoc)
	 * @see de.desy.acop.displayers.AcopChart2#addDisplayerParameters(de.desy.acop.displayers.tools.AcopGraphParameters)
	 */
	@Override
	public void addDisplayerParameters(AcopGraphParameters parameters) throws CommonException, PropertyVetoException {
		super.addDisplayerParameters(parameters);
		for (AcopChartConsumer c : consumersList) {
			if (c instanceof AcopTrendChartConsumer && c.getDisplayerParameters().equals(parameters)) {
				obtainHistory((AcopTrendChartConsumer) c);
			}
		}
	}

	/* (non-Javadoc)
	 * @see de.desy.acop.displayers.AcopChart2#synchronizeGraphParameters(de.desy.acop.displayers.tools.AcopGraphParameters)
	 */
	@Override
	protected AcopGraphParameters synchronizeGraphParameters(AcopGraphParameters parameters) {
		HistoryParameters hp = null;
		if (parameters instanceof AcopGraphHistoryParameters) {
			hp = ((AcopGraphHistoryParameters) parameters).getHistoryParameters();
		}
		AcopGraphHistoryParameters trendParameters = new AcopGraphHistoryParameters(super.synchronizeGraphParameters(parameters), hp);
		return trendParameters;
	}

	/* (non-Javadoc)
	 * @see de.desy.acop.displayers.AcopChart2#updateXScale(de.desy.acop.displayers.tools.AcopChartReorgConsumer)
	 */
	@Override
	public void updateXScale(AcopChartConsumer consumer) {
		// do nothing (XScale is handled by timeSpan)
	}

	/* (non-Javadoc)
	 * @see de.desy.acop.displayers.AcopChart2#updateXScaleOnRemove(de.desy.acop.displayers.tools.AcopChartReorgConsumer)
	 */
	@Override
	protected void updateXScaleOnRemove(AcopChartConsumer consumer) {
		// do nothing (XScale is handled by timeSpan)
	}

	/**
	 * Gets time span for this </code>AcopTrendChart</code>.
	 * @return the timeSpan
	 */
	public double getTimeSpan() {
		return getXRangeMax() - getXRangeMin();
	}

	/**
	 * Sets time span for this </code>AcopTrendChart</code>.
	 * @param timeSpan the timeSpan to set
	 */
	public void setTimeSpan(double timeSpan) {
		if (getTimeSpan() == timeSpan) return;
		double oldSpan = getTimeSpan();
		setXRangeMin(getLastTimestamp() - timeSpan);
		firePropertyChange(TIME_SPAN, oldSpan, getTimeSpan());
		loadHistory();
	}

	/**
	 * Gets last time stamp for this </code>AcopTrendChart</code>.
	 * @return the last timestamp
	 */
	public double getLastTimestamp() {
		return getXRangeMax();
	}

	/**
	 * Sets last time stamp for this </code>AcopTrendChart</code>. Its value can only
	 * be increased.
	 * @param timestamp the timestamp to set
	 */
	public void setLastTimestamp(double timestamp) {
		if (!scaleInitialized) {
			if (isAutoScaleToLiveData()) {
				// workaround for Acop.setXMin method: sca[0].usermax = xMin + 0.5 * Math.abs(xMin);
				setXRangeMax(timestamp+1);
				setXRangeMin(timestamp);
				loadedHistoryStart = timestamp;
			}
			scaleInitialized = true;
		}
		if (getXRangeMax() >= timestamp) return;
		if (!isAutoScaleToLiveData()) {
			setXRangeMin(timestamp - getTimeSpan());
		}
		setXRangeMax(timestamp);
	}

	/**
	 * @return the checkForBadData
	 */
	public boolean isCheckForBadData() {
		return getTHistoryConnector().isCheckForBadData();
	}

	/**
	 * @param checkForBadData the checkForBadData to set
	 */
	public void setCheckForBadData(boolean checkForBadData) {
		boolean oldValue = isCheckForBadData();
		getTHistoryConnector().setCheckForBadData(checkForBadData);
		firePropertyChange(CHECK_FOR_BAD_DATA, oldValue, checkForBadData);
	}

	/**
	 * @return the simulate
	 */
	public boolean isSimulate() {
		return getTHistoryConnector().isSimulate();
	}

	/**
	 * @param simulate the simulate to set
	 */
	public void setSimulate(boolean simulate) {
		getTHistoryConnector().setSimulate(simulate);
	}

	/**
	 * @return the timeout
	 */
	public int getTimeout() {
		return getTHistoryConnector().getTimeout();
	}

	/**
	 * @param timeout the timeout to set
	 */
	public void setTimeout(int timeout) {
		int oldValue = getTimeout();
		getTHistoryConnector().setTimeout(timeout);
		firePropertyChange(TIMEOUT, oldValue, timeout);
	}

	/**
	 * @return the autoScaleToLiveData
	 */
	public boolean isAutoScaleToLiveData() {
		return autoScaleToLiveData;
	}

	/**
	 * @param autoScaleToLiveData the autoScaleToLiveData to set
	 */
	public void setAutoScaleToLiveData(boolean autoScaleToLiveData) {
		boolean oldValue = isAutoScaleToLiveData();
		this.autoScaleToLiveData = autoScaleToLiveData;
		firePropertyChange(AUTOSCALE_TO_LIVE_DATA, oldValue, autoScaleToLiveData);
	}

	/* (non-Javadoc)
	 * @see de.desy.acop.displayers.AcopChart2#removeDisplayerParameters(de.desy.acop.displayers.tools.AcopGraphParameters)
	 */
	@Override
	public DataSource removeDisplayerParameters(AcopGraphParameters parameters) {
		DataSource ds = super.removeDisplayerParameters(parameters);
		if (getDisplayerParameters().length == 0) scaleInitialized = false;
		return ds;
	}

}
