package de.desy.acop.displayers.table;

import java.util.HashMap;
import java.util.Map;

import com.cosylab.gui.components.table.cells.TableCell;
import com.cosylab.gui.displayers.DataConsumer;
import com.cosylab.gui.displayers.DataState;
import com.cosylab.util.ListenerList;

import de.desy.acop.displayers.AcopTable;

/**
 * 
 * @author Jaka Bobnar, Cosylab
 *
 */
public abstract class AcopTableColumn implements DataConsumer {
	
	private String uniqueName;
	private String shortName;
	protected AcopTable table;
	protected AcopTableParameters displayerParameters;
	protected DataState dataState = new DataState(DataState.UNDEFINED);
	protected TableCell[] cells;
	protected int rowCount = 10;
	protected HashMap<String, Object> characteristics = new HashMap<String, Object>();
	
	private ListenerList columnListeners = new ListenerList(TableColumnListener.class);
	
	public AcopTableColumn() {
		this(null, null, "");
	}
	
	public AcopTableColumn(AcopTable table, String uniqueName, String shortName) {
		this.table = table;
		this.uniqueName = uniqueName;
		this.shortName = shortName;
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getName()
	 */
	public String getName() {
		return uniqueName;
	}
	
	public String getHeaderName() {
		return shortName;
	}
	
	/**
	 * Returns the class of the objects that this column presents. 
	 * @return
	 */
	public Class<?> getColumnClass() {
		return TableCell.class;
	}
	
	/**
	 * Returns the value at specified row.
	 * @param rowIndex
	 * @return
	 */
	public abstract Object getValue(int rowIndex);
	
	/**
	 * Sets the number of rows.
	 * @param rowCount
	 */
	public void setRowCount(int rowCount) {
		this.rowCount = rowCount;
	}
	
	/**
	 * Returns the number of rows.
	 * @return
	 */
	public int getRowCount() {
		return rowCount;
	}
	
	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDataConsumer(java.lang.Class)
	 */
	@SuppressWarnings("unchecked")
    public DataConsumer getDataConsumer(Class type) {
		if (type.isAssignableFrom(getClass())) {
			return this;
		}
		return null;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDefaultDataConsumer()
	 */
	public DataConsumer getDefaultDataConsumer() {
		return this;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#updateDataState(com.cosylab.gui.displayers.DataState)
	 */
	public void updateDataState(DataState state) {
		if (dataState != null && dataState.equals(state)) return;
		dataState = state;
		if (cells != null) {
			for (int i = 0; i < cells.length && i < rowCount; i++) {
				if (cells[i] == null) getValue(i);
				setDataStateOnCell(cells[i]);
			}
		}
		fireColumnChanged();
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#setCharacteristics(java.util.Map)
	 */
	public void setCharacteristics(Map characteristics) {
		
		if (cells != null) {
			for (TableCell cell : cells) {
				cell.putAllCharacteristics(characteristics);
			}
		}
		
		this.characteristics.putAll(characteristics);
	}
	
	public void destroy() {
		table = null;
		displayerParameters = null;
	}
	

	public void addTableColumnListener(TableColumnListener l) {
		synchronized (columnListeners) {
			columnListeners.add(l);
		}
	}
	
	public void removeTableColumnListener(TableColumnListener l) {
		synchronized (columnListeners) {
			columnListeners.add(l);
		}		
	}
	
	protected void fireColumnChanged() {
		TableColumnListener[] listeners = (TableColumnListener[]) columnListeners.toArray();
		TableColumnEvent event = new TableColumnEvent(this);
		
		for (TableColumnListener l : listeners) {
			l.columnDataChanged(event);
		}
	}

	public DataState getDataState() {
		return dataState;
	}
	
	protected void setCharacteristicsOnCell(TableCell cell) {
		cell.putCharacteristic(TableCell.TOOLTIP, cell.getIdentifier() + " '" + cell.getValue() + "'");
	}
	
	protected void setDataStateOnCell(TableCell cell) {
		if (cell == null) return;
		DataState dataState = getDataState();
		if (dataState.type.equals(DataState.ERROR)) {
			cell.putCharacteristic(TableCell.SUSPENDED, false);
			cell.putCharacteristic(TableCell.ALARM, false);
			cell.putCharacteristic(TableCell.WARNING, false);
			cell.putCharacteristic(TableCell.TIMEOUT, false);
			cell.putCharacteristic(TableCell.ERROR, true);
		} else if (dataState.type.equals(DataState.ALARM)) {
			cell.putCharacteristic(TableCell.SUSPENDED, false);
			cell.putCharacteristic(TableCell.ALARM, true);
			cell.putCharacteristic(TableCell.WARNING, false);
			cell.putCharacteristic(TableCell.TIMEOUT, false);
			cell.putCharacteristic(TableCell.ERROR, false);
		} else if (dataState.type.equals(DataState.WARNING)) {
			cell.putCharacteristic(TableCell.SUSPENDED, false);
			cell.putCharacteristic(TableCell.ALARM, false);
			cell.putCharacteristic(TableCell.WARNING, true);
			cell.putCharacteristic(TableCell.TIMEOUT, false);
			cell.putCharacteristic(TableCell.ERROR, false);
		} else if (dataState.type.equals(DataState.TIMEOUT)) {
			cell.putCharacteristic(TableCell.SUSPENDED, false);
			cell.putCharacteristic(TableCell.ALARM, false);
			cell.putCharacteristic(TableCell.WARNING, false);
			cell.putCharacteristic(TableCell.TIMEOUT, true);
			cell.putCharacteristic(TableCell.ERROR, false);
		} else if (dataState.type.equals(DataState.TIMELAG)) {
			cell.putCharacteristic(TableCell.SUSPENDED, false);
			cell.putCharacteristic(TableCell.ALARM, false);
			cell.putCharacteristic(TableCell.WARNING, false);
			cell.putCharacteristic(TableCell.TIMEOUT, true);
			cell.putCharacteristic(TableCell.ERROR, false);
		} else if (dataState.type.equals(DataState.UNDEFINED) || dataState.type.equals(DataState.NOT_INITIALIZED)) {
			cell.putCharacteristic(TableCell.SUSPENDED, true);
			cell.putCharacteristic(TableCell.ALARM, false);
			cell.putCharacteristic(TableCell.WARNING, false);
			cell.putCharacteristic(TableCell.TIMEOUT, false);
			cell.putCharacteristic(TableCell.ERROR, false);
		} else if (dataState.type.equals(DataState.NORMAL)) {
			cell.putCharacteristic(TableCell.SUSPENDED, false);
			cell.putCharacteristic(TableCell.ALARM, false);
			cell.putCharacteristic(TableCell.WARNING, false);
			cell.putCharacteristic(TableCell.TIMEOUT, false);
			cell.putCharacteristic(TableCell.ERROR, false);
		}
	}

	public AcopTableParameters getDisplayerParameters() {
		return displayerParameters;
	}

	public void setDisplayerParameters(AcopTableParameters displayerParameters) {
		this.displayerParameters = displayerParameters;
	}
}
