package de.desy.acop.transport;

import java.io.Serializable;

/**
 * This objecct holds all parameters necessary to initiate ACOP transport
 * connection.
 * 
 * @author Igor Kriznar (igor.kriznarATcosylab.com)
 */
public class ConnectionParameters implements Comparable, Serializable
{
	/**
	 * Predefined value for the property size connection parameter, which means that
	 * when property will be connected, then actual remote size for property array will be used.
	 */
	public static int ACTUAL_PROPERTY_SIZE= 0;
	/**
	 * Predefined value for the property size connection parameter, which means that
	 * when property will be connected, size of array will be set according to displayer's
	 * capabilities.
	 */
	public static int AUTO_PROPERTY_SIZE= -1;
	
  private static final long serialVersionUID = -430933461297463309L;
  private String accessProtocol;
  private String deviceContext;
  private String deviceGroup;
  private String deviceName;
  private String deviceProperty;
  private AccessMode accessMode;
  private int accessRate;
  
  private int propertySize = AUTO_PROPERTY_SIZE;
  /**
   * Creates new instance of ConnectionParameters.
   */
  public ConnectionParameters()
  {
  }
  
  public ConnectionParameters(String parameters) {
	  String[] p = parameters.split(",");
	  if (p == null) { 
		  throw new IllegalArgumentException("Invalid string.");
	  } 
	  if (p.length == 4) {
		  setUpRemoteName(p[0]);
		  this.accessMode = AccessMode.valueOf(p[1]);
		  this.accessRate = Integer.parseInt(p[2]);
		  this.propertySize = Integer.parseInt(p[3]);
	  } else if (p.length == 1) {
		  setUpRemoteName(p[0]);
		  this.accessMode = AccessMode.POLL;
		  this.accessRate = 1000;
		  this.propertySize = AUTO_PROPERTY_SIZE;
	  } else if (p.length == 3) {
		  setUpRemoteName(p[0]);
		  this.accessMode = AccessMode.valueOf(p[1]);
		  this.accessRate = Integer.parseInt(p[2]);
	  } else {
		  throw new IllegalArgumentException("Invalid string.");
	  }
	  
	  
  }
  /**
   * Creates new instance of ConnectionParameters.
   * 
   * @param accessProtocol
   * @param deviceContext
   * @param deviceGroup
   * @param deviceName
   * @param deviceProperty
   * @param accessMode
   * @param accessRate
   */
  public ConnectionParameters(String accessProtocol, String deviceContext, String deviceGroup,
      String deviceName, String deviceProperty, AccessMode accessMode, int accessRate)
  {
    super();
    this.accessProtocol = accessProtocol != null ? accessProtocol : "";
    this.deviceContext = deviceContext != null ? deviceContext : "";
    this.deviceGroup = deviceGroup != null ? deviceGroup : "";
    this.deviceName = deviceName != null ? deviceName : "";
    this.deviceProperty = deviceProperty != null ? deviceProperty : "";
    this.accessMode = accessMode;
    this.accessRate = accessRate;
  }
  
  public ConnectionParameters(String accessProtocol, String deviceContext, String deviceGroup,
	      String deviceName, String deviceProperty, AccessMode accessMode, int accessRate, int propertySize)
	  {
	  this(accessProtocol, deviceContext, deviceGroup, deviceName, deviceProperty,
			  accessMode, accessRate);
	  this.propertySize = propertySize;
	  
	  }
  /**
   * Creates new instance of ConnectionParameters.
   * 
   * @param remoteName
   * @param accessMode
   * @param accessRate
   */
  public ConnectionParameters(String remoteName, AccessMode accessMode, int accessRate)
  {
    super();
    setUpRemoteName(remoteName);
    this.accessMode = accessMode;
    this.accessRate = accessRate;
  }
  
  private void setUpRemoteName(String remoteName) {
	String[] s = remoteName.split("/");
	    
	this.accessProtocol = (s.length>0 && s[0]!=null) ? s[0] : "";
	this.deviceContext = (s.length>1 && s[1]!=null) ? s[1] : "";
	this.deviceGroup = (s.length>2 && s[2]!=null) ? s[2] : "";
	this.deviceProperty = (s.length>4 && s[s.length-1]!=null) ? s[s.length-1] : "";
	
	StringBuilder sb= new StringBuilder(128);
	
	if (s.length>3 && s[3]!=null) {
		sb.append(s[3]);
	}
	for (int i = 4; i < s.length-1; i++) {
		sb.append('/');
		if (s[i]!=null) sb.append(s[i]);
	}
	
	this.deviceName = sb.toString();
  }
  
  public ConnectionParameters(String remoteName, AccessMode accessMode, int accessRate, int arraySize)
  {
    this(remoteName, accessMode, accessRate);
    this.propertySize = arraySize;
  }
  /**
   * Returns the accessMode.
   * 
   * @return Returns the accessMode.
   */
  public AccessMode getAccessMode()
  {
    return accessMode;
  }
  /**
   * Returns the accessProtocol.
   * 
   * @return Returns the accessProtocol.
   */
  public String getAccessProtocol()
  {
    return accessProtocol;
  }
  /**
   * Returns the accessRate.
   * 
   * @return Returns the accessRate.
   */
  public int getAccessRate()
  {
    return accessRate;
  }
  /**
   * Returns the deviceContext.
   * 
   * @return Returns the deviceContext.
   */
  public String getDeviceContext()
  {
    return deviceContext;
  }
  /**
   * Returns the deviceGroup.
   * 
   * @return Returns the deviceGroup.
   */
  public String getDeviceGroup()
  {
    return deviceGroup;
  }
  /**
   * Returns the deviceName.
   * 
   * @return Returns the deviceName.
   */
  public String getDeviceName()
  {
    return deviceName;
  }
  /**
   * Returns the deviceProperty.
   * 
   * @return Returns the deviceProperty.
   */
  public String getDeviceProperty()
  {
    return deviceProperty;
  }
  public String toString()
  {
    return getRemoteName() + "," + accessMode + "," + accessRate + "," + propertySize;
  }
  /**
   * Return URI ready name for remote TINE property.
   * 
   * @return remote TINE name delimeted with / characters
   */
  public String getRemoteName()
  {
    return accessProtocol + "/" + deviceContext + "/" + deviceGroup + "/" + deviceName + "/" + deviceProperty;
  }
  /*
   * (non-Javadoc)
   * 
   * @see java.lang.Comparable#compareTo(java.lang.Object)
   */
  public int compareTo(Object o)
  {
    if (!(o instanceof ConnectionParameters)) return -1;
    return this.toString().compareTo(o.toString());
  }
  /*
   * (non-Javadoc)
   * 
   * @see java.lang.Object#hashCode()
   */
  public int hashCode()
  {
    return toString().hashCode();
  }
  /*
   * (non-Javadoc)
   * 
   * @see java.lang.Object#equals(java.lang.Object)
   */
  public boolean equals(Object o)
  {
    if (!(o instanceof ConnectionParameters)) return false;
    return this.toString().equals(o.toString());
//    ConnectionParameters p = (ConnectionParameters) o;
//    return (getRemoteName().equals(p.getRemoteName()) && accessMode.equals(p.getAccessMode()) && accessRate == p
//        .getAccessRate());
  }
  public ConnectionParameters deriveWith(AccessMode mode)
  {
    return new ConnectionParameters(accessProtocol, deviceContext, deviceGroup, deviceName, deviceProperty,
        mode, accessRate, propertySize);
  }
  public ConnectionParameters deriveWith(AccessMode mode, int rate)
  {
    return new ConnectionParameters(accessProtocol, deviceContext, deviceGroup, deviceName, deviceProperty,
        mode, rate, propertySize);
  }
  public ConnectionParameters deriveWith(int rate)
  {
    return deriveWithAccessRate(rate);
  }
  
  public ConnectionParameters deriveWithPropertySize(int propertySize)
  {
    return new ConnectionParameters(accessProtocol, deviceContext, deviceGroup, deviceName, deviceProperty,
        accessMode, accessRate, propertySize);
  }

  /**
   * Size of TINE property. 1 means that property will have one value in array. 
   * -1 or 0 means that when connected, automatic or actual property size will be used.
   * @return size or property if it is array 
   * 
   * @see #ACTUAL_PROPERTY_SIZE
   * @see #AUTO_PROPERTY_SIZE 
   */
	public int getPropertySize() {
		return propertySize;
	}
	
	public static void main(String[] args) {
		System.out.println(new ConnectionParameters("test", AccessMode.POLL, 1000,1));
		System.out.println(AccessMode.valueOf("POLL"));
		
	}

	public ConnectionParameters deriveWithAccessProtocol(String accessProtocol) {
	    return new ConnectionParameters(accessProtocol, deviceContext, deviceGroup, deviceName, deviceProperty,
	            accessMode, accessRate, propertySize);
	}

	public ConnectionParameters deriveWithDeviceContext(String deviceContext) {
	    return new ConnectionParameters(accessProtocol, deviceContext, deviceGroup, deviceName, deviceProperty,
	            accessMode, accessRate, propertySize);
	}

	public ConnectionParameters deriveWithAccessRate(int accessRate) {
	    return new ConnectionParameters(accessProtocol, deviceContext, deviceGroup, deviceName, deviceProperty,
	            accessMode, accessRate, propertySize);
	}

	public ConnectionParameters deriveWithDeviceGroup(String deviceGroup) {
	    return new ConnectionParameters(accessProtocol, deviceContext, deviceGroup, deviceName, deviceProperty,
	            accessMode, accessRate, propertySize);
	}

	public ConnectionParameters deriveWithDeviceName(String deviceName) {
	    return new ConnectionParameters(accessProtocol, deviceContext, deviceGroup, deviceName, deviceProperty,
	            accessMode, accessRate, propertySize);
	}

	public ConnectionParameters deriveWithDeviceProperty(String deviceProperty) {
	    return new ConnectionParameters(accessProtocol, deviceContext, deviceGroup, deviceName, deviceProperty,
	            accessMode, accessRate, propertySize);
	}
}
