package de.desy.tine.alarmUtils;

import de.desy.tine.client.*;
import de.desy.tine.dataUtils.*;
import de.desy.tine.definitions.*;
import de.desy.tine.queryUtils.*;
import de.desy.tine.server.alarms.*;
import de.desy.tine.types.*;
import java.util.Hashtable;

public class TAlarmSystem
{
  private static final int nice_query_size = 100;
  private static NAME16I[] almSysBuf = new NAME16I[nice_query_size];
  private static NAME32[] srvNamesBuf = new NAME32[nice_query_size];
  private static Hashtable<String,NAME16I[]> infoTbl = new Hashtable<String, NAME16I[]>();
  /** 
   * @brief gets the alarm systems available in the given context 
   * 
   * This method gets the available alarm systems in the given context
   * as a simple string array.  For the mapping of alarm system names to
   * alarm system numbers use getAlarmSystemsMap
   *  
   * @param context [input] is the TINE context
   *
   * @return a string array containing the available alarm systems
   * @see getAlarmSystemsMap, getAlarmSystem
   */
  public static String[] getAlarmSystems(String context)
  {
    NAME16I[] n16i = getAlarmSystemsMap(context);
    if (n16i == null || n16i.length == 0) return null;
    String[] as = new String[n16i.length];
    for (int i=0; i<n16i.length; i++) as[i] = n16i[i].getName();
    return as;
  }
  /** 
   * @brief gets the alarm systems available in the given context 
   * 
   * This method gets the available alarm systems in the given context
   * as an array of NAME16I pairs, which gives a mapping of the alarm system
   * name (NAME16 part) to the alarm system number (integer part).
   *  
   * @param context [input] is the TINE context
   *
   * @return a NAME16I array containing the available alarm systems
   * @see getAlarmSystems, getAlarmSystem
   */
  public static synchronized NAME16I[] getAlarmSystemsMap(String context)
  {
    if (context == null) return null;
    NAME16I[] as;
    if ((as=infoTbl.get(context)) != null) return as;
    TDataType ast = new TDataType(almSysBuf);
    String dev = "/" + context + "/CAS";
    TLink tl = new TLink(dev,"ALARMSYSTEMS",ast,null,TAccess.CA_READ);
    int cc = tl.execute(500, true);
    tl.close();
    if (cc == 0)
    {
      int n = ast.getCompletionLength();
      if (n > 0)
      {
        as = new NAME16I[n];
        for (int i=0; i<n; i++) as[i] = new NAME16I(almSysBuf[i]);
        if (as != null) infoTbl.put(context, as);
        return as;
      }
    }
    return null;
  }
  /** 
   * @brief gets the alarm system corresponding the the alarm system number 
   * and context given. 
   * 
   * This method gets the alarm system corresponding the the alarm system number 
   * and context given.
   *  
   * @param context [input] is the TINE context
   * @param alarmSystemNumber [input] is the specificed alarm system number for which
   * the system name is desired. 
   *
   * @return a String containing the corresponding alarm system name.
   * @see getAlarmSystems, getAlarmSystemsMap
   */
  public static String getAlarmSystem(String context,int alarmSystemNumber)
  {
    NAME16I[] as = getAlarmSystemsMap(context);
    if (as != null)
    {
      for (int i=0; i<as.length; i++)
      {
        if (as[i].ival == alarmSystemNumber) return as[i].getName();
      }
    }
    return new String("alarm system " + alarmSystemNumber);
  }
  public static String[] getAlarmServerContexts()
  {
    return TQuery.getContexts("CAS");
  }
  public static String[] getAlarmServerServers(String context,String alarmSystem)
  {
    if (context == null) return null;
    TDataType srvt = new TDataType(srvNamesBuf);
    String srv = "/" + context + "/CAS";
    String dev = srv + "/" + (alarmSystem != null ? "SYS:"+alarmSystem : "#0");
    String prp = alarmSystem != null && alarmSystem.compareToIgnoreCase("ALL") != 0 ? "Server4AlmSystem" : "SERVERS";

    TLink tl = new TLink(dev,prp,srvt,null,TAccess.CA_READ);
    int cc = tl.execute(500, true);
    tl.close();
    if (cc == 0)
    {
      int n = srvt.getCompletionLength();
      if (n > 0)
      {
        String[] as = new String[n];
        for (int i=0; i<n; i++) as[i] = srvNamesBuf[i].getName();
        return as;
      }
    }
    return null;
  }
  public static int getNumberOfAlarms(String context,String server,String alarmSystem,long start,long stop,int severity)
  {
    if (context == null) return -TErrorList.argument_list_error;
    int[] stst = new int[3];
    stst[0] = (int)(start / 1000);
    stst[1] = (int)(stop / 1000);
    stst[2] = severity;
    TDataType sst;
    int[] nalms;
    String srv = server == null ? "CAS" : server;
    String dev = "/" + context + "/" + srv;
    String prp;
    if (srv.compareToIgnoreCase("CAS") == 0)
    {
      sst = new TDataType(stst);
      nalms = new int[1];
      prp = "NUMALARMS";
      if (alarmSystem != null)
      {
        dev += alarmSystem.startsWith("Server:") ? alarmSystem.substring(7) : "/SYS:" + alarmSystem;
      }
    }
    else
    {
      sst = new TDataType();
      nalms = new int[5];
      prp = "NALARMS";
    }
    TDataType nat = new TDataType(nalms);
    TLink tl = new TLink(dev,prp,nat,sst,TAccess.CA_READ);
    int cc = tl.execute(500, true);
    tl.close();
    if (cc == 0) return nalms[0];
    return 0;
  }
  /** 
   * @brief gets the alarm message structure array 
   * 
   * This method gets the alarm message structure array from either
   * the Central Alarm Server (server == null) or the targetted server.
   *  
   * @param context [input] is the TINE context
   * @param server [input] is the targetted server if given.  To get alarms from the 
   * Central Alarm Server, this parameter should be null.
   * @param alarmSystem [input] is the targetted alarm system and is only valid for 
   * call to the Central Alarm Server.  alarmSystem can be a registered alarm system or
   * refer to a specific server when given in the form "Server:<server name>"
   * @param start [input] is the start time for the search as a UTC long
   * @param stop [input] is the stop time for the search as a UTC long
   * @param severity [input] is the minimum severity for the search
   * @param ams [output] is a reference to the legacy Alarm Message Structure array to hold 
   * the returned data.  If ams is null, the call returns the number of alarms.
   * etc.)
   *
   * @return the number of alarms for the input given.
   *
   */
  public static int getAlarms(String context,String server,String alarmSystem,long start,long stop,int severity,TAlarmMessageStructure[] ams)
  {
    if (context == null) return -TErrorList.argument_list_error;
    String srv = server == null ? "CAS" : server;
    String dev = "/" + context + "/" + srv;
    String prp;
    if (srv.compareToIgnoreCase("CAS") == 0)
    {
      prp = "SYSALARMS";
      if (alarmSystem != null && alarmSystem.compareToIgnoreCase("ALL") != 0)
      {
        dev += alarmSystem.startsWith("Server:") ? alarmSystem.substring(7) : "/SYS:" + alarmSystem;
      }
      else
      {
        dev += "/ALL";
      }
    }
    else
    {
      prp = "ALARMS";
    }
    int n = getNumberOfAlarms(context, server, alarmSystem, start, stop, severity);
    if (ams == null || ams.length == 0) return n;
    if (n > 0)
    {
      TAlarmMessageStructure[] a = new TAlarmMessageStructure[n];
      for (int i=0; i<n; i++) a[i] = new TAlarmMessageStructure();
      TDataType amst = new TDataType(a);
      int[] stst = new int[3];
      stst[0] = (int)(start / 1000);
      stst[1] = (int)(stop / 1000);
      stst[2] = severity;
      TDataType sst = new TDataType(stst);
      TLink tl = new TLink(dev,prp,amst,sst,TAccess.CA_READ);
      int cc = tl.execute(500, true);
      tl.close();
      if (cc == 0)
      {
        int nret = amst.getCompletionLength();
        if (nret < n) n = nret; 
      }
      else 
      {
        n = 0;
      }
      int i;
      for (i=0; i<n && i<ams.length; i++)
      {
        ams[i] = a[i];
      }
      n = i;
    }
    return n;
  }
  /** 
   * @brief gets the alarm message structure array 
   * 
   * This method gets the alarm message structure array from 
   * the Central Alarm Server.
   *  
   * @param context [input] is the TINE context
   * @param alarmSystem [input] is the targetted alarm system
   * and can refer to a registered alarm system or
   * to a specific server when given in the form "Server:<server name>"
   * @param start [input] is the start time for the search as a UTC long
   * @param stop [input] is the stop time for the search as a UTC long
   * @param severity [input] is the minimum severity for the search
   * @param ams [output] is a reference to the legacy Alarm Message Structure array to hold 
   * the returned data.  If ams is null, the call returns the number of alarms.
   * etc.)
   *
   * @return the number of alarms for the input given.
   *
   */
  public static int getAlarms(String context,String alarmSystem,long start,long stop,int severity,TAlarmMessageStructure[] ams)
  {
    return getAlarms(context,null,alarmSystem,start,stop,severity,ams);
  }
  /** 
   * @brief gets the alarm message structure array from the targetted server 
   * 
   * This method gets the alarm message structure array from 
   * the targetted server.
   *  
   * @param context [input] is the TINE context
   * @param server [input] is the targetted server name
   * @param start [input] is the start time for the search as a UTC long
   * @param stop [input] is the stop time for the search as a UTC long
   * @param severity [input] is the minimum severity for the search
   * @param ams [output] is a reference to the legacy Alarm Message Structure array to hold 
   * the returned data.  If ams is null, the call returns the number of alarms.
   * etc.)
   *
   * @return the number of alarms for the input given.
   *
   */
  public static int getAlarmsFromServer(String context,String server,long start,long stop,int severity,TAlarmMessageStructure[] ams)
  {
    return getAlarms(context,server,null,start,stop,severity,ams);
  }
  /** 
   * @brief gets the alarm definition structure array 
   * 
   * This method gets the alarm definition structure array as registered at the 
   * Central Alarm Server
   *  
   * @param context [input] is the TINE context
   * @param server [input] is the targetted server. 
   * @param alarmCode [input] is the targetted alarm code.
   * @param ads [output] is a reference to the legacy Alarm Definition Structure array to hold 
   * the returned data.  If ads is null, the call returns the number of alarm definitions.
   * etc.)
   *
   * @return the number of alarm definitions for the input given.
   *
   */
  public static int getAlarmDefinitions(String context,String server,int alarmCode,TAlarmDefinitionStructure[] ads)
  {
    if (context == null || server == null) return -TErrorList.argument_list_error;
    if (ads == null || ads.length == 0) return -TErrorList.argument_list_error;
    String dev = "/" + context + "/CAS/" + server;
    String prp = "SYSALMDEFS";
    int n = 100;
    TAlarmDefinitionStructure[] d = new TAlarmDefinitionStructure[n];
    for (int i=0; i<n; i++) d[i] = new TAlarmDefinitionStructure();
    TDataType adst = new TDataType(d);
    TDataType act;
    int[] almcod = new int[1];
    almcod[0] = alarmCode;
    if (alarmCode > 0)
    {
      act = new TDataType(almcod);
    }
    else
    {
      act = new TDataType();
    }
    TLink tl = new TLink(dev,prp,adst,act,TAccess.CA_READ);
    int cc = tl.execute(500, true);
    tl.close();
    if (cc == 0)
    {
      n = adst.getCompletionLength();
    }
    else 
    {
      n = 0;
    }
    for (int i=0; i<n && i<ads.length; i++)
    {
      ads[i] = d[i];
    }
    return n;
  }
  /** 
   * @brief gets the alarm definition structure array from the targetted server
   * 
   * This method gets the alarm definition structure array directly from the targetted
   * server.
   *  
   * @param context [input] is the TINE context
   * @param server [input] is the targetted server. 
   * @param ads [output] is a reference to the legacy Alarm Definition Structure array to hold 
   * the returned data.  If ads is null, the call returns the number of alarm definitions.
   * etc.)
   *
   * @return the number of alarm definitions for the input given.
   *
   */
  public static int getAlarmDefinitionsFromServer(String context,String server,TAlarmDefinitionStructure[] ads)
  {
    if (context == null || server == null) return -TErrorList.argument_list_error;
    if (ads == null || ads.length == 0) return -TErrorList.argument_list_error;
    String dev = "/" + context + "/" + server;
    String prp = "ALMDEFS";
    int n = 100;
    TAlarmDefinitionStructure[] d = new TAlarmDefinitionStructure[n];
    TDataType adst = new TDataType(d);
    TLink tl = new TLink(dev,prp,adst,null,TAccess.CA_READ);
    int cc = tl.execute(500, true);
    tl.close();
    if (cc == 0)
    {
      n = adst.getCompletionLength();
    }
    else 
    {
      n = 0;
    }
    for (int i=0; i<n && i<ads.length; i++)
    {
      ads[i] = d[i];
    }
    return n;
  }
}
