package de.desy.tine.client;
import java.net.InetAddress;
import java.util.Date;
import java.util.LinkedList;
import de.desy.tine.addrUtils.*;
import de.desy.tine.dataUtils.*;
import de.desy.tine.definitions.*;
import de.desy.tine.headers.*;
import de.desy.tine.io.TBucket;
import de.desy.tine.queryUtils.TQuery;

/** 
 * TLink.java
 *
 * @brief User interface calls to access tine data links.
 *  
 * The methods presented here deal with linking local application data with
 * remote data sources.  Primarily you will need to create a TLink object,
 * passing the necessary identifying parameters of the remote server to which 
 * the link is to be established.  When the TLink object is instantiated, 
 * you will then need to retrieve the data using one of the retrieval 
 * methods, e.g. execute() for synchronous acquistion, or attach() for  
 * asynchronous acquisition, or receive() for the acquisition of network
 * globals.
 */
public final class TLink
{
  public String cntName;
  public String expName;
  public String devName;
  public String devProperty;
  public TDataType dOutput;
  public TDataType dInput;
  public TDataType dError;
  private static int maxOutputBufferSize = 0xffff;
  private String arrayDelimiter;
  public short devAccess;
  public short devTimeout;
  public static int defaultTimeout = 500;  // milliseconds
  public short srvId; // for network subscriptions
  boolean active;
  boolean terminate;
  boolean delayEstablishLink = false;
  boolean needsNotification = false;
  boolean isInsideCallback = false;
  boolean isWildcardLink = false;
  boolean useErrValue = false;
  boolean useErrObject = false;
  TLinkHook hook = null;
  Object hookObj = null;
  public boolean adjustDefaultValues = false;
  public boolean retryOnTimeoutError = true;
  public boolean removeOnClose = false; // was true
  public TModAddress modAddress;
  TContract con;
  TSubscription sub;
  TPublishedData phs;
  Theader hdr;
  TLink boundTo = null;
  boolean cancelledWithDependencies = false;
  private TLinkGroup grp = null;
  public boolean isGrouped() {return grp == null ? false : true; }
  public TLinkGroup getGroup() { return grp; }
  LinkedList dependencies = null;
  public LinkedList getDependencies() { return dependencies; }
  public boolean hasDependencies() 
  { 
    if (dependencies != null && dependencies.size() > 0) return true;
    return false;
  }
  public boolean isBound() { return boundTo == null ? false : true; }
  public TLink getBoundLink() { return boundTo; }
  public boolean isCancelledWithDependencies() { return cancelledWithDependencies; }
  public void addDependency(TLink tlink)
  {
    if (dependencies == null) dependencies = new LinkedList();
    dependencies.add(tlink);
  }
  public void rmvDependency(TLink tlink)
  {
    if (dependencies == null) return;
    dependencies.remove(tlink);
  }
  public int linkStatus;
  public int linkCounter;
  public boolean linkStale;
  public double linkErrValue;
  public String linkErrString;
  public byte[] linkStatusStringBuffer = new byte[32];
  public InetAddress linkPeer;
  TCallback tcb;
  int callbackId;
  TLinkCallback tlcb;
  TWildcardLink twcl;
  private TDataTolerance dtf = null;
  public TDataTolerance getNotificationTolerance() { return dtf; }
  /**
   * @brief sets the tolerance level for asynchrnous event notification
   *
   * Link callback notification can be supressed if changes in the link
   * data do not exceed the tolerance applied in this call.  The tolerance
   * can have an absolute part (in the units of the data) and/or a 
   * percent part. 
   * 
   * @param absoluteValue is the absoulute tolerance any particular data element
   * is allowed to have before an event notification is called.
   * @param percentValue is the percent tolerance any particular data element
   * is allowed to have before an event notification is called.
   */
  public void setNotificationTolerance(float absoluteValue, float percentValue)
  {
    dtf = new TDataTolerance(absoluteValue,percentValue);
  }
  public void setNotificationTolerance(double absoluteValue, double percentValue)
  {
    dtf = new TDataTolerance((float)absoluteValue,(float)percentValue);
  }
  public void setNotificationTolerance(int absoluteValue, int percentValue)
  {
    dtf = new TDataTolerance(absoluteValue,percentValue);
  }
  public boolean isWithinTolerance()
  {
    if (dtf == null) return false; // no tolerance specified => always true
    if (dOutput == null) return false; // no output data object
    if (isGrouped()) return false; // no tolerances for grouped calls
    TDataType dRef = dtf.getReference();
    if (dRef == null) // first call ?
    {
      dRef = new TDataType(dOutput.getArrayLength(),dOutput.getFormat());
      dRef.dataCopy(dOutput);
      dtf.setReference(dRef);
      return false;
    }
    int len = dOutput.getArrayLength();
    float abs = dtf.absolute;
    float pcn = (float)(dtf.percent/100.0);
    switch (dOutput.getFormat())
    {
      case TFormat.CF_BYTE:
      {
        byte[] athis = (byte[])dOutput.getDataObject();
        byte[] alast = (byte[])dRef.getDataObject(); 
        for (int i=0; i<len; i++)
          if (Math.abs(athis[i]-alast[i]) > abs + pcn*(float)alast[i])
          {
            dRef.dataCopy(dOutput);
            return false;
          }
        break;
      }
      case TFormat.CF_SHORT:
      {
        short[] athis = (short[])dOutput.getDataObject();
        short[] alast = (short[])dRef.getDataObject(); 
        for (int i=0; i<len; i++) 
          if (Math.abs(athis[i]-alast[i]) > abs + pcn*(float)alast[i])
          {
            dRef.dataCopy(dOutput);
            return false;
          }
        break;
      }
      case TFormat.CF_LONG:
      {
        int[] athis = (int[])dOutput.getDataObject();
        int[] alast = (int[])dRef.getDataObject(); 
        for (int i=0; i<len; i++) 
          if (Math.abs(athis[i]-alast[i]) > abs + pcn*(float)alast[i])
          {
            dRef.dataCopy(dOutput);
            return false;
          }
        break;
      }
      case TFormat.CF_FLOAT:
      {
        float[] athis = (float[])dOutput.getDataObject();
        float[] alast = (float[])dRef.getDataObject(); 
        for (int i=0; i<len; i++) 
          if (Math.abs(athis[i]-alast[i]) > abs + pcn*(float)alast[i])
          {
            dRef.dataCopy(dOutput);
            return false;
          }
        break;
      }
      case TFormat.CF_DOUBLE:
      {
        double[] athis = (double[])dOutput.getDataObject();
        double[] alast = (double[])dRef.getDataObject(); 
        for (int i=0; i<len; i++) 
          if (Math.abs(athis[i]-alast[i]) > abs + pcn*(float)alast[i])
          {
            dRef.dataCopy(dOutput);
            return false;
          }
        break;
      }
      default:
        return false;
    }
    return true;
  }
  /**
   * @brief Returns the current status for this link
   * 
   * @return The Link status as tine error code
   */
  public int getLinkStatus()
  {
    return linkStatus < 0 ? 0 : linkStatus;
  }
  /**
   * @brief Returns the context of this link
   * 
   * @return The Context of the current link
   */
  public String getContext()
  {
    return cntName;
  }
  /**
   * @brief Returns the device server addressed in this link
   * 
   * @return The device server addressed in the current link
   */
  public String getDeviceServer()
  {
    return expName;
  }
  /**
   * @brief Returns the device (module) name addressed in this link
   * 
   * @return The device (module) name addressed in the current link
   */
  public String getDeviceName()
  {
    return devName;
  }
  /**
   * @brief Returns the device property addressed in this link
   * 
   * @return The device property addressed in the current link
   */
  public String getProperty()
  {
    return devProperty;
  }
  /**
   * @brief Returns the fully qualified device name addressed in this link
   * 
   * @return The the fully qualified device name addressed in the current link
   */
  public String getFullDeviceName()
  {
    return new String("/"+cntName+"/"+expName+"/"+devName);
  }
  /**
   * @brief Returns the output data object as a TDataType
   * 
   * @return The output data type 
   */
  public TDataType getOutputDataObject()
  {
    return dOutput;
  }
  /**
   * @brief Returns the output data input as a TDataType
   * 
   * @return The input data type 
   */
  public TDataType getInputDataObject()
  {
    return dInput;
  }
  /**
   * @brief The last link timestamp
   * 
   * @return The last link timestamp as a UTC java long value (milliseconds
   * since 1.1.1970)
   */
  public long getLastTimeStamp()
  {
    return (long)(dOutput.timestamp) * 1000 + (long)(dOutput.timestampMSEC);
  }

  /**
   * @brief The last link timestamp as a TINE double timestamp value
   * 
   * @return The last link timestamp as a UTC java long value (seconds
   * since 1.1.1970 with decimal fractional) 
   */
  public double getLastDataTimeStamp()
  {
    return (double)(dOutput.timestamp) + ((double)(dOutput.timestampMSEC)/1000);
  }
  /**
   * @brief The last link status in plain text
   * 
   * @return The last link status in plain text
   */
  public String getLastError()
  {
	  return getError(linkStatus);
  }      
  /**
   * @brief The specified error code in plain text
   * 
   * @param code Is the tine error code for which the error text is desired
   * 
   * @return The specified error code in plain text
   */
  public String getError(int code) 
  {
    if (code >= 512)
      return new String(linkStatusStringBuffer).trim();
    else
	    return TErrorList.getErrorString(code);
  }
  Date  date = new Date();
  /**
   * @brief Specifies an error value to fill in associated data buffers 
   * in case of link error.
   * 
   * @param errValue Is the error value to use in case of link errors
   * 
   * 0 if successful
   */
  public int setErrorValue(double errValue)
  {
    if (sub.mode == TMode.CM_CANCEL) return TErrorList.ask_error;
    linkErrValue = errValue;
    useErrValue = true;
    return 0;
  }
  /**
   * @brief Specifies an error string to fill in associated data buffers 
   * in case of link error.
   * 
   * @param errString Is the string to use in case of link errors
   * 
   * 0 if successful
   */
  public int setErrorValue(String errString)
  {
    if (sub.mode == TMode.CM_CANCEL) return TErrorList.ask_error;
    linkErrString = errString;
    useErrValue = true;
    return 0;
  }
  public int setDebugLevel(int level)
  {
    return tf.setDebugLevel(level);
  }
  public int getDebugLevel()
  {
    return tf.getDebugLevel();
  }
  public String getModuleAddressInfo()
  {
    return modAddress.toString(expName,cntName);
  }
  private TLink getExistingTLink(String devname, String devproperty, TDataType dout, TDataType din, short devaccess)
  {
    return tf.getExistingLink(devname,devproperty,dout,din,devaccess);    
  }
  private int TCreateLink(String devname, String devproperty, TDataType dout, TDataType din, short devaccess)
  {
    if (devname == null || devproperty == null) return TErrorList.argument_list_error;
    String fullname;
    tf = TLinkFactory.getInstance();
    setDebugLevel(tf.getDebugLevel());
    if (hook == null) hook = tf.getTLinkHook(); 
    if (hook != null && (fullname=hook.getLinkName(devname)) != null)
    {
      Object objOut = dout == null ? null : dout.getDataObject();
      Object objIn = din == null ? null : din.getDataObject();
      if (dout != null) dOutput = dout; else dOutput = new TDataType();
      if (din != null) dInput = din; else dInput = new TDataType();
      hookObj = hook.create(fullname,devproperty,objOut,objIn);
      devName = new String(fullname);
      devProperty = new String(devproperty);
      expName = "cdi";
      cntName = "localhost";
      return 0;
    }
    fullname = new String(devname);
    active = true;
    if (dout != null) dOutput = dout; else dOutput = new TDataType();
    if (din != null) dInput = din; else dInput = new TDataType();
    if (arrayDelimiter != null) dOutput.setArrayDelimiter(arrayDelimiter);
    
    int indx = fullname.indexOf('/');
    cntName = "DEFAULT";
    if (indx == 0)  // context given
    {
      fullname = fullname.substring(1);
      indx = fullname.indexOf('/');
      if (indx != -1) 
      {
        cntName = fullname.substring(0,indx);
        fullname = fullname.substring(indx+1);
        indx = fullname.indexOf('/');
      }
    }
    if (indx == -1)
    {
      expName = fullname;
      devName = "";
    }
    else
    {
      expName = fullname.substring(0,indx);
      devName = fullname.substring(indx+1);
    }
    fullname = indx == -1 ? "" : fullname.substring(indx+1);
    devProperty = new String(devproperty);
    devAccess = devaccess;
    devTimeout = (short)1000;

    linkId = tf.registerLink(this);
    if (linkId < 0) return TErrorList.resources_exhausted;

    if (expName.compareTo("GLOBAL") != 0)
    {
      modAddress = new TModAddress(tf,expName,cntName);
      if (modAddress.srvAddress == null)
      { 
        if (tf.debugLevel > 0)
          System.out.println("cannot resolve server address for " + "/" + cntName + "/" + expName);
        return TErrorList.host_not_resolved;
      }
    }
    TLink xlnk = getExistingTLink(devname,devproperty,dout,din,devaccess);
    if (xlnk != null)
    {
      xlnk.addDependency(this);
      boundTo = xlnk;
      sub = xlnk.sub;
      con = xlnk.con;
      hdr = xlnk.hdr;
      callbackId = linkId;
      return 0;
    }
    con = new TContract(this);
    sub = new TSubscription(con,this); // going out ...
    hdr = new Theader(tf.getUserName());
    String defaultTimeoutStr;
    if ((defaultTimeoutStr = System.getProperty("default.timeout")) != null)
    {
      defaultTimeout = Integer.parseInt(defaultTimeoutStr);
    }
    tf.linkReady(this);
    return 0;
  }

  /** 
   * @brief Defines a tine Data link
   * 
   * This defines a tine Data Link, verifies that the requested deviceName
   * exists and returns a non-null link object if successful.  This does not
   * establish the link! You must use either
   * the execute method or the attach method to establish the link.
   *  
   * @param devname is the fully specified device name as in /&lt;context&gt;/&lt;server&gt;/&lt;device&gt;
   * or &lt;server&gt;/&lt;device&gt;
   * @param devproperty is the desired property
   * @param dout is a reference to the output data object (which will receive
   * the data from the server)
   * @param din is a reference to the input data object (which will send data
   * to the server)
   * @param devaccess is the requested control access mode (TMode.CA_READ, TMode.CA_WRITE
   * etc.)
   *
   * @return A valid TLink instance if successful or null if not
   *
   * \b Example:
   * 
   * \include eg_TLink1.java
   *
   * @see de.desy.tine.client.TLink#attach
   * @see de.desy.tine.client.TLink#execute
   * @see de.desy.tine.dataUtils.TDataType#TDataType 
   */
  public TLink(String devname, String devproperty, TDataType dout, TDataType din, short devaccess)
  {
    TCreateLink(devname,devproperty,dout,din,devaccess);
  }
  /** 
   * @brief Defines a tine Data link
   * 
   * This defines a tine Data Link, verifies that the requested deviceName
   * exists and returns a non-null link object if successful.  This does not
   * establish the link! You must use either
   * the execute method or the attach method to establish the link.
   *  
   * @param fullDeviceNameAndProperty is the fully specified, property-appended
   * device name as in /&lt;context&gt;/&lt;server&gt;/&lt;device&gt;/&lt;property&gt;
   * or &lt;server&gt;/&lt;device&gt;/&lt;property&gt;
   * @param dout is a reference to the output data object (which will receive
   * the data from the server)
   * @param din is a reference to the input data object (which will send data
   * to the server)
   * @param devaccess is the requested control access mode (TMode.CA_READ, TMode.CA_WRITE
   * etc.)
   *
   * @return A valid TLink instance if successful or null if not
   *
   * \b Example:
   * 
   * \include eg_TLink2.java
   *
   * @see de.desy.tine.client.TLink#attach
   * @see de.desy.tine.client.TLink#execute
   * @see de.desy .tine.dataUtils.TDataType#TDataType 
   */
  public TLink(String fullDeviceNameAndProperty, TDataType dout, TDataType din, short devaccess)
  {
    String devname, devproperty;
    int cnt = 1;
    int indx = fullDeviceNameAndProperty.indexOf('/');
    if (indx == 0) cnt = 3;
    for (int i=0; i<cnt; i++)
    {
      indx +=  fullDeviceNameAndProperty.substring(indx+1).indexOf('/') + 1;
    }
    devname = new String(fullDeviceNameAndProperty.substring(0,indx));
    devproperty = new String(fullDeviceNameAndProperty.substring(indx+1));
    
    TCreateLink(devname,devproperty,dout,din,devaccess);
  }

  /** 
   * @brief Defines a read-only tine Data link with default behavior.
   * 
   * This defines a tine Data Link, verifies that the requested deviceName
   * exists and returns a non-null link object if successful.  This does not
   * establish the link! You must use either
   * the execute method or the attach method to establish the link.
   *  
   * @param devname is the fully specified device name as in /&lt;context&gt;/&lt;server&gt;/&lt;device&gt;
   * or &lt;server&gt;/&lt;device&gt;
   * @param devproperty is the desired property
   *
   * @return A valid TLink instance if successful or null if not
   *
   * \b Example:
   * 
   * \include eg_TLink3.java
   *
   * @see de.desy.tine.client.TLink#attach
   * @see de.desy.tine.client.TLink#execute
   */
  public TLink(String devname, String devproperty)
  {
    TDataType dout = new TDataType(TLink.maxOutputBufferSize,TFormat.CF_DEFAULT);
    TDataType din = new TDataType();
    TCreateLink(devname,devproperty,dout,din,TAccess.CA_READ);
  }
  public static String[] splitDeviceAndPropertyFromName(String fullDeviceNameAndProperty)
  {
    String[] names = new String[2];
    int cnt = 1;
    int indx;
    
    if (fullDeviceNameAndProperty == null) return names;
    indx = fullDeviceNameAndProperty.indexOf('[');
    if (indx > 0)
    { // "check for /<context>/<server>/<device>[<property>]" form
      names[0] = new String(fullDeviceNameAndProperty.substring(0,indx));
      names[1] = new String(fullDeviceNameAndProperty.substring(indx+1,fullDeviceNameAndProperty.length()));      
    }
    else
    { // "check for /<context>/<server>/<device>/<property>" form
      indx = fullDeviceNameAndProperty.indexOf('/');
      if (indx == 0) cnt = 3;
      for (int i=0; i<cnt; i++)
      {
        indx +=  fullDeviceNameAndProperty.substring(indx+1).indexOf('/') + 1;
      }    
      names[0] = new String(fullDeviceNameAndProperty.substring(0,indx));
      names[1] = new String(fullDeviceNameAndProperty.substring(indx+1));
    }
    return names;
  }
  /** 
   * @brief Defines a read-only tine Data link with default behavior.
   * 
   * This defines a tine Data Link, verifies that the requested deviceName
   * exists and returns a non-null link object if successful.  This does not
   * establish the link! You must use either
   * the execute method or the attach method to establish the link.
   *  
   * @param fullDeviceNameAndProperty is the fully specified, property-appended
   * device name as in /&lt;context&gt;/&lt;server&gt;/&lt;device&gt;/&lt;property&gt;
   * or &lt;server&gt;/&lt;device&gt;/&lt;property&gt;
   *
   * @return A valid TLink instance if successful or null if not
   *
   * \b Example:
   * 
   * \include eg_TLink4.java
   *
   * @see de.desy.tine.client.TLink#attach
   * @see de.desy.tine.client.TLink#execute
   */
  public TLink(String fullDeviceNameAndProperty)
  {
    String[] names = splitDeviceAndPropertyFromName(fullDeviceNameAndProperty);
    String devname = names[0], devproperty = names[1];
    TDataType dout = new TDataType(TLink.maxOutputBufferSize,TFormat.CF_DEFAULT);
    TDataType din = new TDataType();
    TCreateLink(devname,devproperty,dout,din,TAccess.CA_READ);
  }

  /** 
   * @brief Defines a tine GLobal Data link
   * 
   * This defines a tine Data Link and assumes that the requested parameter
   * is a network global, available on the tine multicast group.
   * This method does not establish the link! You must use either
   * the execute method or the attach method to establish the link.
   *  
   * @param globalKeyword is the desired network global parameter
   * @param dout is a reference to the output data object (which will receive
   * the network global data).
   *
   * @return A valid TLink instance if successful or null if not
   *
   * \b Example:
   * 
   * \include eg_TLink5.java
   *
   * @see de.desy.tine.client.TLink#attach
   * @see de.desy.tine.client.TLink#receive
   * @see de.desy.tine.dataUtils.TDataType#TDataType 
   */
  public TLink(String globalKeyword, TDataType dout)
  {
    TCreateLink("GLOBAL",globalKeyword,dout,null,TAccess.CA_READ);
  }

  /** 
   * @brief Attaches the data sets to the created link at the default polling rate (1000 msec)
   * 
   * Asynchronous. This method attaches the data sets to the created link using the calling
   * parameters provided at instantiation.  The arriving data will automatically
   * update the output data set and then fire the callback method provided.
   * If the callback is not needed, a null can be passed for this parameter.
   *  
   * @param mode is the tine control mode parameter, that is one of TMode.CM_POLL, 
   * TMode.CM_REFRESH, TMode.SINGLE, etc. in possible combination with
   * TMode.CM_NETWORK or TMode.CM_CONNECT, etc.
   * @param f is the 'callback function' to be called whenever either new data 
   * arrive or a change in link status occurs.
   * The callback function should instantiate a class you create which 
   * implements the TCallback() class interface.
   *
   * @return A positive link handle if successful, otherwise -1.  On 
   * failure the linkStatus property of TLink can be examined.
   * 
   * \b Example:
   * 
   * \include eg_attach1.java
   *
   * @see de.desy.tine.client.TLink#TLink
   * @see de.desy.tine.client.TLink#TLinkCallback
   * @see de.desy.tine.client.TLink#execute
   */
  public int attach(int mode,TCallback f)
  {
    return attach(mode,f,1000,-1);
  }
  public int attach(short mode,TCallback f)
  {
    return attach((int)mode,f); 
  }
  /**
   * Stop thread execution until this link has been signalled.
   */
	   
  public synchronized int Twait()
  {
	 if (sub.id < 0) return -1;
	 try
	 {
		while (!this.terminate)
		{
		  wait();
//		  if (!this.terminate && 
//		       this.needsNotification &&
//		        tcb != null) tcb.callback(sub.id,linkStatus);
//		  this.needsNotification = false;
		}
	 }
	 catch (InterruptedException e) {}
//	 sub.mode = TMode.CM_CANCEL;
	 return -1;
  }            
  /** 
   * @brief Executes a synchronous link using the data sets supplied at  
   * link creation.
   * 
   * Synchronous. This method executes the link using the data sets and 
   * parameters provided at instantiation.  The call will block for 1000 msec
   * or until the call completes.
   *  
   * \note The TLink object created will remain resident until a call to
   * the cancel() method is made. Thus repeated calls to execute() can
   * be made without invoking separate calls to TLink(). 
   *  
   * @return 0 if successful, otherwise a tine return code which can be
   * interpretted using getError() or getLastError().
   *
   * \b Example:
   * 
   * \include eg_execute1.java
   *
   * @see de.desy.tine.client.TLink#TLink
   * @see de.desy.tine.client.TLink#attach
   */
  public int execute()
  {
	 return execute(1000);
  }                    
  /** 
   * @brief Executes a synchronous link using the data sets supplied at  
   * link creation.
   * 
   * Synchronous. This method executes the link using the data sets and 
   * parameters provided at instantiation.  The call will block for the
   * duration of the timeout specified or until the call completes.
   *
   * @param timeout The amount of time in milliseconds to block execution
   * until the call completes.
   * @param appendMode will be appended to the tine Control Mode, which is 
   * TMode.CM_SINGLE. (e.g. TMode.CM_CONNECT).
   *   
   * \note The TLink object created will remain resident until a call to
   * the cancel() method is made. Thus repeated calls to execute() can
   * be made without invoking separate calls to TLink(). 
   *  
   * @return 0 if successful, otherwise a tine return code which can be
   * interpretted using getError() or getLastError().
   * 
   * \b Example:
   * 
   * \include eg_execute2.java
   *
   * @see de.desy.tine.client.TLink#TLink
   * @see de.desy.tine.client.TLink#attach
   */
  public int execute(int timeout, short appendMode)
  {
    return execute(timeout,appendMode,false);
  }                    
  /** 
   * @brief Executes a synchronous link using the data sets supplied at  
   * link creation.
   * 
   * Synchronous. This method executes the link using the data sets and 
   * parameters provided at instantiation.  The call will block for the
   * duration of the timeout specified or until the call completes.
   *
   * @param timeout The amount of time in milliseconds to block execution
   * until the call completes.
   * @param retryOnError if 'true' will instuct the tine engine to issue
   * an automatic retry following a connection timeout. 
   *   
   * \note The TLink object created will remain resident until a call to
   * the cancel() method is made. Thus repeated calls to execute() can
   * be made without invoking separate calls to TLink(). 
   *  
   * @return 0 if successful, otherwise a tine return code which can be
   * interpretted using getError() or getLastError().
   * 
   * \b Example:
   * 
   * \include eg_execute3.java
   *
   * @see de.desy.tine.client.TLink#TLink
   * @see de.desy.tine.client.TLink#attach
   */
  public int execute(int timeout, boolean retryOnError)
  {
    return execute(timeout,(short)0,retryOnError);
  }                    
  /** 
   * @brief Executes a synchronous link using the data sets supplied at  
   * link creation.
   * 
   * Synchronous. This method executes the link using the data sets and 
   * parameters provided at instantiation.  The call will block for the
   * duration of the timeout specified or until the call completes.
   *
   * @param timeout The amount of time in milliseconds to block execution
   * until the call completes.
   * @param appendMode will be appended to the tine Control Mode, which is 
   * TMode.CM_SINGLE. (e.g. TMode.CM_CONNECT).
   * @param retryOnError if 'true' will instuct the tine engine to issue
   * an automatic retry following a connection timeout. 
   *   
   * \note The TLink object created will remain resident until a call to
   * the cancel() method is made. Thus repeated calls to execute() can
   * be made without invoking separate calls to TLink(). 
   *  
   * @return 0 if successful, otherwise a tine return code which can be
   * interpretted using getError() or getLastError().
   * 
   * \b Example:
   * 
   * \include eg_execute4.java
   *
   * @see de.desy.tine.client.TLink#TLink
   * @see de.desy.tine.client.TLink#attach
   */
  public synchronized int execute(int timeout, short appendMode,boolean retryOnError)
  {
    if (tf.debugLevel > 0) System.out.println("execute link " + linkId);
    if (isBound()) 
    { // different TLinks to the same thing in synchronous mode :
      this.sub.id = (short)this.linkId;
      tf.linkReady(this);
    }
    int sts = 0;
    if (hookObj != null) return hook.execute(hookObj,timeout);
    if (this.linkId < 0)
    {
      sts = this.linkStatus = TErrorList.out_of_client_memory;
      return sts;
    }
    if (timeout < 100) timeout = 100;
    devTimeout = (short)timeout;
    if (this.sub == null)
    {
      throw new RuntimeException("Could not resolve address for /" + this.cntName + "/" + this.expName);    
    }
    short bmode = TMode.getBaseMode(this.sub.mode);
    long ltime = System.currentTimeMillis();
    dOutput.timestamp = (int)(ltime/1000);
    this.sub.linkStartTime = ltime;
    this.sub.linkLastTime = ltime;
    if (bmode == TMode.CM_REGISTER || bmode <= TMode.CM_SINGLE)
    { // this could already be an established monitor link ...
      this.sub.mode = TMode.CM_SINGLE;
      devAccess |= TAccess.CA_SYNC;
    }
    if ((appendMode & TMode.CM_CONNECT) == TMode.CM_CONNECT) this.sub.mode |= appendMode;
    retryOnTimeoutError = retryOnError;
    linkStatus = TErrorList.link_not_open;
    //linkStale = true;  <- don't do this here!

    if (this.devName.indexOf("*") >= 0 && !TQuery.isStockProperty(this.devProperty))
    {
      TWildcardLink wc = tf.getWildcardLink();
      wc.parent = this;
      wc.format = this.dOutput.dFormat;
      /* reset these to the wildcard callback  */
      this.tlcb = wc.scb;
      //wait = TRUE;
      this.twcl = wc;
      this.isWildcardLink = true;
    }
    if ((sts=tf.establishLink(this)) != 0) // will set the link to active
    {
      if (tf.debugLevel > 0) System.out.println("can't establish link");
      return sts;
    }
   
    tf.adjustLinkTable(this,TLinkFactory.adjustLinkTableAdd);
    if (tf.debugLevel > 0) System.out.println("link " + linkId + " is " + (removeOnClose ? "non-persistent" : "persistent"));
    for (int i=0; i<3; i++)
    {
      try
      {
        while (true)
        {
          wait();
          if (this.twcl == null || this.twcl.canNotify) break;
        }
        if (bmode == TMode.CM_SINGLE) this.active = false;
        return linkStatus;
      }
      catch (InterruptedException e) 
      {
        linkStatus = TErrorList.net_read_error;
        if (tf.debugLevel > 0) System.out.println("Interrupted io on link " + linkId);
        //terminate = true;
      }
    }
    return linkStatus;
  }                    
  /** 
   * @brief Executes a synchronous link using the data sets supplied at  
   * link creation.
   * 
   * Synchronous. This method executes the link using the data sets and 
   * parameters provided at instantiation.  The call will block for the
   * duration of the timeout specified or until the call completes.
   *
   * @param timeout The amount of time in milliseconds to block execution
   * until the call completes.
   *   
   * \note The TLink object created will remain resident until a call to
   * the cancel() method is made. Thus repeated calls to execute() can
   * be made without invoking separate calls to TLink(). 
   *  
   * @return 0 if successful, otherwise a tine return code which can be
   * interpretted using getError() or getLastError().
   * 
   * \b Example:
   * 
   * \include eg_execute5.java
   *
   * @see de.desy.tine.client.TLink#TLink
   * @see de.desy.tine.client.TLink#attach
   */
  public int execute(int timeout)
  {
    return execute(timeout,(short)0);
  }                            
  /**
   * Terminates an active link
   */
  public int cancel()
  {
    if (hookObj != null) hook.close(hookObj);
    return tf.cancel(this);
  }              
  /**
   * @brief Terminates an active link
   * 
   * Synonymous with cancel().
   */
  public int close()
  {
    return cancel();
  }              
  public int linkId;  
  int linkTimeouts;  
  boolean pending = false;  
  //static TErrorList tel;  
  public TLinkFactory tf; 
  TBucket tb;
  public TBucket getBucket() { return tb; }
  TLink() {}   
  protected TLink(TLinkFactory tfactory, int tid, String devname, String devproperty, TDataType dout, TDataType din, short devaccess)
  {
    String fullname = new String(devname);
    active = true;
    if (dout != null) dOutput = dout; else dOutput = new TDataType();
    dInput = din;
    int indx = fullname.indexOf('/');
    tf = TLinkFactory.getInstance();
    
    cntName = "DEFAULT";
    if (indx == 0)  // context given
    {
      fullname = fullname.substring(1);
      indx = fullname.indexOf('/');
      if (indx != -1) 
      {
        cntName = fullname.substring(0,indx);
        fullname = fullname.substring(indx+1);
        indx = fullname.indexOf('/');
      }
    }
    if (indx == -1)
    {
      expName = fullname;
      devName = "";
    }
    else
    {
      expName = fullname.substring(0,indx);
      devName = fullname.substring(indx+1);
    }
    fullname = indx == -1 ? "" : fullname.substring(indx+1);
//	devName = new String(devname);
    devProperty = new String(devproperty);
    devAccess = devaccess;
    devTimeout = (short)2000;
    
    // an ENS link will return 0 here : 
    linkId = tid = tf.registerLink(this);
    if (tf.debugLevel > 0) System.out.println("assigned link id " + linkId + " to /" + cntName + "/" + expName);

    modAddress = new TModAddress(tfactory,expName,cntName);
    if (modAddress.srvAddress == null && tf.debugLevel > 0) 
      System.out.println("Cannot resolve Server Address for /" + cntName + "/" + expName);
    con = new TContract(this);
    sub = new TSubscription(con,this); // going out ...
//	phs = new TPublishedData(); // coming in ...
    hdr = new Theader(System.getProperty("user.name"));
    tf.linkReady(this);
    if (tid == 0) removeOnClose = false;
  }

/** 
 * @brief Attaches the data sets to the created link at the given polling rate
 * 
 * Asynchronous. This method attaches the data sets to the created link using the calling
 * parameters provided at instantiation.  The arriving data will automatically
 * update the output data set and then fire the callback method provided.
 * If the callback is not needed, a null can be passed for this parameter.
 *  
 * @param mode is the tine control mode parameter, that is one of TMode.CM_POLL, 
 * TMode.CM_REFRESH, TMode.SINGLE, etc. in possible combination with
 * TMode.CM_NETWORK or TMode.CM_CONNECT, etc.
 * @param f is the 'callback function' to be called whenever either new data 
 * arrive or a change in link status occurs.
 * The callback function should instantiate a class you create which 
 * implements the TCallback() class interface.
 * @param pollrate is the desired polling rate (at the server) in milliseconds.
 *
 * @return A positive link handle if successful, otherwise -1.  On 
 * failure the linkStatus property of TLink can be examined.
 * 
 * \b Example:
 * 
 * \include eg_attach2.java
 *
 * @see de.desy.tine.client.TLink#TLink
 * @see de.desy.tine.client.TLink#TLinkCallback
 * @see de.desy.tine.client.TLink#execute
 */
  public int attach(int mode,TCallback f,int pollrate)
  {
	  return attach(mode,f,pollrate,-1);
  }
  public int attach(short mode,TCallback f,int pollrate)
  {
    return attach((int)mode,f,pollrate);
  }
/** 
 * @brief Attaches the data sets to the created link at the given polling rate
 * and signals the callback routine using the give notification parameter. 
 * 
 * Asynchronous. This method attaches the data sets to the created link using the calling
 * parameters provided at instantiation.  The arriving data will automatically
 * update the output data set and then fire the callback method provided.
 * If the callback is not needed, a null can be passed for this parameter.
 *  
 * @param mode is the tine control mode parameter, that is one of TMode.CM_POLL, 
 * TMode.CM_REFRESH, TMode.SINGLE, etc. in possible combination with
 * TMode.CM_NETWORK or TMode.CM_CONNECT, etc.
 * @param f is the 'callback function' to be called whenever either new data 
 * arrive or a change in link status occurs.
 * The callback function should instantiate a class you create which 
 * implements the TCallback() class interface.
 * @param pollrate is the desired polling rate (at the server) in milliseconds.
 * @param notificationId is the desired identifier to be passed to the
 * supplied callback routine when it is fired.
 *
 * @return A positive link handle if successful, otherwise -1.  On 
 * failure the linkStatus property of TLink can be examined.
 * 
 * \b Example:
 * 
 * \include eg_attach3.java
 *
 * @see de.desy.tine.client.TLink#TLink
 * @see de.desy.tine.client.TLink#TCallback
 * @see de.desy.tine.client.TLink#execute
 */
  public synchronized int attach(int mode,TCallback f,int pollrate,int notificationId)
  {
    if (tf.debugLevel > 0) System.out.println("attach link " + linkId + " : " + TMode.toString((short)mode));
    if (hookObj != null) return hook.attach(hookObj,pollrate,(short)mode,f,notificationId);
    if (this.linkId < 0)
    {
      this.linkStatus = TErrorList.out_of_client_memory;
      return -1;
    }
    if (this.sub == null)
    {
      throw new RuntimeException("Could not resolve address for /" + this.cntName + "/" + this.expName);    
    }
    this.sub.mode = (short)mode;
    tcb = f;
    if (notificationId >= 0)
      callbackId = notificationId;
    else if (callbackId <= 0)
      callbackId = sub.id;
    if (pollrate < 100) pollrate = 100;
    devTimeout = (short)pollrate;
    this.sub.pollingRate = devTimeout;
    if (this.isBound()) return sub.id;

    if (this.devName.indexOf("*") >= 0 && !TQuery.isStockProperty(this.devProperty))
    {
      TWildcardLink wc = tf.getWildcardLink();
      wc.tcb = f;
      wc.parent = this;
      wc.format = this.dOutput.dFormat;
      /* reset these to the wildcard callback  */
      this.tlcb = wc.scb;
      //wait = TRUE;
      this.twcl = wc;
      this.isWildcardLink = true;
      this.sub.mode &= ~TMode.CM_GROUPED; // conflicting groups! => turn it off
    }
    
    if (tf.establishLink(this) != 0) return -1;

    long ltime = System.currentTimeMillis();
    dOutput.timestamp = (int)(ltime/1000);
    this.sub.linkStartTime = ltime;
    this.sub.linkLastTime = ltime;
 
    if ((this.sub.mode & TMode.CM_GROUPED) == TMode.CM_GROUPED) 
    {
      if (grp == null) grp = tf.getGroup(f);
      grp.addMember(this);
    }
    if ((mode & TMode.CM_USE_ON_ERROR) == TMode.CM_USE_ON_ERROR) 
    {
      dError = (TDataType)dOutput.clone();
      if (dError != null) useErrObject = true;
    } 
    if ((mode & TMode.CM_WAIT) == TMode.CM_WAIT) waitForLinkCompletion();
 
    tf.adjustLinkTable(this,tf.adjustLinkTableAdd);
    if (tf.debugLevel > 0) System.out.println("link " + linkId + " is " + (removeOnClose ? "non-persistent" : "persistent"));

    return sub.id;
  }
  public int attach(short mode,TCallback f,int pollrate,int notificationId)
  {
    return attach((int)mode,f,pollrate,notificationId);
  }
 
/** 
 * @brief Attaches the data sets to the created link at the given polling rate
 * and signals the callback routine using a reference to the TLink instance. 
 * 
 * Asynchronous. This method attaches the data sets to the created link using the calling
 * parameters provided at instantiation.  The arriving data will automatically
 * update the output data set and then fire the callback method provided.
 * If the callback is not needed, a null can be passed for this parameter.
 * This is the preferred AttachLink method to use.  The Callback instantiates
 * a TLinkCallback() interface which provides a TLink object as parameter.
 *  
 * @param mode is the tine control mode parameter, that is one of TMode.CM_POLL, 
 * TMode.CM_REFRESH, TMode.SINGLE, etc. in possible combination with
 * TMode.CM_NETWORK or TMode.CM_CONNECT, etc.
 * @param f is the 'callback function' to be called whenever either new data 
 * arrive or a change in link status occurs.
 * The callback function should instantiate a class you create which 
 * implements the TLinkCallback() class interface.
 * @param pollrate is the desired polling rate (at the server) in milliseconds.
 *
 * @return A positive link handle if successful, otherwise -1.  On 
 * failure the linkStatus property of TLink can be examined.
 * 
 * \b Example:
 * 
 * \include eg_attach4.java
 *
 * @see de.desy.tine.client.TLink#TLink
 * @see de.desy.tine.client.TLink#TLinkCallback
 * @see de.desy.tine.client.TLink#execute
 */
  public synchronized int attach(int mode,TLinkCallback f,int pollrate)
  {
    if (tf.debugLevel > 0) System.out.println("attach link " + linkId + " : " + TMode.toString((short)mode));
    if (hookObj != null) return hook.attach(hookObj,pollrate,(short)mode,f,this);
    if (this.linkId < 0)
    {
      this.linkStatus = TErrorList.out_of_client_memory;
      return -1;
    }
    if (this.sub == null)
    {
      throw new RuntimeException("Could not resolve address for /" + this.cntName + "/" + this.expName);    
    }
    this.sub.mode = (short)mode;
    tlcb = f;

    if (pollrate < 100) pollrate = 100;
    devTimeout = (short)pollrate;
    this.sub.pollingRate = pollrate;
    if (this.isBound()) return sub.id;

    if (this.devName.indexOf("*") >= 0 && !TQuery.isStockProperty(this.devProperty))
    {
      TWildcardLink wc = tf.getWildcardLink();
      wc.tlcb = f;
      wc.parent = this;
      wc.format = this.dOutput.dFormat;
      /* reset these to the wildcard callback  */
      this.tlcb = wc.scb;
      this.twcl = wc;
      this.isWildcardLink = true;
      this.sub.mode &= ~TMode.CM_GROUPED; // conflicting groups! => turn it off
    }
    
    if (tf.establishLink(this) != 0) return -1;

    long ltime = System.currentTimeMillis();
    dOutput.timestamp = (int)(ltime/1000);
    this.sub.linkStartTime = ltime;
    this.sub.linkLastTime = ltime;
 
    if ((this.sub.mode & TMode.CM_GROUPED) == TMode.CM_GROUPED) 
    {
      if (grp == null) grp = tf.getGroup(f);
      grp.addMember(this);
    }
    if ((mode & TMode.CM_USE_ON_ERROR) == TMode.CM_USE_ON_ERROR) 
    {
      dError = (TDataType)dOutput.clone();
      if (dError != null) useErrObject = true;
    } 
    if ((mode & TMode.CM_WAIT) == TMode.CM_WAIT) waitForLinkCompletion();

    tf.adjustLinkTable(this,TLinkFactory.adjustLinkTableAdd);
    if (tf.debugLevel > 0) System.out.println("link " + linkId + " is " + (removeOnClose ? "non-persistent" : "persistent"));
    return sub.id;
  }
  public int attach(short mode,TLinkCallback f,int pollrate)
  {
    return attach((int)mode,f,pollrate);
  }
  /** 
   * @brief Attaches the data sets to the global link
   * and signals the callback routine using the assigned link id. 
   * 
   * Asynchronous. This method attaches the data sets to the created link using the calling
   * parameters provided at instantiation.  The arriving data will automatically
   * update the output data set and then fire the callback method provided.
   * If the callback is not needed, a null can be passed for this parameter.
   * Using the receive() method assumes that the data requested are global available.
   *  
   * @param f is the 'callback function' to be called whenever either new data 
   * arrive or a change in link status occurs.
   * The callback function should instantiate a class you create which 
   * implements the TCallback() class interface.
   *
   * @return A positive link handle if successful, otherwise -1.  On 
   * failure the linkStatus property of TLink can be examined.
   * 
   * \b Example:
   * 
   * \include eg_receive1.java
   *
   * @see de.desy.tine.client.TLink#TLink
   * @see de.desy.tine.client.TLink#TLinkCallback
   * @see de.desy.tine.client.TLink#receive(TCallback f,int notificationId)
   * @see de.desy.tine.client.TLink#receive(TLinkCallback f)
   */
  public synchronized int receive(TCallback f)
  {
    return attach(TMode.CM_GLOBAL,f,1000);
  }
  /** 
   * @brief Attaches the data sets to the global link
   * and signals the callback routine using a reference to the
   * given notifcation ID in the callback
   * 
   * Asynchronous. This method attaches the data sets to the created link using the calling
   * parameters provided at instantiation.  The arriving data will automatically
   * update the output data set and then fire the callback method provided.
   * If the callback is not needed, a null can be passed for this parameter.
   * Using the receive() method assumes that the data requested are global available.
   *  
   * @param f is the 'callback function' to be called whenever either new data 
   * arrive or a change in link status occurs.
   * The callback function should instantiate a class you create which 
   * implements the TCallback() class interface.
   * @param notificationId is the desired identifier to be passed to the
   * supplied callback routine when it is fired.
   *
   * @return A positive link handle if successful, otherwise -1.  On 
   * failure the linkStatus property of TLink can be examined.
   * 
   * \b Example:
   * 
   * \include eg_receive2.java
   *
   * @see de.desy.tine.client.TLink#TLink
   * @see de.desy.tine.client.TLink#TLinkCallback
   * @see de.desy.tine.client.TLink#receive(TCallback f)
   */
  public synchronized int receive(TCallback f,int notificationId)
  {
    return attach(TMode.CM_GLOBAL,f,1000,notificationId);
  }
  /** 
   * @brief Attaches the data sets to the global link
   * and signals the callback routine using a reference to the
   * referenced TLink object.
   * 
   * Asynchronous. This method attaches the data sets to the created link using the calling
   * parameters provided at instantiation.  The arriving data will automatically
   * update the output data set and then fire the callback method provided.
   * If the callback is not needed, a null can be passed for this parameter.
   * Using the receive() method assumes that the data requested are global available.
   * This is the preferred receive method to use.  The Callback instantiates
   * a TLinkCallback() interface which provides a TLink object as parameter.
   *  
   * @param f is the 'callback function' to be called whenever either new data 
   * arrive or a change in link status occurs.
   * The callback function should instantiate a class you create which 
   * implements the TLinkCallback() class interface.
   *
   * @return A positive link handle if successful, otherwise -1.  On 
   * failure the linkStatus property of TLink can be examined.
   * 
   * \b Example:
   * 
   * \include eg_receive3.java
   *
   * @see de.desy.tine.client.TLink#TLink
   * @see de.desy.tine.client.TLink#TLinkCallback
   * @see de.desy.tine.client.TLink#receive(TCallback f)
   */
  public synchronized int receive(TLinkCallback f)
  {
    return attach(TMode.CM_GLOBAL,f,1000);
  }
  /**
   * @brief Stops thread execution until this link has completed.
   * Useful for grouped asynchronous links.
   */
  public void waitForLinkCompletion() 
  {
    try
    {
  	 while (this.linkStatus < 0)
  	 {
  	   Thread.sleep(100);
  	 }
    }
    catch (InterruptedException e) 
    {
      System.out.println(" waitForLinkCompletion : " + e);
    }
  }
  /**
   * @brief Gets the array delimiter for use in the toString() method 
   * of associated TDataType objects.
   * 
   * @return The delimiter in use in separating data
   * variables in a string dump (default ",")
   */
  public String getArrayDelimiter()
  {
    return arrayDelimiter;
  }
  /**
   * @brief Sets the array delimiter for use in the toString() method 
   * of associated TDataType objects.
   * 
   * @param delimiter is the string delimiter to use in separating data
   * variables in a string dump (default ",")
   */
  public void setArrayDelimiter(String delimiter)
  {
    if (arrayDelimiter != null && arrayDelimiter.compareTo(delimiter) == 0) return;
    arrayDelimiter = new String(delimiter);
    this.dOutput.setArrayDelimiter(arrayDelimiter);
  }
}