package de.desy.tine.dataUtils;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Arrays;
import java.util.Vector;
import de.desy.tine.definitions.*;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.headers.*;
import de.desy.tine.structUtils.*;
import de.desy.tine.types.*;

/**
 * @brief tine Data Type object All data transfered client <-> server must be
 *        contained in a TDataType object. Furthermore all constructors of
 *        TDataType take an array of tine data types as input argument, i.e. a
 *        single element should be passed as an array of length 1
 */
public final class TDataType implements Cloneable
{
  public static int RPCERR_SIZE = 32;
  private static int dBufferSegSize = 1472;
  public int dCompletionLength;
  public int dArrayLength;
  public short dFormat;
  protected Vector dVector;
  private boolean removeTagAtLinkTime = false;
  private String arrayDelimiter = new String("\n");
  private String dTag = new String(new StringBuffer(8));
  private int dSizeInBytes;
  private ByteArrayOutputStream dBuffer;
  private ByteArrayOutputStream[] dBuffers;
  public void pushBytes(byte[] bytestream)
  {
    try 
    {
      dBuffers[0].reset();
      dBuffers[0].write(bytestream);
    }
    catch (Exception e)
    {
    }
    return;
  }
  int debugLevel = 0;
  public long dTimestamp;
  public int timestamp;
  public int timestampMSEC;
  public int numblks;
  public int blknum;
  public int blkid;
  public int blkcnt;
  public int mtu;
  public int blksin;
  public int bytesin;
  private Object hDataObject;
  private Object hStructObject;
  public Object clone()
  {
    Object o = null;
    try
    {
      o = super.clone();
    }
    catch (CloneNotSupportedException e)
    {
      System.err.println("TDataType cannot clone");
    }
    return o;
  }
  void initCounters()
  {
    mtu = (1472 - TContract.hdrSizeInBytes - TSubscription.hdrSizeInBytes - Theader.hdrSizeInBytes);
    if (dFormat != TFormat.CF_NULL) mtu -= (mtu % TFormat.formatSizeOf(dFormat));
    numblks = dSizeInBytes / mtu;
    if (dSizeInBytes == 0 || (dSizeInBytes % mtu) > 0) numblks += 1;
    blkid = 1;
    blknum = 1;
    blkcnt = 0;
  }
  public boolean equals(TDataType dt)
  {
    return equals(true,dt);
  }
  public boolean equals(boolean compareContents,TDataType dt)
  {
    if (dt == null) 
    {
      return dFormat == TFormat.CF_NULL ? true : false;
    }
    if (dt.dFormat == TFormat.CF_NULL && dFormat == TFormat.CF_NULL) return true;
    if (dt.dFormat != dFormat) return false;
    if (dt.dArrayLength == 0) return false;
    if (dt.dArrayLength != dArrayLength) return false;
    if (dt.dBuffers == null || dBuffers == null) return false;
    if (compareContents) 
    {
      for (int i = 0; i < dt.dBuffers.length; i++)
      {
        if (dBufferSegSize * i > dArrayLength) break;
        if (dBuffers[i] == null || dt.dBuffers[i] == null) return false;
        if (!Arrays.equals(dt.dBuffers[i].toByteArray(), dBuffers[i].toByteArray())) return false;
      }
    }
    return true;
  }
  public int copy(TDataType dt)
  {
    if (dt == null) return TErrorList.argument_list_error;
    if (dt.dFormat == TFormat.CF_NULL && dFormat == TFormat.CF_NULL) return 0;
    if (dt.dFormat != dFormat) return TErrorList.illegal_format;
    if (dt.dArrayLength > dArrayLength) return TErrorList.dimension_error;
    if (dt.dBuffers == null || dBuffers == null) return TErrorList.un_allocated;
    try
    {
      for (int i = 0; i < dt.dBuffers.length; i++)
      {
        dBuffers[i].reset();
        dt.dBuffers[i].writeTo(dBuffers[i]);
      }
    }
    catch (Exception e)
    {
    }
    return 0;
  }
  /**
   * @brief Sets the delimiter used in a string representation of an array
   */
  public void setArrayDelimiter(String delimiter)
  {
    if (arrayDelimiter.compareTo(delimiter) == 0) return;
    arrayDelimiter = new String(delimiter);
  }
  /**
   * @brief Get the delimiter used in a string representation of an array
   * @return the delimiter used in a string representation of an array
   */
  public String getArrayDelimiter()
  {
    return arrayDelimiter;
  }
  /**
   * @brief NULL element
   */
  public TDataType()
  {
    dFormat = TFormat.CF_NULL;
    dArrayLength = 0;
    dSizeInBytes = 0;
    initCounters();
  }
  /**
   * @brief An unbound DataType object of the specified size and format.
   * @param dataSize
   *          is the array size of the given de.desy.tine.client format.
   * @param dataFormat
   *          is the tine data type of the object to create.
   * @note A call to getData() must be made in order to retrieve the incoming
   *       data into the application array parameters.
   */
  public TDataType(int dataSize, short dataFormat)
  {
    dFormat = dataFormat >= 0 ? dataFormat : (short) (256 + dataFormat);
    switch (dFormat%256)
    {
      case TFormat.CF_DEFAULT:
        dSizeInBytes = dataSize;
        dArrayLength = dataSize;
        blknum = 1;
        blkcnt = 0;
        return; // no allocation
      case TFormat.CF_DOUBLE:
        hDataObject = new double[dataSize];
        break;
      case TFormat.CF_FLOAT:
        hDataObject = new float[dataSize];
        break;
      case TFormat.CF_SHORT:
        hDataObject = new short[dataSize];
        break;
      case TFormat.CF_BOOLEAN:
      case TFormat.CF_LONG:
        hDataObject = new int[dataSize];
        break;
      case TFormat.CF_NAME16:
        hDataObject = new NAME16[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME16[])hDataObject)[i] = new NAME16();
        break;
      case TFormat.CF_NAME32:
        hDataObject = new NAME32[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME32[])hDataObject)[i] = new NAME32();
        break;
      case TFormat.CF_NAME48:
        hDataObject = new NAME48[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME48[])hDataObject)[i] = new NAME48();
        break;
      case TFormat.CF_NAME64:
        hDataObject = new NAME64[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME64[])hDataObject)[i] = new NAME64();
        break;
      case TFormat.CF_NAME8:
        hDataObject = new NAME8[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME8[])hDataObject)[i] = new NAME8();
        break;
      case TFormat.CF_STRUCT:
      case TFormat.CF_BYTE:
        hDataObject = new byte[dataSize];
        break;
      case TFormat.CF_DBLINT:
        hDataObject = new DBLINT[dataSize];
        break;
      case TFormat.CF_DBLDBL:
        hDataObject = new DBLDBL[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((DBLDBL[])hDataObject)[i] = new DBLDBL();
        break;
      case TFormat.CF_FLTINT:
        hDataObject = new FLTINT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((FLTINT[])hDataObject)[i] = new FLTINT();
        break;
      case TFormat.CF_LNGINT:
        hDataObject = new INTINT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((INTINT[])hDataObject)[i] = new INTINT();
        break;
      case TFormat.CF_NAME16I:
        hDataObject = new NAME16I[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME16I[])hDataObject)[i] = new NAME16I();
        break;
      case TFormat.CF_NAME16FI:
        hDataObject = new NAME16FI[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME16FI[])hDataObject)[i] = new NAME16FI();
        break;
      case TFormat.CF_NAME16II:
        hDataObject = new NAME16II[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME16II[])hDataObject)[i] = new NAME16II();
        break;
      case TFormat.CF_NAME32I:
        hDataObject = new NAME32I[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME32I[])hDataObject)[i] = new NAME32I();
        break;
      case TFormat.CF_NAME48I:
        hDataObject = new NAME48I[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME48I[])hDataObject)[i] = new NAME48I();
        break;
      case TFormat.CF_NAME64I:
        hDataObject = new NAME64I[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME64I[])hDataObject)[i] = new NAME64I();
        break;
      case TFormat.CF_NAME16DBLDBL:
        hDataObject = new NAME16DBLDBL[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME16DBLDBL[])hDataObject)[i] = new NAME16DBLDBL();
        break;
      case TFormat.CF_NAME32DBLDBL:
        hDataObject = new NAME32DBLDBL[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME32DBLDBL[])hDataObject)[i] = new NAME32DBLDBL();
        break;
      case TFormat.CF_NAME64DBLDBL:
        hDataObject = new NAME64DBLDBL[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME64DBLDBL[])hDataObject)[i] = new NAME64DBLDBL();
        break;
      case TFormat.CF_SPECTRUM:
        hDataObject = new SPECTRUM(dataSize);
        break;
      case TFormat.CF_INTFLTINT:
        hDataObject = new INTFLTINT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((INTFLTINT[])hDataObject)[i] = new INTFLTINT();
        break;
      case TFormat.CF_XML:
      case TFormat.CF_TEXT:
        hDataObject = new char[dataSize];
        break;
      case TFormat.CF_USTRING:
        hDataObject = new USTRING[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((USTRING[])hDataObject)[i] = new USTRING();
        break;
      case TFormat.CF_UNAME:
        hDataObject = new UNAME[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((UNAME[])hDataObject)[i] = new UNAME();
        break;
      case TFormat.CF_ADDRESS:
        hDataObject = new INTINTINTINT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((INTINTINTINT[])hDataObject)[i] = new INTINTINTINT();
        break;
      case TFormat.CF_INTINTINT:
        hDataObject = new INTINTINT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((INTINTINT[])hDataObject)[i] = new INTINTINT();
        break;
      case TFormat.CF_POINT:
        hDataObject = new FLTFLT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((FLTFLT[])hDataObject)[i] = new FLTFLT();
        break;
      case TFormat.CF_FILTER:
        hDataObject = new INTFLTFLTFLT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((INTFLTFLTFLT[])hDataObject)[i] = new INTFLTFLTFLT();
        break;
      case TFormat.CF_FLTFLTINT:
        hDataObject = new FLTFLTINT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((FLTFLTINT[])hDataObject)[i] = new FLTFLTINT();
        break;
      case TFormat.CF_FLTINTINT:
        hDataObject = new FLTINTINT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((FLTINTINT[])hDataObject)[i] = new FLTINTINT();
        break;
      case TFormat.CF_FWINDOW:
        hDataObject = new INTINTFLTFLT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((INTINTFLTFLT[])hDataObject)[i] = new INTINTFLTFLT();
        break;
    }
    dArrayLength = dataSize;
    dSizeInBytes = dataSize * TFormat.formatSizeOf(dFormat);
    if (dFormat == TFormat.CF_SPECTRUM) 
    { // add room for the header
      dSizeInBytes += 96;
    }
    initCounters();
    dBuffers = new ByteArrayOutputStream[numblks * 3 / 2]; // room for NIOS calls
    for (int n = 0, i = 0; n < numblks; n++)
    {
      dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
      DataOutputStream ds = new DataOutputStream(dBuffers[n]);
      try
      {
        for (; i < dSizeInBytes && i < (n + 1) * mtu; i++) ds.write(0);
      }
      catch (Exception e)
      {
      }
    }
  }
  /**
   * @brief Returns the associated data timestamp as traditional UNIX timestamp
   *        (seconds since 1.1.1970).
   */
  public int getTimeStamp()
  {
    return timestamp;
  }
  /**
   * @brief Returns the associated data timestamp as JAVA timestamp
   *        (milliseconds since 1.1.1970).
   */
  public long getDataTimeStamp()
  {
    if (dTimestamp > 0) return dTimestamp;
    return (long) timestamp * 1000 + (long) timestampMSEC;
  }
  /**
   * @brief Sets the associated data timestamp from a JAVA timestamp
   *        (milliseconds since 1.1.1970).
   */
  public void setDataTimeStamp(long TimeStamp)
  {
    dTimestamp = TimeStamp;
    timestamp = (int) (TimeStamp / 1000);
    timestampMSEC = (int) (TimeStamp % 1000);
  }
  /**
   * @brief Sets the associated data timestamp from a TINE timestamp
   *        (seconds since 1.1.1970 + fractional seconds).
   */
  public void setDataTimeStamp(double TimeStamp)
  {
    long t = ((long)TimeStamp)*1000 + ((long)(TimeStamp * 1000) % 1000);
    setDataTimeStamp(t);
  }
  /**
   * @brief Returns the associated data completion length
   */
  public int getLastDataSize()
  {
    return dCompletionLength;
  }
  public TDataType(byte[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * @brief A subset of an array of bytes
   * @param data
   *          is a reference to the byte array used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(byte[] data, int length)
  {
    initTDataType(data, length);
  }
  public void initTDataType(byte[] data, int length)
  {
    try
    {
      if (length > data.length) length = data.length;
      dFormat = TFormat.CF_BYTE;
      dArrayLength = length;
      dSizeInBytes = length;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu; i++)
          ds.writeByte(data[i]);
        byte[] b = dBuffers[n].toByteArray();
        dBuffers[n].reset();
        ds.write(b, 0, b.length);
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  /**
   * @brief An array of 'structs' with the given tag or untagged structs if
   *        tag.length == 0 User-defined structures are allowed in tine and not
   *        too difficult to use. It is strongly encouraged to make use of
   *        'tagged' structures whereby the server and client both register the
   *        structure contents with the tine engine and apply a tag. Only then
   *        can issues of 'byte-swapping' and structure boundaries be handled
   *        correctly. As java does not have structures, the associated class is
   *        responsible for providing a toStruct() method which returns a byte
   *        stream containing the contents of the desired structure.
   * @param data
   *          is a reference to the byte array stream containing the 'struct'
   *          used to create and bind to the TDataType object.
   * @param tag
   *          is the structure tag associated with the 'struct.'
   */
  public TDataType(byte[] data, String tag)
  {
    dTag = tag;
    try
    {
      dFormat = TFormat.CF_STRUCT;
      dArrayLength = data.length;
      dSizeInBytes = data.length;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu; i++)
          ds.writeByte(data[i]);
        byte[] b = dBuffers[n].toByteArray();
        dBuffers[n].reset();
        ds.write(b, 0, b.length);
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  /**
   * @brief A de.desy.tine.client Tagged Structure User-defined structures are allowed in tine
   *        and not too difficult to use. It is strongly encouraged to make use
   *        of 'tagged' structures whereby the server and client both register
   *        the structure contents with the tine engine and apply a tag. Only
   *        then can issues of 'byte-swapping' and structure boundaries be
   *        handled correctly. As java does not have structures, the associated
   *        class is responsible for providing a toStruct() method which returns
   *        a byte stream containing the contents of the desired structure.
   * @param data
   *          is a reference to a TTaggedStructure object.
   */
  public TDataType(TTaggedStructure data)
  {
    dTag = data.getName();
    try
    {
      byte[] hb = data.toByteArray();
      dFormat = TFormat.CF_STRUCT;
      dArrayLength = data.getSizeInBytes();
      dSizeInBytes = data.getSizeInBytes();
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu; i++)
          ds.writeByte(hb[i]);
        byte[] b = dBuffers[n].toByteArray();
        dBuffers[n].reset();
        ds.write(b, 0, b.length);
        ds.close();
      }
      hDataObject = data.getByteArray();
      hStructObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  /**
   * @brief A array of de.desy.tine.client Tagged Structures User-defined structures are
   *        allowed in tine and not too difficult to use. It is strongly
   *        encouraged to make use of 'tagged' structures whereby the server and
   *        client both register the structure contents with the tine engine and
   *        apply a tag. Only then can issues of 'byte-swapping' and structure
   *        boundaries be handled correctly. As java does not have structures,
   *        the associated class is responsible for providing a toStruct()
   *        method which returns a byte stream containing the contents of the
   *        desired structure.
   * @param data
   *          is a reference to an array of TTaggedStructure object.
   */
  public TDataType(TTaggedStructure[] data)
  {
    if (data[0] == null)
    {
      System.out.println("Tagged Structure data type unallocated !");
      return;
    }
    for (int k = 1; k < data.length; k++)
      if (data[k] == null) data[k] = (TTaggedStructure) data[0].clone();
    dTag = data[0].getName();
    byte[] hbytes = new byte[data[0].getSizeInBytes() * data.length];
    try
    {
      int dsiz = data[0].getSizeInBytes();
      dFormat = TFormat.CF_STRUCT;
      dArrayLength = dsiz * data.length;
      dSizeInBytes = dsiz * data.length;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int k = 0; k < data.length; k++)
      {
        //System.arraycopy(data[k].getByteArray(), 0, hbytes, k * dsiz, dsiz);
        System.arraycopy(data[k].toByteArray(), 0, hbytes, k * dsiz, dsiz);
      }
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu; i++)
          ds.writeByte(hbytes[i]);
        byte[] b = dBuffers[n].toByteArray();
        dBuffers[n].reset();
        ds.write(b, 0, b.length);
        ds.close();
      }
      hDataObject = hbytes;
      hStructObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  /**
   * @brief An array of shorts (2-byte integers) This corresponds to tine format
   *        CF_SHORT
   * @param data
   *          is a reference to the short array used to create and bind to the
   *          TDataType object
   */
  public TDataType(short[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * @brief A subset of an array of shorts (2-byte integers) This corresponds to
   *        tine format CF_SHORT
   * @param data
   *          is a reference to the short array used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(short[] data, int length)
  {
    initTDataType(data, length);
  }
  public void initTDataType(short[] data, int length)
  {
    try
    {
      if (length > data.length) length = data.length;
      dFormat = TFormat.CF_SHORT;
      dArrayLength = length;
      dSizeInBytes = length * 2;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu / 2; i++)
          ds.writeShort(data[i]);
        byte[] b = dBuffers[n].toByteArray();
        Swap.Bytes(b, 2);
        dBuffers[n].reset();
        ds.write(b, 0, b.length);
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  /**
   * @brief A single short value This method is only useful for quick creation
   *        of data 'down' objects (i.e. data sets sent to the server). As there
   *        is no reference to the argument passed, it cannot be used to bind
   *        returning data to the object passed as the argument.
   * @param data
   *          is a single short value used to create the TDataType object.
   */
  public TDataType(short data)
  {
    try
    {
      dFormat = TFormat.CF_SHORT;
      dArrayLength = 1;
      dSizeInBytes = 2;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffers[0] = new ByteArrayOutputStream(dBufferSegSize);
      DataOutputStream ds = new DataOutputStream(dBuffers[0]);
      ds.writeShort(data);
      byte[] b = dBuffers[0].toByteArray();
      Swap.Bytes(b, 2);
      dBuffers[0].reset();
      ds.write(b, 0, b.length);
      ds.close();
      hDataObject = null;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  /**
   * @brief An array of ints (4-byte integers) This corresponds to tine format
   *        CF_LONG
   * @param data
   *          is a reference to the int array used to create and bind to the
   *          TDataType object
   */
  public TDataType(int[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * @brief A subset of an array of ints (4-byte integers) This corresponds to
   *        tine format CF_LONG
   * @param data
   *          is a reference to the int array used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(int[] data, int length)
  {
    initTDataType(data, length);
  }
  public void initTDataType(int[] data, int length)
  {
    try
    {
      if (length > data.length) length = data.length;
      dFormat = TFormat.CF_LONG;
      dArrayLength = length;
      dSizeInBytes = length * 4;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu / 4; i++)
          ds.writeInt(data[i]);
        byte[] b = dBuffers[n].toByteArray();
        Swap.Bytes(b, 4);
        dBuffers[n].reset();
        ds.write(b, 0, b.length);
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  /**
   * @brief An array of floats This corresponds to tine format CF_FLOAT
   * @param data
   *          is a reference to the float array used to create and bind to the
   *          TDataType object
   */
  public TDataType(float[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * @brief A subset of an array of floats This corresponds to tine format
   *        CF_FLOAT
   * @param data
   *          is a reference to the float array used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(float[] data, int length)
  {
    initTDataType(data, length);
  }
  public void initTDataType(float[] data, int length)
  {
    try
    {
      if (length > data.length) length = data.length;
      dFormat = TFormat.CF_FLOAT;
      dArrayLength = length;
      dSizeInBytes = length * 4;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks * 3 / 2]; // NIOS has mtu =
                                                              // 1024
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu / 4; i++)
          ds.writeFloat(data[i]);
        byte[] b = dBuffers[n].toByteArray();
        Swap.Bytes(b, 4);
        dBuffers[n].reset();
        ds.write(b, 0, b.length);
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  /**
   * @brief An array of tine data objects (e.g. FLTINTs, FLTFLTs, NAME16s, etc.)
   * can be bound to a tine data type object.  Tine data type objects are used
   * in data links between client and server.
   * @param data
   *          is a reference to the data objects used to create and bind to
   *          the TDataType object
   */
  public TDataType(TCompoundDataObject[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * @brief An array of tine data objects (e.g. FLTINTs, FLTFLTs, NAME16s, etc.)
   * can be bound to a tine data type object.  Tine data type objects are used
   * in data links between client and server.
   * @param data
   *          is a reference to the data objects used to create and bind to
   *          the TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(TCompoundDataObject[] data, int length)
  {
    initTDataType(data, length);
  }
  /**
   * @brief An array of FLTINTs (float + integer pairs) This data type is
   *        frequently used to retrieve archived parameters (float value +
   *        timestamp)
   * @param data
   *          is a reference to the FLTINT objects used to create and bind to
   *          the TDataType object
   */
  private void initTDataType(TCompoundDataObject[] data, int length)
  {
    try
    {
      if (length > data.length) length = data.length;
      Object tdo = data[0] != null ? 
        data[0] : data.getClass().getComponentType().newInstance();
      dFormat = ((TCompoundDataObject) tdo).getFormat();
      int elementSize = TFormat.formatSizeOf(dFormat);
      dArrayLength = length;
      dSizeInBytes = length * elementSize;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < length && i < (n + 1) * mtu / elementSize; i++)
          setObjectContents(ds, data, i); //
        byte[] b = dBuffers[n].toByteArray();
        dBuffers[n].reset();
        ds.write(b, 0, b.length);
        ds.close();
      }
      setObjectHandle(data); //
    }
    catch (NullPointerException e)
    {
      e.printStackTrace();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }        
    catch (Exception e)
    {
      e.printStackTrace();      
    }
  }
  public void initTDataType(Object data, int length)
  {
    try
    {
      if (length > getObjectSize(data)) length = getObjectSize(data);
      dFormat = getObjectFormat(data);
      int elementSize = TFormat.formatSizeOf(dFormat);
      dArrayLength = length;
      dSizeInBytes = length * elementSize;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < length && i < (n + 1) * mtu / elementSize; i++)
          setObjectContents(ds, data, i);
        byte[] b = dBuffers[n].toByteArray();
        dBuffers[n].reset();
        ds.write(b, 0, b.length);
        ds.close();
      }
      setObjectHandle(data);
    }
    catch (NullPointerException e)
    {
      e.printStackTrace();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  public TDataType(SPECTRUM data)
  {
    try
    {
      dFormat = TFormat.CF_SPECTRUM;
      dArrayLength = data.length;
      dSizeInBytes = data.length * 4 + 96;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      byte[] bdata = data.toByteArray();
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dSizeInBytes && i < (n + 1) * mtu; i++)
          ds.writeByte(bdata[i]);
        byte[] b = dBuffers[n].toByteArray();
        dBuffers[n].reset();
        ds.write(b, 0, b.length);
        ds.close();
      }
      hDataObject = data;
    }
    catch (NullPointerException e)
    {
      e.printStackTrace();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  private void repackageDataForOutput()
  {
    try
    {
      int fsiz = TFormat.formatSizeOf(dFormat);
      initCounters();
      for (int i = 0, n = 0, m = 0; n < numblks; n++, m = i)
      {
        dBuffers[n].reset();
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < m + dBufferSegSize / fsiz; i++)
        {
          setObjectContents(ds, getDataObject(), i);
          if (dFormat == TFormat.CF_TEXT ||
              dFormat == TFormat.CF_XML) i = dArrayLength;
        }
//        byte[] b = dBuffers[n].toByteArray();
//        dBuffers[n].reset();
//        ds.write(b, 0, b.length);
        ds.close();
      }
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  public int initTDataType(short dataFormat, int dataSize)
  {
    int fmtsiz;
    dFormat = dataFormat >= 0 ? dataFormat : (short) (256 + dataFormat);
    if ((fmtsiz = TFormat.formatSizeOf(dFormat)) < 0) return 2;
    if (fmtsiz == 0) return 0;
    dArrayLength = dataSize;
    dSizeInBytes = dArrayLength * fmtsiz;
    initCounters();
    dBuffers = new ByteArrayOutputStream[numblks];
    for (int n = 0; n < numblks; n++)
    {
      dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
    }
    return 0;
  }
  /**
   * @brief An array of doubles This corresponds to tine format CF_DOUBLE
   * @param data
   *          is a reference to the double array used to create and bind to the
   *          TDataType object
   */
  public TDataType(double[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * @brief A subset of an array of doubles This corresponds to tine format
   *        CF_DOUBLE
   * @param data
   *          is a reference to the double array used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(double[] data, int length)
  {
    initTDataType(data, length);
  }
  public void initTDataType(double[] data, int length)
  {
    try
    {
      if (length > data.length) length = data.length;
      dFormat = TFormat.CF_DOUBLE;
      dArrayLength = length;
      dSizeInBytes = length * 8;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu / 8; i++)
          ds.writeDouble(data[i]);
        byte[] b = dBuffers[n].toByteArray();
        Swap.Bytes(b, 8);
        dBuffers[n].reset();
        ds.write(b, 0, b.length);
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  public void update(byte[] data, int num, int id)
  {
    update(data, 0, data.length, num, id);
  }
  public void update(byte[] data, int off, int len, int num, int id)
  {
    if (debugLevel > 1)
    {
      System.out.println("update: blk " + num + " id	" + id + "(" + data.length + " bytes)");
    }
    blknum = num;
    // check num vs numblks, and id for integrity
    if (id != blkid)
    {
      blkid = id;
      blksin = 0;
      bytesin = 0;
    }
    if (data.length == 0)
    {
      blksin = 1;
      return;
    }
    if (dBuffers == null)
    {
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
      }
    }
    try
    {
      // possible small mtu server (NIOS) ...
      if (dBuffers[num - 1] == null) dBuffers[num - 1] = new ByteArrayOutputStream(dBufferSegSize);
      dBuffers[num - 1].reset();
      DataOutputStream ds = new DataOutputStream(dBuffers[num - 1]);
      ds.write(data, off, len);
      ds.close();
      blksin++;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  public int getData(byte[] data)
  {
    return getData(data,data != null ? data.length : 0, 0); 
  }
  public int getData(byte[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    int len = dArrayLength;  // length of the buffered array
    if (len > length) len = length; // can't get more than is there !
    try
    {
      int i, n, m;
      for (i = 0, n = 0, m = 0; n < numblks; n++, m = i)
      {
        byte[] b = dBuffers[n].toByteArray();
        ByteArrayInputStream bs = new ByteArrayInputStream(b);
        DataInputStream ds = new DataInputStream(bs);
        for (; i < len && i < m + b.length; i++)
          data[i+offset] = ds.readByte();
        ds.close();
        bs.close();
      }
      dCompletionLength = i;
      if (dFormat == TFormat.CF_STRUCT)
      {
        TStructDescription tsd = TStructRegistry.get(dTag);
        if (tsd != null  && tsd.getSize() != 0) 
        {
          dCompletionLength /= tsd.getSize();
        }
      }
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    return 0;
  }
  public int getData(short[] data)
  {
    return getData(data,data != null ? data.length : 0, 0);
  }
  public int getData(short[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    int len = dArrayLength;  // length of the buffered array
    if (len > length) len = length; // can't get more than is there !
    try
    {
      int i, n, m, fmtsiz = TFormat.formatSizeOf(dFormat);
      ByteArrayInputStream bs;
      DataInputStream ds;
      for (i = 0, n = 0, m = 0; n < numblks; n++, m = i)
      {
        if (dBuffers[n] == null) break;
        byte[] b = dBuffers[n].toByteArray();
        Swap.Bytes(b, fmtsiz);
        bs = new ByteArrayInputStream(b);
        ds = new DataInputStream(bs);
        switch (dFormat)
        {
          case TFormat.CF_DOUBLE:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = (short) ds.readDouble();
            break;
          case TFormat.CF_FLOAT:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = (short) ds.readFloat();
            break;
          case TFormat.CF_LONG:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = (short) ds.readInt();
            break;
          case TFormat.CF_SHORT:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = ds.readShort();
            break;
          default:
            ds.close();
            bs.close();
            return TErrorList.illegal_format;
        }
        ds.close();
        bs.close();
      }
      dCompletionLength = i;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    return 0;
  }
  public int putData(short data)
  {
    short[] d = new short[1];
    d[0] = data;
    return putData(d);
  }
  public int putData(short[] data)
  {
    return putData(data,data != null ? data.length : 0,0);
  }
  public int putData(short[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
   if (dArrayLength > length) dArrayLength = length;
    switch (dFormat)
    {
      case TFormat.CF_DOUBLE:
        for (int i = 0; i < dArrayLength; i++)
          ((double[])hDataObject)[i] = (double) data[i+offset];
        break;
      case TFormat.CF_FLOAT:
        for (int i = 0; i < dArrayLength; i++)
          ((float[])hDataObject)[i] = (float) data[i+offset];
        break;
      case TFormat.CF_SHORT:
        for (int i = 0; i < dArrayLength; i++)
          ((short[])hDataObject)[i] = (short) data[i+offset];
        break;
      case TFormat.CF_LONG:
        for (int i = 0; i < dArrayLength; i++)
          ((int[])hDataObject)[i] = (int) data[i+offset];
        break;
      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public int getData()
  {
    switch (dFormat)
    {
      case TFormat.CF_STRUCT:
        if (hDataObject != null) getData(hDataObject);
        if (hStructObject != null && hStructObject instanceof TTaggedStructure)
        {
          ((TTaggedStructure)hStructObject).toStruct((byte[])hDataObject);
        }
        else if (hStructObject instanceof TTaggedStructure[])
        {
          TTaggedStructure[] tts = (TTaggedStructure[])hStructObject;
          int dsiz = tts[0].getSizeInBytes();
          byte[] hb = new byte[dsiz];
          for (int i = 0; i < tts.length; i++)
          {
            System.arraycopy((byte[])hDataObject, i * dsiz, hb, 0, dsiz);
            tts[i].toStruct(hb);
          }
        }
        break;
      default:
        getData(getDataObject());
        break;
    }
    return 0;
  }
  public int getData(int[] data)
  {
    return getData(data,data != null ? data.length : 0, 0); 
  }
  public int getData(int[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    int len = dArrayLength;  // length of the buffered array
    if (len > length) len = length; // can't get more than is there !
    try
    {
      int i, n, m, fmtsiz = TFormat.formatSizeOf(dFormat);
      ByteArrayInputStream bs;
      DataInputStream ds;
      for (i = 0, n = 0, m = 0; n < numblks; n++, m = i)
      {
        if (dBuffers[n] == null) break;
        byte[] b = dBuffers[n].toByteArray();
        Swap.Bytes(b, fmtsiz);
        bs = new ByteArrayInputStream(b);
        ds = new DataInputStream(bs);
        switch (dFormat)
        {
          case TFormat.CF_DOUBLE:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = (int) ds.readDouble();
            break;
          case TFormat.CF_FLOAT:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = (int) ds.readFloat();
            break;
          case TFormat.CF_BOOLEAN:
          case TFormat.CF_LONG:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = ds.readInt();
            break;
          case TFormat.CF_SHORT:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = ds.readShort();
            break;
          default:
            ds.close();
            bs.close();
            return TErrorList.illegal_format;
        }
        ds.close();
        bs.close();
      }
      dCompletionLength = i;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    return 0;
  }
  public int putData(int data)
  {
    int[] d = new int[1];
    d[0] = data;
    return putData(d);
  }
  public int putData(int[] data)
  {
    return putData(data,data != null ? data.length : 0,0);
  }
  public int putData(int[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dArrayLength > length) dArrayLength = length;
    switch (dFormat)
    {
      case TFormat.CF_DOUBLE:
        for (int i = 0; i < dArrayLength; i++)
          ((double[])hDataObject)[i] = (double) data[i+offset];
        break;
      case TFormat.CF_FLOAT:
        for (int i = 0; i < dArrayLength; i++)
          ((float[])hDataObject)[i] = (float) data[i+offset];
        break;
      case TFormat.CF_SHORT:
        for (int i = 0; i < dArrayLength; i++)
          ((short[])hDataObject)[i] = (short) data[i+offset];
        break;
      case TFormat.CF_BOOLEAN:
      case TFormat.CF_LONG:
        for (int i = 0; i < dArrayLength; i++)
          ((int[])hDataObject)[i] = (int) data[i+offset];
        break;
      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public int getData(float[] data)
  {
    return getData(data,data != null ? data.length : 0, 0); 
  }
  public int getData(float[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    int len = dArrayLength;  // length of the buffered array
    if (len > length) len = length; // can't get more than is there !
    try
    {
      int i, n, m, fmtsiz = TFormat.formatSizeOf(dFormat);
      ByteArrayInputStream bs;
      DataInputStream ds;
      for (i = 0, n = 0, m = 0; n < numblks; n++, m = i)
      {
        if (dBuffers[n] == null) break;
        byte[] b = dBuffers[n].toByteArray();
        Swap.Bytes(b, fmtsiz);
        bs = new ByteArrayInputStream(b);
        ds = new DataInputStream(bs);
        switch (dFormat)
        {
          case TFormat.CF_DOUBLE:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = (float) ds.readDouble();
            break;
          case TFormat.CF_FLOAT:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = ds.readFloat();
            break;
          case TFormat.CF_LONG:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = ds.readInt();
            break;
          case TFormat.CF_SHORT:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = ds.readShort();
            break;
          default:
            ds.close();
            bs.close();
            return TErrorList.illegal_format;
        }
        ds.close();
        bs.close();
      }
      dCompletionLength = i;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    return 0;
  }
  public int getData(TTaggedStructure[] data)
  {
    return getData(data,data != null ? data.length : 0, 0);
  }
  public int getData(TTaggedStructure[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dFormat != TFormat.CF_STRUCT) return TErrorList.illegal_format;
    if (dTag.compareTo(data[0].getName()) != 0) return TErrorList.invalid_structure_tag;
    if (hDataObject == null) return TErrorList.invalid_data;
    getData(hDataObject);
    int dsiz = data[0].getSizeInBytes();
    byte[] hb = new byte[dsiz];
    for (int i = 0; i < data.length; i++)
    { // copy the byte stream into the referenced data object
      System.arraycopy(hDataObject, i*dsiz, hb, 0, dsiz);
      data[i].toStruct(hb);
    }
    dCompletionLength /= dsiz;
    return 0;
  }  
  public int getData(TCompoundDataObject[] data)
  {
    return getData(data,data != null ? data.length : 0, 0); 
  }
  public int getData(TCompoundDataObject[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    int len = dArrayLength;  // length of the buffered array
    if (len > length) len = length; // can't get more than is there !
    int siz = 0;  // size in bytes of the TCompoundDataObject
    try
    {
      int i, n, m, j;
      for (i = 0, n = 0, m = 0, j = 0; n < numblks; n++, m = i)
      {
        if (data[i+offset] == null) makeNewCompoundDataObject(data,i+offset);
        if (siz == 0) siz = data[i+offset].getSizeInBytes();
        if (dBuffers[n] == null) break;
        byte[] b = dBuffers[n].toByteArray();
        ByteArrayInputStream bs = new ByteArrayInputStream(b);
        DataInputStream ds = new DataInputStream(bs);
        for (; i < len && i < m + b.length /siz; i++)
        {
          if (data[i+offset] == null) makeNewCompoundDataObject(data,i+offset);
          if (data[i+offset].getByteArray() == null) data[i+offset].toByteArray();
          byte[] ba = data[i+offset].getByteArray();
          for (j = 0; j <siz; j++) ba[j] = ds.readByte();
          data[i+offset].toStruct();
        }
        ds.close();
        bs.close();
      }
      dCompletionLength = i;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    return 0;
  }  
  public int getData(SPECTRUM data)
  {
    if (dFormat != TFormat.CF_SPECTRUM) return TErrorList.illegal_format;
    if (dArrayLength > data.length) return TErrorList.dimension_error;
    int dSize = dArrayLength + 24; // effective data size for SPECTRUM
    try
    {
      int i, n, m;
      for (i = 0, n = 0, m = 0; n < numblks; n++, m = i)
      {
        if (dBuffers[n] == null) break;
        byte[] b = dBuffers[n].toByteArray();
        ByteArrayInputStream bs = new ByteArrayInputStream(b);
        DataInputStream ds = new DataInputStream(bs);
        for (; i < dSize * 4 && i < m + b.length; i++)
        {
          data.hByteArray[i] = ds.readByte();
        }
        data.toStruct();
        ds.close();
        bs.close();
      }
      dCompletionLength = i - 24;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    return 0;
  }
  public int putData(float data)
  {
    float[] d = new float[1];
    d[0] = data;
    return putData(d);
  }
  public int putData()
  {
    repackageDataForOutput();
    return 0;    
  }
  public int putData(float[] data)
  {
    return putData(data,data != null ? data.length : 0,0);
  }
  public int putData(float[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dArrayLength > length) dArrayLength = length;
    switch (dFormat)
    {
      case TFormat.CF_DOUBLE:
        for (int i = 0; i < dArrayLength; i++)
          ((double[])hDataObject)[i] = (double) data[i+offset];
        break;
      case TFormat.CF_FLOAT:
        for (int i = 0; i < dArrayLength; i++)
          ((float[])hDataObject)[i] = (float) data[i+offset];
        break;
      case TFormat.CF_SHORT:
        for (int i = 0; i < dArrayLength; i++)
          ((short[])hDataObject)[i] = (short) data[i+offset];
        break;
      case TFormat.CF_LONG:
        for (int i = 0; i < dArrayLength; i++)
          ((int[])hDataObject)[i] = (int) data[i+offset];
        break;
      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public int putData(FLTINT data)
  {
    FLTINT[] d = new FLTINT[1];
    d[0] = data;
    return putData(d);
  }
  public int putData(FLTINT[] data)
  {
    return putData(data,data != null ? data.length : 0,0);
  }
  public int putData(FLTINT[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dArrayLength > length) dArrayLength = length;
    switch (dFormat)
    {
      case TFormat.CF_LNGINT:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((INTINT[])hDataObject)[i] != null)
            ((INTINT[])hDataObject)[i].setValues((int) data[i+offset].fval, data[i+offset].ival);
          else
            ((INTINT[])hDataObject)[i] = 
              new INTINT((int) data[i+offset].fval, data[i+offset].ival);
        }
        break;
      case TFormat.CF_FLTINT:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          if (((FLTINT[])hDataObject)[i] != null)
            ((FLTINT[])hDataObject)[i].setValues(data[i+offset].fval, data[i+offset].ival);
          else
            ((FLTINT[])hDataObject)[i] = data[i+offset];
        }
        break;
      case TFormat.CF_DBLDBL:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((DBLDBL[])hDataObject)[i] != null)
            ((DBLDBL[])hDataObject)[i].setValues(data[i+offset].fval, data[i+offset].ival);
          else
            ((DBLDBL[])hDataObject)[i] = new DBLDBL(data[i+offset].fval, data[i+offset].ival);
        }
        break;
      case TFormat.CF_FLOAT:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          ((float[])hDataObject)[i] = data[i+offset].fval;
        }
        break;
      case TFormat.CF_DOUBLE:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          ((double[])hDataObject)[i] = data[i+offset].fval;
        }
        break;
      case TFormat.CF_LONG:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          ((int[])hDataObject)[i] = data[i+offset].ival;
        }
        break;
      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public int putData(NAME16FI[] data)
  {
    return putData(data,data != null ? data.length : 0, 0);
  }
  public int putData(NAME16FI[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dArrayLength > length) dArrayLength = length;
    switch (dFormat)
    {
      case TFormat.CF_NAME16FI:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((NAME16FI[])hDataObject)[i] != null)
            ((NAME16FI[])hDataObject)[i].setValues(data[i+offset].name,data[i+offset].fval,data[i+offset].ival);
          else
            ((NAME16FI[])hDataObject)[i] = data[i+offset];
        }
        break;
      case TFormat.CF_NAME16II:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((NAME16II[])hDataObject)[i] != null)
            ((NAME16II[])hDataObject)[i].setValues(data[i+offset].name,(int)data[i+offset].fval,data[i+offset].ival);
          else
            ((NAME16II[])hDataObject)[i] = 
              new NAME16II(data[i+offset].name, (int)data[i+offset].fval, data[i+offset].ival);
        }
        break;
      case TFormat.CF_NAME16DBLDBL:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((NAME16DBLDBL[])hDataObject)[i] != null)
            ((NAME16DBLDBL[])hDataObject)[i].setValues(data[i+offset].name,data[i+offset].fval,data[i+offset].ival);
          else
            ((NAME16DBLDBL[])hDataObject)[i] =
              new NAME16DBLDBL(data[i+offset].name, data[i+offset].fval, data[i+offset].ival);
        }
        break;
      case TFormat.CF_NAME32DBLDBL:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((NAME32DBLDBL[])hDataObject)[i] != null)
            ((NAME32DBLDBL[])hDataObject)[i].setValues(data[i+offset].name,data[i+offset].fval,data[i+offset].ival);
          else
            ((NAME32DBLDBL[])hDataObject)[i] = 
              new NAME32DBLDBL(data[i+offset].name, data[i+offset].fval, data[i+offset].ival);
        }
        break;
      case TFormat.CF_NAME64DBLDBL:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((NAME64DBLDBL[])hDataObject)[i] != null)
            ((NAME64DBLDBL[])hDataObject)[i].setValues(data[i+offset].name,data[i+offset].fval,data[i+offset].ival);
          else
            ((NAME64DBLDBL[])hDataObject)[i] = 
              new NAME64DBLDBL(data[i+offset].name, data[i+offset].fval, data[i+offset].ival);
        }
        break;
      case TFormat.CF_USTRING:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i] == null) continue;
          if (((USTRING[])hDataObject)[i] != null)
            ((USTRING[])hDataObject)[i].setValues(data[i+offset].ival,data[i+offset].fval,0,0,data[i+offset].name);
          else
            ((USTRING[])hDataObject)[i] = 
              new USTRING(data[i+offset].ival,data[i+offset].fval,0,0,data[i+offset].name);
        }
        break;
      case TFormat.CF_UNAME:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((UNAME[])hDataObject)[i] != null)
            ((UNAME[])hDataObject)[i].setValues(data[i+offset].ival,data[i+offset].fval,0,0,data[i+offset].name);
          else
            ((UNAME[])hDataObject)[i] = 
              new UNAME(data[i+offset].ival,data[i+offset].fval,0,0,data[i+offset].name);
        }
        break;
      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public int putData(NAME32DBLDBL[] data)
  {
    return putData(data,data != null ? data.length : 0, 0);
  }
  public int putData(NAME32DBLDBL[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dArrayLength > length) dArrayLength = length;
    switch (dFormat)
    {
      case TFormat.CF_NAME16II:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((NAME16II[])hDataObject)[i] != null)
            ((NAME16II[])hDataObject)[i].setValues(data[i+offset].name,(int)data[i+offset].d1val,(int)data[i+offset].d2val);
          else
            ((NAME16II[])hDataObject)[i] = 
              new NAME16II(data[i+offset].name, (int)data[i+offset].d1val, (int)data[i+offset].d2val);
        }
        break;
      case TFormat.CF_NAME16FI:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((NAME16FI[])hDataObject)[i] != null)
            ((NAME16FI[])hDataObject)[i].setValues(data[i+offset].name,(float)data[i+offset].d1val,(int)data[i+offset].d2val);
          else
            ((NAME16FI[])hDataObject)[i] = 
              new NAME16FI(data[i].name, (float)data[i+offset].d1val, (int)data[i+offset].d2val);
        }
        break;
      case TFormat.CF_NAME16DBLDBL:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((NAME16DBLDBL[])hDataObject)[i] != null)
            ((NAME16DBLDBL[])hDataObject)[i].setValues(data[i+offset].name,data[i+offset].d1val,data[i+offset].d2val);
          else
            ((NAME16DBLDBL[])hDataObject)[i] = 
              new NAME16DBLDBL(data[i+offset].name, data[i+offset].d1val, data[i+offset].d2val);
        }
        break;
      case TFormat.CF_NAME32DBLDBL:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((NAME32DBLDBL[])hDataObject)[i] != null)
            ((NAME32DBLDBL[])hDataObject)[i].setValues(data[i+offset].name,data[i+offset].d1val,data[i+offset].d2val);
          else
            ((NAME32DBLDBL[])hDataObject)[i] = data[i+offset]; 
        }
        break;
      case TFormat.CF_NAME64DBLDBL:
        for (int i = 0; i < dArrayLength; i++) 
        {
          if (data[i+offset] == null) continue;
          if (((NAME64DBLDBL[])hDataObject)[i] != null)
            ((NAME64DBLDBL[])hDataObject)[i].setValues(data[i+offset].name,data[i+offset].d1val,data[i+offset].d2val);
          else
            ((NAME64DBLDBL[])hDataObject)[i] = 
              new NAME64DBLDBL(data[i+offset].name, data[i+offset].d1val, data[i+offset].d2val);
        }
        break;
      case TFormat.CF_USTRING:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((USTRING[])hDataObject)[i] != null)
            ((USTRING[])hDataObject)[i].setValues((int)data[i+offset].d2val,(float)data[i+offset].d1val,0,0,data[i+offset].name);
          else
            ((USTRING[])hDataObject)[i] = 
              new USTRING((int)data[i].d2val,(float)data[i+offset].d1val,0,0,data[i+offset].name);
        }
        break;
      case TFormat.CF_UNAME:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((UNAME[])hDataObject)[i] != null)
            ((UNAME[])hDataObject)[i].setValues((int)data[i+offset].d2val,(float)data[i+offset].d1val,0,0,data[i+offset].name);
          else
            ((UNAME[])hDataObject)[i] = 
              new UNAME((int)data[i+offset].d2val,(float)data[i+offset].d1val,0,0,data[i+offset].name);
        }
        break;
      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public int putData(INTINT data)
  {
    INTINT[] d = new INTINT[1];
    d[0] = data;
    return putData(d);
  }
  public int putData(INTINT[] data)
  {
    return putData(data,data != null ? data.length : 0,0);
  }
  public int putData(INTINT[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dArrayLength > length) dArrayLength = length;
    switch (dFormat)
    {
      case TFormat.CF_LNGINT:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((INTINT[])hDataObject)[i] != null)
            ((INTINT[])hDataObject)[i].setValues(data[i+offset].i1val, data[i+offset].i2val);
          else
            ((INTINT[])hDataObject)[i] = data[i+offset]; 
        }
        break;
      case TFormat.CF_FLTINT:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((FLTINT[])hDataObject)[i] != null)
            ((FLTINT[])hDataObject)[i].setValues(data[i+offset].i1val, data[i+offset].i2val);
          else
            ((FLTINT[])hDataObject)[i] = new FLTINT(data[i+offset].i1val, data[i+offset].i2val);
        }
        break;
      case TFormat.CF_DBLDBL:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((DBLDBL[])hDataObject)[i] != null)
            ((DBLDBL[])hDataObject)[i].setValues(data[i+offset].i1val, data[i+offset].i2val);
          else
            ((DBLDBL[])hDataObject)[i] = new DBLDBL(data[i+offset].i1val, data[i+offset].i2val);
        }
        break;
      case TFormat.CF_FLOAT:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          ((float[])hDataObject)[i] = data[i+offset].i1val;
        }
        break;
      case TFormat.CF_DOUBLE:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          ((double[])hDataObject)[i] = data[i+offset].i1val;
        }
        break;
      case TFormat.CF_LONG:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          ((int[])hDataObject)[i] = data[i+offset].i1val;
        }
        break;
      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public int putData(DBLDBL data)
  {
    DBLDBL[] d = new DBLDBL[1];
    d[0] = data;
    return putData(d);
  }
  public int putData(DBLDBL[] data)
  {
    return putData(data,data != null ? data.length : 0, 0);
  }
  public int putData(DBLDBL[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dArrayLength > length) dArrayLength = length;
    switch (dFormat)
    {
      case TFormat.CF_LNGINT:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((INTINT[])hDataObject)[i] != null)
            ((INTINT[])hDataObject)[i].setValues((int) data[i+offset].d1val, (int) data[i+offset].d2val);
          else
            ((INTINT[])hDataObject)[i] = 
              new INTINT((int) data[i+offset].d1val, (int) data[i+offset].d2val);
        }
        break;
      case TFormat.CF_FLTINT:
        for (int i = 0; i < dArrayLength; i++) 
        {
          if (data[i+offset] == null) continue;
          if (((FLTINT[])hDataObject)[i] != null)
            ((FLTINT[])hDataObject)[i].setValues((float) data[i+offset].d1val, (int) data[i+offset].d2val);
          else
            ((FLTINT[])hDataObject)[i] = 
              new FLTINT((float) data[i+offset].d1val, (int) data[i+offset].d2val);
        }
        break;
      case TFormat.CF_DBLDBL:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((DBLDBL[])hDataObject)[i] != null)
            ((DBLDBL[])hDataObject)[i].setValues(data[i+offset].d1val,data[i+offset].d2val);
          else
            ((DBLDBL[])hDataObject)[i] = data[i+offset]; 
        }
        break;
      case TFormat.CF_FLOAT:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          ((float[])hDataObject)[i] = (float)data[i+offset].d1val;
        }
        break;
      case TFormat.CF_DOUBLE:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          ((double[])hDataObject)[i] = data[i+offset].d1val;
        }
        break;
      case TFormat.CF_LONG:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          ((int[])hDataObject)[i] = (int)data[i+offset].d2val;
        }
        break;      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public int putData(String data)
  {
    int len;
    if (data == null || (len = data.length()) < 1) return 0;
    switch (dFormat)
    {
      case TFormat.CF_NAME8:
        ((NAME8[])hDataObject)[0] = new NAME8(data);
        break;
      case TFormat.CF_NAME16:
        ((NAME16[])hDataObject)[0] = new NAME16(data);
        break;
      case TFormat.CF_NAME32:
        ((NAME32[])hDataObject)[0] = new NAME32(data);
        break;
      case TFormat.CF_NAME48:
        ((NAME48[])hDataObject)[0] = new NAME48(data);
        break;
      case TFormat.CF_NAME64:
        ((NAME64[])hDataObject)[0] = new NAME64(data);
        break;
      case TFormat.CF_XML:
      case TFormat.CF_TEXT:
        if (len < dArrayLength) dArrayLength = len;
        hDataObject = data.toCharArray();
        break;
      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public int putData(TTaggedStructure data)
  {
    if (data == null) return TErrorList.argument_list_error;
    if (dFormat != TFormat.CF_STRUCT) return TErrorList.illegal_format;
    if (dTag.compareTo(data.getName()) != 0) return TErrorList.invalid_structure_tag;
    if (hDataObject == null) return TErrorList.invalid_data;
    data.toByteArray();
    System.arraycopy(data.getByteArray(), 0, (byte[])hDataObject, 0, data.getSizeInBytes());
    if (hStructObject != null && hStructObject instanceof TTaggedStructure)
    { // this TDataType references another tagged structure !
      TTaggedStructure tts = (TTaggedStructure)hStructObject;
      tts.toStruct(data.getByteArray());
    }
    else if (hStructObject instanceof TTaggedStructure[])
    { // this TDataType references an array of tagged structures !
      TTaggedStructure tts[] = (TTaggedStructure[])hStructObject;
      tts[0].toStruct(data.getByteArray());
    }
    repackageDataForOutput();
    return 0;
  }
  public int putData(TTaggedStructure[] data)
  {
    return putData(data,data != null ? data.length : 0,0);
  }
  public int putData(TTaggedStructure[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dFormat != TFormat.CF_STRUCT) return TErrorList.illegal_format;
    if (dTag.compareTo(data[0].getName()) != 0) return TErrorList.invalid_structure_tag;
    if (hDataObject == null) return TErrorList.invalid_data;
    int tsiz = data[0].getSizeInBytes();
    int dsiz = getArrayLength();
    if (dsiz > length) dsiz = length;

    for (int i=0; i<dsiz; i++)
    {
      data[i+offset].toByteArray();
      System.arraycopy(data[i+offset].getByteArray(), 0, (byte[])hDataObject, i*tsiz, tsiz);
    }
    if (hStructObject != null && hStructObject instanceof TTaggedStructure)
    {
      TTaggedStructure tts = (TTaggedStructure)hStructObject;
      tts.toStruct(data[offset].getByteArray());
    }
    else if (hStructObject instanceof TTaggedStructure[])
    {
      TTaggedStructure tts[] = (TTaggedStructure[])hStructObject;
      for (int i = 0; i < dsiz; i++)
      {
        tts[i].toStruct(data[i+offset].getByteArray());
      }
    }
    setArrayLength(dsiz);
    repackageDataForOutput();
    return 0;
  }
  public int putData(byte[] data)
  {
    return putData(data,data != null ? data.length : 0, 0);
  }
  public int putData(byte[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dArrayLength > length) dArrayLength = length;
    switch (dFormat)
    {
      case TFormat.CF_BYTE:
        System.arraycopy(data, offset, (byte[])hDataObject, 0, dArrayLength);
        break;
      case TFormat.CF_SHORT:
        for (int i = 0; i < dArrayLength; i++)
          ((short[])hDataObject)[i] = (short) data[i+offset];
        break;
      case TFormat.CF_LONG:
        for (int i = 0; i < dArrayLength; i++)
          ((int[])hDataObject)[i] = (int) data[i+offset];
        break;
      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public int putData(byte[] data, String tag)
  {
    if (hDataObject != null)
    {
      if (tag != null && TStructRegistry.getSizeInBytes(tag) < 1) return TErrorList.invalid_structure_tag;
      if (data.length < dArrayLength) dArrayLength = data.length;
      System.arraycopy(data, 0, (byte[])hDataObject, 0, dArrayLength);
    }
    else if (hStructObject != null && hStructObject instanceof TTaggedStructure)
    {
      TTaggedStructure tts = (TTaggedStructure)hStructObject;
      if (data.length != tts.getSizeInBytes()) return TErrorList.invalid_structure_tag;
      System.arraycopy(data, 0, tts.getByteArray(), 0, data.length);
    }
    else if (hStructObject instanceof TTaggedStructure[])
    {
      TTaggedStructure tts[] = (TTaggedStructure[])hStructObject;
      int dsiz = tts[0].getSizeInBytes();
      for (int i = 0; i < tts.length; i++)
      {
        System.arraycopy(data, i * dsiz, tts[i].getByteArray(), 0, dsiz);
      }
    }
    else
    {
      return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public int putData(Object data)
  {
    if (dFormat != getObjectFormat(data)) return TErrorList.illegal_format;
    int len = getObjectSize(data);
    if (dFormat != TFormat.CF_TEXT &&
        dFormat != TFormat.CF_XML &&
        len < dArrayLength) dArrayLength = len;
    setObjectHandle(data);
    repackageDataForOutput();
    return 0;
  }
  public int putData(TCompoundDataObject data)
  {
    TCompoundDataObject[] d = new TCompoundDataObject[1];
    d[0] = data;
    return putData(d);    
  }  
  public int putData(TCompoundDataObject[] data)
  {
    return putData(data,data != null ? data.length : 0,0);
  }
  public int putData(TCompoundDataObject[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (dArrayLength > length) dArrayLength = length;
    if (length <= 0) return TErrorList.argument_list_error;
    boolean formatsEqual = dFormat == getObjectFormat(data);

    if ((data instanceof NAME[] && hDataObject instanceof NAME[]) ||
        (data instanceof NAMEINT[] && hDataObject instanceof NAMEINT[]))
    {
      if (formatsEqual)
      {
        for (int i = 0; i < dArrayLength; i++) 
          ((TCompoundDataObject[])hDataObject)[i] = data[i+offset];
      }
      else
      {
        try
        {
          Class[] params = new Class[1];
          params[0] = String.class; 
          Object[] objs = new Object[1];
          for (int i = 0; i < dArrayLength; i++)
          {
            objs[0] = (data[i+offset].toString());
            ((TCompoundDataObject[])hDataObject)[i] = 
              (TCompoundDataObject)data.getClass().getComponentType().getConstructor(params).newInstance(objs); 
          }
        }
        catch (Exception e)
        {
        }
      }
    }
    else if (data instanceof NAME[] && hDataObject instanceof char[])
    {
      hDataObject = data[offset].toString().toCharArray();
    }
    else
    {
      if (!formatsEqual) return TErrorList.illegal_format;
      setObjectHandle(data);
    }
    repackageDataForOutput();
    return 0;
  }
  public int getData(double[] data)
  {
    return getData(data,data != null ? data.length : 0, 0); 
  }
  public int getData(double[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    int len = dArrayLength;  // length of the buffered array
    if (len > length) len = length; // can't get more than is there !
    try
    {
      int i, n, m, fmtsiz = TFormat.formatSizeOf(dFormat);
      ByteArrayInputStream bs;
      DataInputStream ds;
      for (i = 0, n = 0, m = 0; n < numblks; n++, m = i)
      {
        if (dBuffers[n] == null) break;
        byte[] b = dBuffers[n].toByteArray();
        Swap.Bytes(b, fmtsiz);
        bs = new ByteArrayInputStream(b);
        ds = new DataInputStream(bs);
        switch (dFormat)
        {
          case TFormat.CF_DOUBLE:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = ds.readDouble();
            break;
          case TFormat.CF_FLOAT:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = ds.readFloat();
            break;
          case TFormat.CF_LONG:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = ds.readInt();
            break;
          case TFormat.CF_SHORT:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = ds.readShort();
            break;
          default:
            ds.close();
            bs.close();
            return TErrorList.illegal_format;
        }
        ds.close();
        bs.close();
      }
      dCompletionLength = i;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    return 0;
  }
  public int putData(double data)
  {
    double[] d = new double[1];
    d[0] = data;
    return putData(d);
  }
  public int putData(double[] data)
  {
    return putData(data,data != null ? data.length : 0,0);
  }
  public int putData(double[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dArrayLength > length) dArrayLength = length;
    switch (dFormat)
    {
      case TFormat.CF_DOUBLE:
        for (int i = 0; i < dArrayLength; i++)
          ((double[])hDataObject)[i] = (double) data[i+offset];
        break;
      case TFormat.CF_FLOAT:
        for (int i = 0; i < dArrayLength; i++)
          ((float[])hDataObject)[i] = (float) data[i+offset];
        break;
      case TFormat.CF_SHORT:
        for (int i = 0; i < dArrayLength; i++)
          ((short[])hDataObject)[i] = (short) data[i+offset];
        break;
      case TFormat.CF_LONG:
        for (int i = 0; i < dArrayLength; i++)
          ((int[])hDataObject)[i] = (int) data[i+offset];
        break;
      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public byte[] getDataBuffer()
  {
    try
    {
      if (dBuffer == null)
        dBuffer = new ByteArrayOutputStream(dSizeInBytes);
      else
        dBuffer.reset();
      DataOutputStream dos = new DataOutputStream(dBuffer);
      for (int n = 0; n < numblks; n++)
      {
        if (dBuffers[n] == null) break;
        dos.write(dBuffers[n].toByteArray());
      }
      dos.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    return dBuffer.toByteArray();
  }
  public byte[] getDataBuffer(int n)
  {
    try
    {
      if (dBuffer == null)
        dBuffer = new ByteArrayOutputStream();
      else
        dBuffer.reset();
      DataOutputStream dos = new DataOutputStream(dBuffer);
      if (n < numblks && dBuffers[n] != null)
      {
        dos.write(dBuffers[n].toByteArray());
      }
      dos.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    return dBuffer.toByteArray();
  }
  /**
   * @brief Returns the data array length of the associated tine Data Type
   *        object
   * @return Returns the data array length of the associated tine Data Type
   *         object
   */
  public int getArrayLength()
  {
    int dlen = dArrayLength;
    if (dFormat == TFormat.CF_STRUCT) 
    { // if tagged structure, then this call returns the number of structures, not bytes !
      int ssiz = TStructRegistry.getSizeInBytes(dTag);
      if (ssiz < 1) ssiz = 1;
      dlen /= ssiz;
    }
    return dlen;
  }
  /**
   * @brief Returns the data type of the associated tine Data Type object
   * @return Returns the data type of the associated tine Data Type object
   */
  public short getFormat()
  {
    return dFormat >= 0 ? dFormat : (short) (256 + dFormat);
  }
  public void removeTagFromLink()
  {
    removeTagAtLinkTime = true;
  }
  public String getTag()
  {
    return removeTagAtLinkTime ? "" : dTag;
  }
  public int getSizeInBytes()
  {
    return dSizeInBytes;
  }
  public int getDataSize()
  {
    int fmtsiz = 0;
    if ((fmtsiz = TFormat.formatSizeOf(dFormat)) > 0)
    {
      int datasize = fmtsiz * dArrayLength;
      if (dFormat == TFormat.CF_SPECTRUM) datasize += SPECTRUM.hdrSizeInBytes;
      return datasize;
    }
    return 0;
  }
  private void makeNewCompoundDataObject(TCompoundDataObject[] Data, int index)
  {
    try
    {
      Data[index] = (TCompoundDataObject)Data.getClass().getComponentType().newInstance();
    }
    catch (Exception e)
    {
      e.printStackTrace();
    }
  }
  private void setObjectContents(DataOutputStream ds, Object Data, int index)
  {
    try
    {
      if (Data instanceof byte[])
      {
        ds.writeByte(((byte[]) Data)[index]);
      }
      else if (Data instanceof short[])
      {
        ds.writeShort(Swap.Short(((short[]) Data)[index]));
      }
      else if (Data instanceof int[])
      {
        ds.writeInt(Swap.Long(((int[]) Data)[index]));
      }
      else if (Data instanceof float[])
      {
        ds.write(Swap.Float(((float[]) Data)[index]));
      }
      else if (Data instanceof double[])
      {
        ds.write(Swap.Double(((double[]) Data)[index]));
      }
      else if (Data instanceof char[])
      {
        ds.writeBytes(new String((char[])Data));
      }
      else if (Data instanceof StringBuffer)
      {
        ds.writeBytes(((StringBuffer)Data).toString());
      }
      else if (Data instanceof String)
      {
        ds.write(((String) Data).getBytes());
      }
      else if (Data instanceof TCompoundDataObject[])
      {
        if (((TCompoundDataObject[])Data)[index] == null) 
          makeNewCompoundDataObject((TCompoundDataObject[])Data,index); 
        ds.write(((TCompoundDataObject[])Data)[index].toByteArray());
      }
      else if (Data instanceof SPECTRUM)
      {
        ds.write(((SPECTRUM)Data).toByteArray());
      }
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  private void setObjectHandle(TCompoundDataObject[] Data)
  {
    hDataObject = Data;
  }  
  private void setObjectHandle(Object Data)
  {
    hDataObject = Data;    
  }
  public static int getObjectSize(Object Data)
  {
    if (Data instanceof TCompoundDataObject[])
      return ((TCompoundDataObject[]) Data).length;
    else if (Data instanceof byte[])
      return ((byte[]) Data).length;
    else if (Data instanceof short[])
      return ((short[]) Data).length;
    else if (Data instanceof int[])
      return ((int[]) Data).length;
    else if (Data instanceof float[])
      return ((float[]) Data).length;
    else if (Data instanceof double[])
      return ((double[]) Data).length;
    else if (Data instanceof SPECTRUM)
      return ((SPECTRUM) Data).hByteArray.length;
    else if (Data instanceof char[])
      return ((char[]) Data).length;
    else if (Data instanceof String)
      return ((String) Data).length();
    else if (Data instanceof StringBuffer) 
      return ((StringBuffer) Data).toString().length();
    return 0;
  }
  public static short getObjectFormat(Object Data)
  {
    if (Data instanceof TCompoundDataObject[])
    {
      try
      {
        Object tdo = ((TCompoundDataObject[])Data)[0] != null ? 
            ((TCompoundDataObject[])Data)[0] : Data.getClass().getComponentType().newInstance();
        return ((TCompoundDataObject) tdo).getFormat();
      }
      catch (Exception e)
      {
      }
    }
    if (Data instanceof byte[])
      return TFormat.CF_BYTE;
    else if (Data instanceof short[])
      return TFormat.CF_SHORT;
    else if (Data instanceof int[])
      return TFormat.CF_LONG;
    else if (Data instanceof float[])
      return TFormat.CF_FLOAT;
    else if (Data instanceof double[])
      return TFormat.CF_DOUBLE;
    else if (Data instanceof String)
      return TFormat.CF_TEXT;
    else if (Data instanceof StringBuffer)
      return TFormat.CF_TEXT;
    else if (Data instanceof char[]) 
      return TFormat.CF_TEXT;
    return TFormat.CF_NULL;
  }
  public int getData(Object Data)
  {
    if (Data instanceof byte[])
      getData((byte[]) Data);
    else if (Data instanceof short[])
      getData((short[]) Data);
    else if (Data instanceof int[])
      getData((int[]) Data);
    else if (Data instanceof float[])
      getData((float[]) Data);
    else if (Data instanceof double[])
      getData((double[]) Data);
    else if (Data instanceof char[])
      getData((char[]) Data);
    else if (Data instanceof StringBuffer)
      getData((StringBuffer) Data);
    else if (Data instanceof TCompoundDataObject[])
      getData((TCompoundDataObject[]) Data);
    else if (Data instanceof SPECTRUM) 
      getData((SPECTRUM) Data);
    return 0;
  }
  public int dataFill(String fromStr)
  {
    int n = 0;
    switch (dFormat)
    {
      case TFormat.CF_NULL:
        break;
      case TFormat.CF_NAME16FI:
        for (n = 0; n < dArrayLength && ((NAME16FI[])hDataObject)[n] != null; n++)
          ((NAME16FI[])hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_NAME16II:
        for (n = 0; n < dArrayLength && ((NAME16II[])hDataObject)[n] != null; n++)
          ((NAME16II[])hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_NAME16DBLDBL:
        for (n = 0; n < dArrayLength && ((NAME16DBLDBL[])hDataObject)[n] != null; n++)
          ((NAME16DBLDBL[])hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_NAME32DBLDBL:
        for (n = 0; n < dArrayLength && ((NAME32DBLDBL[])hDataObject)[n] != null; n++)
          ((NAME32DBLDBL[])hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_NAME64DBLDBL:
        for (n = 0; n < dArrayLength && ((NAME64DBLDBL[])hDataObject)[n] != null; n++)
          ((NAME64DBLDBL[])hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_NAME8:
      case TFormat.CF_NAME16:
      case TFormat.CF_NAME32:
      case TFormat.CF_NAME48:
      case TFormat.CF_NAME64:
        for (n = 0; n < dArrayLength && ((TCompoundDataObject[])hDataObject)[n] != null; n++)
          ((NAME[])hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_NAME16I:
      case TFormat.CF_NAME32I:
      case TFormat.CF_NAME48I:
      case TFormat.CF_NAME64I:
        for (n = 0; n < dArrayLength && ((TCompoundDataObject[])hDataObject)[n] != null; n++)
          ((NAMEINT[])hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_USTRING:
        for (n = 0; n < dArrayLength && ((USTRING[])hDataObject)[n] != null; n++)
          ((USTRING[])hDataObject)[n].setString(fromStr);
        break;
      case TFormat.CF_UNAME:
        for (n = 0; n < dArrayLength && ((UNAME[])hDataObject)[n] != null; n++)
          ((UNAME[])hDataObject)[n].setString(fromStr);
        break;
      case TFormat.CF_XML:
      case TFormat.CF_TEXT:
        int size = dArrayLength < fromStr.length() ? dArrayLength : fromStr.length();
        char[] s = fromStr.toCharArray();
        for (n = 0; n < size && hDataObject != null; n++)
          ((char[])hDataObject)[n] = s[n];
        break;
    }
    return n;
  }
  public int dataFill(double fromVal)
  {
    return dataFill(fromVal, 0);
  }
  public int dataFill(double fromVal, int offset)
  {
    int n = 0, len;
    if (offset >= dArrayLength) return 0;
    switch (dFormat)
    {
      case TFormat.CF_NULL:
        break;
      case TFormat.CF_NAME16FI:
        if ((len = ((NAME16FI[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((NAME16FI[])hDataObject)[n] != null; n++)
          ((NAME16FI[])hDataObject)[n].fval = (float) fromVal;
        break;
      case TFormat.CF_NAME16II:
        if ((len = ((NAME16II[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((NAME16II[])hDataObject)[n] != null; n++)
          ((NAME16II[])hDataObject)[n].i1val = (int) fromVal;
        break;
      case TFormat.CF_NAME16DBLDBL:
        if ((len = ((NAME16DBLDBL[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((NAME16DBLDBL[])hDataObject)[n] != null; n++)
          ((NAME16DBLDBL[])hDataObject)[n].d1val = fromVal;
        break;
      case TFormat.CF_NAME32DBLDBL:
        if ((len = ((NAME32DBLDBL[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((NAME32DBLDBL[])hDataObject)[n] != null; n++)
          ((NAME32DBLDBL[])hDataObject)[n].d1val = fromVal;
        break;
      case TFormat.CF_NAME64DBLDBL:
        if ((len = ((NAME64DBLDBL[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((NAME64DBLDBL[])hDataObject)[n] != null; n++)
          ((NAME64DBLDBL[])hDataObject)[n].d1val = fromVal;
        break;
      case TFormat.CF_STRUCT:
      case TFormat.CF_BYTE:
        if ((len = ((byte[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((byte[])hDataObject) != null; n++)
          ((byte[])hDataObject)[n] = (byte) fromVal;
        break;
      case TFormat.CF_SHORT:
        if ((len = ((short[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((short[])hDataObject) != null; n++)
          ((short[])hDataObject)[n] = (short) fromVal;
        break;
      case TFormat.CF_LONG:
        if ((len = ((int[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((int[])hDataObject) != null; n++)
          ((int[])hDataObject)[n] = (int) fromVal;
        break;
      case TFormat.CF_FLOAT:
        if ((len = ((float[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((float[])hDataObject) != null; n++)
          ((float[])hDataObject)[n] = (float) fromVal;
        break;
      case TFormat.CF_DOUBLE:
        if ((len = ((double[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((double[])hDataObject) != null; n++)
          ((double[])hDataObject)[n] = fromVal;
        break;
      case TFormat.CF_LNGINT:
        if ((len = ((INTINT[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((INTINT[])hDataObject)[n] != null; n++)
          ((INTINT[])hDataObject)[n].i1val = (int) fromVal;
        break;
      case TFormat.CF_FLTINT:
        if ((len = ((FLTINT[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((FLTINT[])hDataObject)[n] != null; n++)
          ((FLTINT[])hDataObject)[n].fval = (float) fromVal;
        break;
      case TFormat.CF_FLTFLT:
        if ((len = ((FLTFLT[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((FLTFLT[])hDataObject)[n] != null; n++)
          ((FLTFLT[])hDataObject)[n].f1val = (float) fromVal;
        break;
      case TFormat.CF_DBLDBL:
        if ((len = ((DBLDBL[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((DBLDBL[])hDataObject)[n] != null; n++)
          ((DBLDBL[])hDataObject)[n].d1val = fromVal;
        break;
      case TFormat.CF_INTFLTINT:
        if ((len = ((INTFLTINT[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((INTFLTINT[])hDataObject)[n] != null; n++)
          ((INTFLTINT[])hDataObject)[n].fval = (float) fromVal;
        break;
      case TFormat.CF_FLTFLTINT:
        if ((len = ((FLTFLTINT[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((FLTFLTINT[])hDataObject)[n] != null; n++)
          ((FLTFLTINT[])hDataObject)[n].f1val = (float) fromVal;
        break;
      case TFormat.CF_FLTINTINT:
        if ((len = ((FLTINTINT[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((FLTINTINT[])hDataObject)[n] != null; n++)
          ((FLTINTINT[])hDataObject)[n].f1val = (float) fromVal;
        break;
      case TFormat.CF_INTINTINT:
        if ((len = ((INTINTINT[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((INTINTINT[])hDataObject)[n] != null; n++)
          ((INTINTINT[])hDataObject)[n].i1val = (int) fromVal;
        break;
      case TFormat.CF_ADDRESS:
        if ((len = ((INTINTINTINT[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((INTINTINTINT[])hDataObject)[n] != null; n++)
          ((INTINTINTINT[])hDataObject)[n].i1val = (int) fromVal;
        break;
      case TFormat.CF_FILTER:
        if ((len = ((INTFLTFLTFLT[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((INTFLTFLTFLT[])hDataObject)[n] != null; n++)
          ((INTFLTFLTFLT[])hDataObject)[n].f1val = (float) fromVal;
        break;
      case TFormat.CF_FWINDOW:
        if ((len = ((INTINTFLTFLT[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((INTINTFLTFLT[])hDataObject)[n] != null; n++)
          ((INTINTFLTFLT[])hDataObject)[n].f1val = (float) fromVal;
        break;
      case TFormat.CF_USTRING:
        if ((len = ((USTRING[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((USTRING[])hDataObject)[n] != null; n++)
          ((USTRING[])hDataObject)[n].f1val = (float) fromVal;
        break;
      case TFormat.CF_UNAME:
        if ((len = ((UNAME[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((UNAME[])hDataObject)[n] != null; n++)
          ((UNAME[])hDataObject)[n].f1val = (float) fromVal;
        break;
      case TFormat.CF_SPECTRUM:
        for (n = offset; n < dArrayLength && ((SPECTRUM)hDataObject) != null; n++)
          ((SPECTRUM)hDataObject).d_spect_array[n] = (float) fromVal;
        break;
    }
    return n - offset;
  }
  public int dataCopy(TDataType fromData)
  {
    return dataCopy(fromData,0);
  }
  public int dataCopy(TDataType fromData,int offset)
  {
    if (dFormat != fromData.dFormat) return 0;
    int n = 0;
    int size = dArrayLength < fromData.dArrayLength ? dArrayLength : fromData.dArrayLength;

    // copy the innards ...
    copy(fromData);
    
    // copy into the referenced object ...
    if (hDataObject instanceof TCompoundDataObject[])
    {
      System.arraycopy((TCompoundDataObject[])fromData.hDataObject,offset,(TCompoundDataObject[])hDataObject,0,size);
      n = size;
    }
    // TODO: catch SPECTRUM and any other special cases ?
    // TODO: LNGINT and INTINT both have the same format (also TDS and INTFLTINT)
    switch (dFormat)
    {
      case TFormat.CF_NULL:
        break;
      case TFormat.CF_STRUCT:
      case TFormat.CF_BYTE:
        System.arraycopy((byte[])fromData.hDataObject, offset, (byte[])hDataObject, 0, size);
        n = size;
        break;
      case TFormat.CF_SHORT:
        System.arraycopy((short[])fromData.hDataObject, offset, (short[])hDataObject, 0, size);
        n = size;
        break;
      case TFormat.CF_LONG:
        System.arraycopy((int[])fromData.hDataObject, offset, (int[])hDataObject, 0, size);
        n = size;
        break;
      case TFormat.CF_FLOAT:
        System.arraycopy((float[])fromData.hDataObject, offset, (float[])hDataObject, 0, size);
        n = size;
        break;
      case TFormat.CF_DOUBLE:
        System.arraycopy((double[])fromData.hDataObject, offset, (double[])hDataObject, 0, size);
        n = size;
        break;
      case TFormat.CF_SPECTRUM:
        System.arraycopy(((SPECTRUM)fromData.hDataObject).d_spect_array, offset, ((SPECTRUM)hDataObject).d_spect_array, 0, size);
        n = size;
        break;
    }
    return n;
  }
public Object getStructObject() { return hStructObject; }
  /**
   * @brief Retrieves the returned data as an Object container. This method is
   *        useful for obtaining the unbound data when the data type is unkown
   *        at link time. A call to getFormat() can be made to ascertain the
   *        tine data type and hence the array nature of the object returned.
   */
  public Object getDataObject() { return hDataObject; }
  /**
   * @brief Retrieves the returned data as a Vector. This method is useful for
   *        obtaining the unbound data when the data type is unkown at link
   *        time. A call to getFormat() can be made to ascertain the tine data
   *        type and hence the contents of the Vector.
   */
  public Vector getVector()
  {
    // TODO: clean up the individual type allocations and follow the NAME8 example
    // We still have to have the long switch on the format so that we know 
    // what we have to create!
    int i = 0;
    if (dVector == null) dVector = new Vector();
    Vector v = dVector;
    v.clear();
    switch (dFormat)
    {
      case TFormat.CF_NULL:
        break;
      case TFormat.CF_NAME8:
        if (hDataObject == null)
        {
          hDataObject = new NAME8[dArrayLength];
          getData(hDataObject);
        }
        for (i = 0; i < ((TCompoundDataObject[])hDataObject).length; i++)
          v.addElement(((TCompoundDataObject[])hDataObject)[i]);
        break;
      case TFormat.CF_NAME16:
        NAME16 n16data[] = (NAME16[])hDataObject;
        if (n16data == null)
        {
          n16data = new NAME16[dArrayLength];
          getData(n16data);
        }
        for (i = 0; i < n16data.length; i++)
          v.addElement(n16data[i]);
        break;
      case TFormat.CF_NAME32:
        NAME32 n32data[] = (NAME32[])hDataObject;
        if (n32data == null)
        {
          n32data = new NAME32[dArrayLength];
          getData(n32data);
        }
        for (i = 0; i < n32data.length; i++)
          v.addElement(n32data[i]);
        break;
      case TFormat.CF_NAME48:
        NAME48 n48data[] = (NAME48[])hDataObject;
        if (n48data == null)
        {
          n48data = new NAME48[dArrayLength];
          getData(n48data);
        }
        for (i = 0; i < n48data.length; i++)
          v.addElement(n48data[i]);
        break;
      case TFormat.CF_NAME64:
        NAME64 n64data[] = (NAME64[])hDataObject;
        if (n64data == null)
        {
          n64data = new NAME64[dArrayLength];
          getData(n64data);
        }
        for (i = 0; i < n64data.length; i++)
          v.addElement(n64data[i]);
        break;
      case TFormat.CF_NAME16I:
        NAME16I n16idata[] = (NAME16I[])hDataObject;
        if (n16idata == null)
        {
          n16idata = new NAME16I[dArrayLength];
          getData(n16idata);
        }
        for (i = 0; i < n16idata.length; i++)
          v.addElement(n16idata[i]);
        break;
      case TFormat.CF_NAME16FI:
        NAME16FI n16fidata[] = (NAME16FI[])hDataObject;
        if (n16fidata == null)
        {
          n16fidata = new NAME16FI[dArrayLength];
          getData(n16fidata);
        }
        for (i = 0; i < n16fidata.length; i++)
          v.addElement(n16fidata[i]);
        break;
      case TFormat.CF_NAME16II:
        NAME16II n16iidata[] = (NAME16II[])hDataObject;
        if (n16iidata == null)
        {
          n16iidata = new NAME16II[dArrayLength];
          getData(n16iidata);
        }
        for (i = 0; i < n16iidata.length; i++)
          v.addElement(n16iidata[i]);
        break;
      case TFormat.CF_NAME16DBLDBL:
        NAME16DBLDBL n16dddata[] = (NAME16DBLDBL[])hDataObject;
        if (n16dddata == null)
        {
          n16dddata = new NAME16DBLDBL[dArrayLength];
          getData(n16dddata);
        }
        for (i = 0; i < n16dddata.length; i++)
          v.addElement(n16dddata[i]);
        break;
      case TFormat.CF_NAME32DBLDBL:
        NAME32DBLDBL n32dddata[] = (NAME32DBLDBL[])hDataObject;
        if (n32dddata == null)
        {
          n32dddata = new NAME32DBLDBL[dArrayLength];
          getData(n32dddata);
        }
        for (i = 0; i < n32dddata.length; i++)
          v.addElement(n32dddata[i]);
        break;
      case TFormat.CF_NAME64DBLDBL:
        NAME64DBLDBL n64dddata[] = (NAME64DBLDBL[])hDataObject;
        if (n64dddata == null)
        {
          n64dddata = new NAME64DBLDBL[dArrayLength];
          getData(n64dddata);
        }
        for (i = 0; i < n64dddata.length; i++)
          v.addElement(n64dddata[i]);
        break;
      case TFormat.CF_NAME32I:
        NAME32I n32idata[] = (NAME32I[])hDataObject;
        if (n32idata == null)
        {
          n32idata = new NAME32I[dArrayLength];
          getData(n32idata);
        }
        for (i = 0; i < n32idata.length; i++)
          v.addElement(n32idata[i]);
        break;
      case TFormat.CF_NAME48I:
        NAME48I n48idata[] = (NAME48I[])hDataObject;
        if (n48idata == null)
        {
          n48idata = new NAME48I[dArrayLength];
          getData(n48idata);
        }
        for (i = 0; i < n48idata.length; i++)
          v.addElement(n48idata[i]);
        break;
      case TFormat.CF_NAME64I:
        NAME64I n64idata[] = (NAME64I[])hDataObject;
        if (n64idata == null)
        {
          n64idata = new NAME64I[dArrayLength];
          getData(n64idata);
        }
        for (i = 0; i < n64idata.length; i++)
          v.addElement(n64idata[i]);
        break;
      case TFormat.CF_BYTE:
        byte bdata[] = (byte[])hDataObject;
        if (bdata == null)
        {
          bdata = new byte[dArrayLength];
          getData(bdata);
        }
        for (i = 0; i < bdata.length; i++)
          v.addElement(new Byte(bdata[i]));
        break;
      case TFormat.CF_STRUCT:
        break;
      case TFormat.CF_SHORT:
        short sdata[] = (short[])hDataObject;
        if (sdata == null)
        {
          sdata = new short[dArrayLength];
          getData(sdata);
        }
        for (i = 0; i < sdata.length; i++)
          v.addElement(new Short(sdata[i]));
        break;
      case TFormat.CF_LONG:
        int idata[] = (int[])hDataObject;
        if (idata == null)
        {
          idata = new int[dArrayLength];
          getData(idata);
        }
        for (i = 0; i < idata.length; i++)
          v.addElement(new Integer(idata[i]));
        break;
      case TFormat.CF_FLOAT:
        float fdata[] = (float[])hDataObject;
        if (fdata == null)
        {
          fdata = new float[dArrayLength];
          getData(fdata);
        }
        for (i = 0; i < fdata.length; i++)
          v.addElement(new Float(fdata[i]));
        break;
      case TFormat.CF_DOUBLE:
        double ddata[] = new double[dArrayLength];
        getData(ddata);
        for (i = 0; i < ddata.length; i++)
          v.addElement(new Double(ddata[i]));
        break;
      case TFormat.CF_LNGINT:
        INTINT lidata[] = (INTINT[])hDataObject;
        if (lidata == null)
        {
          lidata = new INTINT[dArrayLength];
          getData(lidata);
        }
        for (i = 0; i < lidata.length; i++)
          v.addElement(lidata[i]);
        break;
      case TFormat.CF_FLTINT:
        FLTINT fidata[] = (FLTINT[])hDataObject;
        if (fidata == null)
        {
          fidata = new FLTINT[dArrayLength];
          getData(fidata);
        }
        for (i = 0; i < fidata.length; i++)
          v.addElement(fidata[i]);
        break;
      case TFormat.CF_FLTFLT:
        FLTFLT ffdata[] = (FLTFLT[])hDataObject;
        if (ffdata == null)
        {
          ffdata = new FLTFLT[dArrayLength];
          getData(ffdata);
        }
        for (i = 0; i < ffdata.length; i++)
          v.addElement(ffdata[i]);
        break;
      case TFormat.CF_DBLINT:
        DBLINT didata[] = (DBLINT[])hDataObject;
        if (didata == null)
        {
          didata = new DBLINT[dArrayLength];
          getData(didata);
        }
        for (i = 0; i < didata.length; i++)
          v.addElement(didata[i]);
        break;
      case TFormat.CF_DBLDBL:
        DBLDBL dddata[] = (DBLDBL[])hDataObject;
        if (dddata == null)
        {
          dddata = new DBLDBL[dArrayLength];
          getData(dddata);
        }
        for (i = 0; i < dddata.length; i++)
          v.addElement(dddata[i]);
        break;
      case TFormat.CF_INTFLTINT:
        INTFLTINT ifidata[] = (INTFLTINT[])hDataObject;
        if (ifidata == null)
        {
          ifidata = new INTFLTINT[dArrayLength];
          getData(ifidata);
        }
        for (i = 0; i < ifidata.length; i++)
          v.addElement(ifidata[i]);
        break;
      case TFormat.CF_FLTFLTINT:
        FLTFLTINT ffidata[] = (FLTFLTINT[])hDataObject;
        if (ffidata == null)
        {
          ffidata = new FLTFLTINT[dArrayLength];
          getData(ffidata);
        }
        for (i = 0; i < ffidata.length; i++)
          v.addElement(ffidata[i]);
        break;
      case TFormat.CF_FLTINTINT:
        FLTINTINT fiidata[] = (FLTINTINT[])hDataObject;
        if (fiidata == null)
        {
          fiidata = new FLTINTINT[dArrayLength];
          getData(fiidata);
        }
        for (i = 0; i < fiidata.length; i++)
          v.addElement(fiidata[i]);
        break;
      case TFormat.CF_INTINTINT:
        INTINTINT iiidata[] = (INTINTINT[])hDataObject;
        if (iiidata == null)
        {
          iiidata = new INTINTINT[dArrayLength];
          getData(iiidata);
        }
        for (i = 0; i < iiidata.length; i++)
          v.addElement(iiidata[i]);
        break;
      case TFormat.CF_ADDRESS:
        INTINTINTINT iiiidata[] = (INTINTINTINT[])hDataObject;
        if (iiiidata == null)
        {
          iiiidata = new INTINTINTINT[dArrayLength];
          getData(iiiidata);
        }
        for (i = 0; i < iiiidata.length; i++)
          v.addElement(iiiidata[i]);
        break;
      case TFormat.CF_FILTER:
        INTFLTFLTFLT ifffdata[] = (INTFLTFLTFLT[])hDataObject;
        if (ifffdata == null)
        {
          ifffdata = new INTFLTFLTFLT[dArrayLength];
          getData(ifffdata);
        }
        for (i = 0; i < ifffdata.length; i++)
          v.addElement(ifffdata[i]);
        break;
      case TFormat.CF_FWINDOW:
        INTINTFLTFLT iiffdata[] = (INTINTFLTFLT[])hDataObject;
        if (iiffdata == null)
        {
          iiffdata = new INTINTFLTFLT[dArrayLength];
          getData(iiffdata);
        }
        for (i = 0; i < iiffdata.length; i++)
          v.addElement(iiffdata[i]);
        break;
      case TFormat.CF_USTRING:
        USTRING usdata[] = (USTRING[])hDataObject;;
        if (usdata == null)
        {
          usdata = new USTRING[dArrayLength];
          getData(usdata);
        }
        for (i = 0; i < usdata.length; i++)
          v.addElement(usdata[i]);
        break;
      case TFormat.CF_UNAME:
        UNAME undata[] = (UNAME[])hDataObject;;
        if (undata == null)
        {
          undata = new UNAME[dArrayLength];
          getData(undata);
        }
        for (i = 0; i < undata.length; i++)
          v.addElement(undata[i]);
        break;
    }
    dCompletionLength = i;
    return v;
  }
  /**
   * @brief Outputs the returned data as a String. This method is useful for
   *        displaying the unbound data when the data type is unkown at link
   *        time.
   */
  public String toString()
  {
    StringBuffer sb = new StringBuffer();
    // TODO: clean up
    switch (dFormat)
    {
      case TFormat.CF_NULL:
        return new String("");
      case TFormat.CF_XML:
      case TFormat.CF_TEXT:
        if (hDataObject instanceof char[]) 
          return new String((char[])hDataObject);
        if (hDataObject instanceof StringBuffer) 
          return ((StringBuffer)hDataObject).toString();
        char[] t = new char[dArrayLength];
        getData(t);
        return new String(t);
      case TFormat.CF_USTRING:
        USTRING us[] = (USTRING[])hDataObject;
        if (us == null)
        {
          us = new USTRING[dArrayLength];
          getData(us);
        }
        for (int i = 0; i < dCompletionLength && i < us.length; i++)
        {
          if (us[i] == null) break;
          sb.append(us[i].ival);
          sb.append(" ");
          sb.append(us[i].f1val);
          sb.append(" ");
          sb.append(us[i].f2val);
          sb.append(" ");
          sb.append(us[i].tm);
          sb.append(" ");
          sb.append(us[i].str);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_UNAME:
        UNAME un[] = (UNAME[])hDataObject;
        if (un == null)
        {
          un = new UNAME[dArrayLength];
          getData(un);
        }
        for (int i = 0; i < dCompletionLength && i < un.length; i++)
        {
          if (un[i] == null) break;
          sb.append(un[i].ival);
          sb.append(" ");
          sb.append(un[i].f1val);
          sb.append(" ");
          sb.append(un[i].f2val);
          sb.append(" ");
          sb.append(un[i].f3val);
          sb.append(" ");
          sb.append(un[i].str);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_INTINTINT:
        INTINTINT iii[] = (INTINTINT[])hDataObject;
        if (iii == null)
        {
          iii = new INTINTINT[dArrayLength];
          getData(iii);
        }
        for (int i = 0; i < dCompletionLength && i < iii.length; i++)
        {
          if (iii[i] == null) break;
          sb.append(iii[i].i1val);
          sb.append(" ");
          sb.append(iii[i].i2val);
          sb.append(" ");
          sb.append(iii[i].i3val);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_ADDRESS:
        INTINTINTINT iiii[] = (INTINTINTINT[])hDataObject;
        if (iiii == null)
        {
          iiii = new INTINTINTINT[dArrayLength];
          getData(iiii);
        }
        for (int i = 0; i < dCompletionLength && i < iiii.length; i++)
        {
          if (iiii[i] == null) break;
          sb.append(iiii[i].i1val);
          sb.append(" ");
          sb.append(iiii[i].i2val);
          sb.append(" ");
          sb.append(iiii[i].i3val);
          sb.append(" ");
          sb.append(iiii[i].i4val);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_FILTER:
        INTFLTFLTFLT ifff[] = (INTFLTFLTFLT[])hDataObject;
        if (ifff == null)
        {
          ifff = new INTFLTFLTFLT[dArrayLength];
          getData(ifff);
        }
        for (int i = 0; i < dCompletionLength && i < ifff.length; i++)
        {
          if (ifff[i] == null) break;
          sb.append(ifff[i].i1val);
          sb.append(" ");
          sb.append(ifff[i].f1val);
          sb.append(" ");
          sb.append(ifff[i].f2val);
          sb.append(" ");
          sb.append(ifff[i].f3val);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_FWINDOW:
        INTINTFLTFLT iiff[] = (INTINTFLTFLT[])hDataObject;
        if (iiff == null)
        {
          iiff = new INTINTFLTFLT[dArrayLength];
          getData(iiff);
        }
        for (int i = 0; i < dCompletionLength && i < iiff.length; i++)
        {
          if (iiff[i] == null) break;
          sb.append(iiff[i].i1val);
          sb.append(" ");
          sb.append(iiff[i].i2val);
          sb.append(" ");
          sb.append(iiff[i].f1val);
          sb.append(" ");
          sb.append(iiff[i].f2val);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_INTFLTINT:
        INTFLTINT ifi[] = (INTFLTINT[])hDataObject;
        if (ifi == null)
        {
          ifi = new INTFLTINT[dArrayLength];
          getData(ifi);
        }
        for (int i = 0; i < dCompletionLength && i < ifi.length; i++)
        {
          if (ifi[i] == null) break;
          sb.append(ifi[i].i1val);
          sb.append(" ");
          sb.append(ifi[i].fval);
          sb.append(" ");
          sb.append(ifi[i].i2val);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_FLTFLTINT:
        FLTFLTINT ffi[] = (FLTFLTINT[])hDataObject;
        if (ffi == null)
        {
          ffi = new FLTFLTINT[dArrayLength];
          getData(ffi);
        }
        for (int i = 0; i < dCompletionLength && i < ffi.length; i++)
        {
          if (ffi[i] == null) break;
          sb.append(ffi[i].f1val);
          sb.append(" ");
          sb.append(ffi[i].f2val);
          sb.append(" ");
          sb.append(ffi[i].ival);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_FLTINTINT:
        FLTINTINT fii[] = (FLTINTINT[])hDataObject;
        if (fii == null)
        {
          fii = new FLTINTINT[dArrayLength];
          getData(fii);
        }
        for (int i = 0; i < dCompletionLength && i < fii.length; i++)
        {
          if (fii[i] == null) break;
          sb.append(fii[i].f1val);
          sb.append(" ");
          sb.append(fii[i].i1val);
          sb.append(" ");
          sb.append(fii[i].i2val);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_NAME8:
        NAME8 n8data[] = (NAME8[])hDataObject;
        if (n8data == null)
        {
          n8data = new NAME8[dArrayLength];
          getData(n8data);
        }
        for (int i = 0; i < dCompletionLength && i < n8data.length; i++)
        {
          if (n8data[i] == null) break;
          sb.append(n8data[i].name);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_NAME16:
        NAME16 n16data[] = (NAME16[])hDataObject;
        if (n16data == null)
        {
          n16data = new NAME16[dArrayLength];
          getData(n16data);
        }
        for (int i = 0; i < dCompletionLength && i < n16data.length; i++)
        {
          if (n16data[i] == null) break;
          sb.append(n16data[i].name);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_NAME32:
        NAME32 n32data[] = (NAME32[])hDataObject;
        if (n32data == null)
        {
          n32data = new NAME32[dArrayLength];
          getData(n32data);
        }
        for (int i = 0; i < dCompletionLength && i < n32data.length; i++)
        {
          if (n32data[i] == null) break;
          sb.append(n32data[i].name);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_NAME48:
        NAME48 n48data[] = (NAME48[])hDataObject;
        if (n48data == null)
        {
          n48data = new NAME48[dArrayLength];
          getData(n48data);
        }
        getData(n48data);
        for (int i = 0; i < dCompletionLength && i < n48data.length; i++)
        {
          if (n48data[i] == null) break;
          sb.append(n48data[i].name);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_NAME64:
        NAME64 n64data[] = (NAME64[])hDataObject;
        if (n64data == null)
        {
          n64data = new NAME64[dArrayLength];
          getData(n64data);
        }
        for (int i = 0; i < dCompletionLength && i < n64data.length; i++)
        {
          if (n64data[i] == null) break;
          sb.append(n64data[i].name);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_NAME16I:
        NAME16I n16idata[] = (NAME16I[])hDataObject;
        if (n16idata == null)
        {
          n16idata = new NAME16I[dArrayLength];
          getData(n16idata);
        }
        for (int i = 0; i < dCompletionLength && i < n16idata.length; i++)
        {
          if (n16idata[i] == null) break;
          sb.append(n16idata[i].name);
          sb.append(" ");
          sb.append(n16idata[i].ival);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_NAME16FI:
        NAME16FI n16fidata[] = (NAME16FI[])hDataObject;
        if (n16fidata == null)
        {
          n16fidata = new NAME16FI[dArrayLength];
          getData(n16fidata);
        }
        for (int i = 0; i < dCompletionLength && i < n16fidata.length; i++)
        {
          if (n16fidata[i] == null) break;
          sb.append(n16fidata[i].name);
          sb.append(" ");
          sb.append(n16fidata[i].fval);
          sb.append(" ");
          sb.append(n16fidata[i].ival);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_NAME16II:
        NAME16II n16iidata[] = (NAME16II[])hDataObject;
        if (n16iidata == null)
        {
          n16iidata = new NAME16II[dArrayLength];
          getData(n16iidata);
        }
        for (int i = 0; i < dCompletionLength && i < n16iidata.length; i++)
        {
          if (n16iidata[i] == null) break;
          sb.append(n16iidata[i].name);
          sb.append(" ");
          sb.append(n16iidata[i].i1val);
          sb.append(" ");
          sb.append(n16iidata[i].i2val);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_NAME16DBLDBL:
        NAME16DBLDBL n16dddata[] = (NAME16DBLDBL[])hDataObject;
        if (n16dddata == null)
        {
          n16dddata = new NAME16DBLDBL[dArrayLength];
          getData(n16dddata);
        }
        for (int i = 0; i < dCompletionLength && i < n16dddata.length; i++)
        {
          if (n16dddata[i] == null) break;
          sb.append(n16dddata[i].name);
          sb.append(" ");
          sb.append(n16dddata[i].d1val);
          sb.append(" ");
          sb.append(n16dddata[i].d2val);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_NAME32DBLDBL:
        NAME32DBLDBL n32dddata[] = (NAME32DBLDBL[])hDataObject;
        if (n32dddata == null)
        {
          n32dddata = new NAME32DBLDBL[dArrayLength];
          getData(n32dddata);
        }
        for (int i = 0; i < dCompletionLength && i < n32dddata.length; i++)
        {
          if (n32dddata[i] == null) break;
          sb.append(n32dddata[i].name);
          sb.append(" ");
          sb.append(n32dddata[i].d1val);
          sb.append(" ");
          sb.append(n32dddata[i].d2val);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_NAME64DBLDBL:
        NAME64DBLDBL n64dddata[] = (NAME64DBLDBL[])hDataObject;
        if (n64dddata == null)
        {
          n64dddata = new NAME64DBLDBL[dArrayLength];
          getData(n64dddata);
        }
        for (int i = 0; i < dCompletionLength && i < n64dddata.length; i++)
        {
          if (n64dddata[i] == null) break;
          sb.append(n64dddata[i].name);
          sb.append(" ");
          sb.append(n64dddata[i].d1val);
          sb.append(" ");
          sb.append(n64dddata[i].d2val);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_NAME32I:
        NAME32I n32idata[] = (NAME32I[])hDataObject;
        if (n32idata == null)
        {
          n32idata = new NAME32I[dArrayLength];
          getData(n32idata);
        }
        for (int i = 0; i < dCompletionLength && i < n32idata.length; i++)
        {
          if (n32idata[i] == null) break;
          sb.append(n32idata[i].name);
          sb.append(" ");
          sb.append(n32idata[i].ival);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_NAME48I:
        NAME48I n48idata[] = (NAME48I[])hDataObject;
        if (n48idata == null)
        {
          n48idata = new NAME48I[dArrayLength];
          getData(n48idata);
        }
        for (int i = 0; i < dCompletionLength && i < n48idata.length; i++)
        {
          if (n48idata[i] == null) break;
          sb.append(n48idata[i].name);
          sb.append(" ");
          sb.append(n48idata[i].ival);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_NAME64I:
        NAME64I n64idata[] = (NAME64I[])hDataObject;
        if (n64idata == null)
        {
          n64idata = new NAME64I[dArrayLength];
          getData(n64idata);
        }
        for (int i = 0; i < dCompletionLength && i < n64idata.length; i++)
        {
          if (n64idata[i] == null) break;
          sb.append(n64idata[i].name);
          sb.append(" ");
          sb.append(n64idata[i].ival);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_BYTE:
        byte bdata[] = (byte[])hDataObject;
        if (bdata == null)
        {
          bdata = new byte[dArrayLength];
          getData(bdata);
        }
        for (int i = 0; i < dCompletionLength && i < bdata.length; i++)
        {
          sb.append(bdata[i]);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_STRUCT:
        if (hStructObject != null && hStructObject instanceof TTaggedStructure)
        {
          return ((TTaggedStructure)hStructObject).toString();
        }
        if (hStructObject instanceof TTaggedStructure[])
        {
          for (int i = 0; i < ((TTaggedStructure[])hStructObject).length; i++)
          {
            sb.append(((TTaggedStructure[])hStructObject)[i].toString());
            sb.append(arrayDelimiter);
          }
          return sb.toString();
        }
        if (hDataObject != null)
        {
          return TStructRegistry.toString(dTag, (byte[])hDataObject);
        }
        return new String("user defined structure");
      case TFormat.CF_SHORT:
        short sdata[] = (short[])hDataObject;
        if (sdata == null)
        {
          sdata = new short[dArrayLength];
          getData(sdata);
        }
        for (int i = 0; i < dCompletionLength && i < sdata.length; i++)
        {
          sb.append(sdata[i]);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_LONG:
        int idata[] = (int[])hDataObject;
        if (idata == null)
        {
          idata = new int[dArrayLength];
          getData(idata);
        }
        for (int i = 0; i < dCompletionLength && i < idata.length; i++)
        {
          sb.append(idata[i]);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_BOOLEAN:
        int booldata[] = (int[])hDataObject;
        if (booldata == null)
        {
          booldata = new int[dArrayLength];
          getData(booldata);
        }
        for (int i = 0; i < dCompletionLength && i < booldata.length; i++)
        {
          sb.append(booldata[i] == 0 ? "false" : "true");
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_LNGINT:
        INTINT lidata[] = (INTINT[])hDataObject;
        if (lidata == null)
        {
          lidata = new INTINT[dArrayLength];
          getData(lidata);
        }
        for (int i = 0; i < dCompletionLength && i < lidata.length; i++)
        {
          sb.append(lidata[i].i1val);
          sb.append(" ");
          sb.append(lidata[i].i2val);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_FLOAT:
        float fdata[] = (float[])hDataObject;
        if (fdata == null)
        {
          fdata = new float[dArrayLength];
          getData(fdata);
        }
        for (int i = 0; i < dCompletionLength && i < fdata.length; i++)
        {
          sb.append(fdata[i]);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_FLTINT:
        FLTINT fidata[] = (FLTINT[])hDataObject;
        if (fidata == null)
        {
          fidata = new FLTINT[dArrayLength];
          getData(fidata);
        }
        for (int i = 0; i < dCompletionLength && i < fidata.length; i++)
        {
          if (fidata[i] == null) break;
          sb.append(fidata[i].fval);
          sb.append(" ");
          sb.append(fidata[i].ival);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_FLTFLT:
        FLTFLT ffdata[] = (FLTFLT[])hDataObject;
        if (ffdata == null)
        {
          ffdata = new FLTFLT[dArrayLength];
          getData(ffdata);
        }
        for (int i = 0; i < dCompletionLength && i < ffdata.length; i++)
        {
          if (ffdata[i] == null) break;
          sb.append(ffdata[i].f1val);
          sb.append(" ");
          sb.append(ffdata[i].f2val);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_SPECTRUM:
        SPECTRUM sp = (SPECTRUM)hDataObject;
        if (sp == null)
        {
          sp = new SPECTRUM();
          getData(sp);
        }
        sb.append(sp.status);
        sb.append(" ");
        sb.append(sp.s_start);
        sb.append(" ");
        sb.append(sp.s_inc);
        sb.append(" ");
        sb.append(sp.tm);
        sb.append(" ");
        sb.append(sp.comment);
        sb.append(arrayDelimiter);
        for (int i = 0; i < sp.length; i++)
        {
          sb.append(sp.d_spect_array[i]);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_DOUBLE:
        double ddata[] = (double[])hDataObject;
        if (ddata == null)
        {
          ddata = new double[dArrayLength];
          getData(ddata);
        }
        for (int i = 0; i < dCompletionLength && i < ddata.length; i++)
        {
          sb.append(ddata[i]);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_DBLINT:
        DBLINT didata[] = (DBLINT[])hDataObject;
        if (didata == null)
        {
          didata = new DBLINT[dArrayLength];
          getData(didata);
        }
        for (int i = 0; i < dCompletionLength && i < didata.length; i++)
        {
          if (didata[i] == null) break;
          sb.append(didata[i].dval);
          sb.append(" ");
          sb.append(didata[i].ival);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      case TFormat.CF_DBLDBL:
        DBLDBL dddata[] = (DBLDBL[])hDataObject;
        if (dddata == null)
        {
          dddata = new DBLDBL[dArrayLength];
          getData(dddata);
        }
        for (int i = 0; i < dCompletionLength && i < dddata.length; i++)
        {
          if (dddata[i] == null) break;
          sb.append(dddata[i].d1val);
          sb.append(" ");
          sb.append(dddata[i].d2val);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
    }
    return new String("");
  }
  /**
   * @brief An array of chars (to be interpretted as a text string)
   * @param data
   *          is a reference to the char[] array used to create and bind to the
   *          TDataType object
   */
  public TDataType(char[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * @brief A subset of an array of chars (to be interpretted as a text string)
   * @param data
   *          is a reference to the char[] array used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(char[] data, int length)
  {
    initTDataType(data, length);
  }
  public void initTDataType(char[] data, int length)
  {
    try
    {
      if (length > data.length) length = data.length;
      dFormat = TFormat.CF_TEXT;
      dArrayLength = length;
      dSizeInBytes = length;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        byte bdata;
        for (; i < dArrayLength && i < (n + 1) * mtu; i++)
        {
          bdata = (byte) data[i];
          ds.writeByte(bdata);
        }
        byte[] b = dBuffers[n].toByteArray();
        dBuffers[n].reset();
        ds.write(b, 0, b.length);
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  /**
   * @brief A text string buffer
   * @param data
   *          is a reference to the StringBuffer used to create and bind to the
   *          TDataType object
   */
  public TDataType(StringBuffer data)
  {
    initTDataType(data, data.capacity());
  }
  /**
   * @brief A subset of a text string buffer
   * @param data
   *          is a reference to the StringBuffer used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset length to use.
   */
  public TDataType(StringBuffer data, int length)
  {
    initTDataType(data, length);
  }
  public void initTDataType(StringBuffer data, int length)
  {
    try
    {
      if (length > data.capacity()) length = data.capacity();
      dFormat = TFormat.CF_TEXT;
      dArrayLength = length;
      dSizeInBytes = dArrayLength;
      byte[] bdata = data.toString().getBytes();
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < bdata.length && i < (n + 1) * mtu; i++)
          ds.writeByte(bdata[i]);
        byte[] b = dBuffers[n].toByteArray();
        dBuffers[n].reset();
        ds.write(b, 0, b.length);
        ds.close();
      }
      hDataObject = data;
//      hText = new char[dArrayLength];
//      data.getChars(0, data.length(), hText, 0);
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  /**
   * @brief A text string
   * @param data
   *          is a reference to the String used to create and bind to the
   *          TDataType object
   */
  public TDataType(String data)
  {
    initTDataType(data, data.length());
  }
  /**
   * @brief A text string
   * @param data
   *          is a reference to the StringBuffer used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset length to use.
   */
  public TDataType(String data, int length)
  {
    initTDataType(data, length);
  }
  public void initTDataType(String data, int length)
  {
    try
    {
      if (length > data.length()) length = data.length();
      dFormat = TFormat.CF_TEXT;
      dArrayLength = length;
      dSizeInBytes = dArrayLength;
      byte[] bdata = data.getBytes();
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < bdata.length && i < (n + 1) * mtu; i++)
          ds.writeByte(bdata[i]);
        byte[] b = dBuffers[n].toByteArray();
        dBuffers[n].reset();
        ds.write(b, 0, b.length);
        ds.close();
      }
      hDataObject = new char[dArrayLength];
      if (dArrayLength > 0) data.getChars(0, dArrayLength - 1, (char[])hDataObject, 0);
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  public int getData(char[] data)
  {
    if (dArrayLength > data.length) return TErrorList.string_too_long;
    int i, n, m;
    byte bdata[] = new byte[data.length];
    for (i = 0, n = 0, m = 0; n < numblks; n++, m = i)
    {
      if (dBuffers[n] == null) break;
      byte[] b = dBuffers[n].toByteArray();
      if (m + b.length <= data.length)
      {
        System.arraycopy(b, 0, bdata, m, b.length);
        i += b.length;
      }
    }
    String sdata = new String(bdata,0,i);
    sdata.getChars(0, sdata.length(), data, 0);
    if (hDataObject instanceof StringBuffer) 
      ((StringBuffer)hDataObject).replace(0,sdata.length(),sdata);
    else if (hDataObject instanceof char[])
      System.arraycopy(data,0,(char[])hDataObject,0,sdata.length());
    dCompletionLength = i;
    return 0;
  }
  public int getData(StringBuffer data)
  {
    if (dArrayLength > data.capacity()) return TErrorList.string_too_long;
    int i, n, m;
    byte bdata[] = new byte[data.capacity()];
    for (i = 0, n = 0, m = 0; n < numblks; n++, m = i)
    {
      if (dBuffers[n] == null) break;
      byte[] b = dBuffers[n].toByteArray();
      if (m + b.length <= data.capacity())
      {
        System.arraycopy(b, 0, bdata, m, b.length);
        i += b.length;
      }
    }
    String sdata = new String(bdata,0,i);
    data.replace(0,sdata.length(),sdata);
    if (hDataObject instanceof StringBuffer) 
      ((StringBuffer)hDataObject).replace(0,sdata.length(),sdata);
    else if (hDataObject instanceof char[])
      System.arraycopy(data,0,(char[])hDataObject,0,sdata.length());
    dCompletionLength = i;
    return 0;
  }
  public void setTag(String tag)
  {
    dTag = tag;
  }
  public int getCompletionLength()
  {
    return dCompletionLength;
  }
  public void setArrayLength(int arrayLength)
  {
    int dlen = arrayLength;
    if (dFormat == TFormat.CF_STRUCT) 
    { // if tagged structure, then this call takes the number of structures, not bytes !
      int ssiz = TStructRegistry.getSizeInBytes(dTag);
      if (ssiz < 1) ssiz = 1;
      dlen *= ssiz;
    }
    dArrayLength = dlen;
  }
}