package de.desy.tine.definitions;

/**
 * \brief de.desy.tine.client data access flags
 * 
 * @author duval The access word in a TLink will contain information pertaining
 *         to the type of access requested by the caller or by the subsystem. A
 *         client can specify 'CA_READ' or 'CA_WRITE' and in addition ask for a
 *         data link over a connected (TCP/IP) socket by specifying
 *         'CA_CONNECT'. A server can examine the access word when handling the
 *         call. A server can also determine whether the call is coming from the
 *         local alarm server or local history server. Likewise, a server can
 *         examine the scope of the call, that is whether the link is starting
 *         up (the 'CA_FIRST' bit is set) or going out of scope (the 'CA_LAST'
 *         bit is set).
 */
public final class TAccess
{
  private short mode;
  public static final short CA_NULL = 0x00;
  public static final short CA_READ = 0x01;
  public static final short CA_WRITE = 0x02;
  public static final short CA_ASYNC = 0x04;
  public static final short CA_SYNC = 0x08;
  public static final short CA_ALARM = 0x10;
  public static final short CA_HIST = 0x20;
  public static final short CA_FIRST = 0x40;
  public static final short CA_LAST = 0x80;
  public static final short CA_CONNECT = 0x2000;
  /**
   * Constructor. Used for when CF-constants are or-ed
   * 
   * @param newMode
   */
  public TAccess(int newMode)
  {
    mode = (short) newMode;
  }
  public TAccess(short newMode)
  {
    mode = newMode;
  }
  public TAccess()
  {
    this(CA_NULL);
  }
  public void clear()
  {
    mode = CA_NULL;
  }
  public static short toBase(short access)
  {
    return (short) (access & 0x03);
  }
  
  public static TAccess valueOf(short val) {
	  return new TAccess(val);
  }
  
  // public static final short CA_SYNCNOTIFY = 0x8000;
  public String toString()
  {
    return TAccess.toString(mode);
  }
  public static String toString(short access)
  {
    String add = "";
    if ((access & 0x04) == 0x4) add = "async ";
    if ((access & 0x08) == 0x8) add = "sync ";
    switch (access & 0x03)
    {
      case CA_NULL:
        return "none";
      case CA_READ:
        return add + "read only";
      case CA_WRITE:
        return add + "write";
      default:
        return add + "read/write";
    }
  }
  /**
   * Returns the access mode bits as a short integer.
   * 
   * @return
   */
  public short toShort()
  {
    return this.mode;
  }
  /**
   * Returns true if all bits which are set in pattern are also set in this
   * object.
   * 
   * @param pattern
   *          A Bit pattern.
   * @return
   */
  public boolean includes(short pattern)
  {
    return ((this.mode & pattern) == pattern);
  }
  /**
   * Returns true if read bit is set.
   * 
   * @return
   */
  public boolean isRead()
  {
    return ((this.mode & CA_READ) != 0);
  }
  /**
   * Returns true if write bit is set.
   * 
   * @return
   */
  public boolean isWrite()
  {
    return ((this.mode & CA_WRITE) != 0);
  }
  /**
   * Returns true if no bit is set.
   * 
   * @return
   */
  public boolean isNull()
  {
    return (this.mode == CA_NULL);
  }
  /**
   * Returns true if all bits are equal.
   * 
   * @param acc
   * @return
   */
  public boolean equals(TAccess acc)
  {
    return (this.mode == acc.mode);
  }
  /**
   * Returns true if all mode bits which are set in another object are also set
   * in this object.
   * 
   * @param acc
   * @return
   */
  public boolean includes(TAccess acc)
  {
    return includes(acc.mode);
  }
  public void add(short mode)
  {
    this.mode |= mode;
  }
  public void add(TAccess acc)
  {
    this.mode |= acc.mode;
  }
  // TODO More methods (getSync etc)
}