/*
 * Created on May 10, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package de.desy.tine.structUtils;
import java.util.*;
import java.io.*;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.endianUtils.Swap;
/**
 * @brief de.desy.tine.client Tagged Structure 
 *
 * The transfer of 'user-defined' data structures is allowed in de.desy.tine.client. 
 * In order to avoid nasty problems concerning alignment and byte swapping
 * it is strongly suggested to extend the TTaggedStructure class when
 * creating a de.desy.tine.client data structure.
 */
public class TTaggedStructure implements Cloneable
{
  private String structName;
  private int sizeInBytes;
  private int numOfElements;
  private int numOfFields;
  private ArrayList fieldArrayList = null; 
  private byte[] hByteArray = null;
  private ByteArrayOutputStream boas; 
//  private static boolean initialized = false; // made this static 9.3.07
  private boolean initialized = false; // made this un-static 26.6.07
  private TTaggedStructure instance; 
  public Object clone()
  {
    Object o = null;
    try
    {
      o = super.clone();
    }
    catch (CloneNotSupportedException e)
    {
      System.err.println("TTaggedStructure cannot clone");
    }
    return o;
  }
  /**
   * @brief Creates a tagged structure with the given name.
   * 
   * @param name is the name to be used as the structure tag.
   * Note that only the first 8 characters of the given name will
   * be used as the tag name!
   * 
   * @note The addField() method must be called for each element of
   * the structure data element.
   */
  public TTaggedStructure(String name) 
  {
    if (structName == null) structName = new String(name);
    if (instance == null) instance = this; 
    fieldArrayList = new ArrayList();
    if (TStructRegistry.contains(structName))
    {
      initialized = true;
      sizeInBytes = TStructRegistry.getSizeInBytes(structName);
    }
    numOfElements++;
  }
  /**
   * @brief Creates a tagged structure using the class name as the 
   * structure tag.
   * Note that only the first 8 characters of the class name will
   * be used as the tag name!
   * 
   * @note The addField() method must be called for each element of
   * the structure data element.
   * 
   * \b Example:
   * @include eg_tineTaggedStructure1.java
   */
  public TTaggedStructure() 
  {
    String name = this.getClass().getName();
    StringTokenizer t = new StringTokenizer(name,".$");
    String lastToken = name;
    while (t.hasMoreTokens())
    {
      lastToken  = t.nextToken();
    }
    if (structName == null) structName = new String(lastToken);
    if (instance == null) instance = this; 
    fieldArrayList = new ArrayList();
    if (TStructRegistry.contains(structName))
    {
      initialized = true;
      sizeInBytes = TStructRegistry.getSizeInBytes(structName);
    }
    numOfElements++;
  }
  /**
   * @brief Returns the structure's assigned tag name. 
   * 
   * @return The structure's assigned tag name as a string.
   */
  public String getName() { return structName; }
  /**
   * @brief Returns the structure's data size in bytes. 
   * 
   * @return The structure's data size in bytes.
   */
  public int getSizeInBytes() { return sizeInBytes; }
  /**
   * @brief Returns the string representation of the tagged structure. 
   * 
   * @return The string representation of the tagged structure.
   */
  public String toString()
  {
    StringBuffer sb = new StringBuffer("");
    structField sf;
    if (!initialized) initDone(numOfElements);
    for (int i=0; i<fieldArrayList.size(); i++)
    {
      sf = (structField)fieldArrayList.get(i);
      switch (sf.format)
      {
        case TFormat.CF_BYTE:
          sb.append(sf.name + " ->");
          int ibyte;
          for (int k=0; k<sf.size; k++)
          {
            if ((ibyte = sf.hBytes[k]) < 0) ibyte += 256; 
            sb.append(" " + ibyte);
          }
          sb.append("\n");
          break;
        case TFormat.CF_SHORT:
          sb.append(sf.name + " ->");
          for (int k=0; k<sf.size; k++)
          {
            sb.append(" " + sf.hShorts[k]);
          }
          sb.append("\n");
          break;
        case TFormat.CF_LONG:
          sb.append(sf.name + " ->");
          for (int k=0; k<sf.size; k++)
          {
            sb.append(" " + sf.hLongs[k]);
          }
          sb.append("\n");
          break;
        case TFormat.CF_FLOAT:
          sb.append(sf.name + " ->");
          for (int k=0; k<sf.size; k++)
          {
            sb.append(" " + sf.hFloats[k]);
          }
          sb.append("\n");
          break;
        case TFormat.CF_DOUBLE:
          sb.append(sf.name + " ->");
          for (int k=0; k<sf.size; k++)
          {
            sb.append(" " + sf.hDoubles[k]);
          }
          sb.append("\n");
          break;
        case TFormat.CF_TEXT:
          sb.append(sf.name + " ->");
          sb.append(" " + (new String(sf.hChars)).trim());
          sb.append("\n");
          break;
      }      
    }
    return sb.toString();
  }
  public void toStruct(byte[] data)
  {
    if (!initialized) initDone(numOfElements);
    try
    {
      ByteArrayInputStream dis = new ByteArrayInputStream(data);
      DataInputStream ds = new DataInputStream(dis);
      structField sf;
      int sz;
      byte[] db = new byte[8];
      for (int i=0; i<fieldArrayList.size(); i++)
      {
        sf = (structField)fieldArrayList.get(i);
        sz = TFormat.formatSizeOf(sf.format);
        switch (sf.format)
        {
          case TFormat.CF_BYTE:
            ds.read(data,0,sf.size*sz);
            System.arraycopy(data, 0, sf.hBytes, 0, sf.size*sz);
            break;
          case TFormat.CF_SHORT:
            for (int k=0; k<sf.size; k++)
            {
              sf.hShorts[k] = Swap.Short(ds.readShort());
            }
            break;
          case TFormat.CF_LONG:
            for (int k=0; k<sf.size; k++)
            {
              sf.hLongs[k] = Swap.Long(ds.readInt());
            }
            break;
          case TFormat.CF_FLOAT:
            ds.read(data,0,sf.size*sz);
            for (int k=0; k<sf.size; k++)
            {
              System.arraycopy(data, k*sz, db, 0, sz);
              sf.hFloats[k] = Swap.Float(db);
            }
            break;
          case TFormat.CF_DOUBLE:
            ds.read(data,0,sf.size*sz);
            for (int k=0; k<sf.size; k++)
            {
              System.arraycopy(data, k*sz, db, 0, sz);
              sf.hDoubles[k] = Swap.Double(db);
            }
            break;
          case TFormat.CF_TEXT:
            ds.read(data,0,sf.size*sz);
//            System.arraycopy(data, 0, sf.hChars, 0, sf.size*sz);
            for (int k=0; k<sf.size; k++)
            {
              sf.hChars[k] = (char)data[k];
            }
            break;
        }
      }
      ds.close(); dis.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  public byte[] toByteArray()
  {
    if (!initialized) initDone(numOfElements);
    try
    {
      if (boas == null) boas = new ByteArrayOutputStream(numOfElements*sizeInBytes);
      boas.reset();
      DataOutputStream ds = new DataOutputStream(boas);
      structField sf;
      int sz;
      for (int i=0; i<fieldArrayList.size(); i++)
      {
        sf = (structField)fieldArrayList.get(i);
        sz = TFormat.formatSizeOf(sf.format);
        switch (sf.format)
        {
          case TFormat.CF_BYTE:
            ds.write(sf.hBytes,0,sf.size*sz);
            break;
          case TFormat.CF_SHORT:
            for (int k=0; k<sf.size; k++)
            {
              ds.writeShort(Swap.Short(sf.hShorts[k]));
            }
            break;
          case TFormat.CF_LONG:
            for (int k=0; k<sf.size; k++)
            {
              ds.writeInt(Swap.Long(sf.hLongs[k]));
            }
            break;
          case TFormat.CF_FLOAT:
            for (int k=0; k<sf.size; k++)
            {
              ds.write(Swap.Float(sf.hFloats[k]),0,sz);
            }
            break;
          case TFormat.CF_DOUBLE:
            for (int k=0; k<sf.size; k++)
            {
              ds.write(Swap.Double(sf.hDoubles[k]),0,sz);
            }
            break;
          case TFormat.CF_TEXT:
//             ds.write(sf.hChars.toString().getBytes(),0,sf.size*sz);
//            ds.writeBytes(structName);
            for (int k=0; k<sf.size; k++)
            {
              ds.writeByte(sf.hChars[k]);
            }
            break;
        }
      }
      ds.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      System.out.println("could not prepare byte array");
      return null;
    }  
    hByteArray = boas.toByteArray();
    return hByteArray;
  }
  /**
   * @brief Adds a field to the tagged structure. 
   * 
   * This method plays an integral role in defining the tagged
   * structure.  It should be called once for each field defining the
   * structure and in the designated order, starting with the first 
   * element and ending with the last element.
   * 
   * @param field is the field element in question.  It must be an
   * array of a primitive type.  Even if the field contains only one
   * element, it should be declared as an array of 1 (otherwise java 
   * will pass the contents by value!).  The field can also be one of
   * 'String', 'StringBuffer', or 'char[]' type.
   * 
   * @return 0 upon success, otherwise -1.
   * 
   * \b Example:
   * @include eg_tineTaggedStructure3.java
   * 
   * @note It is a good idea to call this method for each field inside the
   * constructor for the TTaggedStructure object as shown below: 
   * 
   * \b Example:
   * @include eg_tineTaggedStructure2.java
   */
  public int addField(Object field)
  {
    return addField(field,field.getClass().getName());
  }
  /**
   * @brief Adds a field to the tagged structure. 
   * 
   * This method plays an integral role in defining the tagged
   * structure.  It should be called once for each field defining the
   * structure and in the designated order, starting with the first 
   * element and ending with the last element.
   * 
   * @param field is the field element in question.  It must be an
   * array of a primitive type.  Even if the field contains only one
   * element, it should be declared as an array of 1 (otherwise java 
   * will pass the contents by value!).  The field can also be one of
   * 'String', 'StringBuffer', or 'char[]' type.
   * @param name is the name of the field, which is used in the 'toString()'
   * method of displaying the contents.
   * 
   * @return 0 upon success, otherwise -1.
   * 
   * @note It is a good idea to call this method for each field inside the
   * constructor for the TTaggedStructure object. 
   * 
   * \b Example:
   * @include eg_tineTaggedStructure1.java
   */
  public int addField(Object field,String name)
  {
 //   if (initialized) return -1;
    int fieldLength = 0;
    short fieldFormat = TFormat.CF_NULL;
    String fieldName = name;
    if (field instanceof byte[]) 
    {
      fieldLength = ((byte[])field).length;
      fieldFormat = TFormat.CF_BYTE;
    }
    else if (field instanceof short[]) 
    {
      fieldLength = ((short[])field).length;
      fieldFormat = TFormat.CF_SHORT;
    }
    else if (field instanceof int[]) 
    {
      fieldLength = ((int[])field).length;
      fieldFormat = TFormat.CF_LONG;
    }
    else if (field instanceof float[]) 
    {
      fieldLength = ((float[])field).length;
      fieldFormat = TFormat.CF_FLOAT;
    }
    else if (field instanceof double[]) 
    {
      fieldLength = ((double[])field).length;
      fieldFormat = TFormat.CF_DOUBLE;
    }
    else if (field instanceof String) 
    {
      fieldLength = ((String)field).length();
      fieldFormat = TFormat.CF_TEXT;
    }
    else if (field instanceof StringBuffer) 
    {
      fieldLength = ((StringBuffer)field).length();
      fieldFormat = TFormat.CF_TEXT;
    }
    else if (field instanceof char[]) 
    {
      fieldLength = ((char[])field).length;
      fieldFormat = TFormat.CF_TEXT;
    }
    else
    {
      return -1;
    }
    structField sf = instance.new structField(fieldName,fieldLength,field);
    fieldArrayList.add(numOfFields, sf);
    numOfFields++;
    return addField(fieldName,fieldFormat,fieldLength);
  }
  
  private int addField(String fieldName,short fieldFormat,int fieldLength)
  {
    if (initialized) return -1;
    TStructRegistry.fill(structName, fieldFormat, fieldLength);
    sizeInBytes += fieldLength * TFormat.formatSizeOf(fieldFormat);
    return 0;
  }
  
  /**
   * @brief Terminates the tagged structure definition and signals that
   * the tagged structure is ready for use. 
   * 
   * This method plays an integral role in defining the tagged
   * structure.  It should be called following all calls to addField().
   */
  public void initDone() { initDone(1); }
  public void initDone(int numElements)
  {
    if (!initialized)
    {
      initialized = true;
      hByteArray = new byte[numElements*sizeInBytes];
      numOfElements = numElements;
      TStructRegistry.fill(structName,TFormat.CF_NULL,numOfElements);
    }
    if (hByteArray == null) 
    {
      hByteArray = new byte[numElements*sizeInBytes];
    }
  }
  public byte[] getByteArray() 
  { 
    initDone(numOfElements);
    return hByteArray; 
  }
  private class structField
  {
    String name;
    int size;
    short format;
    byte[] hBytes;
    short[] hShorts;
    int[] hLongs;
    float[] hFloats;
    double[] hDoubles;
    char[] hChars;   
    structField(String fieldname,int fieldsize,Object buffer)
    {
      name = fieldname; size = fieldsize; 
      if (buffer instanceof byte[]) {hBytes = (byte[])buffer; format = TFormat.CF_BYTE;} 
      else if (buffer instanceof short[]) {hShorts = (short[])buffer; format = TFormat.CF_SHORT;} 
      else if (buffer instanceof int[]) {hLongs = (int[])buffer; format = TFormat.CF_LONG; }
      else if (buffer instanceof float[]) {hFloats = (float[])buffer; format = TFormat.CF_FLOAT; }
      else if (buffer instanceof double[]) {hDoubles = (double[])buffer; format = TFormat.CF_DOUBLE; }
      else if (buffer instanceof char[]) {hChars = (char[])buffer; format = TFormat.CF_TEXT; }
    }
  }
}
